/*
 *
 * Copyright 2002-2008 Network Appliance
 */

#ifndef _LDAP_OPT_H
#define _LDAP_OPT_H

#include <ldap.h>
#include <ldap/libldap/ldap-int.h>
#include <rpc/types.h>
#include <smb/smb.h>
#include <auth/sec_ldap.h>
#include <net7g/sys/sk_socketvar.h>


#define MAXLDAPBASEDN		1024
#define MAXLDAPSEARCHFILTER	1024
#define MAXLDAPATTRNAME		64
#define MAXNSSMAPOPTLEN		128


/* basedn link list */
typedef struct _ldap_blink
{
	char	*pbasedn;
	int	scope;	
	struct _ldap_blink	*pnext;	
} ldap_blink;

typedef struct {
	ldap_blink	*pheadb;
	ldap_blink	*ptailb;
	sk_rwlock_t	rwlock;
} ldap_blink_header;

typedef enum{
	LDAP_FILTERTYPE_UNAME2PASSWD = 0,
	LDAP_FILTERTYPE_UIDNUMBER2PASSWD,
	LDAP_FILTERTYPE_GRPNAME2GID,
	LDAP_FILTERTYPE_GID2GRPNAME,
	LDAP_FILTERTYPE_UNAME2GRPLIST,
	LDAP_FILTERTYPE_NETGRP,
	LDAP_FILTERTYPE_UNIX2WIN,
	LDAP_FILTERTYPE_WIN2UNIX,
	LDAP_FILTERTYPE_TOTALCNT
} ldap_search_type;

typedef enum{
	LDAP_CHECKOUT_ALL,
	LDAP_CHECKOUT_NONEAD,
	LDAP_CHECKOUT_AD
} ldap_checkout_type;

typedef enum {
	VALID_INPUT = 0,
	INVALID_INPUT,
	HOSTNAME_EXCEEDED_LIMIT,
	INVALID_IPV6_INPUT_FORMAT,
	INVALID_HOST,
	INVALID_PORT
} ldap_servers_input_error;

#define LDAP_FILT_AVGSIZ		256
#define LDAP_FILTER_GENERIC		"(&(objectClass=%s)(%s=%s))"

#define LDAP_FILTER_UNAME2PASSWD	"(&(objectClass=posixAccount)(uid=%s))"
#define LDAP_FILTER_UIDNUMBER2PASSWD	"(&(objectClass=posixAccount)(uidNumber=%s))"
#define LDAP_FILTER_GRPNAME2GID		"(&(objectClass=posixGroup)(cn=%s))"
#define LDAP_FILTER_GID2GRPNAME		"(&(objectClass=posixGroup)(gidNumber=%s))"
#define LDAP_FILTER_UNAME2GRPLIST	"(&(objectClass=posixGroup)(memberUid=%s))"
#define LDAP_FILTER_NETGROUP		"(&(objectClass=nisNetGroup)(cn=%s))"
#define LDAP_FILTER_UNIX2WIN		"(&(objectClass=User)(unixAccount=%s))"
#define LDAP_FILTER_WIN2UNIX		"(&(objectClass=User)(windowsAccount=%s))"

#define LDAP_POSIXACCOUNT	"posixAccount"
#define LDAP_POSIXGROUP		"posixGroup"
#define LDAP_NISNETGROUP	"nisNetgroup"
#define LDAP_UID		"uid"
#define LDAP_UIDNUMBER		"uidNumber"
#define LDAP_GIDNUMBER		"gidNumber"
#define LDAP_USERPASSWORD	"userPassword"
#define LDAP_HOMEDIRECTORY	"homeDirectory"
#define LDAP_LOGINSHELL		"loginShell"
#define LDAP_GECOS		"gecos"
#define LDAP_GROUPNAME		"cn"
#define LDAP_MEMBERUID		"memberUid"
#define LDAP_UNIQUEMEMBER	"uniqueMember"
#define LDAP_GROUPOFUNIQUENAMES "groupOfUniqueNames"
#define LDAP_MEMBERNISNETGROUP	"memberNisNetgroup"
#define LDAP_NISNETGROUPTRIPLE	"nisNetgroupTriple"
#define LDAP_NETGROUPNAME	"cn"

#define LDAP_SAMACCOUNTNAME	"sAMAccountName"
#define LDAP_WINDOWSACCOUNT	"windowsaccount"
#define LDAP_UNIXACCOUNT	"unixaccount"


extern char* getopt_ldap_enable(void*, char*, int);
extern char* getopt_ldap_userid(void*, char*, int);
extern char* getopt_ldap_passwd(void*, char*, int);

extern char* getopt_ldap_rfc2307bis_enable(void*, char*, int);
extern char* getopt_ldap_skip_cn_unescape_enable(void*, char*, int);

extern char* getopt_ldap_addomain(void*, char*, int);
extern char* getopt_ldap_servers(void*, char*, int);
extern char* getopt_ldap_pref_servers(void*, char*, int);
extern char* getopt_ldap_port(void*, char*, int);
extern char* getopt_ldap_timeout(void*, char*, int);
extern char* getopt_ldap_retry_delay(void*, char*, int);
extern char* getopt_ldap_fast_timeout(void*, char*, int);
extern char* getopt_ldap_minimum_bind_level(void*, char*, int);

extern char* getopt_ldap_basedn(void*, char*, int);
extern char* getopt_ldap_passwd_basedn(void *, char*, int);
extern char* getopt_ldap_group_basedn(void *, char*, int);
extern char* getopt_ldap_netgroup_basedn(void *, char*, int);

extern char* getopt_ldap_nssmap_OC_posixAccount(void*, char*, int);
extern char* getopt_ldap_nssmap_OC_posixGroup(void*, char*, int);
extern char* getopt_ldap_nssmap_OC_nisNetgroup(void*, char*, int);
extern char* getopt_ldap_nssmap_OC_groupOfUniqueNames(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_uid(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_uidNumber(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_gidNumber(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_userPassword(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_homeDirectory(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_loginShell(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_gecos(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_groupname(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_memberUid(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_uniqueMember(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_memberNisNetgroup(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_nisNetgroupTriple(void*, char*, int);
extern char* getopt_ldap_nssmap_ATTR_netgroupname(void*, char*, int);

extern char* getopt_ldap_usermap_enable(void *, char *, int);
extern char* getopt_ldap_usermap_basedn(void *, char *, int);
extern char* getopt_ldap_usermap_win2unix_OC(void*, char*, int);
extern char* getopt_ldap_usermap_win2unix_ATTR(void*, char*, int);
extern char* getopt_ldap_usermap_ATTR_unixaccount(void*, char*, int);
extern char* getopt_ldap_usermap_ATTR_windowsaccount(void*, char*, int);
extern char* getopt_ldap_usermap_symmetriclookup(void*, char*, int);


bool_t _ldap_basedn(char *basedn, ldap_blink_header *pblinkh);
void blink_init(ldap_blink_header *pblinkh);
void blink_free(ldap_blink_header *pblinkh);
bool_t blink_insert(ldap_blink_header *pblinkh, ldap_blink *pinsb);
bool_t blink_new(ldap_blink_header **ppbh, char *pDNSDomain);

void _ldap_release_ld(void);
LDAP* _ldap_checkout_and_multi_search_st(LDAP *pld,
					 ldap_checkout_type checkout_type,
					 LDAPConnection **ppLDAPConnection,
					 bool_t *pisadhandle,
					 char *pbh_Domain,
					 ldap_blink_header **ppbh,
					 bool_t free_pbh_inroutine,
					 char *psearchstr,
					 char *pattrs[],
					 LDAPMessage **ppres,
					 int *ldap_resp);

LDAP* _ldap_checkout_and_search_st(ldap_checkout_type checkout_type,
				   LDAPConnection **ppLDAPConnection,
				   bool_t *pisadhandle,
				   char *pbh_Domain,
				   ldap_blink_header *pbh,
				   char *psearchstr,
				   char *pattrs[],
				   LDAPMessage **ppres,
				   int *ldap_resp);
void _ldap_checkin_ld(void);
bool_t ldap_setgenericopt(char **ppoptaddr,const char *poptval,const int vallen);
bool ldap_validateopt(const char *poptval,const int vallen);
uint8_t ldap_get_hostandport_info(char *pinput, char **phostname, int *pport);
void ldap_servers_error_handle(const char *curhost, int status);
void ldap_setschemaopt(const char *poptname, 
			char **ppoptaddr,
			const char *poptval,
			const int vallen);
/* 	
* add the the prototypes for the sldap functions including the ssl functions
*/

unsigned long  ldap_ssl_get_context(void );
char *ldap_get_sslcontext(char *);


char * getopt_ldap_ssl_enable(void *, char *, int );
void setopt_ldap_ssl_enable(void *, char *, uint32_t , uint64_t , char *);
extern void * sslcfgctx_load_ldap_default(unsigned long *);
extern int    sslcfgctx_reload_ldap_default(unsigned long *);
extern char *ldap_config_searchstr(char *, int, const char *, ...);


/*******************************************************************************
 * LDAP handle status verification routine
 */
static inline 
bool_t _ldap_isvalidhandle(LDAP *ld)
{
	
	if ( (ld != NULL) && 
	     (ld->ld_defhost != NULL) && 
	     (strcmp(ld->ld_defhost, "") != 0) )
	{
		return true;
	}

	return false;
}

static inline
bool_t _ldap_hasvalidsocket(LDAP *ld)
{
	 struct socket *so;

	 if (NULL == ld)
	 	return false;
	 
	 so = fio_getsocket(ld->ld_sbp->sb_sd);

	 if ( (so == NULL) ||
	      (so->so_state & SS_CANTRCVMORE) ||
	      (so->so_state & SS_CANTSENDMORE) )
	 {
	 	return false;
	 }
	 else
	 {
	 	return true;
	 }
	 	
}

static inline
bool_t _ldap_resource_timeout(int rc)
{
	if ((rc == LDAP_SERVER_DOWN) ||
	    (rc == LDAP_CONNECT_ERROR) ||
	    (rc == LDAP_TIMEOUT))
	{
		return true;
	}
	else
		return false;
}

#endif
