/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The following files were originally developed by Netscape Communications
 * Corporation
 */
/*
 * The contents of this file are subject to the Netscape Public License
 * Version 1.0 (the "NPL"); you may not use this file except in
 * compliance with the NPL.  You may obtain a copy of the NPL at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the NPL is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the NPL
 * for the specific language governing rights and limitations under the
 * NPL.
 *
 * The Initial Developer of this code under the NPL is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation.  All Rights
 * Reserved.
 */

/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */
/*
 * The contents of these files are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use these files except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is lber-int.h 
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Copyright (C) 1998. All Rights Reserved
 */
/*
 * Portions created by NetApp, Inc. are Copyright (c) 2001.  All Rights
 * Reserved.
 */
/* lbet-int.h - internal header file for liblber */

#ifndef _LBERINT_H
#define _LBERINT_H

#ifdef __cplusplus
extern "C" {
#endif

#include <sys/types.h>
#include <ontap_sizes.h>
#include <sys/endian.h>
#include <sk/sk_common.h>

#ifdef NOT_USED_BY_NETAPP
#ifdef macintosh
# include "ldap-macos.h"
#else /* macintosh */
#if !defined(BSDI)
# include <malloc.h>
#endif
#endif

# include <sys/errno.h>
# include <sys/types.h>
#if defined(SUNOS4) || defined(SCOOS)
# include <sk/time.h>
#endif
#if defined( _WINDOWS )
#  define WIN32_LEAN_AND_MEAN
#  include <windows.h>
#  include <time.h>
/* No stderr in a 16-bit Windows DLL */
#  if defined(_WINDLL) && !defined(_WIN32)
#    define USE_DBG_WIN
#  endif
# else
#if !defined(XP_OS2)
/* #  include <sys/varargs.h> */
#  include <net7g/sys/sk_socket.h>
#  include <net7g/netinet/sk_in.h>
#endif
# endif /* defined( _WINDOWS ) */
#endif /* macintosh */

#include <legacy/string.h>
#include "portable.h"

#ifdef _WINDOWS
#include <winsock.h>
#include <io.h>
#endif /* _WINDOWS */

#ifdef XP_OS2
#include <os2sock.h>
#include <io.h>
#endif /* XP_OS2 */

/* No stderr in a 16-bit Windows DLL */
#if defined(_WINDLL) && !defined(_WIN32)
#define stderr NULL
#endif

#include "lber.h"

/*
 * Malloc tags definitions
 */
MALLOC_DECLARE(M_LDAP);


#define OLD_LBER_SEQUENCE	0x10L	/* w/o constructed bit - broken */
#define OLD_LBER_SET		0x11L	/* w/o constructed bit - broken */

#define LBER_UNINITIALIZED              0x0
#define LBER_INITIALIZED                0x1
#define LBER_VALID_BERELEMENT   0x2
#define LBER_VALID_SOCKBUF              0x3

/* LBER Sockbuf_IO options */
#define LBER_SB_OPT_GET_FD              1
#define LBER_SB_OPT_SET_FD              2
#define LBER_SB_OPT_HAS_IO              3
#define LBER_SB_OPT_SET_NONBLOCK        4
#define LBER_SB_OPT_GET_SSL             7
#define LBER_SB_OPT_DATA_READY          8
#define LBER_SB_OPT_SET_READAHEAD       9
#define LBER_SB_OPT_DRAIN               10
#define LBER_SB_OPT_NEEDS_READ          11
#define LBER_SB_OPT_NEEDS_WRITE         12
#define LBER_SB_OPT_GET_MAX_INCOMING    13
#define LBER_SB_OPT_SET_MAX_INCOMING    14
/* Largest option used by the library */
#define LBER_SB_OPT_OPT_MAX             14

/* LBER IO operations stacking levels */
#define LBER_SBIOD_LEVEL_PROVIDER       10
#define LBER_SBIOD_LEVEL_TRANSPORT      20
#define LBER_SBIOD_LEVEL_APPLICATION    30

#ifdef NOT_USED_BY_NETAPP
#ifndef _IFP
#define _IFP
typedef int (LDAP_C LDAP_CALLBACK *IFP)();
#endif
#endif

extern struct lber_memalloc_fns nslberi_memalloc_fns;

typedef struct seqorset {
	unsigned long	sos_clen;
	unsigned long	sos_tag;
	char		*sos_first;
	char		*sos_ptr;
	struct seqorset	*sos_next;
} Seqorset;
#define NULLSEQORSET	((Seqorset *) 0)

#define SOS_STACK_SIZE 8 /* depth of the pre-allocated sos structure stack */

struct berelement {
	char		*ber_buf;
	char		*ber_ptr;
	char		*ber_end;
	struct seqorset	*ber_sos;
	unsigned long	ber_tag;
	unsigned long	ber_len;
	int		ber_usertag;
	char		ber_options;
	char		*ber_rwptr;
	BERTranslateProc ber_encode_translate_proc;
	BERTranslateProc ber_decode_translate_proc;
	int		ber_flags;
#define LBER_FLAG_NO_FREE_BUFFER	1	/* don't free ber_buf */
	int		ber_sos_stack_posn;
	Seqorset	ber_sos_stack[SOS_STACK_SIZE];
};
#define NULLBER	((BerElement *)NULL)

#define SOCKBUF_VALID( sb ) ((sb)->sb_valid == LBER_VALID_SOCKBUF)

typedef struct sockbuf_buf {
	int	buf_size;
	int	buf_ptr;
	int	buf_end;
	char	*buf_base;
} Sockbuf_Buf;

typedef struct sockbuf_io Sockbuf_IO;

/* Structure for LBER IO operarion descriptor */
typedef struct sockbuf_io_desc {
	int			sbiod_level;
	Sockbuf			*sbiod_sb;
	Sockbuf_IO		*sbiod_io;
	void 			*sbiod_pvt;
	struct sockbuf_io_desc	*sbiod_next;
} Sockbuf_IO_Desc;

/* Structure for LBER IO operation functions */
struct sockbuf_io {
	int (*sbi_setup)( Sockbuf_IO_Desc *sbiod, void *arg );
	int (*sbi_remove)( Sockbuf_IO_Desc *sbiod );
	int (*sbi_ctrl)( Sockbuf_IO_Desc *sbiod, int opt, void *arg);

	int (*sbi_read)( Sockbuf_IO_Desc *sbiod, void *buf,
		int len );
	int (*sbi_write)( Sockbuf_IO_Desc *sbiod, void *buf,
		int len );

	int (*sbi_close)( Sockbuf_IO_Desc *sbiod );
};

/* Helper macros for LBER IO functions */
#define LBER_SBIOD_READ_NEXT( sbiod, buf, len ) \
	( (sbiod)->sbiod_next->sbiod_io->sbi_read( (sbiod)->sbiod_next, \
		buf, len ) )
#define LBER_SBIOD_WRITE_NEXT( sbiod, buf, len ) \
	( (sbiod)->sbiod_next->sbiod_io->sbi_write( (sbiod)->sbiod_next, \
		buf, len ) )
#define LBER_SBIOD_CTRL_NEXT( sbiod, opt, arg ) \
	( (sbiod)->sbiod_next ? \
		( (sbiod)->sbiod_next->sbiod_io->sbi_ctrl( \
		(sbiod)->sbiod_next, opt, arg ) ) : 0 )

struct ldap_sockbuf {
	LBER_SOCKET	sb_sd;
	BerElement	sb_ber;
	int		sb_naddr;	/* > 0 implies using CLDAP (UDP) */
	void		*sb_useaddr;	/* pointer to sockaddr to use next */
	void		*sb_fromaddr;	/* pointer to message source sockaddr */
	void		**sb_addrs;	/* actually an array of pointers to
					   sockaddrs */

	int		sb_options;	/* to support copying ber elements */
	LBER_SOCKET	sb_fd;
	long		sb_max_incoming;
	Sockbuf_IO_Desc         *sb_iod;/* I/O functions */
	short		sb_valid;
	
#ifdef NOT_USED_BY_NETAPP
	LDAP_IOF_READ_CALLBACK *sb_read_fn;
	LDAP_IOF_WRITE_CALLBACK *sb_write_fn;
#endif
};
#define NULLSOCKBUF	((Sockbuf *)NULL)
#define READBUFSIZ	8192


/*
 * macros used to check validity of data structures and parameters
 */
#define NSLBERI_VALID_BERELEMENT_POINTER( ber ) \
	( (ber) != NULLBER )

#define NSLBERI_VALID_SOCKBUF_POINTER( sb ) \
	( (sb) != NULLSOCKBUF )


#if defined(_WIN32) && defined(_ALPHA)

#define LBER_HTONL( _l ) \
	    ((((_l)&0xff)<<24) + (((_l)&0xff00)<<8) + \
	     (((_l)&0xff0000)>>8) + (((_l)&0xff000000)>>24))
#define LBER_NTOHL(_l) LBER_HTONL(_l)

#elif defined(ARCH_IS_64_BIT)
/*
 * htonl and ntohl only swap the
 * lower-order 32-bits of a (64-bit) long, 
 * so we define correct versions here.
 */
#define LBER_HTONL( l )	(((long)htonl( (l) & 0x00000000FFFFFFFF )) << 32 \
    			| htonl( ( (l) & 0xFFFFFFFF00000000 ) >> 32 ))

#define LBER_NTOHL( l )	(((long)ntohl( (l) & 0x00000000FFFFFFFF )) << 32 \
    			| ntohl( ( (l) & 0xFFFFFFFF00000000 ) >> 32 ))
#else

#define LBER_HTONL( l )	htonl( l )
#define LBER_NTOHL( l )	ntohl( l )

#endif


#ifdef NOT_USED_BY_NETAPP
/* function prototypes */
#ifdef LDAP_DEBUG
void ber_dump( BerElement *ber, int inout );
void lber_bprint( char *data, int len );
#endif
void ber_err_print( char *data );
#else
#define ber_err_print(X)
#define lber_bprint(X,Y)
#define ber_dump(X,Y)
#endif

void *nslberi_malloc( size_t size );
void *nslberi_calloc( size_t nelem, size_t elsize );
void *nslberi_realloc( void *ptr, size_t size );
void nslberi_free( void *ptr );


/* blame: dboreham 
 * slapd spends much of its time doing memcpy's for the ber code.
 * Most of these are single-byte, so we special-case those and speed
 * things up considerably. 
 */

#ifdef sunos4
#define THEMEMCPY( d, s, n )	bcopy( s, d, n )
#else /* sunos4 */
#define THEMEMCPY( d, s, n )	memmove( d, s, n )
#endif /* sunos4 */

#ifdef SAFEMEMCPY
#undef SAFEMEMCPY
#define SAFEMEMCPY(d,s,n) if (1 == n) *((char*)d) = *((char*)s); else THEMEMCPY(d,s,n);
#endif

/*
 * Memory allocation done in liblber should all go through one of the
 * following macros. This is so we can plug-in alternative memory
 * allocators, etc. as the need arises.
 */
#define NSLBERI_MALLOC( size )	        sk_kmem_alloc( size, M_LDAP, SKMA_SLEEP )
#define NSLBERI_CALLOC( nelem, elsize )	sk_kmem_zalloc( (nelem) * (elsize), M_LDAP, SKMA_SLEEP )
#define NSLBERI_REALLOC( ptr, size )	sk_kmem_realloc( ptr, size, M_LDAP, SKMA_SLEEP )
#define NSLBERI_FREE( ptr )		sk_kmem_free( ptr )

#ifdef __cplusplus
}
#endif

int ber_sockbuf_add_io (
	Sockbuf *sb,
	Sockbuf_IO *sbio,
	int layer,
	void *arg );

int
ber_sockbuf_remove_io (
	Sockbuf *sb,
	Sockbuf_IO *sbio,
	int layer );

int
ber_sockbuf_ctrl (
	Sockbuf *sb,
	int opt,
	void *arg );

int ber_pvt_sb_do_write ( Sockbuf_IO_Desc *sbiod,
	Sockbuf_Buf *buf_out );

void
ber_pvt_sb_buf_init ( Sockbuf_Buf *buf );

void
ber_pvt_sb_buf_destroy ( Sockbuf_Buf *buf );

int
ber_pvt_sb_grow_buffer ( Sockbuf_Buf *buf,
	int minsize );

int
ber_pvt_sb_copy_out ( Sockbuf_Buf *sbb,
	char *buf, int len );

#ifdef NOT_USED_BY_NETAPP
int
ber_pvt_socket_set_nonblock ( ber_socket_t sd, int nb );
#endif

int
ber_int_sb_init ( Sockbuf *sb );

int
ber_int_sb_close ( Sockbuf *sb );

int
ber_int_sb_destroy ( Sockbuf *sb );

int
ber_int_sb_read ( Sockbuf *sb, void *buf, int len );

int
ber_int_sb_write ( Sockbuf *sb, void *buf, int len );

#endif /* _LBERINT_H */
