/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The following files were originally developed by Netscape Communications
 * Corporation
 */
/*
 * The contents of this file are subject to the Netscape Public License
 * Version 1.0 (the "NPL"); you may not use this file except in
 * compliance with the NPL.  You may obtain a copy of the NPL at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the NPL is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the NPL
 * for the specific language governing rights and limitations under the
 * NPL.
 *
 * The Initial Developer of this code under the NPL is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation.  All Rights
 * Reserved.
 */
/*
 *  Copyright (c) 1995 Regents of the University of Michigan.
 *  All rights reserved.
 */
/*
 * The contents of these files are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use these files except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is os-ip.c
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Copyright (C) 1998. All Rights Reserved
 */
/*
 * Portions created by NetApp, Inc. are Copyright (c) 1998.  All Rights
 * Reserved.
 */
/*
 *  os-ip.c -- platform-specific TCP & UDP related code
 */

#include <sys/types.h>
#include <sk/sk_timeout.h>

#include <net7g/sys/sk_socket.h>
#include <sys/errno.h>
#include <sk/proto.h>
#include <fio/fio.h>
#include <legacy/string.h>
#include <net7g/netinet/sk_in.h>
#include <net7g/netinet/sk_ssl_var.h>
#include <libnet/ext.h>
#include <legacy/stdlib.h>
#include <net7g/sk_netdb.h>
#include <sys/filio.h>
#include <net7g/sys/sk_socketvar.h>
#include <smb/smbutils.h>
#include <sk/sk_errno.h>

static int  ontap_connect( int s, struct sockaddr *a, size_t l );
static int  ontap_open_socket( int af, int t, int p , char *rtag);
static void ontap_close_socket( int s );

static int
connect_timeout(struct socket *so, struct sockaddr *a, size_t l, int timeout_msecs)
{
    sk_domain_t old = sk_domain_switch_begin(sk_domain_network);
    sk_Pid my_pid = sk_my_pid();
    int ret = 0;

    /*
     * Make sure we  end up with a non-zero value for timeout.
     */
    ASSERT(timeout_msecs >= SK_MSEC_PER_TIC);

    sk_timeout(sk_post_fake_event, my_pid, timeout_msecs/SK_MSEC_PER_TIC);
    if (connect(so, a, l) != 0) {
        ret = -1;
    }
    sk_untimeout(sk_post_fake_event, my_pid);


    sk_domain_switch_end(old);

    return ret;
}

static int
ontap_connect( int s, struct sockaddr *a, size_t l )
{
    struct socket *so;

    so = fio_getsocket(s);

    if ( !so ) {
        return -1;
    }

    return connect_timeout( so, a, l, smbflags.smb_ldap_timeout * MSEC_PER_SEC);
}

static int
ontap_open_socket( int af, int t, int p, char *rtag )
{
    struct socket *so;
    int fd;

    so = socket(af, t, p, rtag);

    if (!so)
	return -1;
        
    /* add to fio fdtable with mode 2, read + write */
    fd = fio_addsocket(so, 2);

    if (fd == -1)
	soclose(so);

    return fd;
}

static void
ontap_close_socket( int s )
{
    struct socket *so;

    so = fio_getsocket( s );

    if ( so ) {
        soclose( so );
    }

    fio_close( s );
    return;
}

#define socket(W, X,Y,Z)  ontap_open_socket(W,X,Y,Z)
#define close(X)       ontap_close_socket(X)
#define connect(X,Y,Z) ontap_connect(X,Y,Z)

#if 0
#ifndef lint 
static char copyright[] = "@(#) Copyright (c) 1995 Regents of the University of Michigan.\nAll rights reserved.\n";
#endif
#endif

#include "ldap-int.h"

#define CLOSESOCKET(_s) close((_s))

#ifdef NOT_USED_BY_NETAPP
#ifdef LDAP_CONNECT_MUST_NOT_BE_INTERRUPTED
#include <signal.h>
#endif

#ifdef NSLDAPI_HAVE_POLL
#include <poll.h>
#endif

#ifdef _WINDOWS
#define CLOSESOCKET(_s) closesocket((_s))
#else
#define CLOSESOCKET(_s) close((_s))
#endif

struct selectinfo {
	fd_set		si_readfds;
	fd_set		si_writefds;
	fd_set		si_use_readfds;
	fd_set		si_use_writefds;
#ifdef NSLDAPI_HAVE_POLL
	struct pollfd	*si_pollfds;
	int		si_pollfds_size;
#endif
};

#ifdef NSLDAPI_HAVE_POLL
static int add_to_pollfds( int fd, struct selectinfo *sip, short events );
static int clear_from_pollfds( int fd, struct selectinfo *sip,
    short events );
static int find_in_pollfds( int fd, struct selectinfo *sip, short revents );
#endif

#ifdef irix
/*
 * XXXmcs: on IRIX NSPR's poll() and select() wrappers will crash if NSPR
 * has not been initialized.  We work around the problem by bypassing
 * the NSPR wrapper functions and going directly to the OS' functions.
 */
#define NSLDAPI_POLL		_poll
#define NSLDAPI_SELECT		_select
extern int _poll(struct pollfd *fds, unsigned long nfds, int timeout);
extern int _select(int nfds, fd_set *readfds, fd_set *writefds,
        fd_set *exceptfds, struct sec_micro *timeout);
#else
#define NSLDAPI_POLL		poll
#define NSLDAPI_SELECT		select
#endif

#endif


int
nsldapi_connect_to_host( LDAP *ld, Sockbuf *sb, char *host,
	struct sockaddr_storage *address, int port, int async, int secure )
/*
 * if host == NULL, connect using address
 * "address" and "port" must be in network byte order
 * zero is returned upon success, -1 if fatal error, -2 EINPROGRESS
 * if -1 is returned, ld_errno is set
 * non-zero async means don't wait for connect
 */
{
	int			rc, i, s, connected, use_hp, naddrs, ret;
	struct sockaddr_storage	*pldap_SASaddrs, *ptmpldap_SASaddrs, SASaddr;
#ifdef NOT_USED_BY_NETAPP
	char			**addrlist, *ldhpbuf_allocd;
	char *ldhpbuf;
	LDAPHostEnt		ldhent;
	LDAPHostEnt		*ldhp;
	struct hostent		*hp;
#ifdef GETHOSTBYNAME_BUF_T
	GETHOSTBYNAME_BUF_T	hbuf;
	struct hostent		hent;
#endif /* GETHOSTBYNAME_BUF_T */
#endif /* NOT_USED_BY_NETAPP */
	int			err;
#ifdef LDAP_ASYNC_IO
	int			status;	/* for ioctl call */
#endif /* LDAP_ASYNC_IO */

	LDAPDebug( LDAP_DEBUG_TRACE, "nsldapi_connect_to_host: %s:%d\n",
	    ( host == NULL ) ? "(by address)" : host,
	    ntohs( (unsigned short)port ), 0 );

#ifdef NOT_USED_BY_NETAPP
	if ( secure && ld->ld_ssl_enable_fn == NULL ) {
		LDAP_SET_LDERRNO( ld, LDAP_LOCAL_ERROR, NULL, NULL );
		return( -1 );
	}

	ldhpbuf_allocd = NULL;
	ldhp = NULL;
	hp = NULL;
	addrlist = NULL;
#endif /* NOT_USED_BY_NETAPP */
	s = 0;
	connected = use_hp = 0;
	ptmpldap_SASaddrs = pldap_SASaddrs = NULL;

	if (host == NULL && address == NULL) {
		ASSERT(0);
		return(-1);
	}

	if ( host != NULL ) {
		if (inet_ptosa(host, &SASaddr) != 1) {
#ifdef NOT_USED_BY_NETAPP
		if ( ld->ld_dns_gethostbyname_fn == NULL ) {
#endif /* NOT_USED_BY_NETAPP */
			C_REFERENCE_SWITCH_DOMAIN(smb_getaddrinfo_t);
			if (( ret = smb_getaddrinfo_t(host, AF_UNSPEC, &pldap_SASaddrs, &naddrs, SMB_DNS_TIMEOUT))) {
				LDAP_SET_LDERRNO( ld, LDAP_CONNECT_ERROR, NULL, NULL );
				LDAP_SET_ERRNO( ld, EHOSTUNREACH );  /* close enough */
				return( -1 );
			}
#ifdef NOT_USED_BY_NETAPP
		} else {
			/*
			 * DNS callback installed... use it.
			 */
#ifdef GETHOSTBYNAME_buf_t
			/* avoid allocation by using hbuf if large enough */
			if ( sizeof( hbuf ) < ld->ld_dns_bufsize ) {
				ldhpbuf = ldhpbuf_allocd
					= NSLDAPI_MALLOC( ld->ld_dns_bufsize );
			} else {
				ldhpbuf = (char *)hbuf;
			}
#else /* GETHOSTBYNAME_buf_t */
			ldhpbuf = ldhpbuf_allocd = NSLDAPI_MALLOC(
					ld->ld_dns_bufsize );
#endif /* GETHOSTBYNAME_buf_t */

			if ( ldhpbuf == NULL ) {
				LDAP_SET_LDERRNO( ld, LDAP_NO_MEMORY, NULL,
						NULL );
				return( -1 );
			}

			if (( ldhp = ld->ld_dns_gethostbyname_fn( host,
							&ldhent, ldhpbuf, ld->ld_dns_bufsize, &err,
							ld->ld_dns_extradata )) != NULL ) {
				addrlist = ldhp->ldaphe_addr_list;
			}
		}
#endif /* NOT_USED_BY_NETAPP */
		use_hp = 1;
		} else {
			address = &SASaddr;
		}
	}

	rc = -1;
	for ( i = 0; !use_hp || ( naddrs != 0 ); i++, naddrs-- ) {
#ifdef NOT_USED_BY_NETAPP
		if ( ld->ld_socket_fn == NULL ) {
#endif /* NOT_USED_BY_NETAPP */
		if (use_hp) {
			ptmpldap_SASaddrs = &pldap_SASaddrs[i];	
		} else {
			ptmpldap_SASaddrs = address;
		}
		/* Use ONTAP native SSL socket for secure LDAP */
		if ( !secure || (ld->ld_ssl_context == NULL)) {
			s = socket( SA_FAM(ptmpldap_SASaddrs), SOCK_STREAM, 0, 0 );
		} else {
			s = socket( SA_FAM(ptmpldap_SASaddrs), SOCK_STREAM, IPPROTO_SSL, 0 );
		}

#ifdef NOT_USED_BY_NETAPP
		} else {
			s = ld->ld_socket_fn( AF_INET, SOCK_STREAM, 0 );
		}
#endif /* NOT_USED_BY_NETAPP */

		/*
		 * if the socket() call failed or it returned a socket larger
		 * than we can deal with, return a "local error."
		 */
#ifdef NOT_USED_BY_NETAPP
#ifdef _WINDOWS
		if ( s < 0 ) {
#elif NSLDAPI_HAVE_POLL
		if ( s < 0 || ( ld->ld_select_fn != NULL && s >= FD_SETSIZE )) {
#else	/* not on Windows and do not have poll() */
		if ( s < 0 || s >= FD_SETSIZE ) {
#endif /* _WINDOWS */
#endif /* NOT_USED_BY_NETAPP */
		if ( s < 0 ) {
			char		*errmsg;

			if ( s < 0 ) {
				errmsg = "unable to create a socket";
			} else {
				errmsg = "can't use socket >= FD_SETSIZE";
#ifdef NOT_USED_BY_NETAPP
				if ( ld->ld_close_fn == NULL ) {
#endif /* NOT_USED_BY_NETAPP */
					CLOSESOCKET( s );
#ifdef NOT_USED_BY_NETAPP
				} else {
					ld->ld_close_fn( s );
				}
#endif /* NOT_USED_BY_NETAPP */
			}
			errmsg = nsldapi_strdup( errmsg );
			LDAP_SET_LDERRNO( ld, LDAP_LOCAL_ERROR, NULL, errmsg );
			SAFE_FREE(pldap_SASaddrs);
			return( -1 );
		}

#ifdef LDAP_ASYNC_IO
		status = 1;
		if ( async ) {
			if ( ld->ld_ioctl_fn == NULL ) {
				err = ioctl( s, FIONBIO, (caddr_t)&status );
			} else {
				err = ld->ld_ioctl_fn( s, FIONBIO,
				    (caddr_t)&status );
			}
			if ( err == -1 ) {
				LDAPDebug( LDAP_DEBUG_ANY,
				    "FIONBIO ioctl failed on %d\n", s, 0, 0 );
			}
		}
#endif /* LDAP_ASYNC_IO */

		C_REFERENCE_MPSAFE(smb_SAS_SetPort);
		smb_SAS_SetPort(ptmpldap_SASaddrs, port);

        	/* Setup SSL context for ONTAP native SSL socket */
        	if ( secure && (ld->ld_ssl_context != NULL) ) {
            		err = setsockopt(fio_getsocket(s), IPPROTO_SSL, 
					 SSLOPT_CONTEXT,
                             		 ld->ld_ssl_context, sizeof(char*));
            		if ( err != 0 ) {
                		char		*errmsg;

				errmsg = "Unable to setup ssl context";
				errmsg = nsldapi_strdup( errmsg );
				LDAP_SET_LDERRNO( ld, LDAP_LOCAL_ERROR, 
						  NULL, errmsg );
				SAFE_FREE(pldap_SASaddrs);
				return( -1 );
			}
		}

#ifdef NOT_USED_BY_NETAPP
		if ( secure && ld->ld_ssl_enable_fn( s ) < 0 ) {
			if ( ld->ld_close_fn == NULL ) {
				CLOSESOCKET( s );
			} else {
				ld->ld_close_fn( s );
			}
			LDAP_SET_LDERRNO( ld, LDAP_LOCAL_ERROR, NULL, NULL );
			if ( ldhpbuf_allocd != NULL ) {
				NSLDAPI_FREE( ldhpbuf_allocd );
			}
			return( -1 );
		}
#endif /* NOT_USED_BY_NETAPP */

#ifdef NOT_USED_BY_NETAPP
		if ( ld->ld_connect_fn == NULL ) {
#ifdef LDAP_CONNECT_MUST_NOT_BE_INTERRUPTED
/*
 * Block all of the signals that might interrupt connect() since there
 * is an OS bug that causes connect() to fail if it is restarted.  Look in
 * ns/netsite/ldap/include/portable.h for the definition of
 * LDAP_CONNECT_MUST_NOT_BE_INTERRUPTED
 */
			sigset_t	ints_off, oldset;

			sigemptyset( &ints_off );
			sigaddset( &ints_off, SIGALRM );
			sigaddset( &ints_off, SIGIO );
			sigaddset( &ints_off, SIGCLD );

			sigprocmask( SIG_BLOCK, &ints_off, &oldset );
#endif /* LDAP_CONNECT_MUST_NOT_BE_INTERRUPTED */
#endif /* NOT_USED_BY_NETAPP */
			err = connect( s, (struct sockaddr *)ptmpldap_SASaddrs,
						ptmpldap_SASaddrs->ss_len);
#ifdef NOT_USED_BY_NETAPP
#ifdef LDAP_CONNECT_MUST_NOT_BE_INTERRUPTED
/*
 * restore original signal mask
 */
			sigprocmask( SIG_SETMASK, &oldset, 0 );
#endif /* LDAP_CONNECT_MUST_NOT_BE_INTERRUPTED */
		} else {
		        if (ld->ld_options & LDAP_BITOPT_ASYNC)
			{
			    err = 0;
			}
			else {
			    err = ld->ld_connect_fn( s,
				(struct sockaddr *)&sin,
				sizeof( struct sockaddr_in ));
			}
			
		}
#endif /*NOT_USED_BY_NETAPP */
		/*
		 * The ONTAP connect() routine is not standard in that it
		 * returns positive error values on failure. 
		 */
		if ( err == 0 ) {
			connected = 1;
			rc = 0;
			break;
		} else {
#ifdef LDAP_ASYNC_IO
			err = LDAP_GET_ERRNO( ld );
			if ( NSLDAPI_ERRNO_IO_INPROGRESS( err )) {
				LDAPDebug( LDAP_DEBUG_TRACE,
					"connect would block...\n", 0, 0, 0 );
				rc = -2;
				break;
			}
#endif /* LDAP_ASYNC_IO */

#ifdef LDAP_DEBUG
			if ( ldap_debug & LDAP_DEBUG_TRACE ) {
				SASIPString addrStr;
				SAS_IPString(ptmpldap_SASaddrs, addrStr);
				perror( addrStr);
			}
#endif /* LDAP_DEBUG */
#ifdef NOT_USED_BY_NETAPP
			if ( ld->ld_close_fn == NULL ) {
#endif /* NOT_USED_BY_NETAPP */
				CLOSESOCKET( s );
#ifdef NOT_USED_BY_NETAPP
			    
			} else {
			    ld->ld_close_fn( s );
			}
#endif /* NOT_USED_BY_NETAPP */
			if ( !use_hp ) {
				break;
			}
		}
	}

#ifdef NOT_USED_BY_NETAPP
	if ( ldhpbuf_allocd != NULL ) {
		NSLDAPI_FREE( ldhpbuf_allocd );
	}
#endif /* NOT_USED_BY_NETAPP */
	sb->sb_sd = s;

	if ( connected ) {
		SASIPString addrStr;
		SAS_IPString(ptmpldap_SASaddrs, addrStr);
		LDAPDebug(LDAP_DEBUG_TRACE, "sd %d connected to: %s\n", s, addrStr, 0 );
	}

	if ( rc == -1 ) {
		LDAP_SET_LDERRNO( ld, LDAP_CONNECT_ERROR, NULL, NULL );
	}

	SAFE_FREE(pldap_SASaddrs);

	return( rc );
}


void
nsldapi_close_connection( LDAP *ld, Sockbuf *sb )
{
#ifdef NOT_USED_BY_NETAPP
    if ( ld->ld_close_fn == NULL ) {
#endif /* NOT_USED_BY_NETAPP */
	    CLOSESOCKET( sb->sb_sd );
#ifdef NOT_USED_BY_NETAPP
    } else {
	    ld->ld_close_fn( sb->sb_sd );
	
    }
#endif /* NOT_USED_BY_NETAPP */
}

char *nsldapi_host_connected_to( Sockbuf *sb );

#ifdef NOT_USED_BY_NETAPP
#ifdef KERBEROS
char *
nsldapi_host_connected_to( Sockbuf *sb )
{
	struct hostent		*hp, hbuf;
	int			len, err = 0;
	struct sockaddr_in	sin;
	struct socket		*so;
	char			buf[MAX_LINELENGTH_HOSTS+1];

	(void)memset( (char *)&sin, 0, sizeof( struct sockaddr_in ));
	if ( (sb == NULL) || ((so = fio_getsocket(sb->sb_sd)) == NULL) ) {
		return( NULL );
	}

	len = sizeof( sin );
	if ( getpeername( so, (struct sockaddr *)&sin, &len ) == -1 ) {
		return( NULL );
	}

	/*
	 * do a reverse lookup on the addr to get the official hostname.
	 * this is necessary for kerberos to work right, since the official
	 * hostname is used as the kerberos instance.
	 */
/* XXXmcs: need to use DNS callbacks here XXX */

	if (( hp = gethostbyaddr_r( (char *) &sin.sin_addr,
	    sizeof( sin.sin_addr ), AF_INET, &hbuf, buf,
	    sizeof(buf), &err )) != NULL ) {
		if ( hp->h_name != NULL ) {
			return( nsldapi_strdup( hp->h_name ));
		}
	}

	return( NULL );
}
#endif /* KERBEROS */
#endif /* NOT_USED_BY_NETAPP */

#ifdef KERBEROS
char *
nsldapi_host_connected_to( Sockbuf *sb )
{
	int			len, err = 0;
	struct sockaddr_storage	addr;
	struct socket		*so;
	char			buf[MAX_LINELENGTH_HOSTS+1];
	int flag = 0;

	(void)memset( (char *)&addr, 0, sizeof( struct sockaddr_storage));
	if ( (sb == NULL) || ((so = fio_getsocket(sb->sb_sd)) == NULL) ) {
		return( NULL );
	}

	len = sizeof( addr );
	if ( getpeername( so, (struct sockaddr *)&addr, &len ) == -1 ) {
		return( NULL );
	}

	/*
	 * do a reverse lookup on the addr to get the official hostname.
	 * this is necessary for kerberos to work right, since the official
	 * hostname is used as the kerberos instance.
	 */
	/* If the flag bit NI_NAMEREQD is set, an error shall be returned if the
	 * host's name cannot be located.
	 */

	flag |= NI_NAMEREQD;
	if ((getnameinfo((struct sockaddr *)&addr,
			len, buf, sizeof(buf), NULL, 0, flag)) == 0) {
		return(nsldapi_strdup(buf));
	}

	return( NULL );
}
#endif /* KERBEROS */

#ifdef NOT_USED_BY_NETAPP

void
nsldapi_mark_select_write( LDAP *ld, Sockbuf *sb )
{
	struct selectinfo	*sip;

	LDAP_MUTEX_LOCK( ld, LDAP_SELECT_LOCK );
	sip = (struct selectinfo *)ld->ld_selectinfo;

#ifdef NSLDAPI_HAVE_POLL
	if ( ld->ld_select_fn == NULL ) {
		if ( add_to_pollfds( sb->sb_sd, sip, POLLOUT )) {
			++ld->ld_selectwritecnt;
		}
		LDAP_MUTEX_UNLOCK( ld, LDAP_SELECT_LOCK );
		return;
	}
#endif

	if ( !FD_ISSET( sb->sb_sd, &sip->si_writefds )) {
		FD_SET( sb->sb_sd, &sip->si_writefds );
		++ld->ld_selectwritecnt;
	}
	LDAP_MUTEX_UNLOCK( ld, LDAP_SELECT_LOCK );
}


void
nsldapi_mark_select_read( LDAP *ld, Sockbuf *sb )
{
	struct selectinfo	*sip;

	LDAP_MUTEX_LOCK( ld, LDAP_SELECT_LOCK );
	sip = (struct selectinfo *)ld->ld_selectinfo;

#ifdef NSLDAPI_HAVE_POLL
	if ( ld->ld_select_fn == NULL ) {
		if ( add_to_pollfds( sb->sb_sd, sip, POLLIN )) {
			++ld->ld_selectreadcnt;
		}
		LDAP_MUTEX_UNLOCK( ld, LDAP_SELECT_LOCK );
		return;
	}
#endif

	if ( !FD_ISSET( sb->sb_sd, &sip->si_readfds )) {
		FD_SET( sb->sb_sd, &sip->si_readfds );
		++ld->ld_selectreadcnt;
	}
	LDAP_MUTEX_UNLOCK( ld, LDAP_SELECT_LOCK );
}


void
nsldapi_mark_select_clear( LDAP *ld, Sockbuf *sb )
{
	struct selectinfo	*sip;

	LDAP_MUTEX_LOCK( ld, LDAP_SELECT_LOCK );
	sip = (struct selectinfo *)ld->ld_selectinfo;

#ifdef NSLDAPI_HAVE_POLL
	if ( ld->ld_select_fn == NULL ) {
		if ( clear_from_pollfds( sb->sb_sd, sip, POLLIN )) {
			--ld->ld_selectreadcnt;
		}
		if ( clear_from_pollfds( sb->sb_sd, sip, POLLOUT )) {
			--ld->ld_selectwritecnt;
		}
		LDAP_MUTEX_UNLOCK( ld, LDAP_SELECT_LOCK );
		return;
	}
#endif

	if ( FD_ISSET( sb->sb_sd, &sip->si_writefds )) {
		FD_CLR( sb->sb_sd, &sip->si_writefds );
		--ld->ld_selectwritecnt;
	}
	if ( FD_ISSET( sb->sb_sd, &sip->si_readfds )) {
		FD_CLR( sb->sb_sd, &sip->si_readfds );
		--ld->ld_selectreadcnt;
	}
	LDAP_MUTEX_UNLOCK( ld, LDAP_SELECT_LOCK );
}


int
nsldapi_is_write_ready( LDAP *ld, Sockbuf *sb )
{
	struct selectinfo	*sip;

	LDAP_MUTEX_LOCK( ld, LDAP_SELECT_LOCK );
	sip = (struct selectinfo *)ld->ld_selectinfo;

#ifdef NSLDAPI_HAVE_POLL
	if ( ld->ld_select_fn == NULL ) {
		LDAP_MUTEX_UNLOCK( ld, LDAP_SELECT_LOCK );
		/*
		 * if we are using poll() we do something a little tricky: if
		 * any bits in the socket's returned events field other than
		 * POLLIN (ready for read) are set, we return true.  This
		 * is done so we notice when a server closes a connection
		 * or when another error occurs.  The actual error will be
		 * noticed later when we call write() or send().
		 */
		return( find_in_pollfds( sb->sb_sd, sip, ~POLLIN ));
	}
#endif

	LDAP_MUTEX_UNLOCK( ld, LDAP_SELECT_LOCK );
	return( FD_ISSET( sb->sb_sd, &sip->si_use_writefds ));
}


int
nsldapi_is_read_ready( LDAP *ld, Sockbuf *sb )
{
	struct selectinfo	*sip;

	LDAP_MUTEX_LOCK( ld, LDAP_SELECT_LOCK );
	sip = (struct selectinfo *)ld->ld_selectinfo;

#ifdef NSLDAPI_HAVE_POLL
	if ( ld->ld_select_fn == NULL ) {
		LDAP_MUTEX_UNLOCK( ld, LDAP_SELECT_LOCK );
		/*
		 * if we are using poll() we do something a little tricky: if
		 * any bits in the socket's returned events field other than
		 * POLLOUT (ready for write) are set, we return true.  This
		 * is done so we notice when a server closes a connection
		 * or when another error occurs.  The actual error will be
		 * noticed later when we call read() or recv().
		 */
		return( find_in_pollfds( sb->sb_sd, sip, ~POLLOUT ));
	}
#endif

	LDAP_MUTEX_UNLOCK( ld, LDAP_SELECT_LOCK );
	return( FD_ISSET( sb->sb_sd, &sip->si_use_readfds ));
}


void *
nsldapi_new_select_info(void)
{
	struct selectinfo	*sip;

	if (( sip = (struct selectinfo *)NSLDAPI_CALLOC( 1,
	    sizeof( struct selectinfo ))) != NULL ) {
		FD_ZERO( &sip->si_readfds );
		FD_ZERO( &sip->si_writefds );
	}

	return( (void *)sip );
}


void
nsldapi_free_select_info( void *vsip )
{
	struct selectinfo	*sip = (struct selectinfo *)vsip;

#ifdef NSLDAPI_HAVE_POLL
	if ( sip->si_pollfds != NULL ) {
		NSLDAPI_FREE( sip->si_pollfds );
	}
#endif
	NSLDAPI_FREE( sip );
}


int
nsldapi_do_ldap_select( LDAP *ld, struct sec_micro *timeout )
{
	struct selectinfo	*sip;
	static int		tblsize = 0;

	LDAPDebug( LDAP_DEBUG_TRACE, "nsldapi_do_ldap_select\n", 0, 0, 0 );

	if ( tblsize == 0 ) {
#if defined(_WINDOWS) || defined(XP_OS2)
		tblsize = FOPEN_MAX; /* ANSI spec. */
#else
#ifdef USE_SYSCONF
		tblsize = sysconf( _SC_OPEN_MAX );
#else /* USE_SYSCONF */
		tblsize = getdtablesize();
#endif /* USE_SYSCONF */
#endif /* _WINDOWS */

		if ( tblsize >= FD_SETSIZE ) {
			/*
			 * clamp value so we don't overrun the fd_set structure
			 */
			tblsize = FD_SETSIZE - 1;
		}
	}

	if ( ld->ld_selectreadcnt <= 0 && ld->ld_selectwritecnt <= 0 ) {
		return( 0 );	/* simulate a timeout */
	}

	sip = (struct selectinfo *)ld->ld_selectinfo;

#ifdef NSLDAPI_HAVE_POLL
	if ( ld->ld_select_fn == NULL ) {
		int		to;

		if ( timeout == NULL ) {
			to = -1;
		} else {
			to = timeout->tv_sec * 1000 + timeout->tv_usec / 1000;
		}
		return( NSLDAPI_POLL( sip->si_pollfds, sip->si_pollfds_size,
		    to ));
	}
#endif

	sip->si_use_readfds = sip->si_readfds;
	sip->si_use_writefds = sip->si_writefds;

	if ( ld->ld_select_fn != NULL ) {
		return( ld->ld_select_fn( tblsize, &sip->si_use_readfds,
		    &sip->si_use_writefds, NULL, timeout ));
	} else {
#ifdef HPUX9
		return( NSLDAPI_SELECT( tblsize, (int *)&sip->si_use_readfds,
		    (int *)&sip->si_use_writefds, NULL, timeout ));
#else
		return( NSLDAPI_SELECT( tblsize, &sip->si_use_readfds,
		    &sip->si_use_writefds, NULL, timeout ));
#endif
	}
}

#ifdef NSLDAPI_HAVE_POLL

/*
 * returns 1 if "fd" was added to pollfds.
 * returns 1 if some of the bits in "events" were added to pollfds.
 * returns 0 if no changes were made.
 */
static int
add_to_pollfds( int fd, struct selectinfo *sip, short events )
{
	int	i, openslot;

	/* first we check to see if "fd" is already in our pollfds */
	openslot = -1;
	for ( i = 0; i < sip->si_pollfds_size; ++i ) {
		if ( sip->si_pollfds[ i ].fd == fd ) {
			if (( sip->si_pollfds[ i ].events & events )
			    != events ) {
				sip->si_pollfds[ i ].events |= events;
				return( 1 );
			} else {
				return( 0 );
			}
		}
		if ( sip->si_pollfds[ i ].fd == -1 && openslot == -1 ) {
			openslot = i;	/* remember for later */
		}
	}

	/*
	 * "fd" is not currently being poll'd on -- add to array.
	 * if we need to expand the pollfds array, we do it in increments of 5.
	 */
	if ( openslot == -1 ) {
		struct pollfd	*newpollfds;

		if ( sip->si_pollfds_size == 0 ) {
			newpollfds = (struct pollfd *)NSLDAPI_MALLOC(
			    5 * sizeof( struct pollfd ));
		} else {
			newpollfds = (struct pollfd *)NSLDAPI_REALLOC(
			    sip->si_pollfds, (5 + sip->si_pollfds_size) *
			    sizeof( struct pollfd ));
		}
		if ( newpollfds == NULL ) { /* XXXmcs: no way to return err! */
			return( 0 );
		}
		sip->si_pollfds = newpollfds;
		openslot = sip->si_pollfds_size;
		sip->si_pollfds_size += 5;
		for ( i = openslot + 1; i < sip->si_pollfds_size; ++i ) {
			sip->si_pollfds[ i ].fd = -1;
			sip->si_pollfds[ i ].events =
			    sip->si_pollfds[ i ].revents = 0;
		}
	}
	sip->si_pollfds[ openslot ].fd = fd;
	sip->si_pollfds[ openslot ].events = events;
	sip->si_pollfds[ openslot ].revents = 0;
	return( 1 );
}


/*
 * returns 1 if any "events" from "fd" were removed from pollfds
 * returns 0 of "fd" wasn't in pollfds or if events did not overlap.
 */
static int
clear_from_pollfds( int fd, struct selectinfo *sip,
    short events )
{
	int	i;

	for ( i = 0; i < sip->si_pollfds_size; ++i ) {
		if ( sip->si_pollfds[i].fd == fd ) {
			if (( sip->si_pollfds[ i ].events & events ) != 0 ) {
				sip->si_pollfds[ i ].events &= ~events;
				if ( sip->si_pollfds[ i ].events == 0 ) {
					sip->si_pollfds[i].fd = -1;
				}
				return( 1 );	/* events overlap */
			} else {
				return( 0 );	/* events do not overlap */
			}
		}
	}

	return( 0 );	/* "fd" was not found */
}


/*
 * returns 1 if any "revents" from "fd" were set in pollfds revents field.
 * returns 0 if not.
 */
static int
find_in_pollfds( int fd, struct selectinfo *sip, short revents )
{
	int	i;

	for ( i = 0; i < sip->si_pollfds_size; ++i ) {
		if ( sip->si_pollfds[i].fd == fd ) {
			if (( sip->si_pollfds[ i ].revents & revents ) != 0 ) {
				return( 1 );	/* revents overlap */
			} else {
				return( 0 );	/* revents do not overlap */
			}
		}
	}

	return( 0 );	/* "fd" was not found */
}
#endif /* NSLDAPI_HAVE_POLL */

/******************************************************
 *  ntstubs.c - Stubs needed on NT when linking in
 *  the SSL code. If these stubs were not here, the 
 *  named functions below would not be located at link
 *  time, because there is no implementation of the 
 *  functions for Win32 in cross-platform libraries.
 *
 ******************************************************/
#if defined( _WIN32 ) && defined ( NET_SSL )
/*#include <xp_file.h>*/
typedef enum {xpAddrBook} XP_FileType ;
typedef FILE          * XP_File;
typedef char          * XP_FilePerm;
XP_File XP_FileOpen(const char* name, XP_FileType type, 
		    const XP_FilePerm permissions)
{
    return NULL;
}

char *
WH_FileName (const char *name, XP_FileType type)
{
	return NULL;
}
#endif /* WIN32 && NET_SSL */

#endif
