/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The following files were originally developed by Netscape Communications
 * Corporation
 */
/*
 * The contents of this file are subject to the Netscape Public License
 * Version 1.0 (the "NPL"); you may not use this file except in
 * compliance with the NPL.  You may obtain a copy of the NPL at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the NPL is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the NPL
 * for the specific language governing rights and limitations under the
 * NPL.
 *
 * The Initial Developer of this code under the NPL is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation.  All Rights
 * Reserved.
 */
/*
 *  Copyright (c) 1993 Regents of the University of Michigan.
 *  All rights reserved.
 */
/*
 * The contents of these files are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use these files except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is sbind.c
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Copyright (C) 1998. All Rights Reserved
 */
/*
 * Portions created by NetApp, Inc. are Copyright (c) 1998.  All Rights
 * Reserved.
 */
/*
 *  sbind.c
 */

#if 0
#ifndef lint 
static char copyright[] = "@(#) Copyright (c) 1993 Regents of the University of Michigan.\nAll rights reserved.\n";
#endif
#endif

#include "ldap-int.h"

static int simple_bind_nolock( LDAP *ld, const char *dn, const char *passwd,
	int unlock_permitted );
static int simple_bindifnot_s( LDAP *ld, const char *dn, const char *passwd );
static int simple_bindifnot_st( LDAP *ld, const char *dn, const char *passwd,
	struct sec_micro *timeout );

/*
 * ldap_simple_bind - bind to the ldap server (and X.500).  The dn and
 * password of the entry to which to bind are supplied.  The message id
 * of the request initiated is returned.
 *
 * Example:
 *	ldap_simple_bind( ld, "cn=manager, o=university of michigan, c=us",
 *	    "secret" )
 */
SK_DOMAIN_EXPORT(ldap_simple_bind,SK_REASON_SWITCH_DOMAIN);
int
LDAP_CALL
ldap_simple_bind( LDAP *ld, const char *dn, const char *passwd )
{
	int	rc;
	sk_domain_t	calling_domain;

	LDAPDebug( LDAP_DEBUG_TRACE, "ldap_simple_bind\n", 0, 0, 0 );

	if ( !NSLDAPI_VALID_LDAP_POINTER( ld )) {
		return( -1 );
	}

	calling_domain=sk_domain_switch_begin(sk_domain_kahuna);
	rc = simple_bind_nolock( ld, dn, passwd, 1 );
	sk_domain_switch_end(calling_domain);

	return( rc );
}


static int
simple_bind_nolock( LDAP *ld, const char *dn, const char *passwd,
    int unlock_permitted )
{
	BerElement	*ber;
	int		rc, msgid;

	/*
	 * The bind request looks like this:
	 *	BindRequest ::= SEQUENCE {
	 *		version		INTEGER,
	 *		name		DistinguishedName,	 -- who
	 *		authentication	CHOICE {
	 *			simple		[0] OCTET STRING -- passwd
	 *		}
	 *	}
	 * all wrapped up in an LDAPMessage sequence.
	 */

	LDAP_MUTEX_LOCK( ld, LDAP_MSGID_LOCK );
	msgid = ++ld->ld_msgid;
	LDAP_MUTEX_UNLOCK( ld, LDAP_MSGID_LOCK );

	if ( dn == NULL ) {
		dn = "";
	}
	if ( passwd == NULL ) {
		passwd = "";
	}

#ifdef NOT_USED_BY_NETAPP
	if ( ld->ld_cache_on && ld->ld_cache_bind != NULL ) {
		struct berval	bv;

		bv.bv_val = (char *)passwd;
		bv.bv_len = strlen( passwd );
		/* if ( unlock_permitted ) LDAP_MUTEX_UNLOCK( ld ); */
		LDAP_MUTEX_LOCK( ld, LDAP_CACHE_LOCK );
		rc = (ld->ld_cache_bind)( ld, msgid, LDAP_REQ_BIND, dn, &bv,
		    LDAP_AUTH_SIMPLE );
		LDAP_MUTEX_UNLOCK( ld, LDAP_CACHE_LOCK );
		/* if ( unlock_permitted ) LDAP_MUTEX_LOCK( ld ); */
		if ( rc != 0 ) {
			return( rc );
		}
	}
#endif

	/* create a message to send */
	if (( rc = nsldapi_alloc_ber_with_options( ld, &ber ))
	    != LDAP_SUCCESS ) {
		return( -1 );
	}

	/* fill it in */
	if ( ber_printf( ber, "{it{ists}", msgid, LDAP_REQ_BIND,
	    NSLDAPI_LDAP_VERSION( ld ), dn, LDAP_AUTH_SIMPLE, passwd ) == -1 ) {
		LDAP_SET_LDERRNO( ld, LDAP_ENCODING_ERROR, NULL, NULL );
		ber_free( ber, 1 );
		return( -1 );
	}

	if ( nsldapi_put_controls( ld, NULL, 1, ber ) != LDAP_SUCCESS ) {
		ber_free( ber, 1 );
		return( -1 );
	}

	/* send the message */
	return( nsldapi_send_initial_request( ld, msgid, LDAP_REQ_BIND,
		(char *)dn, ber ));
}


/*
 * ldap_simple_bind - bind to the ldap server (and X.500) using simple
 * authentication.  The dn and password of the entry to which to bind are
 * supplied.  LDAP_SUCCESS is returned upon success, the ldap error code
 * otherwise.
 *
 * Example:
 *	ldap_simple_bind_s( ld, "cn=manager, o=university of michigan, c=us",
 *	    "secret" )
 */
SK_DOMAIN_EXPORT(ldap_simple_bind_s,SK_REASON_SWITCH_DOMAIN); 
int
LDAP_CALL
ldap_simple_bind_s( LDAP *ld, const char *dn, const char *passwd )
{
	int		msgid;
	LDAPMessage	*result;
	int		retVal;
	struct sec_micro	timeout = {smbflags.smb_ldap_timeout, 0};
	sk_domain_t	calling_domain = sk_domain_switch_begin(sk_domain_kahuna)

	LDAPDebug( LDAP_DEBUG_TRACE, "ldap_simple_bind_s\n", 0, 0, 0 );

	if ( NSLDAPI_VALID_LDAP_POINTER( ld ) &&
	    ( ld->ld_options & LDAP_BITOPT_RECONNECT ) != 0 ) {
		retVal = simple_bindifnot_s( ld, dn, passwd );
		goto Done;
	}

	if ( (msgid = ldap_simple_bind( ld, dn, passwd )) == -1 ) {
		retVal =  LDAP_GET_LDERRNO( ld, NULL, NULL );
		goto Done;
	}
	if ( ldap_result( ld, msgid, 1, &timeout, &result ) == -1 ) {
		retVal =  LDAP_GET_LDERRNO( ld, NULL, NULL );
		goto Done;
	}

	retVal =  ldap_result2error( ld, result, 1 );

Done:
	sk_domain_switch_end(calling_domain);
	return retVal;

}

/*
 * ldap_simple_bind_st - bind to the ldap server (and X.500) using simple
 * authentication, with a timeout.  The dn and password of the entry to
 * which to bind are supplied.  LDAP_SUCCESS is returned upon success,
 * the ldap error code otherwise.
 *
 * Example:
 *	ldap_simple_bind_st( ld, "cn=manager, o=university of michigan, c=us",
 *	    "secret", &timeout )
 */
SK_DOMAIN_EXPORT(ldap_simple_bind_st,SK_REASON_SWITCH_DOMAIN); 
int
LDAP_CALL
ldap_simple_bind_st( LDAP *ld, const char *dn, const char *passwd,
    struct sec_micro *timeout )
{
	int		msgid;
	int		ret;
	LDAPMessage	*result;
	sk_domain_t	calling_domain = sk_domain_switch_begin(sk_domain_kahuna)

	LDAPDebug( LDAP_DEBUG_TRACE, "ldap_simple_bind_st\n", 0, 0, 0 );

	if ( NSLDAPI_VALID_LDAP_POINTER( ld ) &&
	    ( ld->ld_options & LDAP_BITOPT_RECONNECT ) != 0 ) {
		ret = simple_bindifnot_st( ld, dn, passwd, timeout );
		goto out;
	}

	if ( (msgid = ldap_simple_bind( ld, dn, passwd )) == -1 ) {
		ret = LDAP_GET_LDERRNO( ld, NULL, NULL );
		goto out;
	}

	if ( ldap_result( ld, msgid, 1, timeout, &result ) == -1 ) {
		ret = LDAP_GET_LDERRNO( ld, NULL, NULL );
		goto out;
	}

	ret = ldap_result2error( ld, result, 1 );

out:
	sk_domain_switch_end(calling_domain);
	return ret;

}

/*
 * simple_bindifnot_s() is like ldap_simple_bind_s() except that it only does
 * a bind if the default connection is not currently bound.
 * If a successful bind using the same DN has already taken place we just
 * return LDAP_SUCCESS without conversing with the server at all.
 */
static int
simple_bindifnot_s( LDAP *ld, const char *dn, const char *passwd )
{
	int		msgid, rc;
	LDAPMessage	*result;
	char		*binddn;

	LDAPDebug( LDAP_DEBUG_TRACE, "simple_bindifnot_s\n", 0, 0, 0 );

	if ( !NSLDAPI_VALID_LDAP_POINTER( ld )) {
		return( LDAP_PARAM_ERROR );
	}

	if ( dn == NULL ) {
		dn = "";	/* to make comparisons simpler */
	}

	/*
	 * if we are already bound using the same DN, just return LDAP_SUCCESS.
	 */
	if ( NULL != ( binddn = nsldapi_get_binddn( ld ))
	    && 0 == strcmp( dn, binddn )) {
		rc = LDAP_SUCCESS;
		LDAP_SET_LDERRNO( ld, rc, NULL, NULL );
		goto unlock_and_return;
	}

	/*
	 * if the default connection has been lost and is now marked dead,
	 * dispose of the default connection so it will get re-established.
	 */
	LDAP_MUTEX_LOCK( ld, LDAP_CONN_LOCK );
	if ( NULL != ld->ld_defconn 
		&& LDAP_CONNST_DEAD == ld->ld_defconn->lconn_status ) {
	    nsldapi_free_connection( ld, ld->ld_defconn, 1, 0 );
	    ld->ld_defconn = NULL;
	}
	LDAP_MUTEX_UNLOCK( ld, LDAP_CONN_LOCK );

	/*
	 * finally, bind (this will open a new connection if necessary)
	 */
	if ( (msgid = simple_bind_nolock( ld, dn, passwd, 0 )) == -1 ) {
		rc = LDAP_GET_LDERRNO( ld, NULL, NULL );
		goto unlock_and_return;
	}

	if ( nsldapi_result_nolock( ld, msgid, 1, 0, (struct sec_micro *) 0,
	    &result ) == -1 ) {
		rc = LDAP_GET_LDERRNO( ld, NULL, NULL );
		goto unlock_and_return;
	}

	rc = ldap_result2error( ld, result, 1 );

unlock_and_return:
	return( rc );
}

/*
 * simple_bindifnot_st() is like ldap_simple_bind_st() except that it only does
 * a bind if the default connection is not currently bound.
 * If a successful bind using the same DN has already taken place we just
 * return LDAP_SUCCESS without conversing with the server at all.
 */
static int
simple_bindifnot_st( LDAP *ld, const char *dn, const char *passwd,
	struct sec_micro *timeout )
{
	int		msgid, rc;
	LDAPMessage	*result;
	char		*binddn;

	LDAPDebug( LDAP_DEBUG_TRACE, "simple_bindifnot_st\n", 0, 0, 0 );

	if ( !NSLDAPI_VALID_LDAP_POINTER( ld )) {
		return( LDAP_PARAM_ERROR );
	}

	if ( dn == NULL ) {
		dn = "";	/* to make comparisons simpler */
	}

	/*
	 * if we are already bound using the same DN, just return LDAP_SUCCESS.
	 */
	if ( NULL != ( binddn = nsldapi_get_binddn( ld ))
	    && 0 == strcmp( dn, binddn )) {
		rc = LDAP_SUCCESS;
		LDAP_SET_LDERRNO( ld, rc, NULL, NULL );
		goto unlock_and_return;
	}

	/*
	 * if the default connection has been lost and is now marked dead,
	 * dispose of the default connection so it will get re-established.
	 */
	LDAP_MUTEX_LOCK( ld, LDAP_CONN_LOCK );
	if ( NULL != ld->ld_defconn 
		&& LDAP_CONNST_DEAD == ld->ld_defconn->lconn_status ) {
	    nsldapi_free_connection( ld, ld->ld_defconn, 1, 0 );
	    ld->ld_defconn = NULL;
	}
	LDAP_MUTEX_UNLOCK( ld, LDAP_CONN_LOCK );

	/*
	 * finally, bind (this will open a new connection if necessary)
	 */
	if ( (msgid = simple_bind_nolock( ld, dn, passwd, 0 )) == -1 ) {
		rc = LDAP_GET_LDERRNO( ld, NULL, NULL );
		goto unlock_and_return;
	}

	if ( nsldapi_result_nolock( ld, msgid, 1, 0, timeout,
	    &result ) == -1 ) {
		rc = LDAP_GET_LDERRNO( ld, NULL, NULL );
		goto unlock_and_return;
	}

	rc = ldap_result2error( ld, result, 1 );

unlock_and_return:
	return( rc );
}
