/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */
/*
 *  Template MIB group implementation - at.c
 *
 */

/* Portions of this file are subject to the following copyright(s).  See
 * the Net-SNMP's COPYING file for more details and other copyrights
 * that may apply:
 */
/*
 * Portions of this file are copyrighted by:
 * Copyright  2003 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms specified in the COPYING file
 * distributed with the Net-SNMP package.
 */

#include <net-snmp/net-snmp-config.h>
#include <sk/ha_host.h>

#include <sk/sk_domain.h>
#include <sk/sk_processor.h>

#if HAVE_STRING_H
#include <legacy/string.h>
#else
#include <strings.h>
#endif
#if HAVE_STDLIB_H
#include <legacy/stdlib.h>
#endif
#if defined(IFNET_NEEDS_KERNEL) && !defined(_KERNEL)
#define _KERNEL 1
#define _I_DEFINED_KERNEL
#endif
#include <sys/types.h>
#if TIME_WITH_SYS_TIME
# ifdef WIN32
#  include <sys/timeb.h>
# else
# include <sk/time.h>
# endif
# include <legacy/time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sk/time.h>
# else
#  include <legacy/time.h>
# endif
#endif
#if HAVE_SYS_SOCKET_H
#include <net7g/sys/sk_socket.h>
#endif

#if HAVE_NETINET_IN_H
#include <net7g/netinet/sk_in.h>
#endif
#if HAVE_NET_IF_H
#include <net7g/net/sk_if.h>
#endif
#if HAVE_NET_IF_VAR_H
#include <net7g/net/sk_if_var.h>
#endif
#ifdef _I_DEFINED_KERNEL
#undef _KERNEL
#endif

#if HAVE_NETINET_IF_ETHER_H
#include <net7g/netinet/sk_if_ether.h>
#endif
#if HAVE_INET_MIB2_H
#include <inet/mib2.h>
#endif
#if HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#if HAVE_SYS_SYSCTL_H
#include <sys/sysctl.h>
#endif
#if HAVE_NET_IF_DL_H
#ifndef dynix
#include <net7g/net/sk_if_dl.h>
#else
#include <sys/net/if_dl.h>
#endif
#endif
#if HAVE_SYS_STREAM_H
#include <sys/stream.h>
#endif
#if HAVE_NET_ROUTE_H
#include <net7g/net/sk_route.h>
#endif
#ifdef solaris2
#include "kernel_sunos5.h"
#endif
#if HAVE_WINSOCK_H
#include <winsock.h>
#endif
#if HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

#ifdef hpux11
#include <sys/mib.h>
#include <netinet/mib_kern.h>
#endif                          /* hpux11 */

#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include <net-snmp/agent/auto_nlist.h>

#include "at.h"
#include <net7g/netinet/sk_in_var.h>
#include <net7g/netinet/sk_if_ether.h>
#include <net7g/netinet/sk_in_systm.h>
#include <net7g/netinet/sk_ip.h>


extern void * ifnet_inaddr(struct ifnet *ifp, int host, int af);

STATIC int	MatchRTMACAddr(struct radix_node * rn, oid * name,
    size_t *length, oid * Current, int exact, oid *lowest, uint32_t *LowAddrp,
    char *LowPhysAddr);
STATIC void	FindARPTEntry(struct radix_node * rn,
    oid * name, size_t *length, oid * Current, int exact, oid *lowest,
    uint32_t *LowAddrp, char *LowPhysAddr);
STATIC int if_index_by_net(uint32_t addr);
STATIC FindVarMethod _var_atEntry;


        /*********************
	 *
	 *  Public interface functions
	 *
	 *********************/

/*
 * define the structure we're going to ask the agent to register our
 * information at 
 */
struct variable1 at_variables[] = {
    {ATIFINDEX, ASN_INTEGER, RONLY, var_atEntry, 1, {1}},
    {ATPHYSADDRESS, ASN_OCTET_STR, RONLY, var_atEntry, 1, {2}},
    {ATNETADDRESS, ASN_IPADDRESS, RONLY, var_atEntry, 1, {3}}
};

/*
 * Define the OID pointer to the top of the mib tree that we're
 * registering underneath 
 */
oid             at_variables_oid[] = { SNMP_OID_MIB2, 3, 1, 1 };

void
init_at(void)
{
    /*
     * register ourselves with the agent to handle our mib tree 
     */
    REGISTER_MIB("mibII/at", at_variables, variable1, at_variables_oid);
}

/*
 * Read the ARP table
 */
STATIC uint8_t *
_var_atEntry(SNMP_FM_ARGS)
{
	/*
	 * object identifier is of form:
	 * 1.3.6.1.2.1.3.1.1.1.interface.1.A.B.C.D,  where A.B.C.D is IP
	 * address. Interface is at offset 10, IPADDR starts at offset 12.
	 */
	oid             lowest[16];
	oid             current[16];
	static char     LowPhysAddr[6];
	uint32_t           LowAddr;
	struct radix_node_head *rnh;

	size_t          oid_length;

	DEBUGMSGTL(("mibII/at", "var_atEntry: "));
	DEBUGMSGOID(("mibII/at", vp->name, vp->namelen));
	DEBUGMSG(("mibII/at", " %d\n", exact));

	/*
	 * fill in object part of name for current (less sizeof instance part) 
	 */
	memcpy((char *) current, (char *) vp->name,
	       (int) vp->namelen * sizeof(oid));

	if (current[6] == 3) {      /* AT group oid */
		oid_length = 16;
	} else {                    /* IP NetToMedia group oid */
		oid_length = 15;
	}

	lowest[0] = '\0';

	LowAddr = UINT_MAX;		/* Don't have one yet */
	SK_DOMAIN_REFERENCE(get_rtable, SK_DOMAIN_NETWORK);
	if ((rnh = get_rtable(AF_INET)) != NULL) {
		FindARPTEntry(rnh->rnh_treetop, name, length,
		    current, exact, lowest, &LowAddr, LowPhysAddr);
	}
	if (LowAddr == UINT_MAX)
		return (NULL);

	memcpy((char *) name, (char *) lowest, oid_length * sizeof(oid));
	*length = oid_length;
	switch (vp->name[9]) {
	case ATIFINDEX:
		*var_len = sizeof long_return;
		long_return = if_index_by_net(LowAddr);
		return (uint8_t *) & long_return;
	case ATPHYSADDRESS:
		*var_len = sizeof(LowPhysAddr);
		return (uint8_t *) LowPhysAddr;
	case ATNETADDRESS:
		*var_len = sizeof long_return;
		long_return = LowAddr;
		return (uint8_t *) & long_return;
	default:
		DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_at\n",
			    vp->magic));
		ERROR("var_atEntry");
	}
	return NULL;
}

/*
 * var_atEntry(...
 * Arguments:
 * vp     IN      - pointer to variable entry that points here
 * name    IN/OUT  - IN/name requested, OUT/name found
 * length  IN/OUT  - length of IN/OUT oid's 
 * exact   IN      - TRUE if an exact match was requested
 * var_len OUT     - length of variable or 0 if function returned
 * write_method
 * 
 */

uint8_t         *
var_atEntry(struct variable *vp,
            oid * name,
            size_t * length,
            int exact, size_t * var_len, WriteMethod ** write_method)
{
    /*
     * Address Translation table object identifier is of form:
     * 1.3.6.1.2.1.3.1.1.1.interface.1.A.B.C.D,  where A.B.C.D is IP address.
     * Interface is at offset 10,
     * IPADDR starts at offset 12.
     *
     * IP Net to Media table object identifier is of form:
     * 1.3.6.1.2.1.4.22.1.1.1.interface.A.B.C.D,  where A.B.C.D is IP address.
     * Interface is at offset 10,
     * IPADDR starts at offset 11.
     */
	DOMAIN_SWITCH_NEST(sk_domain_network, _var_atEntry,
			   (vp, name, length, exact, var_len, write_method)); 
}

SK_DOMAIN_EXPORT(FindARPTEntry, SK_DOMAIN_NETWORK);
void
FindARPTEntry(struct radix_node * rn,
     oid * name, size_t *length, oid * Current, int exact, oid *lowest,
    uint32_t *LowAddrp, char *LowPhysAddr)
{
	ASSERT(sk_in_domain(sk_domain_network));
	for (;;) {
		while (rn->rn_b >= 0)
			rn = rn->rn_l;	/* First time through node, go left */

		if (MatchRTMACAddr(rn, name, length, Current, exact, 
				   lowest, LowAddrp,  LowPhysAddr))
			return;		/* exact match, stop searching */
		while (rn->rn_p->rn_r == rn) {	/* if coming back from right */
			rn = rn->rn_p;	/* go back up */
			if (rn->rn_flags & RNF_ROOT)
				return;
		}
		rn = rn->rn_p->rn_r;	/* otherwise, go right */
	}
	return;

}

SK_DOMAIN_EXPORT(MatchRTMACAddr, SK_DOMAIN_NETWORK);
int
MatchRTMACAddr(struct radix_node * rn, oid * name, size_t *length, 
	       oid * Current, int exact, oid *lowest, uint32_t *LowAddrp, 
	       char *LowPhysAddr)
{
	struct rtentry *rt;
	struct sockaddr_inarp *sin;
	struct sockaddr_dl *sdl;
	uint8_t	*cp;
	oid     *op;

	ASSERT(sk_in_domain(sk_domain_network));
	for (; rn; rn = rn->rn_dupedkey) {
		if (rn->rn_flags & RNF_ROOT)
			continue;
		rt = (struct rtentry *)rn;
		if (!(rt->rt_flags & RTF_LLINFO))
			continue;
		sin = (struct sockaddr_inarp *)rt_key(rt);
		if (sin->sin_family != AF_INET)
			continue;
		sdl = (struct sockaddr_dl *)(rt->rt_gateway);
		if (sdl->sdl_alen == 0)
			continue;
		Current[10] = 1;//ntap net-snmp calls if_index_by_net(sin->sin_addr.s_addr);
		Current[11] = 1;
		cp = (uint8_t *) & sin->sin_addr.s_addr;
		op = Current + 12;
		*op++ = *cp++;
		*op++ = *cp++;
		*op++ = *cp++;
		*op++ = *cp++;

		if (exact) {
			if (snmp_oid_compare(Current, 16, name, *length) == 0) {
				bcopy((char *) Current, (char *) lowest,
						16 * sizeof(oid));
				*LowAddrp = sin->sin_addr.s_addr;
				bcopy(LLADDR(sdl), LowPhysAddr, sdl->sdl_alen);
				return (1);	/* no need to search further */
			}
		} else {
			if ((snmp_oid_compare(Current, 16, name, *length) > 0)
			    && ((*LowAddrp == -1) || 
				(snmp_oid_compare(Current, 16, lowest,	16) < 0))) {
				/*
				 * if new one is greater than input and
				 * closer to input than previous lowest, save
				 * this one as the "next" one.
				 */
				bcopy((char *) Current, (char *) lowest,
					16 * sizeof(oid));
				*LowAddrp = sin->sin_addr.s_addr;
				bcopy(LLADDR(sdl), LowPhysAddr, sdl->sdl_alen);
			}
		}
	}
	return (0);
}

/*
 * return index of interface that input network address corresponds to
 */
SK_DOMAIN_EXPORT(if_index_by_net, SK_DOMAIN_NETWORK);
int
if_index_by_net(uint32_t addr)
{
	int interface;
	struct ifnet *ifp;
	struct in_ifaddr *ia;
	uint32_t addr2;
	int host = sk_gethost();
	
	ASSERT(sk_in_domain(sk_domain_network));

	/*
	 * This is gross, but we have to play the little game with
	 * local/partner in counting up the index values.
	 */
	addr = ntohl(addr) & 0xFFFFFF00;
	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (interface = 1, ifp = ifnet; ifp; ifp = ifp->if_next) {
		/* skip things not in the default ipspace */
		if(ifp->ipsid != 0)
			continue;

		ia = ifnet_inaddr(ifp, sk_gethost(), AF_INET);
		if (ia == NULL) {
			if (host == SK_HOST_LOCAL)
				interface++;
			continue;
		}

		addr2 = (uint32_t) (((struct sockaddr_in *)
				& (ia->ia_addr))->sin_addr.s_addr);
		addr2 = ntohl(addr2) & 0xFFFFFF00;
		if(addr2 == addr)
			return(interface);
		interface++;
	}
	return(0);
}

