/*
 *
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */

#include <legacy/string.h>
#include <legacy/stdlib.h>

#include <sk/sk_common.h>

#include <sys/types.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/sys/sk_socketvar.h>
#include <net7g/sys/sk_sockio.h>


#include <net7g/net/sk_if.h>

#include <net7g/netinet/sk_ip6.h>
#include <net7g/netinet/sk_if_ether.h>

#include <net7g/netinet6/sk_in6_var.h>

#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include "icmp_ipv6.h"
#include "interfaces.h"

STATIC FindVarMethod var_ipv6IfIcmpEntry;

STATIC FindVarMethod _var_ipv6IfIcmpEntry;

#define IPV6IFICMPENTRY_OID     { SNMP_OID_MIB2, 56, 1, 1, 1 }

struct variable1 ipv6IfIcmpEntry_variables[34] = {
    {IPV6IFICMPINMSGS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {1}},
    {IPV6IFICMPINERRORS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {2}},
    {IPV6IFICMPINDESTUNREACHS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {3}},
    {IPV6IFICMPINADMINPROHIBS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {4}},
    {IPV6IFICMPINTIMEEXCDS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {5}},
    {IPV6IFICMPINPARMPROBLEMS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {6}},
    {IPV6IFICMPINPKTTOOBIGS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {7}},
    {IPV6IFICMPINECHOS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {8}},
    {IPV6IFICMPINECHOREPLIES, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {9}},
    {IPV6IFICMPINROUTERSOLICITS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {10}},
    {IPV6IFICMPINROUTERADVERTISEMENTS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {11}},
    {IPV6IFICMPINNEIGHBORSOLICITS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {12}},
    {IPV6IFICMPINNEIGHBORADVERTISEMENTS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {13}},
    {IPV6IFICMPINREDIRECTS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {14}},
    {IPV6IFICMPINGROUPMEMBQUERIES, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {15}},
    {IPV6IFICMPINGROUPMEMBRESPONSES, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {16}},
    {IPV6IFICMPINGROUPMEMBREDUCTIONS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {17}},
    {IPV6IFICMPOUTMSGS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {18}},
    {IPV6IFICMPOUTERRORS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {19}},
    {IPV6IFICMPOUTDESTUNREACHS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {20}},
    {IPV6IFICMPOUTADMINPROHIBS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {21}},
    {IPV6IFICMPOUTTIMEEXCDS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {22}},
    {IPV6IFICMPOUTPARMPROBLEMS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {23}},
    {IPV6IFICMPOUTPKTTOOBIGS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {24}},
    {IPV6IFICMPOUTECHOS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {25}},
    {IPV6IFICMPOUTECHOREPLIES, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {26}},
    {IPV6IFICMPOUTROUTERSOLICITS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {27}},
    {IPV6IFICMPOUTROUTERADVERTISEMENTS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {28}},
    {IPV6IFICMPOUTNEIGHBORSOLICITS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {29}},
    {IPV6IFICMPOUTNEIGHBORADVERTISEMENTS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {30}},
    {IPV6IFICMPOUTREDIRECTS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {31}},
    {IPV6IFICMPOUTGROUPMEMBQUERIES, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {32}},
    {IPV6IFICMPOUTGROUPMEMBRESPONSES, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {33}},
    {IPV6IFICMPOUTGROUPMEMBREDUCTIONS, COUNTER32, RONLY, var_ipv6IfIcmpEntry, 1, {34}}
};
oid ipv6IfIcmpEntry_variables_oid[10] = IPV6IFICMPENTRY_OID;


void init_icmp(void)
{
    REGISTER_MIB("ipv6IfIcmpEntry",
            ipv6IfIcmpEntry_variables, variable1,
            ipv6IfIcmpEntry_variables_oid);
}

/* format: .1.3.6.1.2.1.56.1.1.1.?.interface_id */
uint8_t *
_var_ipv6IfIcmpEntry(SNMP_FM_ARGS)
{
    bool found = false;
    int32_t max = Interface_Scan_Get_Count(AF_INET6);
    oid newname[MAX_OID_LEN], lowest[MAX_OID_LEN];
    struct icmp6_ifstat *ifs6 = NULL;
    struct ifnet *ifp = NULL, *lowin_ifp = NULL;
    struct in6_ifreq ifr;
    struct socket *s = NULL;

    if (!max) {
        return NULL;
    }

    bzero(newname, sizeof(newname));
    bzero(lowest, sizeof(lowest));
    bcopy(vp->name, newname, (int32_t)vp->namelen * sizeof(oid));

    SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
    for (ifp = ifnet; ifp; ifp = ifp->if_next) {
        newname[11] = ifp->if_index;

        if(exact) {
            if(snmp_oid_compare(newname, 12, name, *length) == 0) {
                bcopy((char *) newname,
                        (char *) lowest,
                        12 * sizeof(oid));
                found = true;
                lowin_ifp = ifp;
                break;  /* no need to search further */
            }
        }
        else {
            if ((snmp_oid_compare(newname, 12, name, *length) > 0) &&
                    (!found || (snmp_oid_compare(newname, 12, lowest, 12) < 0))) {
                found = true;
                lowin_ifp = ifp;
                bcopy((char *) newname,
                        (char *) lowest,
                        12 * sizeof(oid));
            }
        }
    }
    if (!found) {
        return NULL;
    }

    bcopy(lowest, name, 12 * sizeof(oid));
    *length = 12;
    *write_method = 0;
    *var_len = sizeof(long_return); /* default */

    memset(&ifr, 0, sizeof(ifr));

    strlcpy(ifr.ifr_name, lowin_ifp->if_fullname, sizeof(ifr.ifr_name));

    if (!(s = socket(AF_INET6, SOCK_DGRAM, 0, NULL))) {
        return NULL;
    }

    SK_DOMAIN_REFERENCE(ifioctl, SK_DOMAIN_NETWORK);
    if (ifioctl(s, SIOCGIFSTAT_ICMP6, (caddr_t)&ifr) != 0) {
        soclose(s);
        return NULL;
    }
    soclose(s);

    ifs6 = &ifr.ifr_ifru.ifru_icmp6stat;

    if(!ifs6) {
        return NULL;
    }

    switch (vp->magic) {

        case IPV6IFICMPINMSGS:
            long_return = ifs6->ifs6_in_msg;
            break;

        case IPV6IFICMPINERRORS:
            long_return = ifs6->ifs6_in_error;
            break;

        case IPV6IFICMPINDESTUNREACHS:
            long_return = ifs6->ifs6_in_dstunreach;
            break;

        case IPV6IFICMPINADMINPROHIBS:
            long_return = ifs6->ifs6_in_adminprohib;
            break;

        case IPV6IFICMPINTIMEEXCDS:
            long_return = ifs6->ifs6_in_timeexceed;
            break;

        case IPV6IFICMPINPARMPROBLEMS:
            long_return = ifs6->ifs6_in_paramprob;
            break;

        case IPV6IFICMPINPKTTOOBIGS:
            long_return = ifs6->ifs6_in_pkttoobig;
            break;

        case IPV6IFICMPINECHOS:
            long_return = ifs6->ifs6_in_echo;
            break;

        case IPV6IFICMPINECHOREPLIES:
            long_return = ifs6->ifs6_in_echoreply;
            break;

        case IPV6IFICMPINROUTERSOLICITS:
            long_return = ifs6->ifs6_in_routersolicit;
            break;

        case IPV6IFICMPINROUTERADVERTISEMENTS:
            long_return = ifs6->ifs6_in_routeradvert;
            break;

        case IPV6IFICMPINNEIGHBORSOLICITS:
            long_return = ifs6->ifs6_in_neighborsolicit;
            break;

        case IPV6IFICMPINNEIGHBORADVERTISEMENTS:
            long_return = ifs6->ifs6_in_neighboradvert;
            break;

        case IPV6IFICMPINREDIRECTS:
            long_return = ifs6->ifs6_in_redirect;
            break;

        case IPV6IFICMPINGROUPMEMBQUERIES:
            long_return = ifs6->ifs6_in_mldquery;
            break;

        case IPV6IFICMPINGROUPMEMBRESPONSES:
            long_return = ifs6->ifs6_in_mldreport;
            break;

        case IPV6IFICMPINGROUPMEMBREDUCTIONS:
            long_return = ifs6->ifs6_in_mlddone;
            break;

        case IPV6IFICMPOUTMSGS:
            long_return = ifs6->ifs6_out_msg;
            break;

        case IPV6IFICMPOUTERRORS:
            long_return = ifs6->ifs6_out_error;
            break;

        case IPV6IFICMPOUTDESTUNREACHS:
            long_return = ifs6->ifs6_out_dstunreach;
            break;

        case IPV6IFICMPOUTADMINPROHIBS:
            long_return = ifs6->ifs6_out_adminprohib;
            break;

        case IPV6IFICMPOUTTIMEEXCDS:
            long_return = ifs6->ifs6_out_timeexceed;
            break;

        case IPV6IFICMPOUTPARMPROBLEMS:
            long_return = ifs6->ifs6_out_paramprob;
            break;

        case IPV6IFICMPOUTPKTTOOBIGS:
            long_return = ifs6->ifs6_out_pkttoobig;
            break;

        case IPV6IFICMPOUTECHOS:
            long_return = ifs6->ifs6_out_echo;
            break;

        case IPV6IFICMPOUTECHOREPLIES:
            long_return = ifs6->ifs6_out_echoreply;
            break;

        case IPV6IFICMPOUTROUTERSOLICITS:
            long_return = ifs6->ifs6_out_routersolicit;
            break;

        case IPV6IFICMPOUTROUTERADVERTISEMENTS:
            long_return = ifs6->ifs6_out_routeradvert;
            break;

        case IPV6IFICMPOUTNEIGHBORSOLICITS:
            long_return = ifs6->ifs6_out_neighborsolicit;
            break;

        case IPV6IFICMPOUTNEIGHBORADVERTISEMENTS:
            long_return = ifs6->ifs6_out_neighboradvert;
            break;

        case IPV6IFICMPOUTREDIRECTS:
            long_return = ifs6->ifs6_out_redirect;
            break;

        case IPV6IFICMPOUTGROUPMEMBQUERIES:
            long_return = ifs6->ifs6_out_mldquery;
            break;

        case IPV6IFICMPOUTGROUPMEMBRESPONSES:
            long_return = ifs6->ifs6_out_mldreport;
            break;

        case IPV6IFICMPOUTGROUPMEMBREDUCTIONS:
            long_return = ifs6->ifs6_out_mlddone;
            break;

        default:
            return NULL;
    }

    return (uint8_t *)&long_return;
}

uint8_t *
var_ipv6IfIcmpEntry(SNMP_FM_ARGS)
{
    if (!ipv6_enabled) {
        return NULL;
    }

    DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6IfIcmpEntry,
            (vp, name, length, exact, var_len, write_method));
}
