/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */
/*
 *  Interfaces MIB group implementation - interfaces.c
 *
 */

/* Portions of this file are subject to the following copyright(s).  See
 * the Net-SNMP's COPYING file for more details and other copyrights
 * that may apply:
 */
/*
 * Portions of this file are copyrighted by:
 * Copyright  2003 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms specified in the COPYING file
 * distributed with the Net-SNMP package.
 */

#include <net-snmp/net-snmp-config.h>
#include <sk/ha_host.h>

#include <cmds/ifconfig.h>
#include <sk/sk_domain.h>
#include <sk/sk_processor.h>

#if defined(IFNET_NEEDS_KERNEL) && !defined(_KERNEL) && !defined(IFNET_NEEDS_KERNEL_LATE)
#define _KERNEL 1
#define _I_DEFINED_KERNEL
#endif

#if HAVE_STRING_H
#include <legacy/string.h>
#else
#include <strings.h>
#endif

#if HAVE_STDLIB_H
#include <legacy/stdlib.h>
#endif
#if HAVE_UNISTD_H
#endif
#if HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#include <sys/types.h>
#if HAVE_WINSOCK_H
#include <winsock.h>
#endif
#if defined(IFNET_NEEDS_KERNEL) && !defined(_KERNEL) && defined(IFNET_NEEDS_KERNEL_LATE)
#define _KERNEL 1
#define _I_DEFINED_KERNEL
#endif
#if HAVE_SYS_SOCKET_H
#include <net7g/sys/sk_socket.h>
#endif
#ifndef STREAM_NEEDS_KERNEL_ISLANDS
#if HAVE_SYS_STREAM_H
#include <sys/stream.h>
#endif
#endif
#if HAVE_SYS_SOCKETVAR_H
#include <net7g/sys/sk_socketvar.h>
#endif

#if TIME_WITH_SYS_TIME
# ifdef WIN32
#  include <sys/timeb.h>
# else
# include <sk/time.h>
# endif
# include <legacy/time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sk/time.h>
# else
#  include <legacy/time.h>
# endif
#endif
#if HAVE_SYS_SOCKIO_H
#include <net7g/sys/sk_sockio.h>
#endif
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#if HAVE_SYS_IOCTL_H
#include <net7g/sys/sk_sockio.h>
#endif
#if HAVE_NETINET_IN_H
#include <net7g/netinet/sk_in.h>
#endif
#if HAVE_NET_IF_H
#include <net7g/net/sk_if.h>
#endif
#if HAVE_NET_IF_VAR_H
#include <net7g/net/sk_if_var.h>
#endif
#ifdef _I_DEFINED_KERNEL
#undef _KERNEL
#endif
#ifdef STREAM_NEEDS_KERNEL_ISLANDS
#if HAVE_SYS_STREAM_H
#include <sys/stream.h>
#endif
#endif
#if HAVE_NET_ROUTE_H
#include <net7g/net/sk_route.h>
#endif
#if HAVE_NETINET_IN_SYSTM_H
#include <net7g/netinet/sk_in_systm.h>
#endif
#if HAVE_SYS_HASHING_H
#include <sys/hashing.h>
#endif
#if HAVE_NETINET_IN_VAR_H
#include <net7g/netinet/sk_in_var.h>
#endif
#if HAVE_NETINET_IP_H
#include <net7g/netinet/sk_ip.h>
#endif
#ifdef INET6
#if HAVE_NETINET_IP6_H
#include <net7g/netinet/sk_ip6.h>
#endif
#endif
#if HAVE_SYS_QUEUE_H
#include <sys/queue.h>
#endif
#if HAVE_NETINET_IP_VAR_H
#include <net7g/netinet/sk_ip_var.h>
#endif
#ifdef INET6
#if HAVE_NETINET6_IP6_VAR_H
#include <net7g/netinet6/sk_ip6_var.h>
#endif
#endif
#if HAVE_NETINET_IN_PCB_H
#include <net7g/netinet/sk_in_pcb.h>
#endif
#if HAVE_NETINET_IF_ETHER_H
#include <net7g/netinet/sk_if_ether.h>
#endif
#if HAVE_NET_IF_TYPES_H
#include <net7g/net/sk_if_types.h>
#endif
#if HAVE_NET_IF_DL_H
#ifndef dynix
#include <net7g/net/sk_if_dl.h>
#else
#include <sys/net/if_dl.h>
#endif
#endif
#if HAVE_INET_MIB2_H
#include <inet/mib2.h>
#endif
#if HAVE_IOCTLS_H
#include <ioctls.h>
#endif
#if HAVE_LINUX_ETHTOOLS_H
#   if HAVE_PCI_PCI_H
#      include <pci/pci.h>
#   endif
#include <linux/ethtools.h>
#endif

#if HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

#ifdef solaris2
# include <sys/errno.h>
#include "kernel_sunos5.h"
#else
//#include "sys/kernel.h"
#endif

#ifdef hpux
#include <sys/mib.h>
#include <net7g/netinet/sk_mib_kern.h>
#endif                          /* hpux */

#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include <net-snmp/agent/auto_nlist.h>

#include "interfaces.h"
#include "mibgroup/util_funcs.h"
#include <net7g/netinet/sk_if_ether.h>
#include "at.h"
#include <stdint.h>
#define IFNUMBER        0

extern struct sec_micro starttime;

struct variable3 interfaces_variables[] = {
    {IFNUMBER, ASN_INTEGER, RONLY, var_interfaces, 1, {1}},
    {IFINDEX, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 1}},
    {IFDESCR, ASN_OCTET_STR, RONLY, var_ifEntry, 3, {2, 1, 2}},
    {IFTYPE, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 3}},
    {IFMTU, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 4}},
    {IFSPEED, ASN_GAUGE, RONLY, var_ifEntry, 3, {2, 1, 5}},
    {IFPHYSADDRESS, ASN_OCTET_STR, RONLY, var_ifEntry, 3, {2, 1, 6}},
    {IFADMINSTATUS, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 7}},
    {IFOPERSTATUS, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 8}},
    {IFLASTCHANGE, ASN_TIMETICKS, RONLY, var_ifEntry, 3, {2, 1, 9}},
    {IFINOCTETS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 10}},
    {IFINUCASTPKTS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 11}},
    {IFINNUCASTPKTS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 12}},
    {IFINDISCARDS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 13}},
    {IFINERRORS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 14}},
    {IFINUNKNOWNPROTOS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 15}},
    {IFOUTOCTETS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 16}},
    {IFOUTUCASTPKTS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 17}},
    {IFOUTNUCASTPKTS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 18}},
    {IFOUTDISCARDS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 19}},
    {IFOUTERRORS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 20}},
    {IFOUTQLEN, ASN_GAUGE, RONLY, var_ifEntry, 3, {2, 1, 21}}//,
//Unsupported    {IFSPECIFIC, ASN_OBJECT_ID, RONLY, var_ifEntry, 3, {2, 1, 22}}
};

/*
 * Define the OID pointer to the top of the mib tree that we're
 * registering underneath, and the OID of the MIB module 
 */
oid             interfaces_variables_oid[] = { SNMP_OID_MIB2, 2 };
oid             interfaces_module_oid[] = { SNMP_OID_MIB2, 31 };

void
init_interfaces(void)
{
    /*
     * register ourselves with the agent to handle our mib tree 
     */
    REGISTER_MIB("mibII/interfaces", interfaces_variables, variable3,
                 interfaces_variables_oid);
//ntap    REGISTER_SYSOR_ENTRY(interfaces_module_oid,
//ntap                         "The MIB module to describe generic objects for network interface sub-layers");

//ntap    snmpd_register_config_handler("interface", parse_interface_config,
//ntap                                  free_interface_config,
//ntap                                  "name type speed");
}

#define ifatoia(ifa)	((struct in_ifaddr *)(ifa))


/* Count number of network interfaces */
SK_DOMAIN_EXPORT(Interface_Scan_Get_Count, SK_REASON_SWITCH_DOMAIN);
int32_t
Interface_Scan_Get_Count(int af)
{
	struct ifnet *ifp;
	int32_t ix = 0;
	int host = sk_gethost();
	sk_domain_t     old_domain = sk_domain_switch_begin(sk_domain_network);
	
	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (ifp = ifnet; ifp; ifp = ifp->if_next) {
		/* skip things not in the default ipspace */
		if(ifp->ipsid != 0)
			continue;
		/*
		 * If in partner-mode, we make sure that we have
		 * a configured, running secondary before exposing it.
		 */
		if ((host == SK_HOST_PARTNER) && 
		    (!ifnet_inaddr(ifp, host, af)))
			continue;
		
		/* XXX6: changes required once we support enabling v6 per 
		   interface */
		ix++;
	}

	sk_domain_switch_end(old_domain);

	return ix;
}

SK_DOMAIN_EXPORT(ifnet_inaddr, SK_DOMAIN_NETWORK);
void *
ifnet_inaddr(struct ifnet *ifp, int host, int af)
{
	struct ifaddr *ifa;

	ASSERT(sk_in_domain(sk_domain_network));
	if (ifp->if_flags & IFF_LOOPBACK)
		/* if loopback return either one */
		host = -1;

	SK_DOMAIN_REFERENCE(ifa_ifpfirstaddrhost, SK_DOMAIN_NETWORK);
	/* should we actually use vfilerized version of this for snmp? */
	ifa = ifa_ifpfirstaddrhost(ifp, af, host);
	return (void *)ifa;
}

/* This routine takes the interface # as an argument (int Index)
	and returns its ifnet ptr, or NULL if there isn't such an
	interface */
SK_DOMAIN_EXPORT(Interface_Scan_By_Index, SK_DOMAIN_NETWORK);
struct ifnet *
Interface_Scan_By_Index(int Index)
{
	struct ifnet *ifp;
	int i = 1;
	int host = sk_gethost();
	sk_domain_t     old_domain = sk_domain_switch_begin(sk_domain_network);

	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (ifp = ifnet; ifp; ifp = ifp->if_next) {
		/* skip things not in the default ipspace */
		if(ifp->ipsid != 0)
			continue;

		/*
		 * If in partner-mode, we make sure that we have
		 * a configured, running secondary before exposing it.
		 */
		if (host == SK_HOST_PARTNER  && 
		    (!ifnet_inaddr(ifp, host, AF_INET)))
			continue;

		if (i == Index)
			break;
		i++;
	}

	sk_domain_switch_end(old_domain);

	return ifp;
}

uint8_t *
var_interfaces(struct variable *vp,
            oid * name,
            size_t * length,
            int exact, size_t * var_len, WriteMethod ** write_method)
{
    if (header_generic(vp, name, length, exact, var_len, write_method) ==
        MATCH_FAILED)
        return NULL;

	switch (vp->magic) {
	case IFNUMBER:
		*var_len = sizeof(long_return);
		long_return = Interface_Scan_Get_Count(AF_INET);
		return (uint8_t *) & long_return;
	default:
		DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_interfaces\n",
			    vp->magic));
	}
	return NULL;
}

#define MIBIFNAMSIZ 2*IFNAMSIZ

uint8_t *
var_ifEntry(struct variable *vp,
            oid * name,
            size_t * length,
            int exact, size_t * var_len, WriteMethod ** write_method)
{
	oid             newname[SNMP_OID_MAX];
	int             interface;
	int             count;
	struct ifnet   *ifp;
	struct arpcom  *ap;
	static char     Name[MIBIFNAMSIZ];
	char *Name_ptr = Name;
	int		host = sk_gethost();
        static uint32_t baudrate;

	bcopy((char *) vp->name, (char *) newname, (int) vp->namelen *
					sizeof(oid));
	/* find "next" interface */
	count = (int)Interface_Scan_Get_Count(AF_INET);

	if (header_simple_table(vp, name, length, exact, var_len, write_method, count)
	    == MATCH_FAILED) {
		return NULL;
	}
	interface = name[*length - 1];

	ifp = Interface_Scan_By_Index(interface);
	if (ifp == NULL)
		return NULL;

	switch (vp->magic) {
	case IFINDEX:
		long_return = interface;
		return (uint8_t *) & long_return;
	case IFDESCR:
		/*
		 * We have to play this gross little game of getting
		 * the correct interface name based on whether we are
		 * in local or takeover mode.
		 */
		if (ifp->if_flags2 & IFF2_MGMT_PORTS){
			strlcpy(Name_ptr, "MGMT_PORT_ONLY ", IFNAMSIZ);
			Name_ptr += strlen(Name_ptr);
		}
		if (host == SK_HOST_LOCAL) {
			strlcpy(Name_ptr, ifp->if_fullname, IFNAMSIZ);
		} else {
			ifconfig_map_t *im =
				ifconfig_getPartner(ifp->if_fullname);
			SK_DOMAIN_REFERENCE(ifconfig_getPartner, SK_REASON_SWITCH_DOMAIN);
			if (!im)
				return NULL;
			strlcpy(Name_ptr, im->im_partner, IFNAMSIZ);
		}
		*var_len = strlen(Name);
		return (uint8_t *) Name;
	case IFTYPE:
		/*
		 * The header file that defines "if_type" values says
		 * "This list is derived from the SNMP list of ifTypes,
		 * currently documented in RFC1573", so we just use
		 * the "if_type" value.
		 */
		long_return = ifp->if_type;
		return (uint8_t *) & long_return;
	case IFMTU:
		return (uint8_t *) & ifp->if_mtu;
	case IFSPEED:
                /*
                 * Per RFCs 2863/3635, interfaces that have
                 * a rate that is requires more than 32-bits to
                 * store, set the value to 4,294,967,295.
                 */
                if (ifp->if_baudrate > UINT32_MAX) {
                  baudrate = UINT32_MAX;
                  return ((uint8_t *)(&baudrate));
                }
		return (uint8_t *) & ifp->if_baudrate;
	case IFPHYSADDRESS:
		/*
		 * XXX - this should arguably be done by:
		 *
		 *	1) putting back the code that puts link-level
		 *	   addresses on the list of addresses for the
		 *	   interface;
		 *
		 *	2) making sure that *all* interfaces get that
		 *	   address (including ATM);
		 *
		 *	3) getting the address from that list.
		 */
		ap = (struct arpcom *)ifp;
		if (ifp->if_addrlen != sizeof(ap->ac_enaddr)) {
			/*
			 * XXX - we only support interfaces with 802.x-style
			 * 6-byte addresses.  This needs to be fixed.
			 */
			long_return = 0;
			return (uint8_t *) &long_return;
		}
		*var_len = ifp->if_addrlen;
		bcopy((char*)&ap->ac_enaddr, return_buf, sizeof(ap->ac_enaddr));
		return (uint8_t *) return_buf;
	case IFADMINSTATUS: {
		long_return = (ifp->if_flags & IFF_PHYS_UP) ? 1 : 2;
		return (uint8_t *) & long_return;
	    }
	case IFOPERSTATUS: {
		long_return = (ifp->if_flags & IFF_LINK_UP) ? 1 : 2;
		return (uint8_t *) & long_return;
	    }
	case IFLASTCHANGE:
		/* rfc 1213 says that this should be the value of
			sysUpTime when the interface entered its
			current operational state.  This is when
			the if was last ifconfiged up.

			That should be the number of 10's of milliseconds
			since the if was last configed up; i.e., if
			"sk_tics" was N when it was last configed up,
			it would be (sk_tics - N)*(100 / SK_TICS_PER_SEC).

			This is *not* "if_lastchange", as it currently
			stands; BSD implements that as the wall-clock time
			when something was last sent out on the interface,
			and we do the same.

			I'm not sure why BSD gives it a rather SNMPish
			name but doesn't give it a value that has
			anything to do with SNMP's ifLastChange value.
		 */
		long_return = 0;
		return (uint8_t *) & long_return;
	case IFINOCTETS:
		long_return = LOW32(ifp->if_ibytes);
		return (uint8_t *) & long_return;
	case IFINUCASTPKTS:
		long_return = ifp->if_ipackets - ifp->if_imcasts;
		return (uint8_t *) & long_return;
	case IFINNUCASTPKTS:
		return (uint8_t *) & ifp->if_imcasts;
	case IFINDISCARDS:
		return (uint8_t *) & ifp->if_iqdrops;
	case IFINERRORS:
		return (uint8_t *) & ifp->if_ierrors;
	case IFINUNKNOWNPROTOS:
		return (uint8_t *) & ifp->if_noproto;
	case IFOUTOCTETS:
		long_return = LOW32(ifp->if_obytes);
		return (uint8_t *) & long_return;
	case IFOUTUCASTPKTS:
		long_return = ifp->if_opackets - ifp->if_omcasts;
		return (uint8_t *) & long_return;
	case IFOUTNUCASTPKTS:
		return (uint8_t *) & ifp->if_omcasts;
	case IFOUTDISCARDS:
		return (uint8_t *) & ifp->if_snd.ifq_drops;
	case IFOUTERRORS:
		return (uint8_t *) & ifp->if_oerrors;
	case IFOUTQLEN:
		return (uint8_t *) ifp->if_snd.ifq_len_p;
	default:
		DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_ifEntry\n",
			    vp->magic));
	}
	return NULL;
}
