/*
 *
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */

#include "ip_ipv6.h"
#include <sk/ha_host.h>


STATIC FindVarMethod var_ipv6MIBObjects;
STATIC FindVarMethod var_ipv6IfEntry;
STATIC FindVarMethod var_ipv6IfStatsEntry;
STATIC FindVarMethod var_ipv6AddrPrefixEntry;
STATIC FindVarMethod var_ipv6AddrEntry;
STATIC FindVarMethod var_ipv6RouteEntry;
STATIC FindVarMethod var_ipv6NetToMediaEntry;

STATIC FindVarMethod _var_ipv6MIBObjects;
STATIC FindVarMethod _var_ipv6IfEntry;
STATIC FindVarMethod _var_ipv6IfStatsEntry;
STATIC FindVarMethod _var_ipv6AddrPrefixEntry;
STATIC FindVarMethod _var_ipv6AddrEntry;
STATIC FindVarMethod _var_ipv6RouteEntry;
STATIC FindVarMethod _var_ipv6NetToMediaEntry;

#define IPV6MIBOBJECTS_OID      { SNMP_OID_MIB2, 55, 1 }
#define IPV6IFENTRY_OID         { SNMP_OID_MIB2, 55, 1, 5, 1 }
#define IPV6IFSTATSENTRY_OID    { SNMP_OID_MIB2, 55, 1, 6, 1 }
#define IPV6ADDRPREFIXENTRY_OID { SNMP_OID_MIB2, 55, 1, 7, 1 }
#define IPV6ADDRENTRY_OID       { SNMP_OID_MIB2, 55, 1, 8, 1 }
#define IPV6ROUTEENTRY_OID      { SNMP_OID_MIB2, 55, 1, 11, 1 }
#define IPV6NETTOMEDIAENTRY_OID { SNMP_OID_MIB2, 55, 1, 12, 1 }

struct variable1 ipv6IfEntry_variables[11] = {
    {IPV6IFINDEX, INTEGER, RONLY, var_ipv6IfEntry, 1, {1}},
    {IPV6IFDESC, STRING, RONLY, var_ipv6IfEntry, 1, {2}},
    {IPV6IFLOWERLAYER, VARIABLEPOINTER, RONLY, var_ipv6IfEntry, 1, {3}},
    {IPV6IFEFFECTIVEMTU, INTEGER, RONLY, var_ipv6IfEntry, 1, {4}},
    {IPV6IFREASMMAXSIZE, INTEGER, RONLY, var_ipv6IfEntry, 1, {5}},
    {IPV6IFIDENTIFIER, INETADDRESS, RONLY, var_ipv6IfEntry, 1, {6}},
    {IPV6IFIDENTIFIERLENGTH, INTEGER, RONLY, var_ipv6IfEntry, 1, {7}},
    {IPV6IFPHYSICALADDRESS, PHYSADDRESS, RONLY, var_ipv6IfEntry, 1, {8}},
    {IPV6IFADMINSTATUS, INTEGER, RONLY, var_ipv6IfEntry, 1, {9}},
    {IPV6IFOPERSTATUS, INTEGER, RONLY, var_ipv6IfEntry, 1, {10}},
    {IPV6IFLASTCHANGE, INTEGER, RONLY, var_ipv6IfEntry, 1, {11}}
};
oid ipv6IfEntry_variables_oid[10] = IPV6IFENTRY_OID;

struct variable1 ipv6IfStatsEntry_variables[20] = {
    {IPV6IFSTATSINRECEIVES, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {1}},
    {IPV6IFSTATSINHDRERRORS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {2}},
    {IPV6IFSTATSINTOOBIGERRORS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {3}},
    {IPV6IFSTATSINNOROUTES, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {4}},
    {IPV6IFSTATSINADDRERRORS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {5}},
    {IPV6IFSTATSINUNKNOWNPROTOS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {6}},
    {IPV6IFSTATSINTRUNCATEDPKTS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {7}},
    {IPV6IFSTATSINDISCARDS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {8}},
    {IPV6IFSTATSINDELIVERS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {9}},
    {IPV6IFSTATSOUTFORWDATAGRAMS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {10}},
    {IPV6IFSTATSOUTREQUESTS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {11}},
    {IPV6IFSTATSOUTDISCARDS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {12}},
    {IPV6IFSTATSOUTFRAGOKS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {13}},
    {IPV6IFSTATSOUTFRAGFAILS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {14}},
    {IPV6IFSTATSOUTFRAGCREATES, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {15}},
    {IPV6IFSTATSREASMREQDS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {16}},
    {IPV6IFSTATSREASMOKS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {17}},
    {IPV6IFSTATSREASMFAILS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {18}},
    {IPV6IFSTATSINMCASTPKTS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {19}},
    {IPV6IFSTATSOUTMCASTPKTS, COUNTER32, RONLY,
                                        var_ipv6IfStatsEntry, 1, {20}}
};
oid ipv6IfStatsEntry_variables_oid[10] = IPV6IFSTATSENTRY_OID;

struct variable1 ipv6AddrPrefixEntry_variables[6] = {
    {IPV6ADDRPREFIX, IPV6ADDRESSPREFIX, RONLY,
                                        var_ipv6AddrPrefixEntry, 1, {1}},
    {IPV6ADDRPREFIXLENGTH, INTEGER, RONLY,
                                        var_ipv6AddrPrefixEntry, 1, {2}},
    {IPV6ADDRPREFIXONLINKFLAG, TRUTHVALUE, RONLY,
                                        var_ipv6AddrPrefixEntry, 1, {3}},
    {IPV6ADDRPREFIXAUTONOMOUSFLAG, TRUTHVALUE, RONLY,
                                        var_ipv6AddrPrefixEntry, 1, {4}},
    {IPV6ADDRPREFIXADVPREFERREDLIFETIME, UNSIGNED32, RONLY,
                                        var_ipv6AddrPrefixEntry, 1, {5}},
    {IPV6ADDRPREFIXADVVALIDLIFETIME, UNSIGNED32, RONLY,
                                        var_ipv6AddrPrefixEntry, 1, {6}}
};
oid ipv6AddrPrefixEntry_variables_oid[10] = IPV6ADDRPREFIXENTRY_OID;

struct variable1 ipv6AddrEntry_variables[5] = {
    {IPV6ADDRADDRESS, INETADDRESS, RONLY, var_ipv6AddrEntry, 1, {1}},
    {IPV6ADDRPFXLENGTH, INTEGER, RONLY, var_ipv6AddrEntry, 1, {2}},
    {IPV6ADDRTYPE, INTEGER, RONLY, var_ipv6AddrEntry, 1, {3}},
    {IPV6ADDRANYCASTFLAG, TRUTHVALUE, RONLY, var_ipv6AddrEntry, 1, {4}},
    {IPV6ADDRSTATUS, INTEGER, RONLY, var_ipv6AddrEntry, 1, {5}}
};
oid ipv6AddrEntry_variables_oid[10] = IPV6ADDRENTRY_OID;

struct variable1 ipv6RouteEntry_variables[14] = {
    {IPV6ROUTEDEST, INETADDRESS, RONLY, var_ipv6RouteEntry, 1, {1}},
    {IPV6ROUTEPFXLENGTH, INTEGER, RONLY, var_ipv6RouteEntry, 1, {2}},
    {IPV6ROUTEINDEX, UNSIGNED32, RONLY, var_ipv6RouteEntry, 1, {3}},
    {IPV6ROUTEIFINDEX, IPV6IFINDEXORZERO, RONLY, var_ipv6RouteEntry, 1, {4}},
    {IPV6ROUTENEXTHOP, INETADDRESS, RONLY, var_ipv6RouteEntry, 1, {5}},
    {IPV6ROUTETYPE, INTEGER, RONLY, var_ipv6RouteEntry, 1, {6}},
    {IPV6ROUTEPROTOCOL, INTEGER, RONLY, var_ipv6RouteEntry, 1, {7}},
    {IPV6ROUTEPOLICY, INTEGER32, RONLY, var_ipv6RouteEntry, 1, {8}},
    {IPV6ROUTEAGE, UNSIGNED32, RONLY, var_ipv6RouteEntry, 1, {9}},
    {IPV6ROUTENEXTHOPRDI, UNSIGNED32, RONLY, var_ipv6RouteEntry, 1, {10}},
    {IPV6ROUTEMETRIC, UNSIGNED32, RONLY, var_ipv6RouteEntry, 1, {11}},
    {IPV6ROUTEWEIGHT, UNSIGNED32, RONLY, var_ipv6RouteEntry, 1, {12}},
    {IPV6ROUTEINFO, ROWPOINTER, RONLY, var_ipv6RouteEntry, 1, {13}},
    {IPV6ROUTEVALID, TRUTHVALUE, RONLY, var_ipv6RouteEntry, 1, {14}}
};

oid ipv6RouteEntry_variables_oid[10] = IPV6ROUTEENTRY_OID;

struct variable1 ipv6NetToMediaEntry_variables[6] = {
    {IPV6NETTOMEDIANETADDRESS, INETADDRESS, RONLY,
                                        var_ipv6NetToMediaEntry, 1, {1}},
    {IPV6NETTOMEDIAPHYSADDRESS, PHYSADDRESS, RONLY,
                                        var_ipv6NetToMediaEntry, 1, {2}},
    {IPV6NETTOMEDIATYPE, INTEGER, RONLY,
                                        var_ipv6NetToMediaEntry, 1, {3}},
    {IPV6IFNETTOMEDIASTATE, INTEGER, RONLY,
                                        var_ipv6NetToMediaEntry, 1, {4}},
    {IPV6IFNETTOMEDIALASTUPDATED, TIMESTAMP, RONLY,
                                        var_ipv6NetToMediaEntry, 1, {5}},
    {IPV6NETTOMEDIAVALID, TRUTHVALUE, RONLY,
                                        var_ipv6NetToMediaEntry, 1, {6}}
};
oid ipv6NetToMediaEntry_variables_oid[10] = IPV6NETTOMEDIAENTRY_OID;

struct variable1 ipv6MIBObjects_variables[6] = {
    {IPV6FORWARDING, INTEGER, RONLY, var_ipv6MIBObjects, 1, {1}},
    {IPV6DEFAULTHOPLIMIT, INTEGER, RONLY, var_ipv6MIBObjects, 1, {2}},
    {IPV6INTERFACES, UNSIGNED32, RONLY, var_ipv6MIBObjects, 1, {3}},
    {IPV6IFTABLELASTCHANGE, TIMESTAMP, RONLY, var_ipv6MIBObjects, 1, {4}},
    {IPV6ROUTENUMBER, GAUGE32, RONLY, var_ipv6MIBObjects, 1, {9}},
    {IPV6DISCARDEDROUTES, COUNTER32, RONLY, var_ipv6MIBObjects, 1, {10}}
};
oid ipv6MIBObjects_variables_oid[8] = IPV6MIBOBJECTS_OID;

struct in6_addr addr;
uint8_t phyaddr[1024];

void init_ipv6(void)
{
    REGISTER_MIB("mib_2/ipv6MIBObjects",
            ipv6MIBObjects_variables, variable1,
            ipv6MIBObjects_variables_oid);
    REGISTER_MIB("ipv6IfEntry",
            ipv6IfEntry_variables, variable1,
            ipv6IfEntry_variables_oid);
    REGISTER_MIB("ipv6IfStatsEntry",
            ipv6IfStatsEntry_variables, variable1,
            ipv6IfStatsEntry_variables_oid);
    REGISTER_MIB("ipv6AddrPrefixEntry",
            ipv6AddrPrefixEntry_variables, variable1,
            ipv6AddrPrefixEntry_variables_oid);
    REGISTER_MIB("ipv6AddrEntry",
            ipv6AddrEntry_variables, variable1,
            ipv6AddrEntry_variables_oid);
    REGISTER_MIB("ipv6RouteEntry",
            ipv6RouteEntry_variables, variable1,
            ipv6RouteEntry_variables_oid);
    REGISTER_MIB("ipv6NetToMediaEntry",
            ipv6NetToMediaEntry_variables, variable1,
            ipv6NetToMediaEntry_variables_oid);
}

uint8_t *
_var_ipv6MIBObjects(SNMP_FM_ARGS)
{
    switch (vp->magic) {

        case IPV6FORWARDING:
            SK_DOMAIN_REFERENCE(ip6_forwarding, SK_DOMAIN_NETWORK);
            if (ip6_forwarding) {
                long_return = 1;        /*GATEWAY*/
            }
            else {
                long_return = 2;        /*HOST*/
            }
            return (uint8_t *)&long_return;

        case IPV6DEFAULTHOPLIMIT:
            SK_DOMAIN_REFERENCE(ip6_defhlim, SK_DOMAIN_NETWORK);
            long_return = ip6_defhlim;
            return (uint8_t *)&long_return;

        case IPV6INTERFACES:
            long_return = Interface_Scan_Get_Count(AF_INET6);
            if (long_return < 0) {
                break;
            }
            return (uint8_t *)&long_return;

        case IPV6IFTABLELASTCHANGE:
        case IPV6ROUTENUMBER:
        case IPV6DISCARDEDROUTES:
            /* not implemented */
            long_return = 0;
            return (uint8_t *) & long_return;

        default:
            break;
    }

    return NULL;
}

uint8_t *
var_ipv6MIBObjects(SNMP_FM_ARGS)
{
    if (!ipv6_enabled) {
        return NULL;
    }

    if (header_generic(vp, name, length, exact, var_len, write_method) ==
            MATCH_FAILED) {
        return NULL;
    }

    DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6MIBObjects,
            (vp, name, length, exact, var_len, write_method));
}

/* format: .1.3.6.1.2.1.55.1.5.1.?.interface_id */
uint8_t *
_var_ipv6IfEntry(SNMP_FM_ARGS)
{
    int32_t max = Interface_Scan_Get_Count(AF_INET6);
    oid newname[MAX_OID_LEN], lowest[MAX_OID_LEN];
    bool found = false;
    struct ifnet *ifp = NULL, *lowin_ifp = NULL;

    if (max == 0) {
        return NULL;
    }

    bzero(newname, sizeof(newname));
    bzero(lowest, sizeof(lowest));	/* Burt545219: Lint 645 warning */
    bcopy(vp->name, newname, (int32_t)vp->namelen * sizeof(oid));

    SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
    for (ifp = ifnet; ifp; ifp = ifp->if_next) {
	/* Skip virtual host (vh) entries */
        if(!(strncmp(ifp->if_fullname, "vh", 2))) {
                continue;
        }
        newname[11] = ifp->if_index;
        if(exact) {
            if(snmp_oid_compare(newname, 12, name, *length) == 0) {
                bcopy((char *) newname,
                        (char *) lowest,
                        12 * sizeof(oid));
                found = true;
                lowin_ifp = ifp;
                break;  /* no need to search further */
            }
        }
        else {
            if ((snmp_oid_compare(newname, 12, name, *length) > 0) &&
                    (!found || (snmp_oid_compare(newname, 12, lowest, 12) < 0))) {
                found = true;
                lowin_ifp = ifp;
                bcopy((char *) newname,
                        (char *) lowest,
                        12 * sizeof(oid));
            }
        }
    }
    if (!lowin_ifp) {
        return NULL;
    }

    bcopy(lowest, name, 12 * sizeof(oid));
    *length = 12;
    *write_method = 0;
    *var_len = sizeof(long_return); /* default */

    switch (vp->magic) {

        case IPV6IFINDEX:
            long_return = lowin_ifp->if_index;
            return (uint8_t *) &long_return;

        case IPV6IFDESC:
        {
#define MIBIFNAMSIZ 2*IFNAMSIZ
            static char if_name[MIBIFNAMSIZ];
	    char *Name_ptr = if_name;

	    if (lowin_ifp->if_flags2 & IFF2_MGMT_PORTS){
		strlcpy(Name_ptr, "MGMT_PORT_ONLY ", IFNAMSIZ);
		Name_ptr += strlen(Name_ptr);
	    }

	    strlcpy(Name_ptr, lowin_ifp->if_fullname, IFNAMSIZ);
            *var_len = strlen(if_name);
            return (uint8_t *) &if_name;
        }

        case IPV6IFLOWERLAYER:
        {
			/* Format: OID: {0, if_index} */
            if(lowin_ifp->if_flags & IFF_UP) {
				static oid val[2];
				val[0] = 0;
				val[1] = lowin_ifp->if_index;
				*var_len = sizeof(val);
				return (uint8_t *) &val;
            }
            else {
                *var_len = nullOidLen;
                return (uint8_t *)&nullOid;
            }
        }

        case IPV6IFEFFECTIVEMTU:
            long_return = lowin_ifp->if_mtu;
            return (uint8_t *) &long_return;

        case IPV6IFREASMMAXSIZE:
            /* ONTAP doesnt support jumbo datagrams yet */
            long_return = IPV6_MAXPACKET;
            return (uint8_t *) &long_return;

        case IPV6IFIDENTIFIER:
        {
			static struct in6_addr addr;
			bzero(&addr, sizeof(addr));
			SK_DOMAIN_REFERENCE(get_hw_ifid, SK_DOMAIN_NETWORK);
			/* Get the interface id */
			get_hw_ifid(lowin_ifp, &addr);
			*var_len = sizeof(addr);
			return (uint8_t *) &addr;
        }

        case IPV6IFIDENTIFIERLENGTH:
        {
            *var_len = sizeof(long_return);
	    /* Interface Identifier length is 64 */
	    long_return = 64;

            return (uint8_t *) &long_return;
        }


        case IPV6IFPHYSICALADDRESS:
        {
            static uint8_t addr[6];
            struct arpcom *ap = NULL;

            ap = (struct arpcom *) lowin_ifp;
            if (lowin_ifp->if_addrlen != sizeof(ap->ac_enaddr)) {
                /*
                * we only support interfaces with 802.x-style
                * 6-byte addresses.  This needs to be fixed.
                */
                long_return = 0;
                return (uint8_t *) &long_return;
            }
            *var_len = lowin_ifp->if_addrlen;
            bcopy((char*)ap->ac_enaddr, addr, sizeof(ap->ac_enaddr));
            return (uint8_t *) addr;
        }

        case IPV6IFADMINSTATUS:
            long_return = (lowin_ifp->if_flags & IFF_RUNNING) ? 1 : 2;
            return (uint8_t *)&long_return;

        case IPV6IFOPERSTATUS:
            long_return = (lowin_ifp->if_flags & IFF_UP) ? 1 : 2;
            return (uint8_t *)&long_return;

        case IPV6IFLASTCHANGE:
            /* not implemented */
            long_return = 0;
            return (uint8_t *)&long_return;

        default:
            break;
    }
    return NULL;
}

uint8_t *
var_ipv6IfEntry(SNMP_FM_ARGS)
{
    if (!ipv6_enabled) {
        return NULL;
    }

    DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6IfEntry,
            (vp, name, length, exact, var_len, write_method));
}

/* format: .1.3.6.1.2.1.55.1.6.1.?.interface_id */
uint8_t *
_var_ipv6IfStatsEntry(SNMP_FM_ARGS)
{
    int32_t max = Interface_Scan_Get_Count(AF_INET6);
    oid newname[MAX_OID_LEN], lowest[MAX_OID_LEN];
    bool found = false;
    struct in6_ifstat *ifs6 = NULL;
    struct in6_ifreq ifr;
    struct ifnet *ifp = NULL, *lowin_ifp = NULL;
    struct socket *s = NULL;

    if (max == 0) {
        return NULL;
    }

    bzero(newname, sizeof(newname));
    bzero(lowest, sizeof(lowest));	/* Burt545220: Lint 645 warning */
    bcopy(vp->name, newname, (int32_t)vp->namelen * sizeof(oid));

    SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
    for (ifp = ifnet; ifp; ifp = ifp->if_next) {
	/* Skip virtual host (vh) entries */
        if(!(strncmp(ifp->if_fullname, "vh", 2))) {
                continue;
        }
        newname[11] = ifp->if_index;
        if(exact) {
            if(snmp_oid_compare(newname, 12, name, *length) == 0) {
                bcopy((char *) newname,
                        (char *) lowest,
                        12 * sizeof(oid));
                found = true;
                lowin_ifp = ifp;
                break;  /* no need to search further */
            }
        }
        else {
            if ((snmp_oid_compare(newname, 12, name, *length) > 0) &&
                    (!found || (snmp_oid_compare(newname, 12, lowest, 12) < 0))) {
                found = true;
                lowin_ifp = ifp;
                bcopy((char *) newname,
                        (char *) lowest,
                        12 * sizeof(oid));
            }
        }
    }
    if (!lowin_ifp) {
        return NULL;
    }

    bcopy(lowest, name, 12 * sizeof(oid));
    *length = 12;
    *write_method = 0;
    *var_len = sizeof(long_return); /* default */

    memset(&ifr, 0, sizeof(ifr));
    strlcpy(ifr.ifr_name, lowin_ifp->if_fullname, sizeof(ifr.ifr_name));

    if (!(s = socket(AF_INET6, SOCK_DGRAM, 0, NULL))) {
        return NULL;
    }

    SK_DOMAIN_REFERENCE(ifioctl, SK_DOMAIN_NETWORK);
    if (ifioctl(s, SIOCGIFSTAT_IN6, (caddr_t)&ifr) != 0) {
        soclose(s);
        return NULL;
    }
    soclose(s);

    ifs6 = &ifr.ifr_ifru.ifru_stat;

    if(!ifs6) {
        return NULL;
    }

    switch (vp->magic) {

        case IPV6IFSTATSINRECEIVES:
            long_return = ifs6->ifs6_in_receive;
            break;

        case IPV6IFSTATSINHDRERRORS:
            long_return = ifs6->ifs6_in_hdrerr;
            break;

        case IPV6IFSTATSINTOOBIGERRORS:
            long_return = ifs6->ifs6_in_toobig;
            break;

        case IPV6IFSTATSINNOROUTES:
            long_return = ifs6->ifs6_in_noroute;
            break;

        case IPV6IFSTATSINADDRERRORS:
            long_return = ifs6->ifs6_in_addrerr;
            break;

        case IPV6IFSTATSINUNKNOWNPROTOS:
            long_return = ifs6->ifs6_in_protounknown;
            break;

        case IPV6IFSTATSINTRUNCATEDPKTS:
            long_return = ifs6->ifs6_in_truncated;
            break;

        case IPV6IFSTATSINDISCARDS:
            long_return = ifs6->ifs6_in_discard;
            break;

        case IPV6IFSTATSINDELIVERS:
            long_return = ifs6->ifs6_in_deliver;
            break;

        case IPV6IFSTATSOUTFORWDATAGRAMS:
            long_return = ifs6->ifs6_out_forward;
            break;

        case IPV6IFSTATSOUTREQUESTS:
            long_return = ifs6->ifs6_out_request;
            break;

        case IPV6IFSTATSOUTDISCARDS:
            long_return = ifs6->ifs6_out_discard;
            break;

        case IPV6IFSTATSOUTFRAGOKS:
            long_return = ifs6->ifs6_out_fragok;
            break;

        case IPV6IFSTATSOUTFRAGFAILS:
            long_return = ifs6->ifs6_out_fragfail;
            break;

        case IPV6IFSTATSOUTFRAGCREATES:
            long_return = ifs6->ifs6_out_fragcreat;
            break;

        case IPV6IFSTATSREASMREQDS:
            long_return = ifs6->ifs6_reass_reqd;
            break;

        case IPV6IFSTATSREASMOKS:
            long_return = ifs6->ifs6_reass_ok;
            break;

        case IPV6IFSTATSREASMFAILS:
            long_return = ifs6->ifs6_reass_fail;
            break;

        case IPV6IFSTATSINMCASTPKTS:
            long_return = ifs6->ifs6_in_mcast;
            break;

        case IPV6IFSTATSOUTMCASTPKTS:
            long_return = ifs6->ifs6_out_mcast;
            break;

        default:
            return NULL;
    }
    return (uint8_t *)&long_return;
}

uint8_t *
var_ipv6IfStatsEntry(SNMP_FM_ARGS)
{
    if (!ipv6_enabled) {
        return NULL;
    }

    DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6IfStatsEntry,
            (vp, name, length, exact, var_len, write_method));
}

/* the format is:
 * .1.3.6.1.2.1.55.1.7.1.?.interface_id.prefix_addr.prefix_len
 */
uint8_t *
_var_ipv6AddrPrefixEntry(SNMP_FM_ARGS)
{
    int32_t max = Interface_Scan_Get_Count(AF_INET6);
    oid current[MAX_OID_LEN], lowest[MAX_OID_LEN];
    int32_t lowest_len = 0;
    struct nd_prefix *ndpr = NULL;
    static struct nd_prefix low_addrpr;
    struct ifnet *ifp  = NULL;

    if ((max == 0) || vp->namelen < 11) {
        return NULL;
    }

    bzero(current, sizeof(current));
    bzero(lowest, sizeof(lowest));
    memcpy((char *)current, (char *)vp->name, 11 * sizeof(oid));

    SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
    for (ifp = ifnet; ifp; ifp = ifp->if_next) {
	/* Skip virtual host (vh) entries */
        if(!(strncmp(ifp->if_fullname, "vh", 2))) {
                continue;
        }
        current[11] = ifp->if_index;

        SK_DOMAIN_REFERENCE(nd_prefix, SK_DOMAIN_NETWORK);
        for (ndpr = nd_prefix[sk_gethost()].lh_first; ndpr;
                                ndpr = ndpr->ndpr_next) {
            int32_t i = 0, j = 12;

            if (ndpr->ndpr_ifp->if_index != ifp->if_index) {
                continue;
            }

#define CEILING(a, b) ((a/b) + ((a % b) ? 1 : 0))
            for (i = 0;
                    i < sizeof(struct in6_addr) &&
                    i < CEILING(ndpr->ndpr_plen, 8);
                    i++) {
                current[j++] = ndpr->ndpr_prefix.sin6_addr.s6_addr[i];
            }
#undef CEILING

            /* insert prefix length */
            current[j++] = ndpr->ndpr_plen;

            ASSERT(j <= 29);

            if (exact) {
                if (snmp_oid_compare(current, j, name, *length) == 0) {
                    bcopy((char *) current,
                            (char *) lowest,
                            j * sizeof(oid));
                    low_addrpr = *ndpr;
                    lowest_len = j;
                    break;  /* no need to search further */
                }
            } else {
                if ((snmp_oid_compare(current, j, name, *length) > 0) &&
                        (!lowest_len ||
                         (snmp_oid_compare(current, j, lowest, lowest_len)
                                                < 0))) {
                    /*
                     * if new one is greater than input and
                     * closer to input than previous lowest
                     * , save this one as the "next" one.
                     */
                    low_addrpr = *ndpr;
                    lowest_len = j;
                    bcopy((char *) current,
                            (char *) lowest,
                            j * sizeof(oid));
                }
            }
        }
    }

    if (lowest_len == 0) {
        return (NULL);
    }

    bcopy((char *) lowest, (char *) name, lowest_len * sizeof(oid));
    *length = lowest_len;
    *write_method = 0;
    *var_len = sizeof(long_return);

    switch (vp->magic) {

        case IPV6ADDRPREFIX:
            *var_len = sizeof(low_addrpr.ndpr_prefix.sin6_addr);
            return (uint8_t *) &low_addrpr.ndpr_prefix.sin6_addr;

        case IPV6ADDRPREFIXLENGTH:
            long_return = low_addrpr.ndpr_plen;
            return (uint8_t *) &long_return;

        case IPV6ADDRPREFIXONLINKFLAG:
            long_return = low_addrpr.ndpr_raf_onlink ? 1 : 2;
            return (uint8_t *) &long_return;

        case IPV6ADDRPREFIXAUTONOMOUSFLAG:
            long_return = low_addrpr.ndpr_raf_auto ? 1: 2;
            return (uint8_t *) &long_return;

        case IPV6ADDRPREFIXADVPREFERREDLIFETIME:
            long_return = low_addrpr.ndpr_pltime;
            return (uint8_t *) &long_return;

        case IPV6ADDRPREFIXADVVALIDLIFETIME:
            long_return = low_addrpr.ndpr_vltime;
            return (uint8_t *) &long_return;

        default:
            break;
    }

    return NULL;
}

uint8_t *
var_ipv6AddrPrefixEntry(SNMP_FM_ARGS)
{
    if (!ipv6_enabled) {
        return NULL;
    }

    DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6AddrPrefixEntry,
            (vp, name, length, exact, var_len, write_method));
}

/* the format is:
 * .1.3.6.1.2.1.55.1.8.1.?.interface_id.ipv6_addr
 */

uint8_t *
_var_ipv6AddrEntry(SNMP_FM_ARGS)
{
    int32_t max = Interface_Scan_Get_Count(AF_INET6);
    oid current[MAX_OID_LEN], lowest[MAX_OID_LEN];
    struct ifnet *ifp = NULL;
    bool found = false;
    struct ifaddr *ia = NULL;
    struct in6_ifaddr *low_ia6 = NULL;
    static struct in6_ifaddr ia6;

    if ((max == 0) || vp->namelen < 11) {
        return NULL;
    }

    bzero(current, sizeof(current));
    bzero(lowest, sizeof(lowest));
    memcpy((char *)current, (char *)vp->name, 11 * sizeof(oid));

    SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
    for (ifp = ifnet; ifp; ifp = ifp->if_next) {
	/* Skip virtual host (vh) entries */
        if(!(strncmp(ifp->if_fullname, "vh", 2))) {
                continue;
        }
        current[11] = ifp->if_index;

        for (ia = ifp->if_addrlist; ia; ia = ia->ifa_next) {
            int32_t i = 0, j = 12;
            struct in6_addr sin6_addr;

            if (ia->ifa_addr->sa_family != AF_INET6) {
                continue;
            }

            sin6_addr = ((struct in6_ifaddr *)ia)->ia_addr.sin6_addr;
            if (IN6_IS_ADDR_LINKLOCAL(&sin6_addr)) {
                *(uint16_t *)&sin6_addr.s6_addr[2] = 0;
            }

            for (i = 0;  i < sizeof(struct in6_addr); i++) {
                current[j++] = sin6_addr.s6_addr[i];
            }

            ASSERT(j == 28);

            if (exact) {
                if (snmp_oid_compare(current, j, name, *length) == 0) {
                    bcopy((char *) current,
                            (char *) lowest,
                            j * sizeof(oid));
                    low_ia6 = (struct in6_ifaddr *) ia;
                    found = true;
                    break;  /* no need to search further */
                }
            } else {
                if ((snmp_oid_compare(current, j, name, *length) > 0) &&
                        (!found ||
                         (snmp_oid_compare(current, j, lowest, j) < 0))) {
                    /*
                     * if new one is greater than input and
                     * closer to input than previous lowest
                     *, save this one as the "next" one.
                     */
                    low_ia6 = (struct in6_ifaddr *) ia;
                    bcopy((char *) current,
                            (char *) lowest,
                            j * sizeof(oid));
                    found = true;
                }
            }
        }

    }
    if (found == false) {
        return NULL;
    }

    ia6 = *low_ia6;
    bcopy((char *) lowest, (char *) name, 28 * sizeof(oid));
    *length = 28;
    *write_method = 0;
    *var_len = sizeof(long_return);

    switch (vp->magic) {

        case IPV6ADDRADDRESS:
        {
            *var_len = sizeof(struct in6_addr);

            if (IN6_IS_ADDR_LINKLOCAL(&ia6.ia_addr.sin6_addr)) {
                *(uint16_t *) &ia6.ia_addr.sin6_addr.s6_addr[2] = 0;
            }

            return (uint8_t *) &ia6.ia_addr.sin6_addr.s6_addr[0];
        }

        case IPV6ADDRPFXLENGTH:
        {
            long_return = in6_mask2len(
                        &satosin6(ia6.ia_ifa.ifa_netmask)->sin6_addr);
            return (uint8_t *) &long_return;
        }

        case IPV6ADDRTYPE:
            if (ia6.ia6_flags & IN6_IFF_STATELESS ||
                    IN6_IS_ADDR_LINKLOCAL(&ia6.ia_addr.sin6_addr)) {
                long_return = 1;
            }
            else {
                long_return = 2;
            }
            return (uint8_t *) &long_return;

        case IPV6ADDRANYCASTFLAG:
            /* ONTAP does not support anycast addresses */
            long_return = (ia6.ia6_flags & IN6_IFF_ANYCAST) ? 1 : 2;
            return (uint8_t *) &long_return;

        case IPV6ADDRSTATUS:
#define err_flags (IN6_IFF_DEPRECATED | IN6_IFF_DETACHED | IN6_IFF_NOTREADY)

            switch (ia6.ia6_flags & err_flags)   {

                case IN6_IFF_DEPRECATED:
                    long_return = 2; /* deprecated */
                    break;

                case IN6_IFF_DETACHED:
                    long_return = 4; /* inaccessible */
                    break;

                case IN6_IFF_NOTREADY:
                case IN6_IFF_TENTATIVE:
                case IN6_IFF_DUPLICATED:
                    long_return = 3; /* invalid */
                    break;
                default:
                    if (ia6.ia6_flags & err_flags) {
                        long_return = 5; /* unknown combo state */
                    }
                    else {
                        long_return = 1; /* preferred */
                    }
            }
#undef err_flags
            return (uint8_t *) &long_return;

        default:
            break;
    }

    return NULL;
}

uint8_t *
var_ipv6AddrEntry(SNMP_FM_ARGS)
{
    if (!ipv6_enabled) {
        return NULL;
    }

    DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6AddrEntry,
            (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_ipv6RouteEntry(SNMP_FM_ARGS)
{
    int32_t max = Interface_Scan_Get_Count(AF_INET6);
    oid newname[MAX_OID_LEN], lowest[MAX_OID_LEN];
    int32_t i = 0;
    bool found = false;
    int32_t always_header = 0;
    int32_t error;
    sk_domain_t old_domain;

    if (max == 0) {
        return NULL;
    }
    bzero(newname, sizeof(newname));
    bzero(lowest, sizeof(lowest));
    bcopy(vp->name, newname, (int32_t)vp->namelen * sizeof(oid));

    /*
     * Actual logic for routing info!!!
     */
    old_domain = sk_domain_switch_begin(sk_domain_network);

    struct rt_head rhead;
    int32_t k = 0;
    struct sockaddr_in6 *sa6 = NULL;
    struct rtbuf *low_rtb = NULL;

    SK_DOMAIN_IMPORT(rtable_get, SK_REASON_SWITCH_DOMAIN);

    error = rtable_get(&rhead, AF_INET6,RT_DEFAULT_GRP);
    
    if(rhead.nroutes > 0 || always_header) {
        for(i = 0; i < rhead.rtindex; i++) {
            int32_t j;
            struct rtbuf *rb = &rhead.rb[i];
            struct sockaddr *sa = rtb_dst(rb);
            struct in6_addr *in6;
 
            k = 11;
	    /* Skip if it's not an IPv6 address */
            if(sa->sa_family != AF_INET6) {
               continue;
            }
            sa6 = (struct sockaddr_in6 *)sa;
            in6 = &sa6->sin6_addr;
	    /* Skip if it's a multicast address */
	    if(IN6_IS_ADDR_MULTICAST(in6)) {
	       continue;
	    }	
            /* Destination ip address */
            for(j = 0; j < sizeof(struct in6_addr); j++) {
                newname[k++] = in6->s6_addr[j];
            }
            /* Address prefix length */
            newname[k++] = in6_mask2len(in6);
            /* Route index */
            newname[k++] = rb->rtb_if_index;
            if(exact) {
                if (snmp_oid_compare(newname, k, name, *length) == 0) {
                    bcopy((char *)newname,
                            (char *)lowest,
                            k * sizeof(oid));
                    low_rtb = rb;
                    found = true;
		    *length = k;
                    break;
                }
            } else {
                if ((snmp_oid_compare(newname, k, name, *length) > 0) &&
                        (!found || (snmp_oid_compare(newname, k, lowest, k)
                                            < 0))) {
                    /*
                     * if new one is greater than input and
                     * closer to input than previous lowest
                     *, save this one as the "next" one.
                     */
                    low_rtb = rb;
                    bcopy((char *)newname,
                            (char *)lowest,
                            k * sizeof(oid));
		    *length = k;
                    found = true;
                }
            }
        } /* end of for loop */
    }

    if ((found == 0) || !low_rtb) {
         rtable_free(&rhead);
        sk_domain_switch_end(old_domain);
        return NULL;
    }

    bcopy(lowest, name, (*length * sizeof(oid)));
    *write_method = 0;
    *var_len = sizeof(long_return); /* default */

    switch (vp->magic) {
        case IPV6ROUTEDEST:
	{
	    static struct in6_addr addr;
	    bzero(&addr, sizeof(addr));

            *var_len = sizeof(struct in6_addr);

            sa6 = (struct sockaddr_in6 *)rtb_dst(low_rtb);

            if (IN6_IS_ADDR_LINKLOCAL(&sa6->sin6_addr)) {
                *(uint16_t *) &sa6->sin6_addr.s6_addr[2] = 0;
            }
   	    bcopy(&sa6->sin6_addr.s6_addr,
			&addr, sizeof(addr));
	     rtable_free(&rhead);
            sk_domain_switch_end(old_domain);
            return (uint8_t *) &addr.s6_addr[0];
	}

        case IPV6ROUTEPFXLENGTH:
            sa6 = (struct sockaddr_in6 *)rtb_dst(low_rtb);
            long_return = in6_mask2len(&sa6->sin6_addr);
            break;

        case IPV6ROUTEINDEX:
            long_return = low_rtb->rtb_if_index;
            break;

        case IPV6ROUTEIFINDEX:
            long_return = low_rtb->rtb_if_index;
            break;

        case IPV6ROUTENEXTHOP:
	{
	    static struct in6_addr addr;
	    bzero(&addr, sizeof(addr));

            *var_len = sizeof(struct in6_addr);

            sa6 = (struct sockaddr_in6 *)rtb_gateway(low_rtb);
            if (IN6_IS_ADDR_LINKLOCAL(&sa6->sin6_addr)) {
                *(uint16_t *) &sa6->sin6_addr.s6_addr[2] = 0;
            }
	    bcopy(&sa6->sin6_addr.s6_addr,
			&addr, sizeof(addr));
	     rtable_free(&rhead);
            sk_domain_switch_end(old_domain);
            return (uint8_t *) &addr.s6_addr[0];
	}
        case IPV6ROUTEMETRIC:
            unsigned_return = (unsigned)-1;
	    rtable_free(&rhead);
            sk_domain_switch_end(old_domain);
            return (uint8_t *) &unsigned_return;

        case IPV6ROUTETYPE:
            if(low_rtb->rtb_flags & RTF_LLINFO) {
                long_return = 3;
            }
            else if(low_rtb->rtb_flags & RTF_REJECT) {
                long_return = 2;
            }
            /* Is this right? */
            else if(low_rtb->rtb_flags & RTF_GATEWAY) {
                long_return = 4;
            }
            else {
                long_return = 1;
            }
            break;

        case IPV6ROUTEPROTOCOL:
            long_return = 1;
            break;

        case IPV6ROUTEPOLICY:
        case IPV6ROUTEAGE:
        case IPV6ROUTENEXTHOPRDI:
        case IPV6ROUTEWEIGHT:
        case IPV6ROUTEINFO:
            long_return = 0;
            break;

        case IPV6ROUTEVALID:
            if(low_rtb->rtb_flags & RTF_REJECT) {
                long_return = 2;
            }
            else {
                long_return = 1;
            }
            break;

        default:
            break;
    }
    rtable_free(&rhead);
    sk_domain_switch_end(old_domain);
    return (uint8_t *) & long_return;

}

uint8_t *
var_ipv6RouteEntry(SNMP_FM_ARGS)
{
    if (!ipv6_enabled) {
        return NULL;
    }

    DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6RouteEntry,
            (vp, name, length, exact, var_len, write_method));
}

/* the format is:
 * .1.3.6.1.2.1.55.1.12.1.?.interface_index.net_address
 */
uint8_t *
_var_ipv6NetToMediaEntry(SNMP_FM_ARGS)
{
    int32_t max = Interface_Scan_Get_Count(AF_INET6);
    oid current[MAX_OID_LEN], lowest[MAX_OID_LEN];
    bool found = false;
    struct ifnet *ifp = NULL, *lowin_ifp = NULL;
    struct in6_ifaddr *lowin_ifaddr = NULL;
    struct ifaddr *ifa = NULL;
    struct arpcom *ap = NULL;

    if ((max == 0) || vp->namelen < 11) {
        return NULL;
    }

    bzero(current, sizeof(current));
    bzero(lowest, sizeof(lowest));
    memcpy((char *)current, (char *)vp->name, 11 * sizeof(oid));

    SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
    for (ifp = ifnet; ifp; ifp = ifp->if_next) {
	/* Skip virtual host (vh) entries */
        if(!(strncmp(ifp->if_fullname, "vh", 2))) {
                continue;
        }
        current[11] = ifp->if_index;

        for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
            int32_t i = 0, j = 12;
            struct in6_addr sin6_addr;

            if (ifa->ifa_addr->sa_family != AF_INET6) {
                continue;
            }

            sin6_addr = ((struct in6_ifaddr *)ifa)->ia_addr.sin6_addr;

            if (IN6_IS_ADDR_LINKLOCAL(&sin6_addr)) {
                *(uint16_t *)&sin6_addr.s6_addr[2] = 0;
            }

            for (i = 0; i < sizeof(struct in6_addr); i++) {
                current[j++] =
                    ((struct in6_ifaddr *)ifa)->ia_addr.sin6_addr.s6_addr[i];
            }

            ASSERT(j == 28);

            if (exact) {
                if (snmp_oid_compare(current, j, name, *length) == 0) {
                    bcopy((char *) current,
                            (char *) lowest,
                            j * sizeof(oid));
                    found = true;
                    lowin_ifaddr = (struct in6_ifaddr *) ifa;
                    lowin_ifp = ifp;
                    break;  /* no need to search further */
                }
            } else {
                if ((snmp_oid_compare(current, j, name, *length) > 0) &&
                        (!found || (snmp_oid_compare(current, j, lowest, j)
                                                < 0))) {
                    /*
                     * if new one is greater than input and
                     * closer to input than previous lowest
                     * , save this one as the "next" one.
                     */
                    found = true;
                    lowin_ifaddr = (struct in6_ifaddr*) ifa;
                    lowin_ifp = ifp;
                    bcopy((char *) current,
                            (char *) lowest,
                            j * sizeof(oid));
                }
            }
        }
    }

    if (found == false) {
        return NULL;
    }

    bcopy((char *) lowest, (char *) name, 28 * sizeof(oid));
    *length = 28;
    *write_method = 0;
    *var_len = sizeof(long_return);

    switch (vp->name[10]) {

        case IPV6NETTOMEDIANETADDRESS:
        {
            bcopy(&lowin_ifaddr->ia_addr.sin6_addr.s6_addr,
                    &addr, sizeof(addr));

            if (IN6_IS_ADDR_LINKLOCAL(&addr)) {
                *(uint16_t *) &addr.s6_addr[2] = 0;
            }

            *var_len = sizeof(addr);
            return (uint8_t *) &addr.s6_addr[0];
        }

        case IPV6NETTOMEDIAPHYSADDRESS:
        {
            ap = (struct arpcom *) lowin_ifp;
            if (lowin_ifp->if_addrlen != sizeof(ap->ac_enaddr)) {
                /*
                * we only support interfaces with 802.x-style
                * 6-byte addresses.  This needs to be fixed.
                */
                long_return = 0;
                return (uint8_t *) &long_return;
            }
            *var_len = lowin_ifp->if_addrlen;
            bcopy((char*)&ap->ac_enaddr, phyaddr, sizeof(ap->ac_enaddr));
            return (uint8_t *) phyaddr;
        }

        case IPV6NETTOMEDIATYPE:
            if (IN6_IS_ADDR_LINKLOCAL(&lowin_ifaddr->ia_addr.sin6_addr)) {
                long_return = 4; /* local */
            } else if (lowin_ifaddr->ia6_flags & IN6_IFF_STATELESS) {
                long_return = 2;  /* dynamic */
            } else {
                long_return = 3; /* static */
            }

            return (uint8_t *) &long_return;

        case IPV6IFNETTOMEDIASTATE:
        {
            /* Not implemented */
            struct in6_nbrinfo *nbi;
            cmd_result_t result;

	    SK_DOMAIN_IMPORT(getnbrinfo, SK_REASON_SWITCH_DOMAIN);

            if ((nbi = getnbrinfo(&lowin_ifaddr->ia_addr.sin6_addr,
                            (int32_t)lowin_ifp->if_index, 1, RT_DEFAULT_GRP, &result))
                    && result == cmdOkReturn) {
                switch (nbi->state) {
                    case ND6_LLINFO_REACHABLE:
                        long_return = 1; /* reachable */
                        break;
                    case ND6_LLINFO_STALE:
                        long_return = 2; /* stale */
                        break;
                    case ND6_LLINFO_DELAY:
                        long_return = 3; /* delay */
                        break;
                    case ND6_LLINFO_PROBE:
                        long_return = 4; /* probe */
                        break;
		    /*case ND6_LLINFO_WAITDELETE: 
			long_return = 5;
			break;*/
                    case ND6_LLINFO_NOSTATE:
                    case ND6_LLINFO_INCOMPLETE:
                    default:
                        long_return = 6; /* unknown */
                        break;
                }
            } else {
                long_return = 6;
            }
            return (uint8_t *) &long_return;
        }

        case IPV6IFNETTOMEDIALASTUPDATED:
            /* not implemented */
            long_return = 0;
            return (uint8_t *) &long_return;

        case IPV6NETTOMEDIAVALID:
            /* not implemented */
            long_return = 0;
            return (uint8_t *) &long_return;

        default:
            break;
    }

    return NULL;
}

uint8_t *
var_ipv6NetToMediaEntry(SNMP_FM_ARGS)
{
    if (!ipv6_enabled) {
        return NULL;
    }

    DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6NetToMediaEntry,
            (vp, name, length, exact, var_len, write_method));
}

