/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */

/*
  ---- Part 1: CMU/UCD copyright notice: (BSD like) -----
  
 
        Copyright 1989, 1991, 1992 by Carnegie Mellon University
 
                   Derivative Work - 1996, 1998-2000
 Copyright 1996, 1998-2000 The Regents of the University of California
 
                          All Rights Reserved
 
 Permission to use, copy, modify and distribute this software and its
 documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appears in all copies and
 that both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of CMU and The Regents of
 the University of California not be used in advertising or publicity
 pertaining to distribution of the software without specific written
 permission.
 
 CMU AND THE REGENTS OF THE UNIVERSITY OF CALIFORNIA DISCLAIM ALL
 WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL CMU OR
 THE REGENTS OF THE UNIVERSITY OF CALIFORNIA BE LIABLE FOR ANY SPECIAL,
 INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 FROM THE LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
 CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
 
 ---- Part 2: Networks Associates Technology, Inc copyright notice (BSD) -----
 
 Copyright (c) 2001-2002, Networks Associates Technology, Inc
 All rights reserved.
  
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
  
 *  Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  
 *  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  
 *  Neither the name of the Networks Associates Technology, Inc nor the
    names of its contributors may be used to endorse or promote
    products derived from this software without specific prior written
    permission.
  
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
 IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 ---- Part 3: Cambridge Broadband Ltd. copyright notice (BSD) -----
 
 Portions of this code are copyright (c) 2001-2002, Cambridge Broadband Ltd.
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
  
 *  Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  
 *  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  
 *  The name of Cambridge Broadband Ltd. may not be used to endorse or
    promote products derived from this software without specific prior
    written permission.
  
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND ANY
 EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
*/

/*
 *  IP MIB group implementation - ipv6.c
 *
 */


#include <legacy/string.h>
#include <legacy/stdlib.h>

#include <sk/sk_common.h>

#include <sys/types.h>
#include <sys/param.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/sys/sk_socketvar.h>
#include <net7g/sys/sk_sockio.h>
#include <net7g/sys/sk_sockio.h>
#include <legacy/sys/syslog.h>


#include <net7g/net/sk_if.h>
#include <net7g/net/sk_if_dl.h>
#include <net7g/net/sk_route.h>

#include <net7g/netinet/sk_in.h>
#include <net7g/netinet/sk_ip6.h>
#include <net7g/netinet/sk_ip.h>
#include <net7g/netinet/sk_ip_var.h>
#include <net7g/netinet/sk_in_pcb.h>
#include <net7g/netinet/sk_udp_var.h>
#include <net7g/netinet/sk_if_ether.h>

#include <net7g/netinet6/sk_in6_var.h>
#include <net7g/netinet6/sk_ip6_var.h>
#include <net7g/netinet6/sk_in6_pcb.h>
#include <net7g/netinet6/sk_tcp6_var.h>
#include <net7g/netinet6/sk_nd6.h>

#include <tcp/tcp.h>
#include <tcp/tcp_timer.h>
#include <tcp/tcp_var.h>
#include <tcp/tcp_fsm.h>

#include "asn1.h"
#include "snmp.h"
#include "snmp_impl.h"
#include "mib.h"
#include "snmp_vars.h"

#include "snmp_impl.h"
#include "mib.h"

#include <cmds/ndp.h>

STATIC find_method_t _var_ipv6MIBObjects;
STATIC find_method_t _var_ipv6IfEntry;
STATIC find_method_t _var_ipv6IfStatsEntry;
STATIC find_method_t _var_ipv6AddrPrefixEntry;
STATIC find_method_t _var_ipv6AddrEntry;
STATIC find_method_t _var_ipv6RouteEntry;
STATIC find_method_t _var_ipv6NetToMediaEntry;
STATIC find_method_t _var_icmpv6Entry;
STATIC find_method_t _var_ipv6TcpConnEntry;
STATIC find_method_t _var_ipv6UdpEntry;

SK_DOMAIN_IMPORT(ipv6_enabled, SK_REASON_READONLY);

uint8_t *
_var_ipv6MIBObjects(
	register struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	*var_len = sizeof(long_return);

	switch (vp->magic) {

	case IPV6FORWARDING:
		SK_DOMAIN_REFERENCE(ip6_forwarding, SK_DOMAIN_NETWORK);
		if (ip6_forwarding)
			long_return = 1;	/*GATEWAY*/
		else
			long_return = 2;	/*HOST*/
		return (uint8_t *)&long_return;
		
	case IPV6DEFAULTHOPLIMIT:
		SK_DOMAIN_REFERENCE(ip6_defhlim, SK_DOMAIN_NETWORK);
		long_return = ip6_defhlim;
		return (uint8_t *)&long_return;
		
	case IPV6INTERFACES:
		/*
		 * not really the right answer... we must count IPv6 capable
		 * interfaces only.
		 */
		long_return = Interface_Scan_Get_Count(AF_INET6);
		if (long_return < 0)
			break;
		return (uint8_t *)&long_return;

	case IPV6IFTABLELASTCHANGE:
		/* XXX6: not implemented */
		long_return = 0;
		return (uint8_t *) & long_return;

	case IPV6ROUTENUMBER:
		/* XXX6: not implemented */
		long_return = 0;
		return (uint8_t *) & long_return;

	case IPV6DISCARDEDROUTES:
		/* XXX6 */
		long_return = 0;
		return (uint8_t *) & long_return;
	default:
		break;
	}

	return NULL;
}

uint8_t *
var_ipv6MIBObjects(
	struct variable *vp,
	oid             *name,
	int             *length,
	int             exact,
	int             *var_len,
	write_method_t  **write_method)
{
	if (!ipv6_enabled)
		return NULL;

	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6MIBObjects,
			   (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_ipv6IfEntry(
	register struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	int result = 0;
	int max = Interface_Scan_Get_Count(AF_INET6);
	oid newname[MAX_OID_LEN];
	struct ifnet *ifp = NULL;

	if (!max)
		return NULL;

	bzero(newname, sizeof(newname));
	bcopy(vp->name, newname, (int)vp->namelen * sizeof(oid));

	/* format: .1.3.6.1.2.1.55.1.5.1.?.interface_id */
	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (ifp = ifnet; ifp; ifp = ifp->if_next) {
		newname[11] = ifp->if_index;
		result = compare(name, *length, newname, (int) vp->namelen);
		if ((exact && (result == 0)) || (!exact && (result < 0)))
			break;
	}
	if (!ifp)
		return NULL;

	bcopy(newname, name, 12 * sizeof(oid));
	*length = 12;
	*write_method = 0;
	*var_len = sizeof(long_return); /* default */
	
	switch (vp->magic) {

	case IPV6IFINDEX:
		long_return = ifp->if_index;
		return (uint8_t *) &long_return;
		
	case IPV6IFDESCR:
	{
#define MIBIFNAMSIZ 2*IFNAMSIZ
		static char if_name[MIBIFNAMSIZ];
		char *Name_ptr = if_name;
                if (ifp->if_flags2 & IFF2_MGMT_PORTS){
                        strlcpy(Name_ptr, "MGMT_PORT_ONLY ", IFNAMSIZ);
                        Name_ptr += strlen(Name_ptr);
                }
		
		strlcpy(Name_ptr, ifp->if_fullname, IFNAMSIZ);
	       
		*var_len = strlen(if_name);
		return (uint8_t *) &if_name;
	}

	case IPV6IFLOWERLAYER:
	{
		static oid nullOid[] = {0,0};
		static int nullOidLen = sizeof(nullOid);
		
		/* XXX6: not implemented */
		/* should check if type, this is a hard one... */
		*var_len = nullOidLen;
		return (uint8_t *)nullOid;
	}

	case IPV6IFEFFECTIVEMTU:
		long_return = ifp->nd_ifinfo.linkmtu;
		return (uint8_t *) &long_return;
	
	case IPV6IFREASMMAXSIZE:
		/* ONTAP doesnt support jumbo datagrams yet */
		long_return = IPV6_MAXPACKET;
		return (uint8_t *) &long_return;

	case IPV6IFIDENTIFIER:
	{
		struct in6_ifaddr *ia6;
		SK_DOMAIN_REFERENCE(in6ifa_ifpforlinklocal, SK_DOMAIN_NETWORK);
		ia6 = in6ifa_ifpforlinklocal(ifp, 0);
		
		if (ia6) {
			static struct in6_addr addr;
			bzero(&addr, sizeof(addr));
			addr.s6_addr32[0] = ia6->ia_addr.sin6_addr.s6_addr32[0]
				& ~ia6->ia_prefixmask.sin6_addr.s6_addr32[0];
			addr.s6_addr32[1] = ia6->ia_addr.sin6_addr.s6_addr32[1]
				& ~ia6->ia_prefixmask.sin6_addr.s6_addr32[1];
			addr.s6_addr32[2] = ia6->ia_addr.sin6_addr.s6_addr32[2]
				& ~ia6->ia_prefixmask.sin6_addr.s6_addr32[2];
			addr.s6_addr32[3] = ia6->ia_addr.sin6_addr.s6_addr32[3]
				& ~ia6->ia_prefixmask.sin6_addr.s6_addr32[3];
			*var_len = sizeof(addr);
			return (uint8_t *) &addr;
		}

		long_return = 0;
		return (uint8_t *) &long_return;
	}

	case IPV6IFIDENTIFIERLENGTH:
	{
		struct in6_ifaddr *ia6 = in6ifa_ifpforlinklocal(ifp, 0);
		*var_len = sizeof(long_return);

		if (ia6)
			long_return = 128 - 
				in6_mask2len(&ia6->ia_prefixmask.sin6_addr);
		else
			long_return = 0;

		return (uint8_t *) &long_return;
	}

		
	case IPV6IFPHYSICALADDRESS:
	{
		static uint8_t addr[1024];
		struct arpcom *ap = NULL;

		ap = (struct arpcom *) ifp;
		if (ifp->if_addrlen != sizeof(ap->ac_enaddr)) {
			/*
			 * XXX - we only support interfaces with 802.x-style
			 * 6-byte addresses.  This needs to be fixed.
			 */
			long_return = 0;
			return (uint8_t *) &long_return;
		}
		*var_len = ifp->if_addrlen;
		bcopy((char*)ap->ac_enaddr, addr, sizeof(ap->ac_enaddr));
		return (uint8_t *) addr;
	}

	case IPV6IFADMINSTATUS:
		long_return = (ifp->if_flags & IFF_RUNNING) ? 1 : 2;
		return (uint8_t *)&long_return;

	case IPV6IFOPERSTATUS:
		long_return = (ifp->if_flags & IFF_UP) ? 1 : 2;
		return (uint8_t *)&long_return;

	case IPV6IFLASTCHANGE:
#if 0 
		/* XXX6: FIXME */
		/*untrustable value returned...*/
		
	{
		struct sec_micro lastchange;
		struct sec_micro now;
		int gotanswer;
		
		gotanswer = 0;
		lastchange.tv_sec = lastchange.tv_usec = 0;
		if (!gotanswer) {
			struct ifmibdata ifmd;
			
			if (if_getifmibdata(ifp->if_index, &ifmd) < 0)
				;
			else {
				lastchange = ifmd.ifmd_data.ifi_lastchange;
				gotanswer++;
			}
		}
		
		if (!gotanswer) {
			lastchange = ifnet.if_lastchange;
			gotanswer++;
		}
		
		SNMP_DBG("lastchange=%d %d\n", lastchange.tv_sec, 
			 lastchange.tv_usec);
		if (lastchange.tv_sec == 0 && lastchange.tv_usec == 0)
			long_return = 0;
		else {
			gettod(&now);
			long_return = (unsigned long)((now.tv_sec - lastchange.tv_sec) * 100);
			long_return += (unsigned long)((now.tv_usec - lastchange.tv_usec) / 10000);
		}
		return (uint8_t *)&long_return;
	}
#endif /*0*/

	break;	

	default:
		break;
	}
	return NULL;
}

uint8_t *
var_ipv6IfEntry(
	struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	if (!ipv6_enabled)
		return NULL;

	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6IfEntry,
			   (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_ipv6IfStatsEntry(
	register struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	int result = 0;
	int max = Interface_Scan_Get_Count(AF_INET6);
	oid newname[MAX_OID_LEN];
	struct in6_ifstat *ifs6 = NULL;
	struct in6_ifreq ifr;
	struct ifnet *ifp = NULL;
	struct socket *s = NULL;
	
	if (!max)
		return NULL;

	bzero(newname, sizeof(newname));
	bcopy(vp->name, newname, (int)vp->namelen * sizeof(oid));

	/* format: .1.3.6.1.2.1.55.1.6.1.?.interface_id */
	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (ifp = ifnet; ifp; ifp = ifp->if_next) {
		newname[11] = ifp->if_index;
		result = compare(name, *length, newname, (int) vp->namelen);
		if ((exact && (result == 0)) || (!exact && (result < 0)))
			break;
	}
	if (!ifp)
		return NULL;

	bcopy(newname, name, 12 * sizeof(oid));
	*length = 12;
	*write_method = 0;
	*var_len = sizeof(long_return); /* default */

	memset(&ifr, 0, sizeof(ifr));
	strlcpy(ifr.ifr_name, ifp->if_fullname, sizeof(ifp->if_fullname));
	
	if (!(s = socket(AF_INET6, SOCK_DGRAM, 0, NULL)))
		return NULL;
	
	SK_DOMAIN_REFERENCE(ifioctl, SK_DOMAIN_NETWORK);
	if (ifioctl(s, SIOCGIFSTAT_IN6, (caddr_t)&ifr) != 0) {
		soclose(s);
		return NULL;
	}
	soclose(s);

	ifs6 = &ifr.ifr_ifru.ifru_stat;
	
	switch (vp->magic) {
		
	case IPV6IFSTATSINRECEIVES:
		long_return = ifs6->ifs6_in_receive;
		break;
		
	case IPV6IFSTATSINHDRERRORS:
		long_return = ifs6->ifs6_in_hdrerr;
		break;
		
	case IPV6IFSTATSINTOOBIGERRORS:
		long_return = ifs6->ifs6_in_toobig;
		break;
		
	case IPV6IFSTATSINNOROUTES:
		long_return = ifs6->ifs6_in_noroute;
		break;
		
	case IPV6IFSTATSINADDRERRORS:
		long_return = ifs6->ifs6_in_addrerr;
		break;
		
	case IPV6IFSTATSINUNKNOWNPROTOS:
		long_return = ifs6->ifs6_in_protounknown;
		break;
		
	case IPV6IFSTATSINTRUNCATEDPKTS:
		long_return = ifs6->ifs6_in_truncated;
		break;
		
	case IPV6IFSTATSINDISCARDS:
		long_return = ifs6->ifs6_in_discard;
		break;
		
	case IPV6IFSTATSINDELIVERS:
		long_return = ifs6->ifs6_in_deliver;
		break;
		
	case IPV6IFSTATSOUTFORWDATAGRAMS:
		long_return = ifs6->ifs6_out_forward;
		break;
		
	case IPV6IFSTATSOUTREQUESTS:
		long_return = ifs6->ifs6_out_request;
		break;
		
	case IPV6IFSTATSOUTDISCARDS:
		long_return = ifs6->ifs6_out_discard;
		break;
		
	case IPV6IFSTATSOUTFRAGOKS:
		long_return = ifs6->ifs6_out_fragok;
		break;
		
	case IPV6IFSTATSOUTFRAGFAILS:
		long_return = ifs6->ifs6_out_fragfail;
		break;
		
	case IPV6IFSTATSOUTFRAGCREATES:
		long_return = ifs6->ifs6_out_fragcreat;
		break;
		
	case IPV6IFSTATSREASMREQDS:
		long_return = ifs6->ifs6_reass_reqd;
		break;
		
	case IPV6IFSTATSREASMOKS:
		long_return = ifs6->ifs6_reass_ok;
		break;
		
	case IPV6IFSTATSREASMFAILS:
		long_return = ifs6->ifs6_reass_fail;
		break;
		
	case IPV6IFSTATSINMCASTPKTS:
		long_return = ifs6->ifs6_in_mcast;
		break;
		
	case IPV6IFSTATSOUTMCASTPKTS:
		long_return = ifs6->ifs6_out_mcast;
		break;
		
	default:
		return NULL;
	}
	return (uint8_t *)&long_return;
}

uint8_t *
var_ipv6IfStatsEntry(
	struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	if (!ipv6_enabled)
		return NULL;
		
	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6IfStatsEntry,
			   (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_ipv6AddrPrefixEntry(
	register struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	int max = Interface_Scan_Get_Count(AF_INET6);
	oid current[MAX_OID_LEN], lowest[MAX_OID_LEN];
	int lowest_len = 0;
	struct nd_prefix *ndpr = NULL;
	static struct nd_prefix low_addrpr;
	struct ifnet *ifp  = NULL;

	if (!max || vp->namelen < 11)
		return NULL;
	
	/* the format is: 
	 * .1.3.6.1.2.1.55.1.7.1.?.interface_id.prefix_addr.prefix_len
	 */
	
	/* fill in object part of name ie (.1.3.6.1.2.1.55.1.7.1.?) */
	bzero(current, sizeof(current));
	bzero(lowest, sizeof(lowest));
	memcpy((char *)current, (char *)vp->name, 11 * sizeof(oid));

	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (ifp = ifnet; ifp; ifp = ifp->if_next) {
		current[11] = ifp->if_index;
				
		SK_DOMAIN_REFERENCE(nd_prefix, SK_DOMAIN_NETWORK);
		for (ndpr = nd_prefix.lh_first; ndpr; ndpr = ndpr->ndpr_next) {
			int i = 0, j = 12;

			if (ndpr->ndpr_ifp->if_index != ifp->if_index)
				continue;

#define CEILING(a, b) ((a/b) + ((a % b) ? 1 : 0))

			for (i = 0; 
			     i < sizeof(struct in6_addr) && 
				     i < CEILING(ndpr->ndpr_plen, 8);
			     i++)
			current[j++] = ndpr->ndpr_prefix.sin6_addr.s6_addr[i];
#undef CEILING

			/* XXX6: we currently dont consider cases of the same 
			 *  prefix with diff.prefix lengths */
			/* insert prefix length */
			current[j++] = ndpr->ndpr_plen;

			ASSERT(j <= 29);
		
			if (exact) {
				if (compare(current, j, name, *length) == 0) {
					bcopy((char *) current, 
					      (char *) lowest,
					      j * sizeof(oid));
					low_addrpr = *ndpr;
					lowest_len = j;
					break;	/* no need to search further */
				}
			} else {
				if ((compare(current, j, name, *length) > 0) &&
				    (!lowest_len ||
				     (compare(current, j, lowest, lowest_len) < 0))) {
					/*
					 * if new one is greater than input and
					 * closer to input than previous lowest
					 * , save this one as the "next" one.
					 */
					low_addrpr = *ndpr;
					lowest_len = j;
					bcopy((char *) current, 
					      (char *) lowest,
					      j * sizeof(oid));
				}
			}
		}
	}

	if (!lowest_len)
		return (NULL);

	bcopy((char *) lowest, (char *) name, lowest_len * sizeof(oid));
	*length = lowest_len;
	*write_method = 0;
	*var_len = sizeof(long_return);

	switch (vp->name[10]) {

	case IPV6ADDRPREFIX:
		*var_len = sizeof(low_addrpr.ndpr_prefix.sin6_addr);
		return (uint8_t *) &low_addrpr.ndpr_prefix.sin6_addr;

	case IPV6ADDRPREFIXLENGTH:
		long_return = low_addrpr.ndpr_plen;
		return (uint8_t *) &long_return;

	case IPV6ADDRPREFIXONLINKFLAG:
		long_return = low_addrpr.ndpr_raf_onlink ? 1 : 2;
		return (uint8_t *) &long_return;
		
	case IPV6ADDRPREFIXAUTONOMOUSFLAG:
		long_return = low_addrpr.ndpr_raf_auto ? 1: 2;
		return (uint8_t *) &long_return;

	case IPV6ADDRPREFIXADVPREFERREDLIFETIME:
		long_return = low_addrpr.ndpr_pltime;
		return (uint8_t *) &long_return;
		
	case IPV6ADDRPREFIXADVVALIDLIFETIME:
		long_return = low_addrpr.ndpr_vltime;
		return (uint8_t *) &long_return;

	default:
		break;
	}
	
	return NULL;
}

uint8_t *
var_ipv6AddrPrefixEntry(
	struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	if (!ipv6_enabled)
		return NULL;

	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6AddrPrefixEntry,
			   (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_ipv6AddrEntry(
	register struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	int max = Interface_Scan_Get_Count(AF_INET6);
	oid current[MAX_OID_LEN], lowest[MAX_OID_LEN];
	int found = 0;
	struct ifnet *ifp = NULL;
	struct ifaddr *ia = NULL;
	struct in6_ifaddr *low_ia6 = NULL;
	static struct in6_ifaddr ia6;

	if (!max || vp->namelen < 11)
		return NULL;
	
	/* the format is:
	 * .1.3.6.1.2.1.55.1.8.1.?.interface_id.ipv6_addr
	 */

	/* fill in object part of name ie .1.3.6.1.2.1.55.1.8.1.? */
	bzero(current, sizeof(current));
	bzero(lowest, sizeof(lowest));
	memcpy((char *)current, (char *)vp->name, 11 * sizeof(oid));

	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (ifp = ifnet; ifp; ifp = ifp->if_next) {
		current[11] = ifp->if_index;

		for (ia = ifp->if_addrlist; ia; ia = ia->ifa_next) {
			int i = 0, j = 12;
			struct in6_addr sin6_addr;
			
			if (ia->ifa_addr->sa_family != AF_INET6)
				continue;

			sin6_addr = ((struct in6_ifaddr *)ia)->ia_addr.sin6_addr;
			if (IN6_IS_ADDR_LINKLOCAL(&sin6_addr))
				*(uint16_t *)&sin6_addr.s6_addr[2] = 0;
			
			for (i = 0;  i < sizeof(struct in6_addr); i++)
				current[j++] = sin6_addr.s6_addr[i];
			
			ASSERT(j == 28);
			
			if (exact) {
				if (compare(current, j, name, *length) == 0) {
					bcopy((char *) current, 
					      (char *) lowest,
					      j * sizeof(oid));
					low_ia6 = (struct in6_ifaddr *) ia;
					found = 1;
					break;	/* no need to search further */
				}
			} else {
				if ((compare(current, j, name, *length) > 0) &&
				    (!found ||
				     (compare(current, j, lowest, j) < 0))) {
					/*
					 * if new one is greater than input and
					 * closer to input than previous lowest
					 *, save this one as the "next" one.
					 */
					low_ia6 = (struct in6_ifaddr *) ia;
					bcopy((char *) current, 
					      (char *) lowest,
					      j * sizeof(oid));
					found = 1;
				}
			}
		}
		
	}
	if (!found)
		return NULL;
	
	ia6 = *low_ia6;
	bcopy((char *) lowest, (char *) name, 28 * sizeof(oid));
	*length = 28;
	*write_method = 0;
	*var_len = sizeof(long_return);

	switch (vp->name[10]) {

	case IPV6ADDRADDRESS:
	{
		*var_len = sizeof(struct in6_addr);
		
		if (IN6_IS_ADDR_LINKLOCAL(&ia6.ia_addr.sin6_addr))
			*(uint16_t *) &ia6.ia_addr.sin6_addr.s6_addr[2] = 0;
		
		return (uint8_t *) &ia6.ia_addr.sin6_addr.s6_addr[0];
	}

	case IPV6ADDRPFXLENGTH:
	{
		long_return = in6_mask2len(&satosin6(ia6.ia_ifa.ifa_netmask)->sin6_addr);
		return (uint8_t *) &long_return;
	}

	case IPV6ADDRTYPE:
		if (ia6.ia6_flags & IN6_IFF_STATELESS ||
		    IN6_IS_ADDR_LINKLOCAL(&ia6.ia_addr.sin6_addr))
			long_return = 1;
		else
			long_return = 2;
		return (uint8_t *) &long_return;
    
	case IPV6ADDRANYCASTFLAG:
		/* ONTAP does not support anycast addresses */
		long_return = (ia6.ia6_flags & IN6_IFF_ANYCAST) ? 1 : 2;
		return (uint8_t *) &long_return;

	case IPV6ADDRSTATUS:
#define err_flags (IN6_IFF_DEPRECATED | IN6_IFF_DETACHED | IN6_IFF_NOTREADY)

		switch (ia6.ia6_flags & err_flags)   { 

		case IN6_IFF_DEPRECATED:
			long_return = 2; /* deprecated */
			break;

		case IN6_IFF_DETACHED:
			long_return = 4; /* inaccessible */
			break;

		case IN6_IFF_NOTREADY:
		case IN6_IFF_TENTATIVE:
		case IN6_IFF_DUPLICATED:
			long_return = 3; /* invalid */
			break;
		default:
			if (ia6.ia6_flags & err_flags)
				long_return = 5; /* unknown combo state */
			else
				long_return = 1; /* preferred */
		}
#undef err_flags
		return (uint8_t *) &long_return;

	default:
		break;
	}
	
	return NULL;
}

uint8_t *
var_ipv6AddrEntry(
	struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	if (!ipv6_enabled)
		return NULL;

	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6AddrEntry,
			   (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_ipv6RouteEntry(
	register struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	int max = Interface_Scan_Get_Count(AF_INET6);
	int result = 0;
	oid newname[MAX_OID_LEN];
	struct ifnet *ifp = NULL;

	if (!max)
		return NULL;

	bzero(newname, sizeof(newname));
	bcopy(vp->name, newname, (int)vp->namelen * sizeof(oid));

	/* format: .1.3.6.1.2.1.55.1.11.1.?.interface_id */
	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (ifp = ifnet; ifp; ifp = ifp->if_next) {
		newname[11] = ifp->if_index;
		result = compare(name, *length, newname, (int) vp->namelen);
		if ((exact && (result == 0)) || (!exact && (result < 0)))
			break;
	}
	if (!ifp)
		return NULL;
	
	bcopy(newname, name, 12 * sizeof(oid));
	*length = 12;
	*write_method = 0;
	*var_len = sizeof(long_return); /* default */
	
	switch (vp->magic) {
	case IPV6ROUTEDEST:
	case IPV6ROUTEPFXLENGTH:
	case IPV6ROUTEINDEX:
	case IPV6ROUTEIFINDEX:
	case IPV6ROUTENEXTHOP:
	case IPV6ROUTETYPE:
	case IPV6ROUTEPROTOCOL:
	case IPV6ROUTEPOLICY:
	case IPV6ROUTEAGE:
	case IPV6ROUTENEXTHOPRDI:
	case IPV6ROUTEMETRIC:
	case IPV6ROUTEWEIGHT:
	case IPV6ROUTEINFO:
	case IPV6ROUTEVALID:
		/* XXX6: not implemented */
		long_return = 0;
		return (uint8_t *) & long_return;
	default:
		break;
	}
	
	return NULL;
}

uint8_t *
var_ipv6RouteEntry(
	struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	if (!ipv6_enabled)
		return NULL;

	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6RouteEntry,
			   (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_ipv6NetToMediaEntry(
	register struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	int max = Interface_Scan_Get_Count(AF_INET6);
	oid current[MAX_OID_LEN], lowest[MAX_OID_LEN];
	int found = 0;
	struct ifnet *ifp = NULL, *lowin_ifp = NULL;
	struct in6_ifaddr *lowin_ifaddr = NULL;
	struct ifaddr *ifa = NULL;
	struct arpcom *ap = NULL;
	
	if (!max || vp->namelen < 11)
		return NULL;
	
	/* the format is:
	 *  .1.3.6.1.2.1.55.1.12.1.?.interface_index.net_address
	 */

	/*
	 * fill in object part of name ie .1.3.6.1.2.1.55.1.12.1.? */ 
	bzero(current, sizeof(current));
	bzero(lowest, sizeof(lowest));
	memcpy((char *)current, (char *)vp->name, 11 * sizeof(oid));
	
	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (ifp = ifnet; ifp; ifp = ifp->if_next) {
		current[11] = ifp->if_index;
		
		for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
			int i = 0, j = 12;
			struct in6_addr sin6_addr;
			
			if (ifa->ifa_addr->sa_family != AF_INET6)
				continue;
			
			sin6_addr = ((struct in6_ifaddr *)ifa)->ia_addr.sin6_addr;
			
			if (IN6_IS_ADDR_LINKLOCAL(&sin6_addr))
				*(uint16_t *)&sin6_addr.s6_addr[2] = 0;
		
			for (i = 0; i < sizeof(struct in6_addr); i++)
				current[j++] = ((struct in6_ifaddr *)ifa)->ia_addr.sin6_addr.s6_addr[i];
		
			ASSERT(j == 28);
			
			if (exact) {
				if (compare(current, j, name, *length) == 0) {
					bcopy((char *) current, 
					      (char *) lowest,
					      j * sizeof(oid));
					found = 1;
					lowin_ifaddr = (struct in6_ifaddr *) ifa;
					lowin_ifp = ifp;
					break;	/* no need to search further */
				}
			} else {
				if ((compare(current, j, name, *length) > 0) &&
				    (!found || (compare(current, j, lowest, j) < 0))) {
					/*
					 * if new one is greater than input and
					 * closer to input than previous lowest
					 * , save this one as the "next" one.
					 */
					found = 1;
					lowin_ifaddr = (struct in6_ifaddr*) ifa;
					lowin_ifp = ifp;
					bcopy((char *) current, 
					      (char *) lowest,
					      j * sizeof(oid));
				}
			}
		}
	}

	if (!found)
		return NULL;
		
	bcopy((char *) lowest, (char *) name, 28 * sizeof(oid));
	*length = 28;
	*write_method = 0;
	*var_len = sizeof(long_return);
	
	switch (vp->name[10]) {

	case IPV6NETTOMEDIANETADDRESS:
	{
		static struct in6_addr addr;
		bcopy(&lowin_ifaddr->ia_addr.sin6_addr.s6_addr,
		      &addr, sizeof(addr));

		if (IN6_IS_ADDR_LINKLOCAL(&addr))
			*(uint16_t *) &addr.s6_addr[2] = 0;
		
		*var_len = sizeof(addr);
		return (uint8_t *) &addr.s6_addr[0];
	}

	case IPV6NETTOMEDIAPHYSADDRESS:
	{
		static uint8_t addr[1024];

		ap = (struct arpcom *) lowin_ifp;
		if (lowin_ifp->if_addrlen != sizeof(ap->ac_enaddr)) {
			/*
			 * XXX - we only support interfaces with 802.x-style
			 * 6-byte addresses.  This needs to be fixed.
			 */
			long_return = 0;
			return (uint8_t *) &long_return;
		}
		*var_len = lowin_ifp->if_addrlen;
		bcopy((char*)&ap->ac_enaddr, addr, sizeof(ap->ac_enaddr));
		return (uint8_t *) addr;
	}

	case IPV6NETTOMEDIATYPE:
		if (IN6_IS_ADDR_LINKLOCAL(&lowin_ifaddr->ia_addr.sin6_addr)) {
			long_return = 4; /* local */
 		} else if (lowin_ifaddr->ia6_flags & IN6_IFF_STATELESS) {
			long_return = 2;  /* dynamic */
		} else
			long_return = 3; /* static */

		return (uint8_t *) &long_return;

	case IPV6IFNETTOMEDIASTATE:
	{
		struct in6_nbrinfo *nbi;
		cmd_result_t result;

		if ((nbi = getnbrinfo(&lowin_ifaddr->ia_addr.sin6_addr,
				     lowin_ifp->if_index, 1, RT_DEFAULT_GRP, &result))
		    && result == cmdOkReturn) {
			
			switch (nbi->state) {
			case ND6_LLINFO_REACHABLE:
				long_return = 1; /* reachable */
				break;
			case ND6_LLINFO_STALE:
				long_return = 2; /* stale */
				break;
			case ND6_LLINFO_DELAY:
				long_return = 3; /* delay */
				break;
			case ND6_LLINFO_PROBE:
				long_return = 4; /* probe */
				break;
			case ND6_LLINFO_WAITDELETE:
				long_return = 5; /* invalid */
			case ND6_LLINFO_NOSTATE:
			case ND6_LLINFO_INCOMPLETE:
			default:
				long_return = 6; /* unknown */
				break;
			}
			
		} else
			long_return = 6;
		return (uint8_t *) &long_return;
	}
	case IPV6IFNETTOMEDIALASTUPDATED:
		/* XXX6: not implemented */
		long_return = 0;
		return (uint8_t *) &long_return;

	case IPV6NETTOMEDIAVALID:
		/* XXX6: not implemented */
		long_return = 0;
		return (uint8_t *) &long_return;

	default:
		break;
	}
	
	return NULL;
}

uint8_t *
var_ipv6NetToMediaEntry(
	struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	if (!ipv6_enabled)
		return NULL;

	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6NetToMediaEntry,
			   (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_icmpv6Entry(
	register struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	int result = 0;
	int max = Interface_Scan_Get_Count(AF_INET6);
	oid newname[MAX_OID_LEN];
	struct icmp6_ifstat *ifs6 = NULL;
	struct ifnet *ifp = NULL;
	struct in6_ifreq ifr;
	struct socket *s = NULL;
	
	if (!max)
		return NULL;

	bzero(newname, sizeof(newname));
	bcopy(vp->name, newname, (int)vp->namelen * sizeof(oid));
	
	/* format: .1.3.6.1.2.1.56.1.1.1.?.interface_id */
	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (ifp = ifnet; ifp; ifp = ifp->if_next) {
		newname[11] = ifp->if_index;
		result = compare(name, *length, newname, (int) vp->namelen);
		if ((exact && (result == 0)) || (!exact && (result < 0)))
			break;
	}
	if (!ifp)
		return NULL;

	bcopy(newname, name, 12 * sizeof(oid));
	*length = 12;
	*write_method = 0;
	*var_len = sizeof(long_return); /* default */

	memset(&ifr, 0, sizeof(ifr));

	strlcpy(ifr.ifr_name, ifp->if_fullname, sizeof(ifp->if_fullname));

	if (!(s = socket(AF_INET6, SOCK_DGRAM, 0, NULL)))
		return NULL;
	
	SK_DOMAIN_REFERENCE(ifioctl, SK_DOMAIN_NETWORK);
	if (ifioctl(s, SIOCGIFSTAT_ICMP6, (caddr_t)&ifr) != 0) {
		soclose(s);
		return NULL;
	}
	soclose(s);

	ifs6 = &ifr.ifr_ifru.ifru_icmp6stat;

	switch (vp->magic) {

	case IPV6IFICMPINMSGS:
	    long_return = ifs6->ifs6_in_msg;
	    break;

	case IPV6IFICMPINERRORS:
	    long_return = ifs6->ifs6_in_error;
	    break;

	case IPV6IFICMPINDESTUNREACHS:
	    long_return = ifs6->ifs6_in_dstunreach;
	    break;

	case IPV6IFICMPINADMINPROHIBS:
	    long_return = ifs6->ifs6_in_adminprohib;
	    break;

	case IPV6IFICMPINTIMEEXCDS:
	    long_return = ifs6->ifs6_in_timeexceed;
	    break;

	case IPV6IFICMPINPARMPROBLEMS:
	    long_return = ifs6->ifs6_in_paramprob;
	    break;

	case IPV6IFICMPINPKTTOOBIGS:
	    long_return = ifs6->ifs6_in_pkttoobig;
	    break;

	case IPV6IFICMPINECHOS:
	    long_return = ifs6->ifs6_in_echo;
	    break;

	case IPV6IFICMPINECHOREPLIES:
	    long_return = ifs6->ifs6_in_echoreply;
	    break;

	case IPV6IFICMPINROUTERSOLICITS:
	    long_return = ifs6->ifs6_in_routersolicit;
	    break;

	case IPV6IFICMPINROUTERADVERTISEMENTS:
	    long_return = ifs6->ifs6_in_routeradvert;
	    break;

	case IPV6IFICMPINNEIGHBORSOLICITS:
	    long_return = ifs6->ifs6_in_neighborsolicit;
	    break;

	case IPV6IFICMPINNEIGHBORADVERTISEMENTS:
	    long_return = ifs6->ifs6_in_neighboradvert;
	    break;

	case IPV6IFICMPINREDIRECTS:
	    long_return = ifs6->ifs6_in_redirect;
	    break;

	case IPV6IFICMPINGROUPMEMBQUERIES:
	    long_return = ifs6->ifs6_in_mldquery;
	    break;

	case IPV6IFICMPINGROUPMEMBRESPONSES:
	    long_return = ifs6->ifs6_in_mldreport;
	    break;

	case IPV6IFICMPINGROUPMEMBREDUCTIONS:
	    long_return = ifs6->ifs6_in_mlddone;
	    break;

	case IPV6IFICMPOUTMSGS:
	    long_return = ifs6->ifs6_out_msg;
	    break;

	case IPV6IFICMPOUTERRORS:
	    long_return = ifs6->ifs6_out_error;
	    break;

	case IPV6IFICMPOUTDESTUNREACHS:
	    long_return = ifs6->ifs6_out_dstunreach;
	    break;

	case IPV6IFICMPOUTADMINPROHIBS:
	    long_return = ifs6->ifs6_out_adminprohib;
	    break;

	case IPV6IFICMPOUTTIMEEXCDS:
	    long_return = ifs6->ifs6_out_timeexceed;
	    break;

	case IPV6IFICMPOUTPARMPROBLEMS:
	    long_return = ifs6->ifs6_out_paramprob;
	    break;

	case IPV6IFICMPOUTPKTTOOBIGS:
	    long_return = ifs6->ifs6_out_pkttoobig;
	    break;

	case IPV6IFICMPOUTECHOS:
	    long_return = ifs6->ifs6_out_echo;
	    break;

	case IPV6IFICMPOUTECHOREPLIES:
	    long_return = ifs6->ifs6_out_echoreply;
	    break;

	case IPV6IFICMPOUTROUTERSOLICITS:
	    long_return = ifs6->ifs6_out_routersolicit;
	    break;

	case IPV6IFICMPOUTROUTERADVERTISEMENTS:
	    long_return = ifs6->ifs6_out_routeradvert;
	    break;

	case IPV6IFICMPOUTNEIGHBORSOLICITS:
	    long_return = ifs6->ifs6_out_neighborsolicit;
	    break;

	case IPV6IFICMPOUTNEIGHBORADVERTISEMENTS:
	    long_return = ifs6->ifs6_out_neighboradvert;
	    break;

	case IPV6IFICMPOUTREDIRECTS:
	    long_return = ifs6->ifs6_out_redirect;
	    break;

	case IPV6IFICMPOUTGROUPMEMBQUERIES:
	    long_return = ifs6->ifs6_out_mldquery;
	    break;

	case IPV6IFICMPOUTGROUPMEMBRESPONSES:
	    long_return = ifs6->ifs6_out_mldreport;
	    break;

	case IPV6IFICMPOUTGROUPMEMBREDUCTIONS:
	    long_return = ifs6->ifs6_out_mlddone;
	    break;

	default:
	    return NULL;
	}

	return (uint8_t *)&long_return;
}

uint8_t *
var_ipv6IfIcmpEntry(
	struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	if (!ipv6_enabled)
		return NULL;

	DOMAIN_SWITCH_NEST(sk_domain_network, _var_icmpv6Entry,
			   (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_ipv6TcpConnEntry(
	struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	oid newname[MAX_OID_LEN], tmp1[MAX_OID_LEN], tmp2[MAX_OID_LEN];
	int result = 0;
	int i = 0, j = 0;
	struct inpcbhead *head = NULL;
        struct inpcb *tpcb = NULL, low_tpcb = {{0}};
	int found = 0, hitnext = 0;
	static int tcpstatemap[TCP_NSTATES];
	static int initialized = 0;

	if (!initialized) {
		tcpstatemap[TCPS_CLOSED] = 1;
		tcpstatemap[TCPS_LISTEN] = 2;
		tcpstatemap[TCPS_SYN_SENT] = 3;
		tcpstatemap[TCPS_SYN_RECEIVED] = 4;
		tcpstatemap[TCPS_ESTABLISHED] = 5;
		tcpstatemap[TCPS_CLOSE_WAIT] = 8;
		tcpstatemap[TCPS_FIN_WAIT_1] = 6;
		tcpstatemap[TCPS_CLOSING] = 10;
		tcpstatemap[TCPS_LAST_ACK] = 9;
		tcpstatemap[TCPS_FIN_WAIT_2] = 7;
		tcpstatemap[TCPS_TIME_WAIT] = 11;
		initialized++;
	}

	/* fill in object part of name
	 * eg: .1.3.6.1.2.1.6.16.1.?
	 * for current part (less sizeof instance part)
	 */
	bzero(newname, sizeof(newname));
	bcopy(vp->name, newname, 10 * sizeof(oid));
	
	SK_DOMAIN_REFERENCE(in_auxpcblist, SK_REASON_SWITCH_DOMAIN);
	SK_DOMAIN_REFERENCE(tcb, SK_REASON_LAX);
	for (head = &tcb[sk_gethost()]; 
	     head != NULL; 
	     head = in_auxpcblist(head)) {

		for (tpcb = head->lh_first;
		     tpcb != NULL;
		     tpcb = tpcb->inp_list.le_next) {

			if ((tpcb->inp_vfiler) 
			   && (!VFILER_IS_DEFAULT(tpcb->inp_vfiler)))
				continue;

			if (!(tpcb->inp_flags & INP_IPV6))
				continue;
			
			j = 10;
			for (i = 0; i < sizeof(struct in6_addr); i++)
				newname[j++] = tpcb->in6p_laddr.s6_addr[i];
		
			newname[j++] = ntohs(tpcb->in6p_lport);
			
			for (i = 0; i < sizeof(struct in6_addr); i++)
				newname[j++] = tpcb->in6p_faddr.s6_addr[i];
		
			newname[j++] = ntohs(tpcb->in6p_fport);
			
			if (IN6_IS_ADDR_LINKLOCAL(&tpcb->in6p_laddr))
				newname[j++] = 
					ntohs(*(uint16_t *)&tpcb->in6p_laddr.s6_addr[2]);
			else
				newname[j++] = 0;	/*XXX*/
			
			ASSERT(j < MAX_OID_LEN);
#if 1	/* this is very odd but sometimes happen, and cause infinite loop */
			if (ntohs(tpcb->in6p_lport) == 0)
				continue;
#endif
		
			if (exact) {
				result = compare(name, *length, newname, j);
				if (result == 0) {
					low_tpcb = *tpcb;
					found++;
					break;
				}
			} else {
				/*
				 * tcb entries may not be ordered as SNMP 
				 * scanning code expect. Therefore, we'll do 
				 * a guess...
				 * 1. "name" has the previously matched entry.
				 *    If "name" is TCP6 MIBs, take the item 
				 *    that is right next to the item exactly 
				 *    matches to "name".
				 * 2. Otherwise, it is the very first time we 
				 *    entered to TCP6 MIBs.
				 */
				
				memcpy(tmp1, name, vp->namelen * sizeof(oid));
				memcpy(tmp2, newname, vp->namelen * sizeof(oid));
				tmp1[vp->namelen] = 0;
				tmp2[vp->namelen] = 0;
				
				if (*length == j
				    && compare(tmp1, vp->namelen, tmp2, 
					       vp->namelen) == 0) {
					
					if (!hitnext) {
						result = compare(name, 
								 *length, 
								 newname, 
								 j);
						if (result == 0)
							hitnext++;
					} else {
						found++;
						low_tpcb = *tpcb;
						break;
					}
				} else {
					if (compare(name, *length, newname, j)
					    < 0) {
						found++;
						low_tpcb = *tpcb;
						break;
					}
				}
			}
		}
	}
	
	if (!found)
		return NULL;
	
	*length = j;
	memcpy((char *)name, (char *)newname, *length * sizeof(oid));
	*write_method = 0;
	*var_len = sizeof(int32_t);	/* default to 'int32_t' results */
	
	switch (vp->name[9]) {

	case IPV6TCPCONNLOCALADDRESS:
	{
		static struct in6_addr laddr;
		*var_len = sizeof(struct in6_addr);
		bcopy(&low_tpcb.in6p_laddr.s6_addr, &laddr, *var_len);
		return (uint8_t *) &laddr;
	}

	case IPV6TCPCONNLOCALPORT:
		long_return = ntohs(low_tpcb.in6p_lport);
		return (uint8_t *)&long_return;

	case IPV6TCPCONNREMADDRESS:
	{
		static struct in6_addr faddr;
		*var_len = sizeof(struct in6_addr);
		bcopy(&low_tpcb.in6p_faddr.s6_addr, &faddr, *var_len);
		return (uint8_t *) &faddr;
	}

	case IPV6TCPCONNREMPORT:
		long_return = ntohs(low_tpcb.in6p_fport);
		return (uint8_t *)&long_return;

	case IPV6TCPCONNIFINDEX:
		if (IN6_IS_ADDR_LINKLOCAL(&low_tpcb.in6p_laddr))
			long_return = ntohs(*(uint16_t *)&low_tpcb.in6p_laddr.s6_addr[2]);
		else
			long_return = 0;	/*XXX*/
		return (uint8_t *)&long_return;

	case IPV6TCPCONNSTATE:
		long_return = 
		     tcpstatemap[((struct tcpcb *)low_tpcb.inp_ppcb)->t_state];
		return (uint8_t *)&long_return;

	default:
		break;
	}
	
	
	return NULL;
}

uint8_t *
var_ipv6TcpConnEntry(
	struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	if (!ipv6_enabled)
		return NULL;

	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6TcpConnEntry,
			   (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_ipv6UdpEntry(
    register struct variable *vp,
    oid *name,
    int *length,
    int exact,
    int *var_len,
    write_method_t **write_method)
{
	oid newname[MAX_OID_LEN];
	oid tmp1[MAX_OID_LEN], tmp2[MAX_OID_LEN];
	int result;
	int i = 0, j = 0;
	struct inpcb *upcb = NULL, low_upcb = {{0}};
	int found = 0, hitnext = 0;

	/* fill in object part of the name
	 * eg. .1.3.6.1.2.1.7.6.1.?
	 * for current part (less sizeof instance part)
	 */
	bzero(newname, sizeof(newname));
	bcopy(vp->name, newname, 10 * sizeof(oid));
	
	SK_DOMAIN_REFERENCE(udb, SK_DOMAIN_NETWORK);
	for (upcb = udb[sk_gethost()].lh_first; upcb != NULL; 
	     upcb = upcb->inp_list.le_next, j = 11) {
		
		if ((upcb->inp_vfiler) 
		    && (!VFILER_IS_DEFAULT(upcb->inp_vfiler)))
			continue;
		
		if (!(upcb->inp_flags & INP_IPV6))
			continue;
		
		j = 10;
		for (i = 0; i < sizeof(struct in6_addr); i++)
			newname[j++] = upcb->in6p_laddr.s6_addr[i];
		
		newname[j++] = ntohs(upcb->in6p_lport);
		
		if (IN6_IS_ADDR_LINKLOCAL(&upcb->in6p_laddr))
			newname[j++] = ntohs(*(uint16_t *)&upcb->in6p_laddr.s6_addr[2]);
		else
			newname[j++] = 0;	/*XXX*/
		
		if (exact) {
			result = compare(name, *length, newname, j);
			if (result == 0) {
				found++;
				low_upcb = *upcb;
				break;
			}
		} else {
			/*
			 * udb entries may not be ordered as SNMP scanning 
			 * code expect. Therefore, we'll do a guess...
			 * 1. "name" has the previously matched entry.
			 *    If "name" is UDP6 MIBs, take the item that is 
			 *    right next to the item exactly matches to "name".
			 * 2. Otherwise, it is the very first time we entered 
			 *    to UDP6 MIBs.
			 */
			
			memcpy(tmp1, name, vp->namelen * sizeof(oid));
			memcpy(tmp2, newname, vp->namelen * sizeof(oid));
			tmp1[vp->namelen] = 0;
			tmp2[vp->namelen] = 0;
			
			if (*length == j
			    && compare(tmp1, vp->namelen, tmp2, vp->namelen) == 0) {
				if (!hitnext) {
					result = compare(name, *length, newname, j);
					if (result == 0)
						hitnext++;
				} else {
					found++;
					low_upcb = *upcb;
					break;
				}
			} else {
				if (compare(name, *length, newname, j) < 0) {
					found++;
					low_upcb = *upcb;
					break;
				}
			}
		}
	}

	if (!found)
		return NULL;

	*length = j;
	memcpy((char *)name, (char *)newname, *length * sizeof(oid));
	*write_method = 0;
	*var_len = sizeof(int32_t);	/* default to 'int32_t' results */
	
	
	switch (vp->name[10]) {
		
	case IPV6UDPLOCALADDRESS:
	{
		static struct in6_addr laddr;
		*var_len = sizeof(struct in6_addr);
		bcopy(&low_upcb.in6p_laddr.s6_addr, &laddr, *var_len);
		return (uint8_t *) &laddr;
	}
		
	case IPV6UDPLOCALPORT:
		long_return = ntohs(low_upcb.in6p_lport);
		return (uint8_t *)&long_return;

	case IPV6UDPIFINDEX:
		if (IN6_IS_ADDR_LINKLOCAL(&upcb->in6p_laddr))
			long_return = ntohs(*(uint16_t *)&low_upcb.in6p_laddr.s6_addr[2]);
		else
			long_return = 0;	/*XXX*/

		return (uint8_t *)&long_return;

	default:
		break;

	}

	return NULL;
}

uint8_t *
var_ipv6UdpEntry(
	struct variable *vp,
	oid *name,
	int *length,
	int exact,
	int *var_len,
	write_method_t **write_method)
{
	if (!ipv6_enabled)
		return NULL;

	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipv6UdpEntry,
			   (vp, name, length, exact, var_len, write_method));
}
