/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */

/*
 *  System MIB group implementation - system.c
 *
 */
/* Portions of this file are subject to the following copyright(s).  See
 * the Net-SNMP's COPYING file for more details and other copyrights
 * that may apply:
 */
/*
 * Portions of this file are copyrighted by:
 * Copyright  2003 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms specified in the COPYING file
 * distributed with the Net-SNMP package.
 */

#include <net-snmp/net-snmp-config.h>
#include <sk/sk_host.h>

#include <sk/sk_param.h>


#if HAVE_STDLIB_H
#include <legacy/stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#endif
#if HAVE_STRING_H
#include <legacy/string.h>
#else
#include <strings.h>
#endif
#include <sys/types.h>
#if HAVE_WINSOCK_H
#include <winsock.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sk/time.h>
#endif

#include <legacy/ctype.h>
#if HAVE_UTSNAME_H
#include <utsname.h>
#else
#if HAVE_SYS_UTSNAME_H
#include <sys/utsname.h>
#endif
#endif
#if HAVE_NETINET_IN_H
#include <net7g/netinet/sk_in.h>
#endif

#if HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include <net-snmp/agent/snmp_vars.h>

#include "mibgroup/util_funcs.h"
#include "system_mib.h"

#include <sk/ha_host.h>
#include <sk/sk_stats.h>
#include <ha/fm.h>
#include <ha/fmmisc.h>

#include <sk/sk_vers.h>

        /*********************
	 *
	 *  Kernel & interface information,
	 *   and internal forward declarations
	 *
	 *********************/

#define SYS_STRING_LEN	(SYSTEM_NAMELEN + 1)
char            system_descr[SYS_STRING_LEN] = {VERS_DESC};
char            system_contact[ONTAP_NHOSTS][SYS_STRING_LEN] = {SYS_CONTACT, SYS_CONTACT};
char            system_name[ONTAP_NHOSTS][SYS_STRING_LEN] = {SYS_NAME, SYS_NAME};
char            system_location[ONTAP_NHOSTS][SYS_STRING_LEN] = {SYS_LOC, SYS_LOC};
oid             sysObjectID[MAX_OID_LEN];
size_t          sysObjectIDLength;
oid          *  version_sysoid;
oid             clustered_sysoid[] = {SYSTEM_HA_MIB};
oid             filer_sysoid[] = {SYSTEM_MIB};

int             system_services = 72;
int             sysServicesConfiged = 0;

extern oid      version_id[];
extern int      version_id_len;


int             header_system(struct variable *, oid *, size_t *, int,
                              size_t *, WriteMethod **);

        /*********************
	 *
	 *  Initialisation & common implementation functions
	 *
	 *********************/

/*
 * define the structure we're going to ask the agent to register our
 * information at 
 */
#ifdef USING_MIBII_SYSORTABLE_MODULE
struct variable1 system_variables[8] = {
#else
struct variable1 system_variables[7] = {
#endif
    {VERSION_DESCR, ASN_OCTET_STR, RONLY, var_system, 1, {1}},
    {VERSIONID, ASN_OBJECT_ID, RONLY, var_system, 1, {2}},
    {UPTIME, ASN_TIMETICKS, RONLY, var_system, 1, {3}},
    {SYSCONTACT, ASN_OCTET_STR, RONLY, var_system, 1, {4}},
    {SYSTEMNAME, ASN_OCTET_STR, RONLY, var_system, 1, {5}},
    {SYSLOCATION, ASN_OCTET_STR, RONLY, var_system, 1, {6}},
    {SYSSERVICES, ASN_INTEGER, RONLY, var_system, 1, {7}},
#ifdef USING_MIBII_SYSORTABLE_MODULE
    {SYSORLASTCHANGE, ASN_TIMETICKS, RONLY, var_system, 1, {8}}
#endif
};
/*
 * Define the OID pointer to the top of the mib tree that we're
 * registering underneath 
 */
oid             system_variables_oid[] = { SNMP_OID_MIB2, 1 };
oid             system_module_oid[] = { SNMP_OID_SNMPMODULES, 1 };
int             system_module_oid_len =
    sizeof(system_module_oid) / sizeof(oid);
int             system_module_count = 0;

void
init_system_mib(void)
{

	strlcpy(system_descr, ontap_version.version, SYS_STRING_LEN);

	/* System name, contact, and location are set via the command line. */
	
	version_sysoid = filer_sysoid;

	/* default sysObjectID */
	memcpy(sysObjectID, version_sysoid, sizeof(filer_sysoid));
	sysObjectIDLength = OID_LENGTH(filer_sysoid);

    /*
     * register ourselves with the agent to handle our mib tree 
     */
    REGISTER_MIB("mibII/system", system_variables, variable1,
                 system_variables_oid);

#ifdef USING_MIBII_SYSORTABLE_MODULE
    if (++system_module_count == 3)
        REGISTER_SYSOR_ENTRY(system_module_oid,
                             "The MIB module for SNMPv2 entities");
#endif
}

void snmp_get_sys_object_id(oid ** oidp, int * lenp);

void
snmp_get_sys_object_id(oid ** oidp, int * lenp)
{
	*oidp = sysObjectID;
	*lenp = sysObjectIDLength;
}


        /*********************
	 *
	 *  System specific implementation functions
	 *
	 *********************/

#ifdef USING_MIBII_SYSORTABLE_MODULE
extern struct sec_micro sysOR_lastchange;
#endif

uint8_t         *
var_system(struct variable *vp,
           oid * name,
           size_t * length,
           int exact, size_t * var_len, WriteMethod ** write_method)
{
    static uint32_t   ulret;
    char dname[ONTAP_MAXDOMAINNAMELEN];
    int len;
    int host = sk_gethost();

    if (header_generic(vp, name, length, exact, var_len, write_method) ==
        MATCH_FAILED)
        return NULL;

    switch (vp->magic) {
    case VERSION_DESCR:
        *var_len = strlen(system_descr);
        return (uint8_t *) system_descr;
    case VERSIONID:
	    /* need to check this after HA initialized. */
	    if (fm_is_HAmode()) {
		    version_sysoid = clustered_sysoid;
		    memcpy(sysObjectID, version_sysoid, sizeof(clustered_sysoid));
		    sysObjectIDLength = OID_LENGTH(clustered_sysoid);
	    }
        *var_len = sysObjectIDLength * sizeof(sysObjectID[0]);
        return (uint8_t *)sysObjectID;
    case UPTIME:	
    /* return uptime in 10's of milliseconds */
	    /*
	     * Return a value based on whether we are in takeover mode
	     * or not.
	     */
	    if (host == SK_HOST_LOCAL) {
                    /* 
                     * sk_msecs returns milli seconds in unsigned long, so
                     * divide by ten to return centiseconds.
                     */
                    long_return = (int32_t)
				((sk_msecs - sk_msecs_initial)/10);
                    *var_len = sizeof(long_return);
                    return (uint8_t *) &long_return;
            }
	    else {	/* SK_HOST_PARTNER */
		    fm_status_t status;
		    fm_time_t current =
			    fm_currentTime();
		    fm_status(&status);
		    /*
		     * Note: fm returns these times in msecs, so
		     * divide by ten to return centiseconds.
		     */
		    ulret = (int32_t)
			    ((current - status.s_timeEntered) / 10);
                    return ((uint8_t *) & ulret);
	    };
    case SYSCONTACT:
        *var_len = strlen(system_contact[host]);
        return (uint8_t *) system_contact[host];
    case SYSTEMNAME:
	    sk_gethostname(system_name[host],
			  sizeof(system_name[host]) - 1);
	    /*
	     * Append the domain name to the host name only if 
	     * there is a host name to append to.
	     */
	    if (*system_name[host]) {
		    sk_getdomainname(dname, sizeof(dname));
		    if(*dname) {
			    len = strlen(system_name[host]);
			    system_name[host][len++] = '.';
			    system_name[host][len] = '\0';
				/*
				 * Truncate the domain name, if necessary, to prevent
				 * overflowing system_name[].
				 */
			    if (len + strlen(dname) + 1 > sizeof(system_name[host]))
				    dname[sizeof(system_name[host]) - len - 1] = '\0';
			    strcat(system_name[host], dname);
		    }
	    }
	    *var_len = strlen(system_name[host]);
	    return (uint8_t *) system_name[host];
    case SYSLOCATION:
        *var_len = strlen(system_location[host]);
        return (uint8_t *) system_location[host];
    case SYSSERVICES:
        ulret = system_services;
        return ((uint8_t *) & ulret);

#ifdef USING_MIBII_SYSORTABLE_MODULE
    case SYSORLASTCHANGE:
        ulret = netsnmp_timeval_uptime(&sysOR_lastchange);
        return ((uint8_t *) & ulret);
#endif

    default:
        DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_system\n",
                    vp->magic));
    }
    return NULL;
}
