/*
 *
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */


#include <net-snmp/net-snmp-config.h>
#include <sk/ha_host.h>


#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include <agent/mibgroup/util_funcs.h>
#include <sk/sk_stats.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/netinet/sk_in.h>
#include <net7g/net/sk_route.h>
#include <sk/sk_synch.h>
#include <net7g/netinet/sk_in_var.h>
#include <net7g/netinet/sk_if_ether.h>
#include <net7g/netinet/sk_in_systm.h>
#include <net7g/netinet/sk_ip.h>
#include <net7g/netinet/sk_in_pcb.h>
#include <net7g/netinet/sk_ip_var.h>     /* struct ipstat */
#include <tcp/tcp_var.h>
#include <tcp/tcp_fsm.h>        /* TCPS_ESTABLISHED, TCPS_CLOSE_WAIT */
#include <lib/rsrc_mgr.h>
#include <net7g/netinet/sk_ip_icmp.h>
#include "tcp_ipv6.h"
#include "interfaces.h"
#include <psm/psm.h>
//#include <net/ipsec/ontap.h>

#define getpid()	((unsigned long)(sk_my_pid()))

STATIC FindVarMethod var_tcp;
STATIC FindVarMethod var_tcpConnectionEntry;
STATIC FindVarMethod var_tcpListenerEntry;
STATIC FindVarMethod var_tcpConnEntry;

STATIC FindVarMethod _var_tcp;
STATIC FindVarMethod _var_tcpConnectionEntry;
STATIC FindVarMethod _var_tcpListenerEntry;
STATIC FindVarMethod _var_tcpConnEntry;

#define TCP_OID                     { SNMP_OID_MIB2, 6 }
#define TCPCONNECTIONENTRY_OID      { SNMP_OID_MIB2, 6, 19, 1 }
#define TCPLISTENERENTRY_OID        { SNMP_OID_MIB2, 6, 20, 1 }
#define TCPCONNENTRY_OID            { SNMP_OID_MIB2, 6, 13, 1 }

struct variable1 tcpConnectionEntry_variables[8] = {
    {TCPCONNECTIONLOCALADDRESSTYPE, INETADDRESSTYPE, RONLY,
                                        var_tcpConnectionEntry, 1, {1}},
    {TCPCONNECTIONLOCALADDRESS, INETADDRESS, RONLY,
                                        var_tcpConnectionEntry, 1, {2}},
    {TCPCONNECTIONLOCALPORT, INTEGER, RONLY,
                                        var_tcpConnectionEntry, 1, {3}},
    {TCPCONNECTIONREMADDRESSTYPE, INETADDRESSTYPE, RONLY,
                                        var_tcpConnectionEntry, 1, {4}},
    {TCPCONNECTIONREMADDRESS, INETADDRESS, RONLY,
                                        var_tcpConnectionEntry, 1, {5}},
    {TCPCONNECTIONREMPORT, INTEGER, RONLY,
                                        var_tcpConnectionEntry, 1, {6}},
    {TCPCONNECTIONSTATE, INTEGER, RONLY,
                                        var_tcpConnectionEntry, 1, {7}},
    {TCPCONNECTIONPROCESS, INTEGER, RONLY,
                                        var_tcpConnectionEntry, 1, {8}}
};
oid tcpConnectionEntry_variables_oid[9] = TCPCONNECTIONENTRY_OID;

struct variable1 tcpListenerEntry_variables[4] = {
    {TCPLISTENERLOCALADDRESSTYPE, INETADDRESSTYPE, RONLY,
                                        var_tcpListenerEntry, 1, {1}},
    {TCPLISTENERLOCALADDRESS, INETADDRESS, RONLY,
                                        var_tcpListenerEntry, 1, {2}},
    {TCPLISTENERLOCALPORT, INTEGER, RONLY,
                                        var_tcpListenerEntry, 1, {3}},
    {TCPLISTENERPROCESS, INTEGER, RONLY,
                                        var_tcpListenerEntry, 1, {4}}
};
oid tcpListenerEntry_variables_oid[9] = TCPLISTENERENTRY_OID;

struct variable1 tcp_variables[16] = {
    {TCPRTOALGORITHM, INTEGER, RONLY, var_tcp, 1, {1}},
    {TCPRTOMIN, INTEGER, RONLY, var_tcp, 1, {2}},
    {TCPRTOMAX, INTEGER, RONLY, var_tcp, 1, {3}},
    {TCPMAXCONN, INTEGER, RONLY, var_tcp, 1, {4}},
    {TCPACTIVEOPENS, COUNTER, RONLY, var_tcp, 1, {5}},
    {TCPPASSIVEOPENS, COUNTER, RONLY, var_tcp, 1, {6}},
    {TCPATTEMPTFAILS, COUNTER, RONLY, var_tcp, 1, {7}},
    {TCPESTABRESETS, COUNTER, RONLY, var_tcp, 1, {8}},
    {TCPCURRESTAB, GAUGE, RONLY, var_tcp, 1, {9}},
    {TCPINSEGS, COUNTER, RONLY, var_tcp, 1, {10}},
    {TCPOUTSEGS, COUNTER, RONLY, var_tcp, 1, {11}},
    {TCPRETRANSSEGS, COUNTER, RONLY, var_tcp, 1, {12}},
    {TCPINERRS, COUNTER, RONLY, var_tcp, 1, {14}},
    {TCPOUTRSTS, COUNTER, RONLY, var_tcp, 1, {15}},
    {TCPHCINSEGS, COUNTER64, RONLY, var_tcp, 1, {17}},
    {TCPHCOUTSEGS, COUNTER64, RONLY, var_tcp, 1, {18}}
};
oid tcp_variables_oid[7] = TCP_OID;

struct variable1 tcpConnEntry_variables[5] = {
    {TCPCONNSTATE, INTEGER, RONLY, var_tcpConnEntry, 1, {1}},
    {TCPCONNLOCALADDRESS, IPADDRESS, RONLY, var_tcpConnEntry, 1, {2}},
    {TCPCONNLOCALPORT, INTEGER32, RONLY, var_tcpConnEntry, 1, {3}},
    {TCPCONNREMADDRESS, IPADDRESS, RONLY, var_tcpConnEntry, 1, {4}},
    {TCPCONNREMPORT, INTEGER32, RONLY, var_tcpConnEntry, 1, {5}}
};
oid tcpConnEntry_variables_oid[9] = TCPCONNENTRY_OID;

static int32_t
TCP_Count_Connections(void)
{
        int32_t             Established = 0;
        struct inpcbhead *head;
        struct inpcb   *inp;
        struct tcpcb   *tcp;
        int32_t i, host;
        sk_domain_t old_domain;

        ASSERT(sk_in_domain(sk_domain_network));

        /*
         * Scan the control blocks
         */
        host = sk_gethost();
        old_domain = sk_domain_switch_begin(sk_domain_nwk_exclusive);
        SK_DOMAIN_REFERENCE(tcb_list_iterator, SK_DOMAIN_NETWORK);
        for (i = 0; (head = tcb_list_iterator (&i, host)) != NULL;) {
            for (inp = head->lh_first;
                 inp != NULL;
                 inp = inp->inp_list.le_next) {
                /* exclude things not in the default vfiler
                   by design, for Redhook.
                 */
                if((inp->inp_vfiler)
                    && (!VFILER_IS_DEFAULT(inp->inp_vfiler))) {
                    continue;
                }

                if (inet_lnaof(inp->inp_laddr) == INADDR_ANY) {
                    continue;
                }
                tcp = (struct tcpcb *) inp->inp_ppcb;
                if ((tcp->t_state == TCPS_ESTABLISHED) ||
                    (tcp->t_state == TCPS_CLOSE_WAIT)) {
                    Established++;
                }
            }
        }
        sk_domain_switch_end(old_domain);
        return (Established);
}



void init_tcp_ipv6(void)
{
    REGISTER_MIB("mib_2/tcp",
                  tcp_variables, variable1,
                  tcp_variables_oid);
    REGISTER_MIB("tcpConnectionTable/tcpConnectionEntry",
                  tcpConnectionEntry_variables, variable1,
                  tcpConnectionEntry_variables_oid);
    REGISTER_MIB("tcpListenerTable/tcpListenerEntry",
                  tcpListenerEntry_variables, variable1,
                  tcpListenerEntry_variables_oid);
    REGISTER_MIB("tcpConnTable/tcpConnEntry",
                  tcpConnEntry_variables, variable1,
                  tcpConnEntry_variables_oid);
}

uint8_t *
_var_tcp(SNMP_FM_ARGS)
{

    int32_t temp_return;
    int32_t noprocs = psm_installed_processor_count();

    switch(vp->magic) {
        case TCPRTOALGORITHM:
            long_return = 4; /* Van Jacobsen's algorithm */
            return (uint8_t *) & long_return;
        case TCPRTOMIN:
            long_return = TCPTV_MIN;
            return (uint8_t *) & long_return;
        case TCPRTOMAX:
            long_return = TCPTV_REXMTMAX * MSEC_PER_SEC;
            return (uint8_t *) & long_return;
        case TCPMAXCONN:
            long_return = -1;
            return (uint8_t *) & long_return;
        case TCPACTIVEOPENS:
            TCP_STAT(tcps_connattempt, long_return);
            return (uint8_t *) &long_return;
        case TCPPASSIVEOPENS:
            TCP_STAT(tcps_accepts, long_return);
            return (uint8_t *) &long_return;
        case TCPATTEMPTFAILS:
            TCP_STAT(tcps_conndrops, long_return);
            return (uint8_t *) &long_return;
        case TCPESTABRESETS:
            TCP_STAT(tcps_drops, long_return);
            return (uint8_t *) &long_return;
        case TCPCURRESTAB:
            long_return = TCP_Count_Connections();
            return (uint8_t *) & long_return;
        case TCPINSEGS:
            TCP_STAT(tcps_rcvtotal, long_return);
            return (uint8_t *) &long_return;
        case TCPOUTSEGS:
            TCP_STAT(tcps_sndtotal, long_return);
            return (uint8_t *) &long_return;
        case TCPRETRANSSEGS:
            TCP_STAT(tcps_sndrexmitpack, long_return);
            return (uint8_t *) &long_return;
        case TCPINERRS:
            TCP_STAT(tcps_rcvbadsum, long_return);
            TCP_STAT(tcps_rcvbadoff, temp_return);
            long_return += temp_return;
            TCP_STAT(tcps_rcvshort, temp_return);
            long_return += temp_return;
            return (uint8_t *) & long_return;
        case TCPOUTRSTS:
            TCP_STAT(tcps_sndctrl, long_return);
            return (uint8_t *) &long_return;
        case TCPHCINSEGS:
            TCP_STAT(tcps_rcvtotal, big_return);
	    counter64_return.high = HIGH32(big_return);
            counter64_return.low  = LOW32(big_return);
            *var_len = sizeof(counter64_return);
            return (uint8_t *)&counter64_return;
        case TCPHCOUTSEGS:
            TCP_STAT(tcps_sndtotal, big_return);
	    counter64_return.high = HIGH32(big_return);
            counter64_return.low  = LOW32(big_return);
            *var_len = sizeof(counter64_return);
            return (uint8_t *)&counter64_return;
        default:
                ERROR("var_tcp");
    }
        return NULL;
}

uint8_t *
var_tcp(SNMP_FM_ARGS)
{
        if (header_generic(vp, name, length, exact, var_len, write_method) ==
            MATCH_FAILED) {
                return NULL;
        }
        DOMAIN_SWITCH_NEST(sk_domain_network, _var_tcp,
                           (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_tcpConnectionEntry(SNMP_FM_ARGS)
{
        oid             lowest[SNMP_OID_MAX], *op;
        oid             current[SNMP_OID_MAX];
        uint8_t         *cp;
        int32_t             State = 0, LowState, i, host;
        struct inpcb   *inp;
        static struct inpcb Lowinpcb;
        /* Info about a particular connection */
        struct inpcbhead *head;
        sk_domain_t old_domain;
        int32_t j = 0;

        lowest[0] = 0xff;
        bcopy((char *) vp->name, (char *) current,
              (int32_t) vp->namelen * sizeof(oid));
        /* find "next" connection */
        LowState = -1;  /* Don't have one yet */
        host = sk_gethost();
        old_domain = sk_domain_switch_begin(sk_domain_nwk_exclusive);
        SK_DOMAIN_REFERENCE(tcb_list_iterator, SK_DOMAIN_NETWORK);
        for (i = 0; (head = tcb_list_iterator (&i, host)) != NULL;) {
            for (inp = head->lh_first;
                 inp != NULL;
                 inp = inp->inp_list.le_next) {
                /* exclude things not in the default vfiler */
                if((inp->inp_vfiler)
                    && (!VFILER_IS_DEFAULT(inp->inp_vfiler))) {
                    continue;
                }
                if(((struct tcpcb *)inp->inp_ppcb)->t_state ==
                               TCPS_LISTEN) {
                    continue;
                }
                if(!(inp->inp_flags & INP_IPV6)) {
                     cp = (uint8_t *) & inp->inp_laddr.s_addr;
                     op = current + vp->namelen;

                     /* Address type */
                     *op++ = 1;

                     /* IP address */
                     *op++ = *cp++;
                     *op++ = *cp++;
                     *op++ = *cp++;
                     *op++ = *cp++;

                     *op++ = ntohs(inp->inp_lport);

                     /* remote address type */
                     *op++ = 1;

                     cp = (uint8_t *) & inp->inp_faddr.s_addr;

                     *op++ = *cp++;
                     *op++ = *cp++;
                     *op++ = *cp++;
                     *op++ = *cp++;

                     State = ((struct tcpcb *)inp->inp_ppcb)->t_state;

                     *op = ntohs(inp->inp_fport);

                     /* Add one to the size to get the OID length right */
                     j = (op - current) + 1;
                }
                else {
                     j = vp->namelen;

                     current[j++] = 2;
                     for(i = 0; i < sizeof(struct in6_addr); i++) {
                        current[j++] = inp->in6p_laddr.s6_addr[i];
                     }

                     current[j++] = ntohs(inp->in6p_lport);

                     current[j++] = 2;
                     for(i = 0; i < sizeof(struct in6_addr); i++) {
                        current[j++] = inp->in6p_faddr.s6_addr[i];
                     }

                     current[j] = ntohs(inp->in6p_fport);

		     State = ((struct tcpcb *)inp->inp_ppcb)->t_state;

                     /* Increment by one to get the OID length right */
                     j++;
                }

                if (exact) {
                    if (snmp_oid_compare(current, j, name, *length)
                                    == 0) {
                        bcopy((char *) current, (char *) lowest,
                                   j * sizeof(oid));
                        LowState = State;
                        Lowinpcb = *inp;
                        break; /* no need to search further */
                    }
                } else {
                    if((snmp_oid_compare(current, j, name, *length)
                               > 0) && ((LowState < 0) ||
                          ((snmp_oid_compare(current, j, lowest, j)
                               < 0)))) {
                        /*
                         * if new one is greater than
                         * input and closer to input
                         * than previous lowest, save
                         * this one as the "next"
                         * one.
                         */
                         bcopy((char *) current,
                               (char *) lowest,
                                j * sizeof(oid));
                         LowState = State;
                         Lowinpcb = *inp;
                    }
                }
            }
        }
        sk_domain_switch_end(old_domain);
        if(LowState < 0) {
            return NULL;
        }
        if((Lowinpcb.inp_flags & INP_IPV6)) {
            j = 46;
        }
        else {
            j = 22;
        }
        bcopy((char *) lowest, (char *) name,
              (int32_t) j * sizeof(oid));
        *length = j;
        *write_method = 0;
        *var_len = sizeof(int32_t);

        switch (vp->magic) {
            case TCPCONNECTIONSTATE:{
                static int32_t   StateMap[] =
                {1, 2, 3, 4, 5, 8, 6, 10, 9, 7, 11};

                return (uint8_t *) & StateMap[LowState];
            }
            case TCPCONNECTIONLOCALADDRESS:
                if(!(Lowinpcb.inp_flags & INP_IPV6)) {
                    return (uint8_t *) & Lowinpcb.inp_laddr.s_addr;
                }
                else {
                    *var_len = sizeof(struct in6_addr);
                    return (uint8_t *)& Lowinpcb.in6p_laddr.s6_addr;
                }
            case TCPCONNECTIONLOCALPORT:
                long_return = ntohs(Lowinpcb.inp_lport);
                return (uint8_t *) & long_return;
            case TCPCONNECTIONREMADDRESS:
                if(!(Lowinpcb.inp_flags & INP_IPV6)) {
                    return (uint8_t *) & Lowinpcb.inp_faddr.s_addr;
                }
                else {
                    *var_len = sizeof(struct in6_addr);
                    return (uint8_t *)& Lowinpcb.in6p_faddr.s6_addr;
                }
            case TCPCONNECTIONREMPORT:
                long_return = ntohs(Lowinpcb.inp_fport);
                return (uint8_t *) & long_return;
            case TCPCONNECTIONLOCALADDRESSTYPE:
            case TCPCONNECTIONREMADDRESSTYPE:
                if((Lowinpcb.inp_flags & INP_IPV6)) {
                    long_return = 2;
                }
                else {
                    long_return = 1;
                }
                return (uint8_t *) & long_return;
            case TCPCONNECTIONPROCESS:
                /* Don't have anything to return yet */
                //long_return = 0;
				long_return = (int32_t)getpid();
                return (uint8_t *) & long_return;
        }
        return NULL;
}

uint8_t *
var_tcpConnectionEntry(SNMP_FM_ARGS)
{
        DOMAIN_SWITCH_NEST(sk_domain_network, _var_tcpConnectionEntry,
                           (vp, name, length, exact, var_len, write_method));
}


/*
 * Print INTERNET connections
 */
//SK_DOMAIN_EXPORT(TCP_Count_Connections, SK_DOMAIN_NETWORK);
uint8_t *
_var_tcpListenerEntry(SNMP_FM_ARGS)
{
        oid             lowest[SNMP_OID_MAX], *op;
        oid             current[SNMP_OID_MAX];
        uint8_t         *cp;
        int32_t             i, host;
        struct inpcb   *inp;
        static struct inpcb Lowinpcb;
        /* Info about a particular connection */
        struct inpcbhead *head;
        sk_domain_t old_domain;
        int32_t j = 0;
        bool found = false;

        lowest[0] = 0xff;
        bcopy((char *) vp->name, (char *) current,
              (int32_t) vp->namelen * sizeof(oid));
		
        host = sk_gethost();
        old_domain = sk_domain_switch_begin(sk_domain_nwk_exclusive);
        SK_DOMAIN_REFERENCE(tcb_list_iterator, SK_DOMAIN_NETWORK);
        for (i = 0; (head = tcb_list_iterator (&i, host)) != NULL;) {
            for (inp = head->lh_first;
                 inp != NULL;
                 inp = inp->inp_list.le_next) {
                /* exclude things not in the default vfiler
                   by design, for Redhook.
                 */
                if((inp->inp_vfiler)
                    && (!VFILER_IS_DEFAULT(inp->inp_vfiler))) {
                    continue;
                }
                if(((struct tcpcb *)inp->inp_ppcb)->t_state !=
                               TCPS_LISTEN) {
                    continue;
                }
                if(!(inp->inp_flags & INP_IPV6)) {
                    cp = (uint8_t *) & inp->inp_laddr.s_addr;
                    op = current + vp->namelen;

                    /* Address type */
                    *op++ = 1;

                    /* IP address */
                    *op++ = *cp++;
                    *op++ = *cp++;
                    *op++ = *cp++;
                    *op++ = *cp++;

                    *op = ntohs(inp->inp_lport);

                    /* Add one to the size to get the OID length right */
                    j = (op - current) + 1;
                }
                else {
                    j = vp->namelen;

                    current[j++] = 2;
                    for(i = 0; i < sizeof(struct in6_addr); i++) {
                        current[j++] = inp->in6p_laddr.s6_addr[i];
                    }

                    current[j] = ntohs(inp->in6p_lport);

                    /* Increment by one to get the OID length right */
                    j++;
                }

                if (exact) {
                    if (snmp_oid_compare(current, j, name, *length)
                               == 0) {
                        bcopy((char *) current, (char *) lowest,
                                   j * sizeof(oid));
                        found = true;
                        Lowinpcb = *inp;
                        break; /* no need to search further */
                    }
                } else {
                    if((snmp_oid_compare(current, j, name, *length)
                              > 0) && ((!found) ||
                              ((snmp_oid_compare(current, j, lowest, j)
                              < 0)))) {
                        /*
                         * if new one is greater than
                         * input and closer to input
                         * than previous lowest, save
                         * this one as the "next"
                         * one.
                         */
                        bcopy((char *) current,
                                  (char *) lowest,
                                  j * sizeof(oid));
                        found = true;
                        Lowinpcb = *inp;
                    }
                }
            }
        }
        sk_domain_switch_end(old_domain);
        if(!found) {
                return (NULL);
        }
        if(!(Lowinpcb.inp_flags & INP_IPV6)) {
                j = 16;
        }
        else {
                j = 28;
        }
        bcopy((char *) lowest, (char *) name,
              (int32_t) j * sizeof(oid));
        *length = j;
        *write_method = 0;
        *var_len = sizeof(int32_t);

        switch (vp->magic) {
        case TCPLISTENERLOCALADDRESS:
           if(!(Lowinpcb.inp_flags & INP_IPV6)) {
                return (uint8_t *) & Lowinpcb.inp_laddr.s_addr;
           }
           else {
                *var_len = sizeof(struct in6_addr);
                 return (uint8_t *)& Lowinpcb.in6p_laddr.s6_addr;
           }
        case TCPLISTENERLOCALPORT:
                long_return = ntohs(Lowinpcb.inp_lport);
                return (uint8_t *) & long_return;
        case TCPLISTENERLOCALADDRESSTYPE:
           if(!(Lowinpcb.inp_flags & INP_IPV6)) {
                long_return = 1;
           }
           else {
                long_return = 2;
           }
           return (uint8_t *) & long_return;
        case TCPLISTENERPROCESS:
           /* Don't have anything to return yet. */
                //long_return = 0;
				long_return = (int32_t)getpid();
                return (uint8_t *) & long_return;
        }
        return NULL;
}

uint8_t *
var_tcpListenerEntry(SNMP_FM_ARGS)
{
        DOMAIN_SWITCH_NEST(sk_domain_network, _var_tcpListenerEntry,
                           (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_tcpConnEntry(SNMP_FM_ARGS)
{
        oid             lowest[SNMP_OID_MAX], *op;
        oid             current[SNMP_OID_MAX];
        uint8_t         *cp;
        int32_t         State, LowState, i, host;
		int32_t			j = 0;
        struct inpcb   *inp;
        static struct inpcb Lowinpcb;
        /* Info about a particular connection */
        struct inpcbhead *head;
        sk_domain_t old_domain;

        lowest[0] = 0xff;
        bcopy((char *) vp->name, (char *) current,
              (int32_t) vp->namelen * sizeof(oid));
				
        LowState = -1;  /* Don't have one yet */
        host = sk_gethost();
        old_domain = sk_domain_switch_begin(sk_domain_nwk_exclusive);
        SK_DOMAIN_REFERENCE(tcb_list_iterator, SK_DOMAIN_NETWORK);
        for (i = 0; (head = tcb_list_iterator (&i, host)) != NULL;) {
            for (inp = head->lh_first;
                 inp != NULL;
                 inp = inp->inp_list.le_next) {
                /* exclude things not in the default vfiler
                 * by design, for Redhook.
                 */

                if((inp->inp_vfiler)
                    && (!VFILER_IS_DEFAULT(inp->inp_vfiler))) {
                    continue;
                }
                /* We'll consider only ipv4 addresses */
                if((inp->inp_flags & INP_IPV6)) {
                    continue;
                }

                cp = (uint8_t *) & inp->inp_laddr.s_addr;
                op = current + vp->namelen;
                *op++ = *cp++;
                *op++ = *cp++;
                *op++ = *cp++;
                *op++ = *cp++;

                *op++ = ntohs(inp->inp_lport);

                cp = (uint8_t *) & inp->inp_faddr.s_addr;
                
				*op++ = *cp++;
                *op++ = *cp++;
                *op++ = *cp++;
                *op++ = *cp++;

                State = ((struct tcpcb *)inp->inp_ppcb)->t_state;

                *op = ntohs(inp->inp_fport);

				j = (op - current) + 1;

                if (exact) {
                    if (snmp_oid_compare(current, j, name, *length)
                            == 0) {
                        bcopy((char *) current, (char *) lowest,
                                j * sizeof(oid));
                        LowState = State;
                        Lowinpcb = *inp;
                        break; /* no need to search further */
                    }
                } else {
                    if ((snmp_oid_compare(current, j, name, *length) > 0) &&
                            ((LowState < 0) ||
                            (snmp_oid_compare(current, j, lowest, j) < 0))) {
                        /*
                         * if new one is greater than
                         * input and closer to input
                         * than previous lowest, save
                         * this one as the "next"
                         * one.
                         */
                        bcopy((char *) current,
                              (char *) lowest,
                              j * sizeof(oid));
                        LowState = State;
                        Lowinpcb = *inp;
                    }
                }
            }
        }
        sk_domain_switch_end(old_domain);
        if (LowState < 0) {
            return (NULL);
        }
        bcopy((char *) lowest, (char *) name,
              (int32_t) j * sizeof(oid));
        *length = j;
        *write_method = 0;
        *var_len = sizeof(int32_t);

        switch (vp->magic) {
        case TCPCONNSTATE:{
                static int32_t   StateMap[] =
                {1, 2, 3, 4, 5, 8, 6, 10, 9, 7, 11};

                return (uint8_t *) & StateMap[LowState];
        }
        case TCPCONNLOCALADDRESS:
                return (uint8_t *) & Lowinpcb.inp_laddr.s_addr;
        case TCPCONNLOCALPORT:
                long_return = ntohs(Lowinpcb.inp_lport);
                return (uint8_t *) & long_return;
        case TCPCONNREMADDRESS:
                return (uint8_t *) & Lowinpcb.inp_faddr.s_addr;
        case TCPCONNREMPORT:
                long_return = ntohs(Lowinpcb.inp_fport);
                return (uint8_t *) & long_return;
        }
        return NULL;
}

uint8_t *
var_tcpConnEntry(SNMP_FM_ARGS)
{
        DOMAIN_SWITCH_NEST(sk_domain_network, _var_tcpConnEntry,
                           (vp, name, length, exact, var_len, write_method));
}
