/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */

/*
 *  Template MIB group implementation - udp, tcp, icmp.c, and ip. Most of the
 *  code for this file was taken from the old netapp snmp_vars.c code.
 *
 */

#include <net-snmp/net-snmp-config.h>
#include <sk/ha_host.h>


#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include <agent/mibgroup/util_funcs.h>
#include <sk/sk_stats.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/netinet/sk_in.h>
#include <sys/endian.h>
#include <net7g/net/sk_route.h>
#include <sk/sk_synch.h>
#include <net7g/netinet/sk_in_var.h>
#include <net7g/netinet/sk_if_ether.h>
#include <net7g/netinet/sk_in_systm.h>
#include <net7g/netinet/sk_ip.h>
#include <net7g/netinet/sk_in_pcb.h>
#include <net7g/netinet/sk_ip_var.h>	/* struct ipstat */
#include <tcp/tcp_var.h>
#include <tcp/tcp_fsm.h>	/* TCPS_ESTABLISHED, TCPS_CLOSE_WAIT */
#include <lib/rsrc_mgr.h>
#include <net7g/netinet/sk_udp_var.h> 	/* struct udpstat */
#include <net7g/netinet/sk_ip_icmp.h>
#include <net7g/netinet/sk_icmp_var.h>	/* struct icmpstat */
#include "udp_icmp_tcp_ip.h"
#include "interfaces.h"
#include <psm/psm.h>

extern struct icmpstat icmpstat;
extern int      ipforwarding;

STATIC FindVarMethod var_icmp;
STATIC FindVarMethod var_ip;
STATIC FindVarMethod var_ipNetToMediaEntry;
STATIC FindVarMethod var_ipAddrEntry;

STATIC FindVarMethod _var_icmp;
STATIC FindVarMethod _var_ip;
STATIC FindVarMethod _var_ipNetToMediaEntry;
STATIC FindVarMethod _var_ipAddrEntry;

#define ICMP_OID              { SNMP_OID_MIB2, 5 }
#define IPADDRENTRY_OID       { SNMP_OID_MIB2, 4, 20, 1 }
#define IP_OID                { SNMP_OID_MIB2, 4 }
#define IPNETTOMEDIAENTRY_OID { SNMP_OID_MIB2, 4, 22, 1 }

struct variable1 icmp_variables[26] = {
	{ICMPINMSGS, COUNTER, RONLY, var_icmp, 1, {1}},
	{ICMPINERRORS, COUNTER, RONLY, var_icmp, 1, {2}},
	{ICMPINDESTUNREACHS, COUNTER, RONLY, var_icmp, 1, {3}},
	{ICMPINTIMEEXCDS, COUNTER, RONLY, var_icmp, 1, {4}},
	{ICMPINPARMPROBS, COUNTER, RONLY, var_icmp, 1, {5}},
	{ICMPINSRCQUENCHS, COUNTER, RONLY, var_icmp, 1, {6}},
	{ICMPINREDIRECTS, COUNTER, RONLY, var_icmp, 1, {7}},
	{ICMPINECHOS, COUNTER, RONLY, var_icmp, 1, {8}},
	{ICMPINECHOREPS, COUNTER, RONLY, var_icmp, 1, {9}},
	{ICMPINTIMESTAMPS, COUNTER, RONLY, var_icmp, 1, {10}},
	{ICMPINTIMESTAMPREPS, COUNTER, RONLY, var_icmp, 1, {11}},
	{ICMPINADDRMASKS, COUNTER, RONLY, var_icmp, 1, {12}},
	{ICMPINADDRMASKREPS, COUNTER, RONLY, var_icmp, 1, {13}},
	{ICMPOUTMSGS, COUNTER, RONLY, var_icmp, 1, {14}},
	{ICMPOUTERRORS, COUNTER, RONLY, var_icmp, 1, {15}},
	{ICMPOUTDESTUNREACHS, COUNTER, RONLY, var_icmp, 1, {16}},
	{ICMPOUTTIMEEXCDS, COUNTER, RONLY, var_icmp, 1, {17}},
	{ICMPOUTPARMPROBS, COUNTER, RONLY, var_icmp, 1, {18}},
	{ICMPOUTSRCQUENCHS, COUNTER, RONLY, var_icmp, 1, {19}},
	{ICMPOUTREDIRECTS, COUNTER, RONLY, var_icmp, 1, {20}},
	{ICMPOUTECHOS, COUNTER, RONLY, var_icmp, 1, {21}},
	{ICMPOUTECHOREPS, COUNTER, RONLY, var_icmp, 1, {22}},
	{ICMPOUTTIMESTAMPS, COUNTER, RONLY, var_icmp, 1, {23}},
	{ICMPOUTTIMESTAMPREPS, COUNTER, RONLY, var_icmp, 1, {24}},
	{ICMPOUTADDRMASKS, COUNTER, RONLY, var_icmp, 1, {25}},
	{ICMPOUTADDRMASKREPS, COUNTER, RONLY, var_icmp, 1, {26}},
};
oid icmp_variables_oid[7] = ICMP_OID;

struct variable1 ipAddrEntry_variables[5] = {
	{IPADENTADDR, IPADDRESS, RONLY, var_ipAddrEntry, 1, {1}},
	{IPADENTIFINDEX, INTEGER, RONLY, var_ipAddrEntry, 1, {2}},
	{IPADENTNETMASK, IPADDRESS, RONLY, var_ipAddrEntry, 1, {3}},
	{IPADENTBCASTADDR, INTEGER, RONLY, var_ipAddrEntry, 1, {4}},
	{IPADENTREASMMAXSIZE, INTEGER, RONLY, var_ipAddrEntry, 1, {5}},
};
oid ipAddrEntry_variables_oid[9] = IPADDRENTRY_OID;

struct variable1 ip_variables[20] = {
	{IPFORWARDING, INTEGER, RWRITE, var_ip, 1, {1}},
	{IPDEFAULTTTL, INTEGER, RWRITE, var_ip, 1, {2}},
	{IPINRECEIVES, COUNTER, RONLY, var_ip, 1, {3}},
	{IPINHDRERRORS, COUNTER, RONLY, var_ip, 1, {4}},
	{IPINADDRERRORS, COUNTER, RONLY, var_ip, 1, {5}},
	{IPFORWDATAGRAMS, COUNTER, RONLY, var_ip, 1, {6}},
	{IPINUNKNOWNPROTOS, COUNTER, RONLY, var_ip, 1, {7}},
	{IPINDISCARDS, COUNTER, RONLY, var_ip, 1, {8}},
	{IPINDELIVERS, COUNTER, RONLY, var_ip, 1, {9}},
	{IPOUTREQUESTS, COUNTER, RONLY, var_ip, 1, {10}},
	{IPOUTDISCARDS, COUNTER, RONLY, var_ip, 1, {11}},
	{IPOUTNOROUTES, COUNTER, RONLY, var_ip, 1, {12}},
	{IPREASMTIMEOUT, INTEGER, RONLY, var_ip, 1, {13}},
	{IPREASMREQDS, COUNTER, RONLY, var_ip, 1, {14}},
	{IPREASMOKS, COUNTER, RONLY, var_ip, 1, {15}},
	{IPREASMFAILS, COUNTER, RONLY, var_ip, 1, {16}},
	{IPFRAGOKS, COUNTER, RONLY, var_ip, 1, {17}},
	{IPFRAGFAILS, COUNTER, RONLY, var_ip, 1, {18}},
	{IPFRAGCREATES, COUNTER, RONLY, var_ip, 1, {19}},
	{IPROUTINGDISCARDS, COUNTER, RONLY, var_ip, 1, {23}},
};
oid ip_variables_oid[7] = IP_OID;

struct variable1 ipNetToMediaEntry_variables[4] = {
	{IPNETTOMEDIAIFINDEX, INTEGER, RWRITE, var_ipNetToMediaEntry, 1, {1}},
	{IPNETTOMEDIAPHYSADDRESS, PHYSADDRESS, RWRITE, var_ipNetToMediaEntry, 1, {2}},
	{IPNETTOMEDIANETADDRESS, IPADDRESS, RWRITE, var_ipNetToMediaEntry, 1, {3}},
	{IPNETTOMEDIATYPE, INTEGER, RWRITE, var_ipNetToMediaEntry, 1, {4}},
};
oid ipNetToMediaEntry_variables_oid[9] = IPNETTOMEDIAENTRY_OID;

void
init_icmp_ip(void)
{
    /*
     * register ourselves with the agent to handle our mib tree 
     */
     REGISTER_MIB("mib_2/icmp",
                  icmp_variables, variable1,
                  icmp_variables_oid); 
     REGISTER_MIB("mib_2/ip",
                  ip_variables, variable1,
                  ip_variables_oid); 
     REGISTER_MIB("ipAddrTable/ipAddrEntry",
                  ipAddrEntry_variables, variable1,
                  ipAddrEntry_variables_oid); 
     REGISTER_MIB("ipNetToMediaTable/ipNetToMediaEntry",
                  ipNetToMediaEntry_variables, variable1,
                  ipNetToMediaEntry_variables_oid); 
}

uint8_t *
_var_icmp(SNMP_FM_ARGS)
{
	register int    i;

	ASSERT(sk_in_domain(sk_domain_network));
	SK_DOMAIN_REFERENCE(icmpstat, SK_DOMAIN_NETWORK);

	switch (vp->magic) {
	case ICMPINMSGS:
		long_return = icmpstat.icps_badcode + icmpstat.icps_tooshort +
			icmpstat.icps_checksum + icmpstat.icps_badlen;
		for (i = 0; i <= ICMP_MAXTYPE; i++)
			long_return += icmpstat.icps_inhist[i];
		return (uint8_t *) & long_return;
	case ICMPINERRORS:
		long_return = icmpstat.icps_badcode + icmpstat.icps_tooshort +
			icmpstat.icps_checksum + icmpstat.icps_badlen;
		return (uint8_t *) & long_return;
	case ICMPINDESTUNREACHS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_UNREACH];
	case ICMPINTIMEEXCDS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_TIMXCEED];
	case ICMPINPARMPROBS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_PARAMPROB];
	case ICMPINSRCQUENCHS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_SOURCEQUENCH];
	case ICMPINREDIRECTS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_REDIRECT];
	case ICMPINECHOS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_ECHO];
	case ICMPINECHOREPS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_ECHOREPLY];
	case ICMPINTIMESTAMPS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_TSTAMP];
	case ICMPINTIMESTAMPREPS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_TSTAMPREPLY];
	case ICMPINADDRMASKS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_MASKREQ];
	case ICMPINADDRMASKREPS:
		return (uint8_t *) & icmpstat.icps_inhist[ICMP_MASKREPLY];
	case ICMPOUTMSGS:
		long_return = icmpstat.icps_oldshort + icmpstat.icps_oldicmp;
		for (i = 0; i <= ICMP_MAXTYPE; i++)
			long_return += icmpstat.icps_outhist[i];
		return (uint8_t *) & long_return;
	case ICMPOUTERRORS:
		long_return = icmpstat.icps_oldshort + icmpstat.icps_oldicmp;
		return (uint8_t *) & long_return;
	case ICMPOUTDESTUNREACHS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_UNREACH];
	case ICMPOUTTIMEEXCDS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_TIMXCEED];
	case ICMPOUTPARMPROBS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_PARAMPROB];
	case ICMPOUTSRCQUENCHS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_SOURCEQUENCH];
	case ICMPOUTREDIRECTS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_REDIRECT];
	case ICMPOUTECHOS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_ECHO];
	case ICMPOUTECHOREPS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_ECHOREPLY];
	case ICMPOUTTIMESTAMPS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_TSTAMP];
	case ICMPOUTTIMESTAMPREPS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_TSTAMPREPLY];
	case ICMPOUTADDRMASKS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_MASKREQ];
	case ICMPOUTADDRMASKREPS:
		return (uint8_t *) & icmpstat.icps_outhist[ICMP_MASKREPLY];
	default:
		ERROR("var_icmp");
	}
	return NULL;
}

uint8_t *
var_icmp(SNMP_FM_ARGS)
{
	if (header_generic(vp, name, length, exact, var_len, write_method) == 
	    MATCH_FAILED) {
		return NULL;
	}                                                             
	DOMAIN_SWITCH_NEST(sk_domain_network, _var_icmp,
			   (vp, name, length, exact, var_len, write_method)); 
}

uint8_t *
_var_ip(SNMP_FM_ARGS)
{
	int temp_return;
	int noprocs = psm_installed_processor_count();

	ASSERT(sk_in_domain(sk_domain_network));
	SK_DOMAIN_REFERENCE(_ipstats, SK_DOMAIN_NETWORK);

	switch (vp->magic) {
	case IPFORWARDING:
		SK_DOMAIN_REFERENCE(ipforwarding, SK_DOMAIN_NETWORK);
		if (ipforwarding)
			long_return = 1;	/* GATEWAY */
		else
			long_return = 2;	/* GATEWAY configured as HOST */
		return (uint8_t *) & long_return;
	case IPDEFAULTTTL:
		SK_DOMAIN_REFERENCE(ip_defttl, SK_DOMAIN_NETWORK);
		return (uint8_t *) & ip_defttl;
	case IPINRECEIVES:
		IP_STAT(ips_total, long_return);
		return (uint8_t *) &long_return;
	case IPINHDRERRORS:
		IP_STAT(ips_badsum, long_return);
		IP_STAT(ips_tooshort, temp_return);
		long_return += temp_return;
		IP_STAT(ips_toosmall, temp_return);
		long_return += temp_return;
		IP_STAT(ips_badhlen, temp_return);
		long_return += temp_return;
		IP_STAT(ips_badlen, temp_return);
		long_return += temp_return;
		return (uint8_t *) & long_return;
	case IPINADDRERRORS:
		IP_STAT(ips_cantforward, long_return);
		return (uint8_t *) &long_return;
	case IPFORWDATAGRAMS:
		IP_STAT(ips_forward, long_return);
		return (uint8_t *) &long_return;
	case IPINUNKNOWNPROTOS:
		IP_STAT(ips_noproto, long_return);
		return (uint8_t *) &long_return;
	case IPINDISCARDS:
		/* XXX */
		long_return = 0;
		return (uint8_t *) & long_return;
	case IPINDELIVERS:
		IP_STAT(ips_total, long_return);
		IP_STAT(ips_badsum, temp_return);
		long_return -= temp_return;
		IP_STAT(ips_tooshort, temp_return);
		long_return -= temp_return;
		IP_STAT(ips_toosmall, temp_return);
		long_return -= temp_return;
		IP_STAT(ips_badhlen, temp_return);
		long_return -= temp_return;
		IP_STAT(ips_badlen, temp_return);
		long_return -= temp_return;
		return (uint8_t *) & long_return;
	case IPOUTREQUESTS:
		IP_STAT(ips_localout, long_return);
		return (uint8_t *) &long_return;
	case IPOUTDISCARDS:
		IP_STAT(ips_odropped, long_return);
		return (uint8_t *) &long_return;
	case IPOUTNOROUTES:
		IP_STAT(ips_cantforward, long_return);
		return (uint8_t *) &long_return;
	case IPREASMTIMEOUT:
		long_return = IPFRAGTTL;
		return (uint8_t *) & long_return;
	case IPREASMREQDS:
		IP_STAT(ips_fragments, long_return);
		return (uint8_t *) &long_return;
	case IPREASMOKS:
		IP_STAT(ips_fragments, long_return);
		return (uint8_t *) &long_return;
	case IPREASMFAILS:
		IP_STAT(ips_fragdropped, long_return);
		IP_STAT(ips_fragtimeout, temp_return);
		long_return += temp_return;
		return (uint8_t *) & long_return;
	case IPFRAGOKS:
		IP_STAT(ips_fragmented, long_return);
		return (uint8_t *) &long_return;
	case IPFRAGFAILS:
		IP_STAT(ips_cantfrag, long_return);
		return (uint8_t *) &long_return;
	case IPFRAGCREATES:
		IP_STAT(ips_ofragments, long_return);
		return (uint8_t *) &long_return;
	case IPROUTINGDISCARDS:
		/* XXX */
		long_return = 0;
		return (uint8_t *) & long_return;
	default:
		ERROR("var_ip");
	}
	return NULL;
}

uint8_t *
var_ip(SNMP_FM_ARGS)
{
	if (header_generic(vp, name, length, exact, var_len, write_method) == 
	    MATCH_FAILED) {
		return NULL;
	}                                                             
	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ip,
			   (vp, name, length, exact, var_len, write_method)); 
}

uint8_t *
_var_ipAddrEntry(SNMP_FM_ARGS)
{
	/*
	 * object identifier is of form: 1.3.6.1.2.1.4.20.1.?.A.B.C.D,  where
	 * A.B.C.D is IP address. IPADDR starts at offset 10.
	 */
	oid             lowest[SNMP_OID_MAX];
	oid             current[SNMP_OID_MAX], *op;
	uint8_t         *cp;
	int             interface, lowinterface = -1;
	struct ifnet *ifp;
	static struct in_ifaddr *ia, lowin_ifaddr;
	int host = sk_gethost();

	lowest[0] = 0xff;
	/*
	 * fill in object part of name for current (less sizeof instance
	 * part)
	 */

	bcopy((char *) vp->name, (char *) current, (int) vp->namelen * sizeof(oid));

	ASSERT(sk_in_domain(sk_domain_network));
	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);

	/*
	 * Playing the host game with the interface index ...
	 */
	for (interface = 1, ifp = ifnet; ifp; ifp = ifp->if_next) {
		/* skip things not in the default ipspace */
		if(ifp->ipsid != 0)
			continue;

		ia = ifnet_inaddr(ifp, sk_gethost(), AF_INET);
		if (ia == NULL) {
			if (host == SK_HOST_LOCAL)
				interface++;
			continue;
		}

		cp = (uint8_t *) & (((struct sockaddr_in *)
				& (ia->ia_addr))->sin_addr.s_addr);
		op = &current[10];
		*op++ = *cp++;
		*op++ = *cp++;
		*op++ = *cp++;
		*op++ = *cp++;

		if (exact) {
			if (snmp_oid_compare(current, 14, name, *length) == 0) {
				bcopy((char *) current, (char *) lowest,
					14 * sizeof(oid));
				lowinterface = interface;
				lowin_ifaddr = *ia;
				break;	/* no need to search further */
			}
		} else {
			if ((snmp_oid_compare(current, 14, name, *length) > 0) &&
			    ((lowinterface == -1) ||
			    (snmp_oid_compare(current, 14, lowest, 14) < 0))) {
				/*
				 * if new one is greater than input and
				 * closer to input than previous lowest, save
				 * this one as the "next" one.
				 */
				lowinterface = interface;
				lowin_ifaddr = *ia;
				bcopy((char *) current, (char *) lowest,
					14 * sizeof(oid));
			}
		}
		interface++;
	}
	if (lowinterface == -1)
		return (NULL);
	bcopy((char *) lowest, (char *) name, 14 * sizeof(oid));
	*length = 14;
	*write_method = 0;
	*var_len = sizeof(long_return);
	switch (vp->magic) {
	case IPADENTADDR:
		return (uint8_t *) & ((struct sockaddr_in *)
			& lowin_ifaddr.ia_addr)->sin_addr.s_addr;
	case IPADENTIFINDEX:
		long_return = lowinterface;
		return (uint8_t *) & long_return;
	case IPADENTNETMASK:
		long_return = ntohl(lowin_ifaddr.ia_subnetmask);
		return (uint8_t *) & long_return;
	case IPADENTBCASTADDR:
		long_return = ntohl(((struct sockaddr_in *)
			& lowin_ifaddr.ia_addr)->sin_addr.s_addr) & 1;
		return (uint8_t *) & long_return;
	case IPADENTREASMMAXSIZE:
		long_return = 0;  /* Should this be IP_MAXPACKET? Not obvious
					from ip_reass() */
		return (uint8_t *) & long_return;
	default:
		ERROR("var_ipAddrEntry");
	}
	return NULL;
}
uint8_t *
var_ipAddrEntry(SNMP_FM_ARGS)
{
	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipAddrEntry,
			   (vp,name,length,exact,var_len,write_method)); 
}

uint8_t *
_var_ipNetToMediaEntry(SNMP_FM_ARGS)
{
	/*
	 * object identifier is of form: 1.3.6.1.2.1.4.22.1.?.I.A.B.C.D,  where
	 * A.B.C.D is IP address, and I is the ifIndex of the interface that
	 * has been configured with that interface.  The ifIndex starts at
	 * offset 10.
	 */
	oid             lowest[SNMP_OID_MAX];
	oid             current[SNMP_OID_MAX], *op;
	uint8_t         *cp;
	int             interface, lowinterface = -1;
	struct ifnet *ifp, *lowin_ifp = (void *) 0;
	static struct in_ifaddr *ia, lowin_ifaddr;
	struct arpcom *ap;
	char	Name[IFNAMSIZ];
	int host = sk_gethost();
	static const int entry_oidlen = 15;

	/*
	 * fill in object part of name for current (less sizeof instance
	 * part)
	 */

	bcopy((char *) vp->name, (char *) current, (int) vp->namelen *
							sizeof(oid));
	lowest[0] = 0xff;

	ASSERT(sk_in_domain(sk_domain_network));
	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);

	/*
	 * Playing the host game with the interface index ...
	 */
	for (interface = 1, ifp = ifnet; ifp; ifp = ifp->if_next ) {
		/* skip things not in the default ipspace */
		if(ifp->ipsid != 0)
			continue;

		ia = ifnet_inaddr(ifp, sk_gethost(), AF_INET);
		if (ia == NULL) {
			if (host == SK_HOST_LOCAL)
				interface++;
			continue;
		}

		cp = (uint8_t *) & (((struct sockaddr_in *)
				& (ia->ia_addr))->sin_addr.s_addr);
		op = &current[10];
		*op++ = interface;
		*op++ = *cp++;
		*op++ = *cp++;
		*op++ = *cp++;
		*op++ = *cp++;

		if (exact) {
			if (snmp_oid_compare(current, entry_oidlen, 
					     name, *length) == 0) {
				bcopy((char *) current, (char *) lowest,
				      entry_oidlen * sizeof(oid));
				lowinterface = interface;
				lowin_ifaddr = *ia;
				lowin_ifp = ifp;
				break;	/* no need to search further */
			}
		} else {
			if ((snmp_oid_compare(current, entry_oidlen,
					      name, *length) > 0) &&
			    ((lowinterface == -1) ||
			    (snmp_oid_compare(current, entry_oidlen, 
					      lowest, entry_oidlen) < 0))) {
				/*
				 * if new one is greater than input and
				 * closer to input than previous lowest, save
				 * this one as the "next" one.
				 */
				lowinterface = interface;
				lowin_ifaddr = *ia;
				lowin_ifp = ifp;
				bcopy((char *) current, (char *) lowest,
				      entry_oidlen * sizeof(oid));
			}
		}
		interface++;
	}
	if (lowinterface == -1)
		return (NULL);
	bcopy((char *) lowest, (char *) name, entry_oidlen * sizeof(oid));
	*length = entry_oidlen;
	*write_method = 0;
	*var_len = sizeof(long_return);
	strcpy(Name, lowin_ifp->if_fullname);
	switch (name[9]) {
	case IPNETTOMEDIAIFINDEX:
		long_return = lowinterface;
		return (uint8_t *) & long_return;
	case IPNETTOMEDIAPHYSADDRESS:
		ap = (struct arpcom *)lowin_ifp;
		if (lowin_ifp->if_addrlen != sizeof(ap->ac_enaddr)) {
			/*
			 * XXX - we only support interfaces with 802.x-style
			 * 6-byte addresses.  This needs to be fixed.
			 */
			long_return = 0;
			return (uint8_t *) &long_return;
		}
		*var_len = lowin_ifp->if_addrlen;
		bcopy((char*)&ap->ac_enaddr, return_buf, sizeof(ap->ac_enaddr));
		return (uint8_t *) return_buf;
	case IPNETTOMEDIANETADDRESS:
		return (uint8_t *) & ((struct sockaddr_in *)
			& lowin_ifaddr.ia_addr)->sin_addr.s_addr;
	case IPNETTOMEDIATYPE:
		long_return = 4;
		return (uint8_t *) & long_return;
	default:
		ERROR("var_ipNet2Media");
	}
	return NULL;
}

uint8_t *
var_ipNetToMediaEntry(SNMP_FM_ARGS)
{
	DOMAIN_SWITCH_NEST(sk_domain_network, _var_ipNetToMediaEntry,
			   (vp,name,length,exact,var_len,write_method)); 
}
