/*
 *
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */

#include <net-snmp/net-snmp-config.h>
#include <sk/ha_host.h>


#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include <agent/mibgroup/util_funcs.h>
#include <sk/sk_stats.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/netinet/sk_in.h>
#include <net7g/net/sk_route.h>
#include <sk/sk_synch.h>
#include <net7g/netinet/sk_in_var.h>
#include <net7g/netinet/sk_if_ether.h>
#include <net7g/netinet/sk_in_systm.h>
#include <net7g/netinet/sk_ip.h>
#include <net7g/netinet/sk_in_pcb.h>
#include <net7g/netinet/sk_ip_var.h>     /* struct ipstat */
#include <tcp/tcp_var.h>
#include <tcp/tcp_fsm.h>        /* TCPS_ESTABLISHED, TCPS_CLOSE_WAIT */
#include <lib/rsrc_mgr.h>
#include <net7g/netinet/sk_udp_var.h>    /* struct udpstat */
#include <net7g/netinet/sk_ip_icmp.h>
#include <net7g/netinet/sk_icmp_var.h>   /* struct icmpstat */
#include "udp_ipv6.h"
#include "interfaces.h"
#include <psm/psm.h>

#define getpid()    ((unsigned long)(sk_my_pid()))

SK_DOMAIN_IMPORT(udpstat, SK_REASON_LAX);

STATIC FindVarMethod var_udp;
STATIC FindVarMethod var_udpEndpointEntry;
STATIC FindVarMethod var_udpEntry;

STATIC FindVarMethod _var_udp;
STATIC FindVarMethod _var_udpEndpointEntry;
STATIC FindVarMethod _var_udpEntry;

#define UDP_OID                { SNMP_OID_MIB2, 7 }
#define UDPENDPOINTENTRY_OID   { SNMP_OID_MIB2, 7, 7, 1 }
#define UDPENTRY_OID           { SNMP_OID_MIB2, 7, 5, 1 }

struct variable1 udp_variables[6] = {
    {UDPINDATAGRAMS, COUNTER, RONLY, var_udp, 1, {1}},
    {UDPNOPORTS, COUNTER, RONLY, var_udp, 1, {2}},
    {UDPINERRORS, COUNTER, RONLY, var_udp, 1, {3}},
    {UDPOUTDATAGRAMS, COUNTER, RONLY, var_udp, 1, {4}},
    {UDPHCINDATAGRAMS, COUNTER64, RONLY, var_udp, 1, {8}},
    {UDPHCOUTDATAGRAMS, COUNTER64, RONLY, var_udp, 1, {9}}
};
oid udp_variables_oid[7] = UDP_OID;

struct variable1 udpEndpointEntry_variables[8] = {
    {UDPENDPOINTLOCALADDRESSTYPE, INTEGER, RONLY,
                                        var_udpEndpointEntry, 1, {1}},
    {UDPENDPOINTLOCALADDRESS, INETADDRESS, RONLY,
                                        var_udpEndpointEntry, 1, {2}},
    {UDPENDPOINTLOCALPORT, INTEGER, RONLY,
                                        var_udpEndpointEntry, 1, {3}},
    {UDPENDPOINTREMOTEADDRESSTYPE, INTEGER, RONLY,
                                        var_udpEndpointEntry, 1, {4}},
    {UDPENDPOINTREMOTEADDRESS, INETADDRESS, RONLY,
                                        var_udpEndpointEntry, 1, {5}},
    {UDPENDPOINTREMOTEPORT, INTEGER, RONLY,
                                        var_udpEndpointEntry, 1, {6}},
    {UDPENDPOINTINSTANCE, INTEGER, RONLY,
                                        var_udpEndpointEntry, 1, {7}},
    {UDPENDPOINTPROCESS, INTEGER, RONLY,
                                        var_udpEndpointEntry, 1, {8}}
};
oid udpEndpointEntry_variables_oid[9] = UDPENDPOINTENTRY_OID;

struct variable1 udpEntry_variables[2] = {
    {UDPLOCALADDRESS, IPADDRESS, RONLY, var_udpEntry, 1, {1}},
    {UDPLOCALPORT, INTEGER32, RONLY, var_udpEntry, 1, {2}}
};
oid udpEntry_variables_oid[9] = UDPENTRY_OID;

void init_udp_ipv6(void)
{
    REGISTER_MIB("mib_2/udp",
                  udp_variables, variable1,
                  udp_variables_oid);
    REGISTER_MIB("udpEndpointTable/udpEndpointEntry",
                  udpEndpointEntry_variables, variable1,
                  udpEndpointEntry_variables_oid);
    REGISTER_MIB("udpTable/udpEntry",
          udpEntry_variables, variable1,
          udpEntry_variables_oid);
}

uint8_t *
_var_udp(SNMP_FM_ARGS)
{
    switch (vp->magic) {
        case UDPINDATAGRAMS:
            long_return = udpstat.udps_ipackets;
            return (uint8_t *) &long_return;
        case UDPNOPORTS:
            long_return = udpstat.udps_noport;
            return (uint8_t *) &long_return;
        case UDPOUTDATAGRAMS:
            long_return = udpstat.udps_opackets;
            return (uint8_t *) &long_return;
        case UDPINERRORS:
            long_return = udpstat.udps_hdrops + udpstat.udps_badsum +
            udpstat.udps_badlen;
            return (uint8_t *) &long_return;
        case UDPHCINDATAGRAMS:
            big_return = (uint64_t)udpstat.udps_ipackets;
            counter64_return.high = HIGH32(big_return);
            counter64_return.low = LOW32(big_return);
            *var_len = sizeof(counter64_return);
            return (uint8_t *) &counter64_return;
        case UDPHCOUTDATAGRAMS:
            big_return = (uint64_t)udpstat.udps_opackets;
            counter64_return.high = HIGH32(big_return);
            counter64_return.low = LOW32(big_return);
            *var_len = sizeof(counter64_return);
            return (uint8_t *) &counter64_return;
        default:
            ERROR("_var_udp");
    }
    return NULL;
}

uint8_t *
var_udp(SNMP_FM_ARGS)
{
    if (header_generic(vp, name, length, exact, var_len, write_method) ==
            MATCH_FAILED) {
        return NULL;
    }

    DOMAIN_SWITCH_NEST(sk_domain_network, _var_udp,
      (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_udpEndpointEntry(SNMP_FM_ARGS)
{
    oid             newname[SNMP_OID_MAX], lowest[SNMP_OID_MAX], *op;
    uint8_t         *cp;
    struct inpcb   *inp;
    static struct inpcb Lowinpcb;
    bool            found;
    int32_t i = 0, j = 0;

    ASSERT(sk_in_domain(sk_domain_network));

    if (*length < 7) {
        return NULL;
    }

    bcopy((char *) vp->name, (char *) newname,
        (int32_t) vp->namelen * sizeof(oid));

    lowest[0] = 0xFF;

    found = false;

    ASSERT(sk_in_domain(sk_domain_network));
    SK_DOMAIN_REFERENCE(udb, SK_DOMAIN_NETWORK);
    for (inp = udb[sk_gethost()].lh_first; inp != NULL;
                                    inp = inp->inp_list.le_next) {
        if((inp->inp_vfiler)
            && (!VFILER_IS_DEFAULT(inp->inp_vfiler))) {
            continue;
        }

        if(!(inp->inp_flags & INP_IPV6)) {

            cp = (uint8_t *) & inp->inp_laddr.s_addr;
            op = newname + vp->namelen;

            /* address type */
            *op++ = 1;

            /* IP address */
            *op++ = *cp++;
            *op++ = *cp++;
            *op++ = *cp++;
            *op++ = *cp++;

            /* port */
            *op++ = ntohs(inp->inp_lport);

            /* remote address type */
            *op++ = 1;

            /* IP address */
            cp = (uint8_t *) & inp->inp_faddr.s_addr;

            *op++ = *cp++;
            *op++ = *cp++;
            *op++ = *cp++;
            *op++ = *cp++;

            /* remote port */
            *op++ = ntohs(inp->inp_fport);

            /* instance - Don't have anything concrete to return */
            *op = 0;

            /* Increment size by one to get the OID length right */
            j = (op - newname) + 1;
        }
        else {
            j = vp->namelen;

            newname[j++] = 2;
            for (i = 0; i < sizeof(struct in6_addr); i++) {
                newname[j++] = inp->in6p_laddr.s6_addr[i];
            }

            newname[j++] = ntohs(inp->in6p_lport);

            newname[j++] = 2;
            for (i = 0; i < sizeof(struct in6_addr); i++){
                newname[j++] = inp->in6p_faddr.s6_addr[i];
            }

            newname[j++] = ntohs(inp->in6p_fport);

            newname[j++] = 0;

            if (IN6_IS_ADDR_LINKLOCAL(&inp->in6p_laddr)) {
                newname[j] = ntohs(*(uint16_t *)&inp->in6p_laddr.s6_addr[2]);
            }
            else {
                newname[j] = 0;
            }
            /* Increment by one to get the OID length right */
            j++;
        }

        if (exact) {
            if (snmp_oid_compare(newname, j, name, *length) == 0) {
                bcopy((char *) newname,
                    (char *) lowest, j * sizeof(oid));
                Lowinpcb = *inp;
                found = true;
                break; /* no need to search further */
            }
        } else {
            if ((snmp_oid_compare(newname, j, name, *length) > 0) &&
                (snmp_oid_compare(newname, j, lowest, j) < 0)) {
            /*
            * if new one is greater than input
            * and closer to input than previous
            * lowest, save this one as the
            * "next" one.
            */
                bcopy((char *) newname,
                    (char *) lowest,
                    j * sizeof(oid));
                Lowinpcb = *inp;
                found = true;
            }
        }
    }

    if (!found) {
        return (NULL);
    }
    if(!(Lowinpcb.inp_flags & INP_IPV6)) {
        j = 23;
    }
    else {
        j = 48;
    }
    bcopy((char *) lowest, (char *) name,
        j * sizeof(oid));
    *length = j;
    *write_method = 0;
    *var_len = sizeof(int32_t);

    switch (vp->magic) {
        case UDPENDPOINTLOCALADDRESSTYPE:
        case UDPENDPOINTREMOTEADDRESSTYPE:
            if((Lowinpcb.inp_flags & INP_IPV6)) {
                long_return = 2;
            }
            else {
                long_return = 1;
            }
            return (uint8_t *) & long_return;
        case UDPENDPOINTLOCALADDRESS:
            if((Lowinpcb.inp_flags & INP_IPV6)) {
                *var_len = sizeof(struct in6_addr);
                return (uint8_t *)& Lowinpcb.in6p_laddr.s6_addr;
            }
            else {
                return (uint8_t *) & Lowinpcb.inp_laddr.s_addr;
            }
        case UDPENDPOINTLOCALPORT:
            long_return = ntohs(Lowinpcb.inp_lport);
            return (uint8_t *) & long_return;
        case UDPENDPOINTREMOTEADDRESS:
            if((Lowinpcb.inp_flags & INP_IPV6)) {
                *var_len = sizeof(struct in6_addr);
                return (uint8_t *)& Lowinpcb.in6p_faddr.s6_addr;
            }
            else {
                return (uint8_t *) & Lowinpcb.inp_faddr.s_addr;
            }
        case UDPENDPOINTREMOTEPORT:
            long_return = ntohs(Lowinpcb.inp_fport);
            return (uint8_t *) & long_return;
        case UDPENDPOINTINSTANCE:
        case UDPENDPOINTPROCESS:
        /*  Don't have anything to return yet. */
            //long_return = 0;
			long_return = (int32_t)getpid();
            return (uint8_t *) & long_return;
        default:
            return NULL;
    }
    return NULL;
}

uint8_t *
var_udpEndpointEntry(SNMP_FM_ARGS)
{
    DOMAIN_SWITCH_NEST(sk_domain_network, _var_udpEndpointEntry,
                           (vp, name, length, exact, var_len, write_method));
}

uint8_t *
_var_udpEntry(SNMP_FM_ARGS)
{
    oid             lowest[16], *op;
    oid             current[16];
    uint8_t         *cp;
    struct inpcb   *inp;
    static struct inpcb Lowinpcb;
    bool            found;

    ASSERT(sk_in_domain(sk_domain_network));

    memcpy((char *) current, (char *) vp->name,
            (int32_t) vp->namelen * sizeof(oid));

    lowest[0] = 0xFF;
    found = false;

    SK_DOMAIN_REFERENCE(udb, SK_DOMAIN_NETWORK);
    for (inp = udb[sk_gethost()].lh_first; inp != NULL;
                                    inp = inp->inp_list.le_next) {

        if((inp->inp_vfiler)
            && (!VFILER_IS_DEFAULT(inp->inp_vfiler))) {
                continue;
        }

        /* We consider ipv4 addresses only */
        if (IN6_IS_ADDR_LINKLOCAL(&inp->in6p_laddr)) {
            continue;
        }

        cp = (uint8_t *) & inp->inp_laddr.s_addr;
        op = &current[10];
        *op++ = *cp++;
        *op++ = *cp++;
        *op++ = *cp++;
        *op++ = *cp++;

        current[14] = ntohs(inp->inp_lport);

        if (exact) {
            if (snmp_oid_compare(current, 15, name, *length) == 0) {
                bcopy((char *) current,
                    (char *) lowest, 15 * sizeof(oid));
                Lowinpcb = *inp;
                found = true;
                break;  /* no need to search further */
            }
        } else {
            if ((snmp_oid_compare(current, 15, name, *length) > 0) &&
                    (snmp_oid_compare(current, 15, lowest, 15) < 0)) {
                    /*
                    * if new one is greater than input
                    * and closer to input than previous
                    * lowest, save this one as the
                    * "next" one.
                    */
                    bcopy((char *) current, (char *) lowest,
                            15 * sizeof(oid));
                    Lowinpcb = *inp;
                    found = true;
            }
        }
    }

    if (!found) {
        return (NULL);
    }
    bcopy((char *) lowest, (char *) name, 15 * sizeof(oid));
    *length = 15;
    *write_method = 0;
    *var_len = sizeof(int32_t);

    switch (vp->magic) {
    case UDPLOCALADDRESS:
        return (uint8_t *) & Lowinpcb.inp_laddr.s_addr;
    case UDPLOCALPORT:
        long_return = ntohs(Lowinpcb.inp_lport);
        return (uint8_t *) & long_return;
    default:
        return NULL;
    }
    return NULL;
}

uint8_t *
var_udpEntry(SNMP_FM_ARGS)
{
    DOMAIN_SWITCH_NEST(sk_domain_network, _var_udpEntry,
                        (vp, name, length, exact, var_len, write_method));
}
