/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */

/*
 * var_route.c - return a pointer to the named variable.
 *
 *
 */
/***********************************************************
	Copyright 1988, 1989 by Carnegie Mellon University
	Copyright 1989	TGV, Incorporated

		      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU and TGV not be used
in advertising or publicity pertaining to distribution of the software
without specific, written prior permission.

CMU AND TGV DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL CMU OR TGV BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#include <sys/param.h>
#include <sk/ha_host.h>


#include <sys/cdefs.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/sys/sk_mbuf.h>
#include <net7g/net/sk_if.h>
#include <net7g/netinet/sk_in.h>
#include <net7g/net/sk_route.h>
#include <legacy/string.h>

#include "net-snmp/net-snmp-config.h"
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <sk/sk_common.h>
#include <sk/sk_domain.h>
#include <sk/sk_processor.h>
#include "var_route.h"

CSMP_STATIC struct rtentry * FindRTEntry(struct radix_node *,
    oid *, int, oid *, int, oid *, uint32_t *);
CSMP_STATIC struct rtentry * MatchRTIPAddr(struct radix_node *, oid *, int,
    oid *, int, oid *, uint32_t *);
CSMP_STATIC int32_t Interface_Scan_By_Ifnet(struct ifnet *);
CSMP_STATIC uint8_t * _var_ipRouteEntry(SNMP_FM_ARGS);
CSMP_STATIC uint8_t * var_ipRouteEntry(SNMP_FM_ARGS);

#define IPROUTEENTRY_OID      { SNMP_OID_MIB2, 4, 21, 1 }
struct variable1 ipRouteEntry_variables[13] = {
	{IPROUTEDEST, IPADDRESS, RWRITE, var_ipRouteEntry, 1, {1}},
	{IPROUTEIFINDEX, INTEGER, RWRITE, var_ipRouteEntry, 1, {2}},
	{IPROUTEMETRIC1, INTEGER, RWRITE, var_ipRouteEntry, 1, {3}},
	{IPROUTEMETRIC2, INTEGER, RWRITE, var_ipRouteEntry, 1, {4}},
	{IPROUTEMETRIC3, INTEGER, RWRITE, var_ipRouteEntry, 1, {5}},
	{IPROUTEMETRIC4, INTEGER, RWRITE, var_ipRouteEntry, 1, {6}},
	{IPROUTENEXTHOP, IPADDRESS, RWRITE, var_ipRouteEntry, 1, {7}},
	{IPROUTETYPE, INTEGER, RWRITE, var_ipRouteEntry, 1, {8}},
	{IPROUTEPROTO, INTEGER, RONLY, var_ipRouteEntry, 1, {9}},
	{IPROUTEAGE, INTEGER, RWRITE, var_ipRouteEntry, 1, {10}},
	{IPROUTEMASK, IPADDRESS, RWRITE, var_ipRouteEntry, 1, {11}},
	{IPROUTEMETRIC5, INTEGER, RWRITE, var_ipRouteEntry, 1, {12}},
	{IPROUTEINFO, ASN_OBJECT_ID, RONLY, var_ipRouteEntry, 1, {13}},
};
oid ipRouteEntry_variables_oid[9] = IPROUTEENTRY_OID;

void init_var_route()
{
     REGISTER_MIB("ipRouteTable/ipRouteEntry",
                  ipRouteEntry_variables, variable1,
                  ipRouteEntry_variables_oid); 
}

SK_DOMAIN_EXPORT(_var_ipRouteEntry,SK_DOMAIN_NETWORK);
CSMP_STATIC uint8_t *
_var_ipRouteEntry(SNMP_FM_ARGS)
{
	/*
	 * object identifier is of form: 1.3.6.1.2.1.4.21.1.?.A.B.C.D,  where
	 * A.B.C.D is IP address. IPADDR starts at offset 10.
	 */
	struct radix_node_head *rnh;
	struct rtentry *rn = NULL;
	static oid      Current[SNMP_OID_MAX];
	static struct sockaddr_in *sin;
	static oid	routeinfo_id[] = {0, 0};
	oid             lowest[SNMP_OID_MAX];
	uint32_t           LowAddr;

	/* runtime check of domain-- SK_DOMAIN_EXPORT() is not enough */
	ASSERT(sk_in_domain(sk_domain_network));

	bcopy((char *) vp->name, (char *) Current, vp->namelen * sizeof(oid));

	lowest[0] = '\0';

	LowAddr = UINT_MAX;		/* Don't have one yet */

	SK_DOMAIN_REFERENCE(get_rtable, SK_DOMAIN_NETWORK);
	if ((rnh = get_rtable(AF_INET)) != NULL) {
		rn = FindRTEntry(rnh->rnh_treetop, name, *length,
		    Current, exact, lowest, &LowAddr);
	}
	if (rn == NULL) 
		return (NULL);
	

	bcopy((char *) lowest, (char *) name, 14 * sizeof(oid));
	*length = 14;

	*write_method = 0;
	*var_len = sizeof(long_return);
	switch (name[9]) {
	case IPROUTEDEST:
		sin = (struct sockaddr_in *) rt_key(rn);
		unsigned_return = (sin->sin_addr.s_addr);
		return (uint8_t *) & unsigned_return;
	case IPROUTEIFINDEX:
		long_return = (int32_t)Interface_Scan_By_Ifnet(rn->rt_ifp);
		return (uint8_t *) & long_return;
	case IPROUTEMETRIC1:
		long_return = (rn->rt_flags & RTF_GATEWAY) ? 1 : 0;
		return (uint8_t *) & long_return;
	case IPROUTEMETRIC2:
		long_return = -1;
		return (uint8_t *) & long_return;
	case IPROUTEMETRIC3:
		long_return = -1;
		return (uint8_t *) & long_return;
	case IPROUTEMETRIC4:
		long_return = -1;
		return (uint8_t *) & long_return;
	case IPROUTENEXTHOP:
		if (((struct sockaddr *)rn->rt_gateway)->sa_family == AF_LINK) {
			struct ifaddr *ifa;
			int host = sk_gethost();
			/*
			 * According to the mib:
			 *  "In the case of a route bound to an interface which
			 *  is realized via a broadcast media, the value of this
			 *  field is the agent's IP address on that interface."
			 */
			for (ifa = rn->rt_ifp->if_addrlist; ifa != NULL;
			  ifa = ifa->ifa_next) {
				if (IF_MULTIHOST() && ifa->ifa_host != host)
					continue;

				if (((struct sockaddr_in *)
				  ifa->ifa_addr)->sin_family == AF_INET) {
					unsigned_return = (
					    (struct sockaddr_in *)ifa->ifa_addr)
					        ->sin_addr.s_addr;
					return (uint8_t *)&unsigned_return;
				}
			}
			/*
			 * No AF_INET addr was found on the ifp.  I don't
			 * see how this can happen but just in case there is
			 * a way, I don't want to panic.  Just fall through
			 * and return junk.
			 */
		}
		/* If not AF_LINK we assume it is AF_INET */
		unsigned_return = ((struct sockaddr_in *) rn->rt_gateway)
		                   ->sin_addr.s_addr;
		return (uint8_t *) & unsigned_return;
	case IPROUTETYPE:
		/* For now we only do static routing */
		long_return = (rn->rt_flags & RTF_GATEWAY) ? 4 : 3;
		return (uint8_t *) & long_return;
	case IPROUTEPROTO:
		long_return = (rn->rt_flags & RTF_DYNAMIC) ? 4 : 2;
		return (uint8_t *) & long_return;
	case IPROUTEAGE:
		long_return = 0;
		return (uint8_t *) & long_return;
	case IPROUTEMASK:
		sin = (struct sockaddr_in *) rt_mask(rn);
		if(sin == 0) {
			/* some routes may not have netmasks */
			long_return = -1;  /* netmask is all ones */
			return (uint8_t *) & long_return;
		} else {
			unsigned_return = sin->sin_addr.s_addr;
			return (uint8_t *) & unsigned_return;
		}
	case IPROUTEMETRIC5:
		long_return = -1;
		return (uint8_t *) & long_return;
	case IPROUTEINFO:
		*var_len = sizeof(routeinfo_id);
		return (uint8_t *) routeinfo_id;	
	default:
		ERROR("var_ipRouteEntry");
	}
	return NULL;
}

SK_DOMAIN_EXPORT(var_ipRouteEntry,SK_REASON_SWITCH_DOMAIN);
CSMP_STATIC uint8_t *
var_ipRouteEntry(SNMP_FM_ARGS)
{
	sk_domain_t     old_domain = sk_domain_switch_begin(sk_domain_network);
	uint8_t		*value;

	value = _var_ipRouteEntry(vp,name,length,exact,var_len,write_method);

	sk_domain_switch_end(old_domain);
	return value;
}

SK_DOMAIN_EXPORT(FindRTEntry,SK_DOMAIN_NETWORK);
CSMP_STATIC struct rtentry *
FindRTEntry(struct radix_node * rn,
    oid * name, int length, oid * Current, int exact, oid * lowest, uint32_t * LowAddrp)
{
	struct rtentry *rt;

	/* runtime check of domain-- SK_DOMAIN_EXPORT() is not enough */
	ASSERT(sk_in_domain(sk_domain_network));

	for             (;;) {
		while (rn->rn_b >= 0)
			rn = rn->rn_l;	/* First time through node, go left */

		if ((rt = MatchRTIPAddr(rn, name, length, Current, exact, lowest, LowAddrp)) != NULL)
			return (rt);	/* Process Leaf */
/* XXX This bit above is wrong.  Why don't I take the other approach of
filling up rthost[] and rtnet[] and reading it off from there? See the orig
snmp_var_route.c

*/
		while (rn->rn_p->rn_r == rn) {	/* if coming back from right */
			rn = rn->rn_p;	/* go back up */
			if (rn->rn_flags & RNF_ROOT)
				return 0;
		}
		rn = rn->rn_p->rn_r;	/* otherwise, go right */
	}
	return 0;
}

SK_DOMAIN_EXPORT(MatchRTIPAddr,SK_DOMAIN_NETWORK);
CSMP_STATIC struct rtentry *
MatchRTIPAddr(struct radix_node * rn, oid * name, int length, oid * Current,
    int exact, oid * lowest, uint32_t *LowAddrp)
{
	struct rtentry *rt = NULL;
	uint8_t	*cp;
	oid     *op;
	struct sockaddr *sa;
	struct sockaddr_in *sin;

	/* runtime check of domain-- SK_DOMAIN_EXPORT() is not enough */
	ASSERT(sk_in_domain(sk_domain_network));

	for (; rn; rn = rn->rn_dupedkey) {
		if (rn->rn_flags & RNF_ROOT)
			continue;
		rt = (struct rtentry *)rn;
		sa = (struct sockaddr*) rt_key(rt);
		if(sa->sa_family != AF_INET)
			continue;
		sin = (struct sockaddr_in *) sa;
		cp = (uint8_t *) & sin->sin_addr.s_addr;
		op = Current + 10;
		*op++ = *cp++;
		*op++ = *cp++;
		*op++ = *cp++;
		*op++ = *cp++;

		if (exact) {
			if (snmp_oid_compare(Current, 14, name, length) == 0) {
				bcopy((char *) Current, (char *) lowest,
						14 * sizeof(oid));
				*LowAddrp = sin->sin_addr.s_addr;
				return (rt);	/* no need to search further */
			}
		} else {
			if ((snmp_oid_compare(Current, 14, name, length) > 0)
			    && ((*LowAddrp == -1) || 
				(snmp_oid_compare(Current, 14, lowest,	14) < 0))) {
				/*
				 * if new one is greater than input and
				 * closer to input than previous lowest, save
				 * this one as the "next" one.
				 */
				bcopy((char *) Current, (char *) lowest,
					14 * sizeof(oid));
				*LowAddrp = sin->sin_addr.s_addr;
			}
		}
	}
	if (*LowAddrp != -1) {
		return rt;
	}
	return 0;
}

/* Create a function here that returns the interface
index given the ifnet addr */
SK_DOMAIN_EXPORT(Interface_Scan_By_Ifnet,SK_DOMAIN_NETWORK);
CSMP_STATIC int32_t
Interface_Scan_By_Ifnet(struct ifnet *Retifnet)
{
	int32_t i = 1;
	struct ifnet *ifp;

	/* runtime check of domain-- SK_DOMAIN_EXPORT() is not enough */
	ASSERT(sk_in_domain(sk_domain_network));

	SK_DOMAIN_REFERENCE(ifnet, SK_DOMAIN_NETWORK);
	for (ifp = ifnet; ifp; ifp = ifp->if_next, i++) {
		if(ifp == Retifnet)
			return(i);
	}
	return -1;
}
