/*
 *
 * Copyright (c) 2008-2013 NetApp, Inc.
 * All rights reserved.
 */

#include <sk/sk_stats.h>
#include <sk/ha_host.h>

#include <sys/param.h>
#include <rpc/types.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/netinet/sk_in.h>
#include <net7g/arpa/sk_inet.h>

#include <wafl/typedefs.h>
#include <smb/smb.h>
#include <smb/smbmsgs.h>
#include <smb/smbstat.h>
#include <cmds/smbsetup.h>
#include <cmds/ext.h>
#include <cmds/enable.h>
#include <legacy/string.h>
#include <fio/fio.h>
#include <rpc/sunrpc_vfiler.h>
#include <nis/client.h>
#include <nis/nis_vfiler.h>

#include <snmpv3/agent/mibgroup/mibparser_structs.h>
#include <snmpv3/agent/mibgroup/mibparser_defines.h>
#include "net-snmp/agent/snmp_vars.h"
#include "cifs_vars_aux.h"

SK_DOMAIN_IMPORT(is_service_licensed, SK_REASON_SWITCH_DOMAIN);
C_IMPORT_CIFS(CIFSRuntimeStats__);
C_IMPORT_CIFS(gSMBstats__);

#define RETBUFLEN       256
static uint8_t           retbuf[RETBUFLEN];
/*
 * I'm declaring both a signed and unsigned return variable,
 * because in ASN.1 an INTEGER is a signed quantity and a COUNTER is unsigned.
 */
static int32_t             longReturn;
static uint32_t            ulongReturn;

/*
 * boilerplate macros
 *
 * (note to VCC users: run this file through dos2unix
 * before compiling: the ^M's after the '\'s make gcc gakk)
 */

/*
 * declare local vars, initialize return parameters. 
 */

extern int             header_generic(struct variable *, oid *, size_t *, int,
                               size_t *, WriteMethod **);
#define MATCH_FAILED        (-1)
#define VAR_INIT() \
if (header_generic(vp, name, length, exact, var_len, write_method) == \
    MATCH_FAILED) {                                                   \
		return NULL;                                          \
	}                                                             

/*
 * macros for frequently used switch statement return phrases 
 */

#define RETURN(x)                               \
{                                               \
        C_DOMAIN_SWITCH_END(calling_domain); \
        return (uint8_t*) &x;                    \
}

#define RETURN_IF_CLUSTERMODE()			\
{						\
	if (is_cluster_capable()) {		\
		ulongReturn = 0;		\
		return (uint8_t*) &ulongReturn;	\
	}					\
}

/*
 * hack: add one to smb_total_calls() to avoid divide-by-zero problems
 */
#define LONGPCTRETURN(i)                                                        \
{                                                                               \
	C_REFERENCE_CIFS(smb_total_calls);					\
	C_REFERENCE_CIFS(ulpct);						\
        longReturn = ulpct(smbstatGetGlobalValue (i), (smb_total_calls() + 1)); \
        RETURN(longReturn);                                                     \
}

#define LONGBOOLRETURN(i)               \
{                                       \
        longReturn = (i) ? 2 : 1;       \
        RETURN(longReturn);             \
}

#define ULONGSTATRETURN(i)                              \
{                                                       \
        ulongReturn = smbstatGetGlobalValue (i);        \
        RETURN(ulongReturn);                            \
}

#define ULONGRETURN(n)          \
{                               \
        ulongReturn = n;        \
        RETURN(ulongReturn);    \
}

#define STRINGRETURN(s)                         \
{                                               \
        *var_len = strlen(s);                   \
        C_DOMAIN_SWITCH_END(calling_domain); \
	return (uint8_t*) s;                     \
}

#define LONGRETURN(n)                   \
{                                       \
        *var_len = sizeof(longReturn);  \
	longReturn = n;                 \
        RETURN(longReturn);             \
}


/*
 * In all the following handlers, be careful to return the correct
 * kind of quantity: use LONG* macros to return quantities declared
 * in the MIB as INTEGERs, and ULONG* macros to return those declared
 * as Counters.
 */

uint8_t *
var_cifsOptions(SNMP_FM_ARGS)
{
        int             UNIXSTYLE = 1;
        int             PCSTYLE = 2;
        struct in_addr  WINSservers[MAX_NBNS_SERVERS];
        sk_domain_t     calling_domain;

        VAR_INIT();

	RETURN_IF_CLUSTERMODE();

        calling_domain = C_DOMAIN_SWITCH_BEGIN(sk_domain_cifs);

        retbuf[0] = 0;

        switch (vp->magic)
        {
        case CIFSISENABLED:
                LONGBOOLRETURN(SMB_IsStarted());
        case CIFSISLOGINENABLED:
                LONGBOOLRETURN(SMBAccess(i_SMBLoginEnable));
        case CIFSHOSTNAME:
              	C_REFERENCE_CIFS(bufprint_UC); 
	       	bufprint_UC(ParamAccess(c_CsUncServerName), retbuf, RETBUFLEN);
                STRINGRETURN(retbuf);
        case CIFSALTNAMES:
                C_REFERENCE_CIFS(bufprint_altnames);
		bufprint_altnames(retbuf, RETBUFLEN);
                STRINGRETURN(retbuf);
        case CIFSDOMAINJOINED:
		 C_REFERENCE_CIFS(is_domain_joined);
                LONGBOOLRETURN(is_domain_joined());
        case CIFSDOMAINNAME:
		C_REFERENCE_CIFS(is_domain_joined);
                if (is_domain_joined())
                {
			C_REFERENCE_CIFS(bufprint_UC);
                        bufprint_UC(ParamAccess(c_CsUncDomainName), retbuf, RETBUFLEN);
                }
                STRINGRETURN(retbuf);
        case CIFSWGNAME:
		C_REFERENCE_CIFS(is_domain_joined);
                if (!is_domain_joined())
                {
                        sprintf((char*)retbuf, (const char*)ParamAccess(c_NBWorkgroupName));
                }
                STRINGRETURN(retbuf);
        case CIFSDCNAME:
		C_REFERENCE_CIFS(is_domain_joined);
                if (is_domain_joined())
                {
			C_REFERENCE_CIFS(bufprint_DCName);
                        bufprint_DCName(retbuf, RETBUFLEN);
                }
                STRINGRETURN(retbuf);
        case CIFSISWINSENABLED:
                SK_DOMAIN_REFERENCE(nbns_get_servers, SK_REASON_SWITCH_DOMAIN);
                LONGBOOLRETURN( nbns_get_servers(WINSservers));         
        case CIFSWINSSERVERS:
                C_REFERENCE_CIFS(bufprint_winsservers);
		bufprint_winsservers(retbuf, RETBUFLEN);
                STRINGRETURN(retbuf);           
        case CIFSSECURITYMODEL:
                if (ParamAccess(i_SecurityMode) == 0)
                {
                        LONGRETURN(UNIXSTYLE);
                }
                else
                {
                        LONGRETURN(PCSTYLE);
                }
        case CIFSPCGENERICUSER:
                strncpy((char*) retbuf,
                        (const char*) SMBAccess(i_Default_UNIX_AccountName),
                        RETBUFLEN-1);
                retbuf[RETBUFLEN-1] = 0;
                STRINGRETURN(retbuf);
        case CIFSOPLOCKSENABLED:
                LONGBOOLRETURN(SMBAccess(c_EnableOplocks));      
        case CIFSLEVEL2OPLOCKSENABLED:
                LONGBOOLRETURN(SMBAccess(c_Cap_Lvl2OpLocks));    
        case CIFSPRESERVECASE:
                LONGBOOLRETURN(SMBAccess(c_CasePreserve));
        case CIFSSYMLINKSENABLED:
                LONGBOOLRETURN(SMBAccess(i_EnableSymlinks));    
        case CIFSSYMLINKCYCLEPROTENABLED:
                LONGBOOLRETURN(SMBAccess(i_SymlinkCycleGuard)); 
        case CIFSISLICENSED:
                LONGBOOLRETURN(is_service_licensed(LICENSE_CIFS));      
        case CIFSPERCLIENTSTATSENABLED:
                LONGBOOLRETURN(SMBAccess(c_PerCredStats));
        default:
                LONGRETURN(0);
        } // switch
        
        ASSERT(false); /* should never come here */
}


C_IMPORT_CIFS(secs_to_shutdown);
extern int secs_to_shutdown;

uint8_t *
var_cifsInfo(SNMP_FM_ARGS)
{
        sk_domain_t     calling_domain;
	int		numIDs;

        VAR_INIT();

	RETURN_IF_CLUSTERMODE();

        calling_domain = C_DOMAIN_SWITCH_BEGIN(sk_domain_cifs);

        longReturn = 0;
        switch (vp->magic)
        {
        case CIFSSTATUS:
                LONGRETURN(SMBAccess(c_SMBStartStat));
/*
        case CIFSNBTMODE:
                bufprint_NBTMode(retbuf, RETBUFLEN);
                STRINGRETURN(retbuf);
*/
        case CIFSMAXCONNECTIONS:
		C_REFERENCE_CIFS(SMB_GetNActiveVfilers);
                LONGRETURN(SMB_GetNActiveVfilers() * SMBAccess(c_MaxConnections));
        case CIFSMAXTREES:
		C_REFERENCE_CIFS(SMB_GetNActiveVfilers);
                LONGRETURN(SMB_GetNActiveVfilers() * SMBAccess(c_MaxTree));
        case CIFSMAXSHARES:
		C_REFERENCE_CIFS(SMB_GetNActiveVfilers);
                LONGRETURN(SMB_GetNActiveVfilers() * SMBAccess(c_MaxShares));
        case CIFSMAXFILES:
		C_REFERENCE_CIFS(SMB_GetNActiveVfilers);
                LONGRETURN(SMB_GetNActiveVfilers() * SMBAccess(c_MaxFiles));
        case CIFSMAXACLS:
                LONGRETURN(0);   /* Deprecated.  MaxACLs is never used */
        case CIFSCONNECTEDUSERS:
                ULONGRETURN(HH_CIFSRuntimeStats.fCurrentConnectedUsers);        
        case CIFSNTREES:
                ULONGRETURN(HH_CIFSRuntimeStats.fCurrentTrees);
        case CIFSNSHARES:
                ULONGRETURN(HH_CIFSRuntimeStats.fCurrentShares);
        case CIFSNSESSIONS:
                ULONGRETURN(HH_CIFSRuntimeStats.fCurrentSessions);
        case CIFSNOPENFILES:
                ULONGRETURN(HH_CIFSRuntimeStats.fSMBOpenFiles);
        case CIFSNOPENDIRS:
                ULONGRETURN(HH_CIFSRuntimeStats.fSMBOpenDirs);
        case CIFSNOPLOCKBREAKWAITS:
		C_REFERENCE_CIFS(SMB_GetNOplockBreakWaits);
		longReturn = SMB_GetNOplockBreakWaits();
                RETURN(longReturn);
        case CIFSNOPLOCKACKWAITS:
		C_REFERENCE_CIFS(SMB_GetNOplockAckWaits);
                longReturn = SMB_GetNOplockAckWaits();
                RETURN(longReturn);
        case CIFSNEEDPW:
                if (!VV_nis_enabled && !fio_exist("/etc/passwd", false))
                {
                        ulongReturn = 1;
                }
                else
                {
                        ulongReturn = 0;
                }
                RETURN(ulongReturn);                                   
        case CIFSTIMETOSHUTDOWN:
                C_REFERENCE_CIFS(secs_to_shutdown);
                ULONGRETURN(secs_to_shutdown);
        case CIFSSUSPECTOPS:
                ULONGRETURN(SMB_GetSuspectOps());
	case CIFSNDOMAINCONTROLLERS:
                C_REFERENCE_CIFS(sec_GetDCnumIDs);
		numIDs = sec_GetDCnumIDs(ParamAccess(c_WorkgroupName));
		if (numIDs < 0) {
			/* If the bucket is not found. Possible if cifs 
			 * setup is not yet done.
			 */
			ULONGRETURN(0);
		} else {
			ULONGRETURN(numIDs);
		}
	default:
                ULONGRETURN(0); 
        }

        ASSERT(false); /* should never come here */
}               

/*
 * Summary stats aren't currently accumulated on the fly, so they have to be 
 * calculated every time this function is called.  
 */
                
uint8_t *                
var_cifsOps(SNMP_FM_ARGS)               
{       
        int             i, totalops = 0;
        int             host, ohost;
        int             summarystats[kSMBSummarySize];
        sk_domain_t     calling_domain;

        VAR_INIT();

	RETURN_IF_CLUSTERMODE();

        calling_domain = C_DOMAIN_SWITCH_BEGIN(sk_domain_cifs);
        memset(summarystats, 0, sizeof(summarystats));

        /*
         * XXX This function will have to be fixed if/when SNMP is
         * vfilerized because you can't set the host like this unless we
         * are in a default vfiler.
         */
        ASSERT(VFILER_IS_DEFAULT(sk_get_vfiler()));
        ohost = sk_gethost();

        for (host=0; host<=1; host++)
        {
                sk_sethost(host);
                for (i=0; i<kSMBStatNumber; i++)
                {
                        uint64_t val = smbstatGetGlobalValue(i);
                        summarystats[smbstatGetMajorType(i)] += val;
                        totalops += val;
                } // for (i)
        } // for (host)

        sk_sethost(ohost);

        switch (vp->magic)
        {
        case CIFSTOTALOPS:      ULONGRETURN(totalops - summarystats[kSMBStatskip]);
        case CIFSTOTALCALLS:    ULONGRETURN(totalops);
        case CIFSBADCALLS:      ULONGRETURN(smbstatGetGlobalValue (kSMBStatreject));
        case CIFSGETATTRS:      ULONGRETURN(summarystats[kSMBStatGetAttr]);
	case CIFSSETATTRS:	ULONGRETURN(summarystats[kSMBStatSetAttr]);
        case CIFSREADS:         ULONGRETURN(summarystats[kSMBStatRead]);
        case CIFSWRITES:        ULONGRETURN(summarystats[kSMBStatWrite]);
        case CIFSLOCKS:         ULONGRETURN(summarystats[kSMBStatLock]);
        case CIFSOPENS:         ULONGRETURN(summarystats[kSMBStatOpenClose]);
        case CIFSDIROPS:        ULONGRETURN(summarystats[kSMBStatDir]);
        case CIFSOTHERS:        ULONGRETURN(summarystats[kSMBStatOther]);
        default:                ULONGRETURN(0);
        } // switch

        ASSERT(false); /* should never come here */
}

uint8_t *
var_cifsReqs(SNMP_FM_ARGS)
{
        sk_domain_t     calling_domain;

        VAR_INIT();

	RETURN_IF_CLUSTERMODE();

        calling_domain = C_DOMAIN_SWITCH_BEGIN(sk_domain_cifs);

        switch (vp->magic)
        {
        case SMBNEGPROTS:
                ULONGSTATRETURN(kSMBStatnegprot);
        case SMBSESSIONSETUPANDXS:
                ULONGSTATRETURN(kSMBStatsesssetupX);
        case SMBLOGOFFANDXS:
                ULONGSTATRETURN(kSMBStatulogoffX);
        case SMBTREECONNECTANDXS:
                ULONGSTATRETURN(kSMBStattconX);
        case SMBTREEDISCONNECTS:
                ULONGSTATRETURN(kSMBStattdis);
        case SMBTRANS2QUERYFSINFOS:
                ULONGSTATRETURN(kSMBStatTRANS2_QUERY_FS_INFORMATION);
        case SMBECHOS:
                ULONGSTATRETURN(kSMBStatecho);
        case SMBNTCANCELS:
                ulongReturn = (smbstatGetGlobalValue (kSMBStatNTCancelCN) + 
                               smbstatGetGlobalValue (kSMBStatNTCancelOther));
                RETURN(ulongReturn);
        case SMBNTCREATEANDXS:
                ULONGSTATRETURN(kSMBStatNTCreate);
        case SMBNTTRANSACTCREATES:
                ULONGSTATRETURN(kSMBStatNTTransCreate);
        case SMBCREATETEMPORARIES:
                ULONGRETURN(0);  /* not supported */
        case SMBREADANDXS:
                ULONGSTATRETURN(kSMBStatreadX);
        case SMBWRITEANDXS:
                ULONGSTATRETURN(kSMBStatwriteX);
        case SMBLOCKINGANDXS:
                ULONGSTATRETURN(kSMBStatlockingX);
        case SMBSEEKS:
                ULONGSTATRETURN(kSMBStatlseek);
        case SMBFLUSHES:
                ULONGSTATRETURN(kSMBStatflush);
        case SMBCLOSES:
                ULONGSTATRETURN(kSMBStatclose);
        case SMBDELETES:
                ULONGSTATRETURN(kSMBStatunlink);
        case SMBRENAMES:
                ULONGSTATRETURN(kSMBStatmv);
        case SMBMOVES:
                ULONGRETURN(0); /* ioctl ops not supported */
        case SMBCOPIES:
                ULONGRETURN(0); /* "" */
        case SMBTRANS2QUERYPATHINFOS:
                ULONGSTATRETURN(kSMBStatTRANS2_QUERY_PATH_INFORMATION);
        case SMBTRANS2QUERYFILEINFOS:
                ULONGSTATRETURN(kSMBStatTRANS2_QUERY_FILE_INFORMATION);
        case SMBTRANS2SETPATHINFOS:
                ULONGSTATRETURN(kSMBStatTRANS2_SET_PATH_INFORMATION);
        case SMBTRANS2SETFILEINFOS:
                ULONGSTATRETURN(kSMBStatTRANS2_SET_FILE_INFORMATION);
        case SMBDELETEDIRS:
                ULONGSTATRETURN(kSMBStatrmdir);
        case SMBCHECKDIRS:
                ULONGRETURN(0); /* not supported */
        case SMBTRANS2FINDFIRST2S:
                ULONGSTATRETURN(kSMBStatTRANS2_FIND_FIRST2);
        case SMBTRANS2FINDNEXT2S:
                ULONGSTATRETURN(kSMBStatTRANS2_FIND_NEXT2);
        case SMBFINDCLOSE2S:
                ULONGSTATRETURN(kSMBStatfindclose);
        case SMBNTTRANSACTNOTIFYCHGS:
                ULONGSTATRETURN(kSMBStatNTTransNotify);
        case SMBTRANS2GETDFSREFERRALS:
                ULONGRETURN(kSMBStatTRANS2_DFS_REFERRAL);
        case SMBTRANS2REPORTDFSINCS:
                ULONGRETURN(kSMBStatTRANS2_DFS_REPORT_INCONSISTENCY);
        case SMBOPENPRINTFILES:
                ULONGRETURN(0);
        case SMBGETPRINTQUEUES:
                ULONGRETURN(0);
        case SMBNTTRANSACTIOCTLS:
                ULONGSTATRETURN(kSMBStatNTTransIoctl);
        case SMBNTTRANSACTQUERYSECDESCS:
                ULONGSTATRETURN(kSMBStatNTTransQuerySec);
        case SMBNTTRANSACTSETSECDESCS:
                ULONGSTATRETURN(kSMBStatNTTransSetSec);
        case SMBTRANS2CREATEDIRS:
                ULONGSTATRETURN(kSMBStatTRANS2_CREATE_DIRECTORY);
        case SMBNTCANCELCNS:
                ULONGSTATRETURN(kSMBStatNTCancelCN);
        case SMBNTCANCELOTHERS:
                ULONGSTATRETURN(kSMBStatNTCancelOther);
	case SMB2ECHOS:
		ULONGSTATRETURN(kSMB2StatEcho);
	case SMB2NEGPROTS:
		ULONGSTATRETURN(kSMB2StatNegProt);
	case SMB2SESSIONSETUPS:
		ULONGSTATRETURN(kSMB2StatSessSetup);
	case SMB2LOGOFFS:
		ULONGSTATRETURN(kSMB2StatLogoff);
	case SMB2TREECONNECTS:
		ULONGSTATRETURN(kSMB2StatTreeConnect);
	case SMB2TREEDISCONNECTS:
		ULONGSTATRETURN(kSMB2StatTreeDisconnect);
	case SMB2CANCELS:
		ULONGSTATRETURN(kSMB2StatCancel);
	case SMB2CREATES:
		ULONGSTATRETURN(kSMB2StatCreate);
	case SMB2IPCCREATES:
		ULONGSTATRETURN(kSMB2IpcCreate);
	case SMB2READS:
		ULONGSTATRETURN(kSMB2StatRead);
	case SMB2WRITES:
		ULONGSTATRETURN(kSMB2StatWrite);
	case SMB2IPCREADS:
		ULONGSTATRETURN(kSMB2IpcRead);
	case SMB2IPCWRITES:
		ULONGSTATRETURN(kSMB2IpcWrite);
	case SMB2LOCKS:
		ULONGSTATRETURN(kSMB2StatLock);
	case SMB2UNLOCKS:
		ULONGSTATRETURN(kSMB2StatUnlock);
	case SMB2OPLKBRKACK:
		ULONGSTATRETURN(kSMB2StatOplkBrkAck);
	case SMB2CHGNFYS:
		ULONGSTATRETURN(kSMB2StatChgNfy);
	case SMB2FLUSHS:
		ULONGSTATRETURN(kSMB2StatFlush);
	case SMB2CLOSES:
		ULONGSTATRETURN(kSMB2StatClose);
	case SMB2QUERYFILEBASICINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFileBasicInfo);
	case SMB2QUERYFILESTANDARDINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFileStandardInfo);
	case SMB2QUERYFILEINTERNALINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFileInternalInfo);
	case SMB2QUERYFILEEAINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFileEAInfo);
	case SMB2QUERYFILEFULLEAINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFileFullEAInfo);
	case SMB2QUERYFILEALTNAMES:
		ULONGSTATRETURN(kSMB2StatQueryFileAltNameInfo);
	case SMB2QUERYFILESTREAMINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFileStreamInfo);
	case SMB2QUERYFILENETOPENINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFileNetOpenInfo);
	case SMB2QUERYFILEATTRTAGINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFileAttrTagInfo);
	case SMB2QUERYFILEUNSUPPORTEDS:
		ULONGSTATRETURN(kSMB2StatQueryFileUnsupported);
	case SMB2QUERYFILEINVALIDS:
		ULONGSTATRETURN(kSMB2StatQueryFileInvalid);
	case SMB2QUERYFSVOLINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFsVolInfo);
	case SMB2QUERYFSSIZEINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFsSizeInfo);
	case SMB2QUERYFSDEVICEINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFsDeviceInfo);
	case SMB2QUERYFSATTRINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFsAttrInfo);
	case SMB2QUERYFSFULLSIZEINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFsFullSizeInfo);
	case SMB2QUERYFSOBJECTIDINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFsObjectIdInfo);
	case SMB2QUERYFSINVALIDS:
		ULONGSTATRETURN(kSMB2StatQueryFsInvalid);
	case SMB2SETBASICINFOS:
		ULONGSTATRETURN(kSMB2StatSetBasicInfo);
	case SMB2SETRENAMEINFOS:
		ULONGSTATRETURN(kSMB2StatSetRenameInfo);
	case SMB2SETFILELINKINFOS:
		ULONGSTATRETURN(kSMB2StatSetFileLinkInfo);
	case SMB2SETFILEDISPINFOS:
		ULONGSTATRETURN(kSMB2StatSetFileDispInfo);
	case SMB2SETFULLEAINFOS:
		ULONGSTATRETURN(kSMB2StatSetFullEAInfo);
	case SMB2SETMODEINFOS:
		ULONGSTATRETURN(kSMB2StatSetModeInfo);
	case SMB2SETALLOCINFOS:
		ULONGSTATRETURN(kSMB2StatSetAllocInfo);
	case SMB2SETEOFINFOS:
		ULONGSTATRETURN(kSMB2StatSetEOFInfo);
	case SMB2SETINFOUNSUPPORTEDS:
		ULONGSTATRETURN(kSMB2StatSetInfoUnsupported);
	case SMB2SETINFOINVALIDS:
		ULONGSTATRETURN(kSMB2StatSetInfoInvalid);
	case SMB2FSCTLPIPETRANSCEIVES:
		ULONGSTATRETURN(kSMB2StatFsctlPipeTransceive);
	case SMB2FSCTLPIPEPEEKS:
		ULONGSTATRETURN(kSMB2StatFsctlPipePeek);
	case SMB2FSCTLENUMSNAPSHOTSS:
		ULONGSTATRETURN(kSMB2StatFsctlEnumSnapshots);
	case SMB2FSCTLDFSREFERRALSS:
		ULONGSTATRETURN(kSMB2StatFsctlDfsReferrals);
	case SMB2FSCTLSETSPARSES:
		ULONGSTATRETURN(kSMB2StatFsctlSetSparse);
	case SMB2FSCTLSECURESHARES:
		ULONGSTATRETURN(kSMB2StatFsctlSecureShare);
	case SMB2FSCTLFILEUNSUPPORTEDS:
		ULONGSTATRETURN(kSMB2StatFsctlFileUnsupported);
	case SMB2FSCTLIPCUNSUPPORTEDS:
		ULONGSTATRETURN(kSMB2StatFsctlIpcUnsupported);
	case SMB2QUERYDIRS:
		ULONGSTATRETURN(kSMB2StatQueryDir);
	case SMB2QUERYSECURITYINFOS:
		ULONGSTATRETURN(kSMB2StatQuerySecurityInfo);
	case SMB2SETSECURITYINFOS:
		ULONGSTATRETURN(kSMB2StatSetSecurityInfo);
	case SMB2QUERYFILEACCESSINFOS:
		ULONGSTATRETURN(kSMB2StatQueryFileAccessInfo);
	case SMB21SESSIONSETUPS:
		RETURN(smbstatGetGlobalValue(kSMB2_1StatSessSetup));
        default:
                ULONGRETURN(0);
        } // switch

        ASSERT(false); /* should never come here */
}

uint8_t *
var_cifsObsReqs(SNMP_FM_ARGS)
{
        sk_domain_t     calling_domain;

        VAR_INIT();

	RETURN_IF_CLUSTERMODE();

        calling_domain = C_DOMAIN_SWITCH_BEGIN(sk_domain_cifs);

        switch (vp->magic)
        {
        case OBSSMBCREATES:
                ULONGSTATRETURN(kSMBStatcreate);
        case OBSSMBCREATEDIRS:
                ULONGSTATRETURN(kSMBStatmkdir);
        case OBSSMBCREATENEWS:
                ULONGSTATRETURN(kSMBStatmknew);
        case OBSSMBLOCKANDREADS:
                ULONGSTATRETURN(kSMBStatlockread);
        case OBSSMBLOCKBYTERANGES:
                ULONGSTATRETURN(kSMBStatlock);
        case OBSSMBOPENS:
                ULONGSTATRETURN(kSMBStatopen);
        case OBSSMBOPENANDXS:
                ULONGSTATRETURN(kSMBStatopenX);
        case OBSSMBPROCESSEXITS:
                ULONGSTATRETURN(kSMBStatexit);
        case OBSSMBQUERYINFOS:
                ULONGSTATRETURN(kSMBStatgetatr);
        case OBSSMBQUERYINFO2S:
                ULONGSTATRETURN(kSMBStatgetattre);
        case OBSSMBREADS:
                ULONGSTATRETURN(kSMBStatCoreread);
        case OBSSMBREADMPXS:
                ULONGRETURN(0);
        case OBSSMBREADRAWS:
                ULONGSTATRETURN(kSMBStatreadbraw);
        case OBSSMBSEARCHS:
                ULONGSTATRETURN(kSMBStatsearch);
        case OBSSMBSETINFOS:
                ULONGSTATRETURN(kSMBStatsetatr);
        case OBSSMBSETINFO2S:
                ULONGSTATRETURN(kSMBStatsetattre);
        case OBSSMBQUERYINFODISKS:
                ULONGSTATRETURN(kSMBStatdskattr);
        case OBSSMBTRANS2OPEN2S:
                ULONGSTATRETURN(kSMBStatTRANS2_OPEN2);
        case OBSSMBTREECONNECTS:
                ULONGSTATRETURN(kSMBStattcon);
        case OBSSMBUNLOCKBYTERANGES:
                ULONGSTATRETURN(kSMBStatunlock);
        case OBSSMBWRITES:
                ULONGSTATRETURN(kSMBStatCorewrite);
        case OBSSMBWRITEANDUNLOCKS:
                ULONGSTATRETURN(kSMBStatwriteunlock);
        case OBSSMBWRITEANDCLOSES:
                ULONGSTATRETURN(kSMBStatwriteclose);
        case OBSSMBWRITEMPXS:
                ULONGRETURN(0);
        case OBSSMBWRITEPRINTFILES:
                ULONGRETURN(0);
        case OBSSMBWRITERAWS:
                ULONGSTATRETURN(kSMBStatwritebraw);        
        default:
                ULONGRETURN(0);
        } // switch
        
        ASSERT(false); /* should never come here */
}

uint8_t *
var_cifsPercent(SNMP_FM_ARGS)
{
        sk_domain_t     calling_domain;

        VAR_INIT();

	RETURN_IF_CLUSTERMODE();

        calling_domain = C_DOMAIN_SWITCH_BEGIN(sk_domain_cifs);

        switch (vp->magic)
        {
        case SMBNEGPROTPCT:
                LONGPCTRETURN(kSMBStatnegprot);
        case SMBSESSIONSETUPANDXPCT:
                LONGPCTRETURN(kSMBStatsesssetupX);
        case SMBLOGOFFANDXPCT:
                LONGPCTRETURN(kSMBStatulogoffX);
        case SMBTREECONNECTANDXPCT:
                LONGPCTRETURN(kSMBStattconX);
        case SMBTREEDISCONNECTANDXPCT:
                LONGPCTRETURN(kSMBStattdis);
        case SMBTRANS2QUERYFSINFOPCT:
                LONGPCTRETURN(kSMBStatTRANS2_QUERY_FS_INFORMATION);
        case SMBECHOPCT:
                LONGPCTRETURN(kSMBStatecho);
        case SMBNTCANCELPCT:
		C_REFERENCE_CIFS(smb_total_calls);
		C_REFERENCE_CIFS(ulpct);
                ulongReturn = (smbstatGetGlobalValue (kSMBStatNTCancelCN) + 
                               smbstatGetGlobalValue (kSMBStatNTCancelOther));
                ulongReturn = ulpct(ulongReturn, smb_total_calls()+1);
                RETURN(ulongReturn);
        case SMBCREATEANDXPCT:
                LONGPCTRETURN(kSMBStatNTCreate);
        case SMBTRANSACTCREATEPCT:
                LONGPCTRETURN(kSMBStatNTTransCreate);
        case SMBCREATETEMPORARYPCT:
                LONGRETURN(0);  /* not supported */
        case SMBREADANDXPCT:
                LONGPCTRETURN(kSMBStatreadX);
        case SMBWRITEANDXPCT:
                LONGPCTRETURN(kSMBStatwriteX);
        case SMBLOCKINGANDXPCT:
                LONGPCTRETURN(kSMBStatlockingX);
        case SMBSEEKPCT:
                LONGPCTRETURN(kSMBStatlseek);
        case SMBFLUSHPCT:
                LONGPCTRETURN(kSMBStatflush);
        case SMBCLOSEPCT:
                LONGPCTRETURN(kSMBStatclose);
        case SMBDELETEPCT:
                LONGPCTRETURN(kSMBStatunlink);
        case SMBRENAMEPCT:
                LONGPCTRETURN(kSMBStatmv);
        case SMBMOVEPCT:
                LONGRETURN(0);  /* ioctl ops not supported */
        case SMBCOPYPCT:
                LONGRETURN(0);  /* "" */
        case SMBTRANS2QUERYPATHINFOPCT:
                LONGPCTRETURN(kSMBStatTRANS2_QUERY_PATH_INFORMATION);
        case SMBTRANS2QUERYFILEINFOPCT:
                LONGPCTRETURN(kSMBStatTRANS2_QUERY_FILE_INFORMATION);
        case SMBTRANS2SETPATHINFOPCT:
                LONGPCTRETURN(kSMBStatTRANS2_SET_PATH_INFORMATION);
        case SMBTRANS2SETFILEINFOPCT:
                LONGPCTRETURN(kSMBStatTRANS2_SET_FILE_INFORMATION);
        case SMBDELETEDIRPCT:
                LONGPCTRETURN(kSMBStatrmdir);
        case SMBCHECKDIRPCT:
                LONGRETURN(0);  /* not supported */
        case SMBTRANS2FINDFIRST2PCT:
                LONGPCTRETURN(kSMBStatTRANS2_FIND_FIRST2);
        case SMBTRANS2FINDNEXT2PCT:
                LONGPCTRETURN(kSMBStatTRANS2_FIND_NEXT2);
        case SMBFINDCLOSE2PCT:
                LONGPCTRETURN(kSMBStatfindclose);
        case SMBNTTRANSACTNOTIFYCHGPCT:
                LONGPCTRETURN(kSMBStatNTTransNotify);
        case SMBTRANS2GETDFSREFERRALPCT:
                LONGPCTRETURN(kSMBStatTRANS2_DFS_REFERRAL);
        case SMBTRANS2REPORTDFSINCPCT:
                LONGPCTRETURN(kSMBStatTRANS2_DFS_REPORT_INCONSISTENCY);
        case SMBOPENPRINTFILEPCT:
                LONGRETURN(0);
        case SMBGETPRINTQUEUEPCT:
                LONGRETURN(0);
        case SMBNTTRANSACTIOCTLPCT:
                LONGPCTRETURN(kSMBStatNTTransIoctl);
        case SMBNTTRANSACTQUERYSECDESCPCT:
                LONGPCTRETURN(kSMBStatNTTransQuerySec);
        case SMBNTTRANSACTSETSECDESCPCT:
                LONGPCTRETURN(kSMBStatNTTransSetSec);
        case SMBNTCANCELCNPCT:
                LONGPCTRETURN(kSMBStatNTCancelCN);
        case SMBNTCANCELOTHERPCT:
                LONGPCTRETURN(kSMBStatNTCancelOther);
	case SMB2ECHOPCT:
		LONGPCTRETURN(kSMB2StatEcho);
	case SMB2NEGPROTPCT:
		LONGPCTRETURN(kSMB2StatNegProt);
	case SMB2SESSIONSETUPPCT:
		LONGPCTRETURN(kSMB2StatSessSetup);
	case SMB2LOGOFFPCT:
		LONGPCTRETURN(kSMB2StatLogoff);
	case SMB2TREECONNECTPCT:
		LONGPCTRETURN(kSMB2StatTreeConnect);
	case SMB2TREEDISCONNECTPCT:
		LONGPCTRETURN(kSMB2StatTreeDisconnect);
	case SMB2CANCELPCT:
		LONGPCTRETURN(kSMB2StatCancel);
	case SMB2CREATEPCT:
		LONGPCTRETURN(kSMB2StatCreate);
	case SMB2IPCCREATEPCT:
		LONGPCTRETURN(kSMB2IpcCreate);
	case SMB2READPCT:
		LONGPCTRETURN(kSMB2StatRead);
	case SMB2WRITEPCT:
		LONGPCTRETURN(kSMB2StatWrite);
	case SMB2IPCREADPCT:
		LONGPCTRETURN(kSMB2IpcRead);
	case SMB2IPCWRITEPCT:
		LONGPCTRETURN(kSMB2IpcWrite);
	case SMB2LOCKPCT:
		LONGPCTRETURN(kSMB2StatLock);
	case SMB2UNLOCKPCT:
		LONGPCTRETURN(kSMB2StatUnlock);
	case SMB2OPLKBRKACKPCT:
		LONGPCTRETURN(kSMB2StatOplkBrkAck);
	case SMB2FLUSHPCT:
		LONGPCTRETURN(kSMB2StatFlush);
	case SMB2CLOSEPCT:
		LONGPCTRETURN(kSMB2StatClose);
	case SMB2QUERYFILEBASICINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFileBasicInfo);
	case SMB2QUERYFILESTANDARDINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFileStandardInfo);
	case SMB2QUERYFILEINTERNALINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFileInternalInfo);
	case SMB2QUERYFILEEAINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFileEAInfo);
	case SMB2QUERYFILEFULLEAINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFileFullEAInfo);
	case SMB2QUERYFILEALTNAMEPCT:
		LONGPCTRETURN(kSMB2StatQueryFileAltNameInfo);
	case SMB2QUERYFILESTREAMINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFileStreamInfo);
	case SMB2QUERYFILENETOPENINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFileNetOpenInfo);
	case SMB2QUERYFILEATTRTAGINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFileAttrTagInfo);
	case SMB2QUERYFILEUNSUPPORTEDPCT:
		LONGPCTRETURN(kSMB2StatQueryFileUnsupported);
	case SMB2QUERYFILEINVALIDPCT:
		LONGPCTRETURN(kSMB2StatQueryFileInvalid);
	case SMB2QUERYFSVOLINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFsVolInfo);
	case SMB2QUERYFSSIZEINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFsSizeInfo);
	case SMB2QUERYFSDEVICEINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFsDeviceInfo);
	case SMB2QUERYFSATTRINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFsAttrInfo);
	case SMB2QUERYFSFULLSIZEINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFsFullSizeInfo);
	case SMB2QUERYFSOBJECTIDINFOPCT:
		LONGPCTRETURN(kSMB2StatQueryFsObjectIdInfo);
	case SMB2QUERYFSINVALIDPCT:
		LONGPCTRETURN(kSMB2StatQueryFsInvalid);
	case SMB2SETBASICINFOPCT:
		LONGPCTRETURN(kSMB2StatSetBasicInfo);
	case SMB2SETRENAMEINFOPCT:
		LONGPCTRETURN(kSMB2StatSetRenameInfo);
	case SMB2SETFILELINKINFOPCT:
		LONGPCTRETURN(kSMB2StatSetFileLinkInfo);
	case SMB2SETFILEDISPINFOPCT:
		LONGPCTRETURN(kSMB2StatSetFileDispInfo);
	case SMB2SETFULLEAINFOPCT:
		LONGPCTRETURN(kSMB2StatSetFullEAInfo);
	case SMB2SETMODEINFOPCT:
		LONGPCTRETURN(kSMB2StatSetModeInfo);
	case SMB2SETALLOCINFOPCT:
		LONGPCTRETURN(kSMB2StatSetAllocInfo);
	case SMB2SETEOFINFOPCT:
		LONGPCTRETURN(kSMB2StatSetEOFInfo);
	case SMB2SETINFOUNSUPPORTEDPCT:
		LONGPCTRETURN(kSMB2StatSetInfoUnsupported);
	case SMB2SETINFOINVALIDPCT:
		LONGPCTRETURN(kSMB2StatSetInfoInvalid);
	case SMB2FSCTLPIPETRANSCEIVEPCT:
		LONGPCTRETURN(kSMB2StatFsctlPipeTransceive);
	case SMB2FSCTLPIPEPEEKPCT:
		LONGPCTRETURN(kSMB2StatFsctlPipePeek);
	case SMB2FSCTLENUMSNAPSHOTSPCT:
		LONGPCTRETURN(kSMB2StatFsctlEnumSnapshots);
	case SMB2FSCTLDFSREFERRALSPCT:
		LONGPCTRETURN(kSMB2StatFsctlDfsReferrals);
	case SMB2FSCTLSETSPARSEPCT:
		LONGPCTRETURN(kSMB2StatFsctlSetSparse);
	case SMB2FSCTLSECURESHAREPCT:
		LONGPCTRETURN(kSMB2StatFsctlSecureShare);
	case SMB2FSCTLFILEUNSUPPORTEDPCT:
		LONGPCTRETURN(kSMB2StatFsctlFileUnsupported);
	case SMB2FSCTLIPCUNSUPPORTEDPCT:
		LONGPCTRETURN(kSMB2StatFsctlIpcUnsupported);
	case SMB2QUERYDIRPCT:
		LONGPCTRETURN(kSMB2StatQueryDir);
	default:
                LONGRETURN(0);
        } // switch

        ASSERT(false); /* should never come here */
}

uint8_t *
var_cifsObsPercent(SNMP_FM_ARGS)
{
        sk_domain_t     calling_domain;

        VAR_INIT();

	RETURN_IF_CLUSTERMODE();

        calling_domain = C_DOMAIN_SWITCH_BEGIN(sk_domain_cifs);

        switch (vp->magic)
        {
        case OBSSMBCLOSEPRINTFILEPCT:
                LONGPCTRETURN(kSMBStatcreate);
        case OBSSMBCREATEPCT:
                LONGPCTRETURN(kSMBStatmkdir);
        case OBSSMBCREATEDIRPCT:
                LONGPCTRETURN(kSMBStatmknew);
        case OBSSMBCREATENEWPCT:
                LONGPCTRETURN(kSMBStatlockread);
        case OBSSMBLOCKANDREADPCT:
                LONGPCTRETURN(kSMBStatlock);
        case OBSSMBLOCKBYTERANGEPCT:
                LONGPCTRETURN(kSMBStatopen);
        case OBSSMBOPENPCT:
                LONGPCTRETURN(kSMBStatopenX);
        case OBSSMBOPENANDXPCT:
                LONGPCTRETURN(kSMBStatexit);
        case OBSSMBPROCESSEXITPCT:
                LONGPCTRETURN(kSMBStatgetatr);
        case OBSSMBQUERYINFOPCT:
                LONGPCTRETURN(kSMBStatgetattre);
        case OBSSMBQUERYINFO2PCT:
                LONGPCTRETURN(kSMBStatCoreread);
        case OBSSMBREADPCT:
                LONGRETURN(0);
        case OBSSMBREADMPXPCT:
                LONGPCTRETURN(kSMBStatreadbraw);
        case OBSSMBREADRAWPCT:
                LONGPCTRETURN(kSMBStatsearch);
        case OBSSMBSEARCHPCT:
                LONGPCTRETURN(kSMBStatsetatr);
        case OBSSMBSETINFOPCT:
                LONGPCTRETURN(kSMBStatsetattre);
        case OBSSMBSETINFO2PCT:
                LONGPCTRETURN(kSMBStatdskattr);
        case OBSSMBQUERYINFODISKPCT:
                LONGPCTRETURN(kSMBStatTRANS2_OPEN2);
        case OBSSMBTRANS2OPEN2PCT:
                LONGPCTRETURN(kSMBStattcon);
        case OBSSMBTREECONNECTPCT:
                LONGPCTRETURN(kSMBStatunlock);
        case OBSSMBUNLOCKBYTERANGEPCT:
                LONGPCTRETURN(kSMBStatCorewrite);
        case OBSSMBWRITEPCT:
                LONGPCTRETURN(kSMBStatwriteunlock);
        case OBSSMBWRITEANDUNLOCKPCT:
                LONGPCTRETURN(kSMBStatwriteclose);
        case OBSSMBWRITEANDCLOSEPCT:
                LONGRETURN(0);
        case OBSSMBWRITEMPXPCT:
                LONGRETURN(0);
        case OBSSMBWRITEPRINTFILEPCT:
                LONGRETURN(0);
        case OBSSMBWRITERAWPCT:
                LONGPCTRETURN(kSMBStatwritebraw);
        default:
                LONGRETURN(0);
        } // switch
        
        ASSERT(false); /* should never come here */
}

uint8_t *
var_cifsMisc(SNMP_FM_ARGS)
{
        sk_domain_t     calling_domain;

        VAR_INIT();

	RETURN_IF_CLUSTERMODE();

        calling_domain = C_DOMAIN_SWITCH_BEGIN(sk_domain_cifs);

        switch (vp->magic)
        {
        case CIFSCANCELLOCKS:
                ULONGSTATRETURN(kSMBStatlockingXCancel);
        case CIFSWAITLOCKS:
                ULONGSTATRETURN(kSMBStatlockingXwait);
        case CIFSCOPYTOALIGNS:
                ULONGSTATRETURN(kSMBCopyAlign);
        case CIFSALIGNEDSMALLS:
                ULONGSTATRETURN(kalignedBufferSmall);
        case CIFSALIGNEDLARGES:
                ULONGSTATRETURN(kalignedBufferLarge);
        case CIFSALIGNEDSMALLRELS:
                ULONGSTATRETURN(kalignedBufferSmallRel);
        case CIFSALIGNEDLARGERELS:
                ULONGSTATRETURN(kalignedBufferLargeRel);
        case CIFSMBUFWAITS:
                ULONGSTATRETURN(kmBufWait);
        case CIFSNBTWAITS:
                ULONGSTATRETURN(kmSendDelay);
        case CIFSCWAWAITS:
                ULONGSTATRETURN(kcwaWait);
        case CIFSMULTIPLEVCS:
                ULONGSTATRETURN(kMultipleVCs);
        case CIFSPDCUPCALLS:
                ULONGSTATRETURN(kSMBStatPDCupcalls);
        case CIFSQUEUEDWRITERAWS:
                ULONGSTATRETURN(kSMBStatQueueWriteRaw);
        case CIFSNBTDISCONNECTS:
                ULONGSTATRETURN(kNBTDisconnect);
        case CIFSSMBDISCONNECTS:
                ULONGSTATRETURN(kSMBDisconnect);
        case CIFSDUPDISCONNECTS:
                ULONGSTATRETURN(kDupDisconnect);
        case CIFSOPLKBATCHTOL2S:
                ULONGSTATRETURN(kOpLkBkXorBatchToL2);
        case CIFSOPLKBATCHTONONES:
                ULONGSTATRETURN(kOpLkBkXorBatchToNone);
        case CIFSOPLKL2TONONES:
                ULONGSTATRETURN(kOpLkBkL2ToNone);
        case CIFSOPLKNOBREAKACKS:
                ULONGSTATRETURN(kOpLkBkNoBreakAck);
        case CIFSOPLKIGNOREDACKS:
                ULONGSTATRETURN(kOpLkBkIgnoredAck);
        case CIFSOPLKWAITERTIMEDOUTS:
                ULONGSTATRETURN(kOpLkBkWaiterTimedOut);
        case CIFSOPLKDELAYEDBREAKS:
                ULONGSTATRETURN(kOpLkBkDelayedBreak);
        case CIFSSHARINGERRORRETRIES:
                ULONGSTATRETURN(kSharingErrorRetries);
        case CIFSCLEARTEXTPASSWD:
                ULONGRETURN(((ParamAccess(i_AuthOption) == kUseEtcPasswd)? 2 : 1));
	case CIFSSMB2MIDUSEDHASH:
		ULONGSTATRETURN(kSMB2StatMidUsedHash);
	case CIFSSMB2MIDUNUSEDHASH:
		ULONGSTATRETURN(kSMB2StatMidUnusedHash);
	case CIFSSMB2MIDINVREJECT:
		ULONGSTATRETURN(kSMB2StatMidInvReject);
	case CIFSSMB2DURABLECREATERECEIVED:
		ULONGSTATRETURN(kSMB2StatDurableCreateReceived);
	case CIFSSMB2DURABLECREATESUCCEEDED:
		ULONGSTATRETURN(kSMB2StatDurableCreateSucceeded);
	case CIFSSMB2DURABLERECLAIMRECEIVED:
		ULONGSTATRETURN(kSMB2StatDurableReclaimReceived);
	case CIFSSMB2DURABLERECLAIMSUCCEEDED:
		ULONGSTATRETURN(kSMB2StatDurableReclaimSucceeded);
	case CIFSSMB2DURABLEHANDLEPRESERVED:
		ULONGSTATRETURN(kSMB2StatDurableHandlePreserved);
	case CIFSSMB2DURABLEHANDLEPURGED:
		ULONGSTATRETURN(kSMB2StatDurableHandlePurged);
	case CIFSSMB2DURABLEHANDLEEXPIRED:
		ULONGSTATRETURN(kSMB2StatDurableHandleExpired);
	case CIFSSMB2FILEDIRINFO:
		ULONGSTATRETURN(kSMB2StatFileDirInfo);
	case CIFSSMB2FILEFULLDIRINFO:
		ULONGSTATRETURN(kSMB2StatFileFullDirInfo);
	case CIFSSMB2FILEIDFULLDIRINFO:
		ULONGSTATRETURN(kSMB2StatFileIdFullDirInfo);
	case CIFSSMB2FILEBOTHDIRINFO:
		ULONGSTATRETURN(kSMB2StatFileBothDirInfo);
	case CIFSSMB2FILEIDBOTHDIRINFO:
		ULONGSTATRETURN(kSMB2StatFileIdBothDirInfo);
	case CIFSSMB2FILENAMESINFO:
		ULONGSTATRETURN(kSMB2StatFileNamesInfo);
	case CIFSSMB2FILEDIRUNSUPPORTED:
		ULONGSTATRETURN(kSMB2StatFileDirUnsupported);
	case CIFSSMB2QUERYINFO:
		ULONGSTATRETURN(kSMB2StatQueryInfo);
	case CIFSSMB2SETINFO:
		ULONGSTATRETURN(kSMB2StatSetInfo);
	case CIFSSMB2IOCTL:
		ULONGSTATRETURN(kSMB2StatIoctl);
	case CIFSSMB2RELATEDCOMPREQUEST:
		ULONGSTATRETURN(kSMB2StatRelatedCompRequest);
	case CIFSSMB2UNRELATEDCOMPREQUEST:
		ULONGSTATRETURN(kSMB2StatUnRelatedCompRequest);
	case CIFSSMB2FILEREQUEST:
		ULONGSTATRETURN(kSMB2StatFileRequest);
	case CIFSSMB2PIPEREQUEST:
		ULONGSTATRETURN(kSMB2StatPipeRequest);
	case CIFSSMB2UNSUPPORTED:
		ULONGSTATRETURN(kSMB2StatNotSupport);
	case CIFSSMB2INVALIDSIGNATURES:
		ULONGSTATRETURN(kSMB2StatInvalidSignature);
        default:
                ULONGRETURN(0);
        } // switch
        
        ASSERT(false); /* should never come here */
}

/*---------------------------------------------------------------------------*/

