/*
 *
 * Copyright (c) 1992-2007 Network Appliance, Inc.
 * All rights reserved.
 */

#include <sk/sk_debug.h>
#include <sk/ha_host.h>


#include <smb/smbauthen.h>
#include <sys/types.h>
#include <net7g/netinet/sk_in.h>
#include <nbt/nbns.h>
#include <net7g/arpa/sk_inet.h>
#include <lib/sgmltext.h>
#include <smb/smbservers.h>
#include <smb/smb.h>
#include <smb/smbstat.h>

#include <auth/sec_dc.h>

#include "cifs_vars_aux.h"

/************************************************************************/

void bufprint_winsservers(uint8_t* buf, int buflen)
	/* 
	 * print the names of the WINS servers
	 * the filer knows about, one per line
	 */
{
	int		i, n, len;
	struct in_addr	WINSservers[MAX_NBNS_SERVERS];

	DECLARE_IN_CIFS();
	ASSERT(buf);
	buf[0] = 0;

		/* get number of WINS servers */
	SK_DOMAIN_REFERENCE(nbns_get_servers, SK_REASON_SWITCH_DOMAIN);
	n = nbns_get_servers(WINSservers);
		/* print each on a separate line */
		/* careful not to overflow buf */
	for(i=0; i<n; i++) {
		char addrStr[INETNTOABUFSIZE];
		(void) inet_ntoa_r(WINSservers[i], addrStr);
		len = STRLEN(buf);
		STRNCAT(buf, addrStr, buflen-len);
		STRNCAT(buf, "\n", buflen-len);
	}
	buf[buflen-1] = 0;
}
/************************************************************************/

void bufprint_altnames(uint8_t* buf, int buflen)
	/* 
	 * print the alternate names the filer is
	 * known by to buf.
	 */
{
	DECLARE_IN_CIFS();
	ASSERT(buf);
	buf[0] = 0;
	if (ParamAccess(c_aliasUC) == NULL)
		return;
	else
		ConvertUCtoSGML(ParamAccess(c_aliasUC), (SGML*) buf,
			buflen, false);
}

/************************************************************************/

void bufprint_UC(WCHAR* src, uint8_t* buf, int buflen)
	/* copy Unicode src into buf as an SGML string */
{
	DECLARE_IN_CIFS();
	ASSERT(buf);
	buf[0] = 0;
	
	if (src == NULL)
		return;
	ConvertUCtoSGML(src, (SGML*)buf, buflen, false);
}

/************************************************************************/

void bufprint_DCName(uint8_t* buf, int buflen)
	/*
	 * print the name of the Primary Domain
	 * Controller to buf, translating to
	 * SGML.
	 */
{
	DECLARE_IN_CIFS();
	ASSERT(buf);
	buf[0] = 0;

	C_REFERENCE_CIFS(sec_GetDCNames_SGML);
	sec_GetDCNames_SGML(NULL,buf, buflen);

	return;
}

/************************************************************************/

int is_domain_joined(void)
	/* has the filer joined a domain? */
{
	int rc;

        DECLARE_IN_CIFS();

	C_REFERENCE_CIFS(sec_HaveDCConnection);
	rc = sec_HaveDCConnection(NULL, false);

	return(rc);
}

/************************************************************************/

void bufprint_NBTMode(uint8_t* buf, int buflen)
	/*
	 * print the NetBios Transfer mode
	 * to buf.
	 */
{
	DECLARE_IN_CIFS();
	ASSERT(buf);

	/* XXX to do */
	buf[0] = 0;
}

/************************************************************************/

unsigned long smb_total_calls(void)
{
	int		i;
	unsigned long	n = 0;

	DECLARE_IN_CIFS();
        
	C_REFERENCE_LAX(gSMBstats__);
	for(i=0; i<kSMBStatPercent; i++)
		n += smbstatGetGlobalValue (i);
	return n;
}
	
/************************************************************************/

unsigned long ulpct(unsigned long num, unsigned long denom)
	/*
	 * return num/denom as a percentage, taking steps
	 * to deal with integer overflow.  
	 */
{
	uint64_t	tmp;

	DECLARE_IN_CIFS();
        
	tmp = num * 100;
	tmp /= denom;
	return tmp & 0xffffffff;
}

/************************************************************************/


uint64_t 
SMB_GetNActiveVfilers(void)
{
	uint64_t nActive = 0;
	vfiler *vfp;
	vfiler *ovfp = sk_get_vfiler();
	
	DECLARE_IN_CIFS();
        
	sk_lock(&HH_vfiler_list_lock);

	for (vfp = TAILQ_FIRST(&HH_vfiler_list);
	     vfp != NULL;
	     vfp = vfp->vlink.tqe_next) {
		/*
		 * Only count the ops for consistent, initialized vfilers.
		 */
		if (!VFILER_IS_UP(vfp))
			continue;
		sk_set_vfiler(vfp);
		if(SMB_IsStarted())
			nActive++;
	}
	sk_set_vfiler(ovfp);

	sk_unlock(&HH_vfiler_list_lock);
	return nActive;
}


uint64_t 
SMB_GetNOplockBreakWaits(void)
{
	uint64_t retval = 0;
	FileCB*	fcb;
	vfiler *vfp;
	vfiler *ovfp = sk_get_vfiler();
	
	DECLARE_IN_CIFS();

        sk_lock(&HH_vfiler_list_lock);

	for (vfp = TAILQ_FIRST(&HH_vfiler_list);
	     vfp != NULL;
	     vfp = vfp->vlink.tqe_next) {
		/*
		 * Only count the ops for consistent, initialized vfilers.
		 */
		if (!VFILER_IS_UP(vfp))
			continue;
		sk_set_vfiler(vfp);
		C_LOCK(&SMBAccess(i_WaitSendList_spinlock));
		for (fcb = TAILQ_FIRST(&SMBAccess(i_OplkBk_WaitSendList));
		     fcb != NULL; fcb = TAILQ_NEXT(fcb, u.f.o.ol.OplkBk_Link))
			retval++; 
		C_UNLOCK(&SMBAccess(i_WaitSendList_spinlock));
	}
	sk_set_vfiler(ovfp);

	sk_unlock(&HH_vfiler_list_lock);
	return retval;
}


uint64_t 
SMB_GetNOplockAckWaits(void)
{
	uint64_t retval = 0;
	FileCB*	fcb;
	vfiler *vfp;
	vfiler *ovfp = sk_get_vfiler();
	
	DECLARE_IN_CIFS();

        sk_lock(&HH_vfiler_list_lock);

	for (vfp = TAILQ_FIRST(&HH_vfiler_list);
	     vfp != NULL;
	     vfp = vfp->vlink.tqe_next) {
		/*
		 * Only count the ops for consistent, initialized vfilers.
		 */
		if (!VFILER_IS_UP(vfp))
			continue;
		sk_set_vfiler(vfp);
		C_LOCK(&SMBAccess(i_WaitSendList_spinlock));
		for (fcb = TAILQ_FIRST(&SMBAccess(c_OplkBk_WaitAckList));
		     fcb != NULL; fcb = fcb->u.f.o.ol.OplkBk_Link.tqe_next)
			retval++;
		C_UNLOCK(&SMBAccess(i_WaitSendList_spinlock));
	}
	sk_set_vfiler(ovfp);

	sk_unlock(&HH_vfiler_list_lock);
	return retval;
}




