/*
 *
 * Copyright (c) 1992-2010 NetApp, Inc.
 * All rights reserved.
 */



#include <sk/sk_stats.h>
#include <sys/param.h>
#include <rpc/types.h>
#include <net7g/sys/sk_socket.h>
#include <sk/sk.h>
#include <sk/typedefs.h>
#include <sk/proto.h>
#include <sk/sk_processor.h>
#include <vm/vm_machdep.h>
#include <ha/typedefs.h>
#include <net7g/netinet/sk_in.h>
#include <net7g/netinet/sk_in_cksum_buf.h>
#include <net7g/net/sk_if.h>
#include <net7g/arpa/sk_inet.h>
#include <net7g/net/sk_route.h>
#include <sk/sk_synch.h>
#include <net7g/netinet/sk_in_var.h>
#include <net7g/netinet/sk_if_ether.h>
#include <net7g/netinet/sk_in_systm.h>
#include <net7g/netinet/sk_ip.h>
#include <net7g/netinet/sk_in_pcb.h>
#include <net7g/netinet/sk_ip_var.h>   
#include <tcp/tcp_var.h>
#include <tcp/tcp_fsm.h>        
#include <lib/rsrc_mgr.h>
#include <wafl/buf.h>
#include <net7g/sys/sk_mbuf.h>
#include <wafl/typedefs.h>
#include <smb/smb.h>
#include <smb/smbmsgs.h>
#include <smb/smbstat.h>
#include <cmds/smbsetup.h>
#include <cmds/ext.h>
#include <cmds/enable.h>
#include <legacy/string.h>
#include <fio/fio.h>
#include <rpc/sunrpc_vfiler.h>
#include <net7g/netinet/sk_udp_var.h>
#include <adminapi/zapi.h>
#include <driver/scsi/aha_phys_dev.h>
#include <driver/scsi/scsidef.h>
#include <driver/scsi/wwn.h>
#include <driver/scsitarget/tha.h>
#include <driver/scsitarget/scsi_tgt_proto.h>
#include <driver/scsitarget/iscsi_protocol.h>
#include <driver/scsitarget/iscsimgt/node/iscsimgt_node_mapi_extern.h>
#include "net-snmp/net-snmp-config.h"
#include "net-snmp/net-snmp-includes.h"
#include "net-snmp/agent/snmp_vars.h"
#include "iscsi_vars.h"
#include "netapp_vars.h"
#include "agent/mibgroup/mibparser_defines.h"
#include "agent/mibgroup/mibparser_structs.h"
#include "agent/mibgroup/util_funcs.h"

/*
 * Internal functions
 */
static uint32_t
iscsi_snmp_get_portal_count(void);

static iscsimgt_mapi_portal_t *
iscsi_snmp_get_nth_portal(iscsimgt_mapi_portal_t *, int);

static uint32_t
iscsi_snmp_get_session_count(void);

static iscsimgt_mapi_session_t *
iscsi_snmp_get_nth_session(iscsimgt_mapi_session_t *, int);

static uint32_t
iscsi_snmp_get_session_conn_count(iscsimgt_mapi_session_t *);

static iscsimgt_mapi_session_conn_t *
iscsi_snmp_get_nth_session_conn(iscsimgt_mapi_session_t *, int);

static snmpv2_TruthValue_t Bool_to_TruthValue(bool param);

#define HIGH32(x)	((uint32_t)((x)>>32))
#define LOW32(x)	((uint32_t)(x))

SK_DOMAIN_IMPORT(iscsimgt_mapi_stats_get, SK_REASON_SWITCH_DOMAIN);
SK_DOMAIN_IMPORT(iscsimgt_mapi_helper_stats_free, SK_REASON_SWITCH_DOMAIN);
SK_DOMAIN_IMPORT(iscsimgt_mapi_portal_get_list, SK_REASON_SWITCH_DOMAIN);
SK_DOMAIN_IMPORT(iscsimgt_mapi_helper_portal_list_free, SK_REASON_SWITCH_DOMAIN);
SK_DOMAIN_IMPORT(iscsimgt_mapi_targetname_get, SK_REASON_SWITCH_DOMAIN);
SK_DOMAIN_IMPORT(iscsimgt_mapi_helper_targetname_free, SK_REASON_SWITCH_DOMAIN);
SK_DOMAIN_IMPORT(iscsimgt_mapi_targetalias_get, SK_REASON_SWITCH_DOMAIN);
SK_DOMAIN_IMPORT(iscsimgt_mapi_helper_targetalias_free, SK_REASON_SWITCH_DOMAIN);
SK_DOMAIN_IMPORT(iscsimgt_mapi_session_get_list, SK_REASON_SWITCH_DOMAIN);
SK_DOMAIN_IMPORT(iscsimgt_mapi_helper_session_list_free, SK_REASON_SWITCH_DOMAIN);
SK_DOMAIN_IMPORT(iscsimgt_mapi_helper_util_conn_snmp_state_get, SK_REASON_SWITCH_DOMAIN);


/*
 * var_iscsiInstanceAttributesEntry()
 *
 * Invoked by the snmp agent to get the OID values
 * of the MIB objects in the Instance Attributes Table.
 * We support only one Instance.
 * 9999(iscsiModule).1(iscsiObjects).2(iscsiInstance).
 * 1(iscsiInstanceAttributesTable).1(iscsiInstanceAttributesEntry)
 * 9999.1.2.1.1.{1-12}.1
 */
/* format is .1.3.6.1.3.9999.1.2.1.1.magic.index */
uint8_t *
var_iscsiInstanceAttributesEntry(SNMP_FM_ARGS)
{
	uint8_t *                rval;
	int                     rval_len;
	iscsimgt_mapi_stats_t * stats = NULL;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	/*
	 * We support only one instance (index always equals 1)
	 */
	HEADER_SIMPLE_TABLE(1);

	/* Used by some of the cases below */
	stats = iscsimgt_mapi_stats_get((zapi_error_info_t *)0,
					iscsimgt_transport_type_all);

	switch(vp->magic) {

	case ISCSIINSTINDEX: 
		long_return = NAME_INDEX;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIINSTDESCR:
		/* sprintf((char*)return_buf, ISCSI_SNMP_INST_DESCR); */
		return_buf[0] = (char)0; /* Avoid zero-len cc warn */
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSIINSTVERSIONMIN:
		long_return = ISCSI_PROTOCOL_VERSION_MIN;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIINSTVERSIONMAX:
		long_return = ISCSI_PROTOCOL_VERSION_MAX;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIINSTVENDORID:
		sprintf((char*)return_buf, ISCSI_SNMP_VENDOR_ID);
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSIINSTVENDORVERSION:
		sprintf((char*)return_buf, ISCSI_SNMP_VENDOR_VERSION);
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	/*
	 * The number of rows in the iscsiPortalAttributesTable which
	 * are currently associated with this iSCSI Instance.
	 * Get all the portals that are used for iscsi communication
	 * and return that.
	 */
	case ISCSIINSTPORTALNUMBER:
		long_return = iscsi_snmp_get_portal_count();
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIINSTNODENUMBER:
		long_return = ISCSI_SNMP_NODE_NUMBER;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIINSTSESSIONNUMBER:
		long_return = iscsi_snmp_get_session_count();
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	/*
	 * No of times a session beloning to this instance has failed.
	 */
	case ISCSIINSTSSNFAILURES:
		long_return =
			((stats != (iscsimgt_mapi_stats_t *)0) ?
			 (stats->sc_login_failed +
			  stats->sc_taskmgmt_failed +
			  stats->sc_logout_failed +
			  stats->sc_textcmd_failed +
			  stats->sc_protocol_errors +
			  stats->sc_hdr_digest_errors +
			  stats->sc_data_digest_errors +
			  stats->sc_outside_cmdsn_window_pdus +
			  stats->sc_unknown_pdu_rcvd) :
			 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIINSTLASTSSNFAILURETYPE:
		if ((stats != (iscsimgt_mapi_stats_t *)0) &&
		    (stats->lastssn_failure_var ==
		     ISCSI_SNMP_SSN_DIGEST_ERROR)) {
			sprintf((char*)return_buf,
			        ISCSI_SNMP_SSN_DIGEST_ERROR_OID);
		} else {
			sprintf((char*)return_buf,
			        ISCSI_SNMP_SSN_OTHER_ERROR_OID);
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSIINSTLASTSSNRMTNODENAME:
		if (stats != (iscsimgt_mapi_stats_t *)0) {
			sprintf((char*)return_buf,
			        stats->lastssn_remote_nodename);
		} else {
			return_buf[0] = (char)0; /* Avoid zero-len cc warn */
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;
	}

return_label:
	if (stats != (iscsimgt_mapi_stats_t *)0) {
		iscsimgt_mapi_helper_stats_free(stats);
		stats = (iscsimgt_mapi_stats_t *)0;
	}

	*var_len = rval_len;
	return(rval);
}

/*
 * var_iscsiInstanceSsnErrorStatsEntry()
 * 
 * Invoked by the snmp agent to get the OID values of 
 * MIB objects in the Instance Session Failure Stats Table.
 * 9999(iscsiModule).1(iscsiObjects).2(iscsiInstance).
 * 2(iscsiInstanceSsnErrorStatsTable).1(iscsiInstanceSsnErrorStatsEntry)
 * 9999.1.2.2.1.{1-3}.1
 */
/* format is .1.3.6.1.3.9999.1.2.2.1.magic.index */
uint8_t *
var_iscsiInstanceSsnErrorStatsEntry(SNMP_FM_ARGS)
{
	uint8_t *                rval;
	int                     rval_len;
	iscsimgt_mapi_stats_t * stats = NULL;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	/*
	 * We support only one instance (index always equals 1)
	 */
	HEADER_SIMPLE_TABLE(1);

	/* Used by some of the cases below */
	stats = iscsimgt_mapi_stats_get((zapi_error_info_t *)0,
					iscsimgt_transport_type_all);

	switch(vp->magic) {

	/*
	 * Count of header or data digest errors detected.
	 */
	case ISCSIINSTSSNDIGESTERRORS:
		if (stats != (iscsimgt_mapi_stats_t *)0) {
			long_return = (int)
			   (stats->sc_hdr_digest_errors +
			    stats->sc_data_digest_errors);
		} else {
			long_return = 0;
		}
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIINSTSSNCXNTIMEOUTERRORS:
	case ISCSIINSTSSNFORMATERRORS:
	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;
	}

return_label:
	if (stats != (iscsimgt_mapi_stats_t *)0) {
		iscsimgt_mapi_helper_stats_free(stats);
		stats = (iscsimgt_mapi_stats_t *)0;
	}
	*var_len = rval_len;
	return(rval);
}

/*
 * var_iscsiPortalAttributesEntry()
 *
 * Invoked by the snmp agent to get the OID values
 * of MIB objects in the Portal Attributes Table.
 * 9999(iscsiModule).1(iscsiObjects).3(iscsiPortal).
 * 1(iscsiPortalAttributesTable).1(iscsiPortalAttributesEntry)
 * 9999.1.3.1.1.{1-12}.1
 */
/* format is .1.3.6.1.3.9999.1.3.1.1.magic.inst_index.portal_index */
uint8_t *
var_iscsiPortalAttributesEntry(SNMP_FM_ARGS)
{
	oid                      newname[SNMP_OID_MAX];
	int                      result;
	int		         inst_index = 1;
	uint8_t *                 rval;
	int                      rval_len;
	iscsimgt_mapi_portal_t * portal_list = NULL;
	iscsimgt_mapi_portal_t * portal = NULL;
	struct sockaddr_storage *sa;
	int			portalctr,portalno;
	int			done = 0;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	if (compare(vp->name, vp->namelen, name, vp->namelen) != 0) {
		HEADER_SIMPLE_TABLE(10000);
	}

	if (*length < 12) {
		goto return_null_label;
	}

	bcopy((char *) vp->name, (char *) newname,
	      (int) vp->namelen * sizeof(oid));

	newname[11] = inst_index;

	if (*length >= 13) {
		portalno = name[12];
	} else {
		newname[12] = portalno = 1;
	}
	
	portal_list = iscsimgt_mapi_portal_get_list((zapi_error_info_t *)0);

	for (portalctr = 1; ((!done) && (portalctr <= portalno)); portalctr++) {

		portal = iscsi_snmp_get_nth_portal(portal_list, portalctr);

		if (!portal) {
			goto return_null_label;
		}

		if (portalctr < portalno) {
			continue;
		}
	
		newname[12] = (oid) portalctr;

		result = compare(name, *length, newname, 13);
		done = ((exact && result == 0) || (!exact && result < 0));
	
		if ((!done) && (!exact)) {
			portalno++;
			name[12] = 0;
		}

	}

	if (!done || (portal == NULL)) {
		goto return_null_label;
	}
	portalctr--;
	*length = 13;
	bcopy((char *) newname, (char *) name, (int)*length  * sizeof(oid));
	*write_method = 0;

	switch(vp->magic) {

	case ISCSIPORTALINDEX:
		long_return = portalctr;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIPORTALROWSTATUS:
		long_return = ISCSI_SNMP_STATUS_ACTIVE;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIPORTALROLES:
		sprintf((char*)return_buf, ISCSI_SNMP_TARGET_TYPE_PORTAL);
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSIPORTALADDRTYPE:
		sa = &portal->sockaddr;
		long_return = (SA_FAM(sa) == AF_INET6) ?
			      ISCSI_SNMP_INET_ADDRESS_IPv6 :
			      ISCSI_SNMP_INET_ADDRESS_IPv4;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIPORTALADDR:
		sa = &portal->sockaddr;
		iscsimgt_util_inet_satop(SA(sa),
					&return_buf[0],
					MAX_IPADDRSTR_LEN);
		if (!return_buf[0]) {
			sprintf((char*)return_buf, "none");
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSIPORTALPROTOCOL:
		long_return = ISCSI_SNMP_TRANSPORT_PROTOCOL_TCP;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIPORTALMAXRECVDATASEGLENGTH:
		long_return = ISCSIMGT_MAX_RECV_DSEG_LENGTH_DEFAULT;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIPORTALPRIMARYHDRDIGEST:
	case ISCSIPORTALPRIMARYDATADIGEST:
		long_return = ISCSI_SNMP_DEFAULT_DIGEST;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIPORTALSECONDARYHDRDIGEST:
	case ISCSIPORTALSECONDARYDATADIGEST:
		long_return = ISCSI_SNMP_ALTERNATE_DIGEST;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSIPORTALRECVMARKER:
		long_return = ISCSI_SNMP_MARKER_SUPPORT;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	}

return_label:
	if (portal_list != (iscsimgt_mapi_portal_t *)0) {
		iscsimgt_mapi_helper_portal_list_free(portal_list);
		portal_list = (iscsimgt_mapi_portal_t *)0;
		portal = (iscsimgt_mapi_portal_t *)0;
	}
	*var_len = rval_len;
	return(rval);

return_null_label:
	if (portal_list != (iscsimgt_mapi_portal_t *)0) {
		iscsimgt_mapi_helper_portal_list_free(portal_list);
		portal_list = (iscsimgt_mapi_portal_t *)0;
		portal = (iscsimgt_mapi_portal_t *)0;
	}
	return(NULL);
}

/*
 * var_iscsiTgtPortalAttributesEntry()
 *
 * Invoked by the snmp agent to get the OID values
 * of MIB Objects in the Target Portal Attributes
 * Table.
 * 9999(iscsiModule).1(iscsiObjects).4(iscsiTargetPortal).
 * 1(iscsiTargetPortalAttributesTable).1(iscsiTargetPortalAttributesEntry)
 * 9999.1.4.1.1.{1-2}.1
 */
/* format is .1.3.6.1.3.9999.1.4.1.1.magic.inst_index.portal_index */
uint8_t *
var_iscsiTgtPortalAttributesEntry(SNMP_FM_ARGS)
{
	oid                      newname[SNMP_OID_MAX];
	int                      result;
	int		         inst_index = 1;
	uint8_t *                 rval;
	int                      rval_len;
	iscsimgt_mapi_portal_t * portal_list = NULL;
	iscsimgt_mapi_portal_t * portal = NULL;
	int			 portalctr,portalno;
	int			 done = 0;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	if (compare(vp->name, vp->namelen, name, vp->namelen) != 0) {
		HEADER_SIMPLE_TABLE(10000);
	}

	if (*length < 12) {
		goto return_null_label;
	}

	bcopy((char *) vp->name, (char *) newname,
	      (int) vp->namelen * sizeof(oid));

	newname[11] = inst_index;

	if (*length >= 13) {
		portalno = name[12];
	} else {
		newname[12] = portalno = 1;
	}

	portal_list = iscsimgt_mapi_portal_get_list((zapi_error_info_t *)0);

	for (portalctr = 1; ((!done) && (portalctr <= portalno)); portalctr++) {

		portal = iscsi_snmp_get_nth_portal(portal_list, portalctr);

		if (!portal) {
			goto return_null_label;
		}

		if (portalctr < portalno) {
			continue;
		}
	
		newname[12] = (oid) portalctr;

		result = compare(name, *length, newname, 13);
		done = ((exact && result == 0) || (!exact && result < 0));
	
		if ((!done) && (!exact)) {
			portalno++;
			name[12] = 0;
		}

	}

	if (!done || (portal == NULL)) {
		goto return_null_label;
	}
	portalctr--;
	*length = 13;
	bcopy((char *) newname, (char *) name, (int)*length  * sizeof(oid));
	*write_method = 0;

	switch(vp->magic) {

	case ISCSITGTPORTALPORT:
		long_return = SA_PORT(&portal->sockaddr);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSITGTPORTALTAG: 
		long_return = portal->tpgtag;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	}

return_label:
	if (portal_list != (iscsimgt_mapi_portal_t *)0) {
		iscsimgt_mapi_helper_portal_list_free(portal_list);
		portal_list = (iscsimgt_mapi_portal_t *)0;
		portal = (iscsimgt_mapi_portal_t *)0;
	}
	*var_len = rval_len;
	return(rval);

return_null_label:
	if (portal_list != (iscsimgt_mapi_portal_t *)0) {
		iscsimgt_mapi_helper_portal_list_free(portal_list);
		portal_list = (iscsimgt_mapi_portal_t *)0;
		portal = (iscsimgt_mapi_portal_t *)0;
	}
	return(NULL);
}

/*
 * var_iscsiNodeAttributesEntry()
 *
 * Invoked by the snmp agent to get the OID values of the
 * MIB objects in the Node Attributes Table.
 * 9999(iscsiModule).1(iscsiObjects).6(iscsiNode).
 * 1(iscsiNodeAttributesTable).1(iscsiNodeAttributesEntry)
 * 9999.1.6.1.1.{1-16}.1
 */
/* format is .1.3.6.1.3.9999.1.6.1.1.magic.index */
uint8_t *
var_iscsiNodeAttributesEntry(SNMP_FM_ARGS)
{
	uint8_t *       rval;
	int            rval_len;
	char *         targetname;
	char *         targetalias;
	bool      targetalias_assigned;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	/*
	 * We support only one instance (index always equals 1)
	 */
	HEADER_SIMPLE_TABLE(1);

	switch(vp->magic) {

	case ISCSINODEINDEX:	
		long_return = 1;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODENAME:
		targetname = iscsimgt_mapi_targetname_get(
					(zapi_error_info_t *)0);
		if (targetname != (char *)0) {
			sprintf((char*)return_buf, targetname);
			iscsimgt_mapi_helper_targetname_free(targetname);
		} else {
			sprintf((char*)return_buf, "none");
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSINODEALIAS:
		if ((iscsimgt_mapi_targetalias_get(
				&targetalias_assigned,
				&targetalias, (zapi_error_info_t *)0)) &&
		    (targetalias_assigned)) {
			sprintf((char*)return_buf, targetalias);
			iscsimgt_mapi_helper_targetalias_free(targetalias);
		} else {
			sprintf((char*)return_buf, "none");
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSINODEROLES:
		sprintf((char*)return_buf, ISCSI_SNMP_TARGET_TYPE_PORTAL);
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSINODETRANSPORTTYPE:	
		long_return = ISCSI_SNMP_NODE_TRANSPORT_TYPE;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEINITIALR2T:
		long_return = 
			Bool_to_TruthValue(ISCSIMGT_INITIALR2T_DEFAULT);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEIMMEDIATEDATA:
		long_return =
			Bool_to_TruthValue(ISCSIMGT_IMMEDIATE_DATA_DEFAULT);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEMAXOUTSTANDINGR2T:
		long_return = ISCSIMGT_MAX_OUTSTANDING_R2T_DEFAULT;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEFIRSTBURSTLENGTH:
		long_return = ISCSIMGT_FIRST_BURST_LENGTH_DEFAULT;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEMAXBURSTLENGTH:
		long_return = ISCSIMGT_MAX_BURST_LENGTH_DEFAULT;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEMAXCONNECTIONS:
		long_return = ISCSIMGT_MAX_CONNECTIONS_PER_SESSION_DEFAULT;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEDATASEQUENCEINORDER:
		long_return = 
			Bool_to_TruthValue(ISCSIMGT_DATA_SEQUENCE_IN_ORDER_DEFAULT);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEDATAPDUINORDER:
		long_return = 
			Bool_to_TruthValue(ISCSIMGT_DATA_PDU_IN_ORDER_DEFAULT);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEDEFAULTTIME2WAIT:	
		long_return = ISCSIMGT_DEFAULT_TIME_TO_WAIT_DEFAULT;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEDEFAULTTIME2RETAIN:
		long_return = ISCSIMGT_DEFAULT_TIME_TO_RETAIN_DEFAULT;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSINODEERRORRECOVERYLEVEL:
        	long_return = ISCSIMGT_ERROR_RECOVERY_LEVEL_DEFAULT;
        	rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

      
    default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;
	}

return_label:
	*var_len = rval_len;
	return(rval);
}


/*
 * var_iscsiTargetAttributesEntry()
 *
 * Invoked by the snmp agent to get the OID values of
 * MIB objects in the Target Attributes Table.
 * 9999(iscsiModule).1(iscsiObjects).7(iscsiTarget).
 * 1(iscsiTargetAttributesTable).1(iscsiTargetAttributesEntry)
 * 9999.1.7.1.1.{1-6}.1
 */
/* format is .1.3.6.1.3.9999.1.7.1.1.magic.index */
uint8_t *
var_iscsiTargetAttributesEntry(SNMP_FM_ARGS)
{
	uint8_t *                rval;
	int                     rval_len;
	iscsimgt_mapi_stats_t * stats = NULL;
	struct sockaddr_storage *sa;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	/*
	 * We support only one instance (index always equals 1)
	 */
	HEADER_SIMPLE_TABLE(1);

	/* Used by some of the cases below */
	stats = iscsimgt_mapi_stats_get((zapi_error_info_t *)0,
					iscsimgt_transport_type_all);

	switch(vp->magic) {

	case ISCSITGTLOGINFAILURES:
		long_return =
		    ((stats != (iscsimgt_mapi_stats_t *)0) ?
		     stats->sc_login_failed : 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSITGTLASTFAILURETIME:
		long_return =
		    ((stats != (iscsimgt_mapi_stats_t *)0) ?
		     (stats->last_failure_time.tv_sec * 100) : 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSITGTLASTFAILURETYPE: 
		if ((stats == (iscsimgt_mapi_stats_t *)0) ||
		    (stats->last_failure_time.tv_sec == 0)) {
			sprintf((char*)return_buf,
			        ISCSI_SNMP_LOGIN_NO_FAILURE_OID);
		} else if (stats->last_failure_var == ISCSI_SNMP_AUTH_FAILURE) {
			sprintf((char*)return_buf,
			        ISCSI_SNMP_LOGIN_AUTH_FAILURE_OID);
		} else {
			sprintf((char*)return_buf,
				ISCSI_SNMP_LOGIN_OTHER_FAILURE_OID);
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSITGTLASTINTRFAILURENAME:
		if ((stats != (iscsimgt_mapi_stats_t *)0) &&
		    (stats->last_failure_time.tv_sec != 0)) {
			sprintf((char*)return_buf,
			        stats->last_intr_failure_name);
		} else {
			return_buf[0] = (char)0; /* Avoid zero-len cc warn */
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	/*
	 * We support ipv4 and ipv6
	 */
	case ISCSITGTLASTINTRFAILUREADDRTYPE:
		if (stats == (iscsimgt_mapi_stats_t *)0) {
			long_return = 0;
		} else {
			sa = &stats->last_intr_failure_sockaddr;
			long_return = (SA_FAM(sa) == AF_INET6) ?
				      ISCSI_SNMP_INET_ADDRESS_IPv6 :
				      ISCSI_SNMP_INET_ADDRESS_IPv4;
		}
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSITGTLASTINTRFAILUREADDR:
		if (stats == (iscsimgt_mapi_stats_t *)0) {
			long_return = 0;
			rval_len = sizeof(long_return);
			rval = (uint8_t *)&long_return;
		} else {
			sa = &stats->last_intr_failure_sockaddr;
			if (SA_INITTED(sa)) {
				iscsimgt_util_inet_satop(SA(sa),
						&return_buf[0],
						MAX_IPADDRSTR_LEN);
			} else {
				return_buf[0] = 0;
			}

			if (!return_buf[0]) {
				sprintf((char*)return_buf, "none");
			}
			rval_len = strlen((const char *)return_buf);
			rval = (uint8_t *)return_buf;
		}
		goto return_label;

	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	}

return_label:
	if (stats != (iscsimgt_mapi_stats_t *)0) {
		iscsimgt_mapi_helper_stats_free(stats);
		stats = (iscsimgt_mapi_stats_t *)0;
	}
	*var_len = rval_len;
	return(rval);
}

/*
 * var_iscsiTargetLoginStatsEntry()
 *
 * Invoked by the snmp agent to get the OID values of
 * MIB objects in the Target Login Stats Table.
 * 9999(iscsiModule).1(iscsiObjects).7(iscsiTarget).
 * 2(iscsiTargetLoginStatsTable).1(iscsiTargetLoginStatsEntry)
 * 9999.1.7.2.1.{1-6}.1
 */
/* format is .1.3.6.1.3.9999.1.7.2.1.magic.index */
uint8_t *
var_iscsiTargetLoginStatsEntry(SNMP_FM_ARGS)
{
	uint8_t *                rval;
	int                     rval_len;
	iscsimgt_mapi_stats_t * stats = NULL;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	/*
	 * We support only one instance (index always equals 1)
	 */
	HEADER_SIMPLE_TABLE(1);

	/* Used by some of the cases below */
	stats = iscsimgt_mapi_stats_get((zapi_error_info_t *)0,
					iscsimgt_transport_type_all);

	switch(vp->magic) {

	case ISCSITGTLOGINACCEPTS:
		long_return =
		    ((stats != (iscsimgt_mapi_stats_t *)0) ?
		     stats->sc_login_accepts : 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSITGTLOGINOTHERFAILS:
		long_return =
		    ((stats != (iscsimgt_mapi_stats_t *)0) ?
		     stats->sc_login_other_failures : 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSITGTLOGINREDIRECTS:
		long_return =
		    ((stats != (iscsimgt_mapi_stats_t *)0) ?
		     stats->sc_login_redirects : 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSITGTLOGINAUTHORIZEFAILS:
		long_return =
		    ((stats != (iscsimgt_mapi_stats_t *)0) ?
		     stats->sc_login_authorization_errors : 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSITGTLOGINAUTHENTICATEFAILS:
		long_return =
		    ((stats != (iscsimgt_mapi_stats_t *)0) ?
		     stats->sc_login_authentication_errors : 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSITGTLOGINNEGOTIATEFAILS:
		long_return =
		    ((stats != (iscsimgt_mapi_stats_t *)0) ?
		     stats->sc_login_negotiation_fails : 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;
	}

return_label:
	if (stats != (iscsimgt_mapi_stats_t *)0) {
		iscsimgt_mapi_helper_stats_free(stats);
		stats = (iscsimgt_mapi_stats_t *)0;
	}
	*var_len = rval_len;
	return(rval);
}


/*
 * var_iscsiTargetLogoutStatsEntry()
 *
 * Invoked by the snmp agent to get the OID values of
 * MIB objects in the Target Logout Stats Table.
 * 9999(iscsiModule).1(iscsiObjects).7(iscsiTarget).
 * 3(iscsiTargetLogoutStatsTable).1(iscsiTargetLogoutStatsEntry)
 * 9999.1.7.3.1.{1-2}.1
 */
/* format is .1.3.6.1.3.9999.1.7.2.1.magic.index */
uint8_t *
var_iscsiTargetLogoutStatsEntry(SNMP_FM_ARGS)
{
	uint8_t *                rval;
	int                     rval_len;
	iscsimgt_mapi_stats_t * stats = NULL;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	/*
	 * We support only one instance (index always equals 1)
	 */
	HEADER_SIMPLE_TABLE(1);

	/* Used by some of the cases below */
	stats = iscsimgt_mapi_stats_get((zapi_error_info_t *)0,
					iscsimgt_transport_type_all);

	switch(vp->magic) {
    	case ISCSITGTLOGOUTNORMALS:	
		long_return =
		    ((stats != (iscsimgt_mapi_stats_t *)0) ?
		     stats->sc_logout_normals : 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSITGTLOGOUTOTHERS:
		long_return =
		    ((stats != (iscsimgt_mapi_stats_t *)0) ?
		     stats->sc_logout_others : 0);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	}

return_label:
	if (stats != (iscsimgt_mapi_stats_t *)0) {
		iscsimgt_mapi_helper_stats_free(stats);
		stats = (iscsimgt_mapi_stats_t *)0;
	}
	*var_len = rval_len;
	return(rval);
}

/*
 * var_iscsiSessionAttributesEntry()
 *
 * Invoked by the snmp agent to get the OID values of MIB
 * objects in the Session Attributes Table.
 * 9999(iscsiModule).1(iscsiObjects).11(iscsiSession).
 * 1(iscsiSessionAttributesTable).1(iscsiSessionAttributesEntry)
 * 9999.1.11.1.1.{1-20}.1
 */
/* format is .1.3.6.1.3.9999.1.11.1.1.magic.inst_index.portal_index */
uint8_t *
var_iscsiSessionAttributesEntry(SNMP_FM_ARGS)
{
	oid                       newname[SNMP_OID_MAX];
	int		          inst_index = 1;
	int		          node_index = 1;
	uint8_t *                  rval;
	int                       rval_len;
	int			  result;
	iscsimgt_mapi_session_t * session_list = NULL;
	iscsimgt_mapi_session_t * session = NULL;
	char *                    targetname;
	bool                      targetalias_assigned;
	char *                    targetalias;
	int 			  done=0;
	int 			  sessctr,sessionno;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	if (compare(vp->name, vp->namelen, name, vp->namelen) != 0) {
		HEADER_SIMPLE_TABLE(10000);
	}

	if (*length < 12) {
		goto return_null_label;
	}

	bcopy((char *) vp->name, (char *) newname, (int) vp->namelen *
					sizeof(oid));

	newname[11] = inst_index;
	newname[12] = node_index;


	if (*length >= 14) {
		sessionno = name[13];
	} else {
		newname[13] = sessionno = 1;
	}

	session_list = iscsimgt_mapi_session_get_list((zapi_error_info_t *)0);

	for (sessctr = 1; ((!done) && (sessctr <= sessionno)); sessctr++) {

		session = iscsi_snmp_get_nth_session(session_list, sessctr);

		if (!session) {
			goto return_null_label;
		}

		if (sessctr < sessionno) {
			continue;
		}

		newname[13] = (oid) sessctr;

		result = compare(name, *length, newname, 14);
		done = ((exact && result == 0) || (!exact && result < 0));

		if ((!done) && (!exact)) {
			sessionno++;
			name[13] = 0;
		}
	}

	if (!done || (session == NULL)) {
		goto return_null_label;
	}


	sessctr--;

	*length = 14;
	bcopy((char *) newname, (char *) name, (int) *length * sizeof(oid));
	*write_method = 0;


	switch(vp->magic) {

	case ISCSISSNNODEINDEX:
		long_return = node_index;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSISSNINDEX:
		long_return = sessctr;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	/*
	 * inboundSession(1) - session is established from an external
	 *			initiator to a target within this iSCSI
	 *			instance.
	 * OutboundSession(2) - session is established from an initiator
	 *			within this iscsi instance to an external
	 *			target.
	 * Ours is always inbound
	 */
    	case ISCSISSNDIRECTION:
		long_return = ISCSI_SNMP_INBOUND_SESSION;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;
		
    	case ISCSISSNINITIATORNAME:
		sprintf((char*)return_buf, session->initiator_nodename);
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

    	case ISCSISSNTARGETNAME:
		targetname = iscsimgt_mapi_targetname_get(
					(zapi_error_info_t *)0);
		if (targetname != (char *)0) {
			sprintf((char*)return_buf, targetname);
			iscsimgt_mapi_helper_targetname_free(targetname);
		} else {
			sprintf((char*)return_buf, "none");
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

    	case ISCSISSNTSIH:
		long_return = session->tsih;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSISSNISID: 
		sprintf((char*)return_buf,
			"%02x:%02x:%02x:%02x:%02x:%02x",
			session->isid[0], session->isid[1],
			session->isid[2], session->isid[3],
			session->isid[4], session->isid[5]);
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

    	case ISCSISSNINITIATORALIAS:
		sprintf((char*)return_buf, session->initiator_alias);
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

    	case ISCSISSNTARGETALIAS:
		if ((iscsimgt_mapi_targetalias_get(
				&targetalias_assigned,
				&targetalias, (zapi_error_info_t *)0)) &&
		    (targetalias_assigned)) {
			sprintf((char*)return_buf, targetalias);
			iscsimgt_mapi_helper_targetalias_free(targetalias);
		} else {
			sprintf((char*)return_buf, "none");
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

    	case ISCSISSNINITIALR2T:
		long_return = Bool_to_TruthValue(session->initial_r2t);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSISSNIMMEDIATEDATA:
		long_return = Bool_to_TruthValue(session->immediate_data);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSISSNTYPE:
		long_return =
		    ((session->session_var ==
		      Iscsimgt_Mapi_Session_Var_Discovery) ?
		     ISCSI_SNMP_SESSION_TYPE_DISCOVERY :
		     ISCSI_SNMP_SESSION_TYPE_NORMAL);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;
		
    	case ISCSISSNMAXOUTSTANDINGR2T:
		long_return = session->max_outstanding_r2t;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSISSNFIRSTBURSTLENGTH:
		long_return = session->first_burst_length;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSISSNMAXBURSTLENGTH:
		long_return = session->max_burst_length;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSISSNCONNECTIONNUMBER: 
		long_return = session->max_connections;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSISSNAUTHIDENTITY:
		sprintf((char*)return_buf, ISCSI_SNMP_IPS_AUTH_IDENTITY);
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

    	case ISCSISSNDATASEQUENCEINORDER:
		long_return = Bool_to_TruthValue(session->data_sequence_in_order);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSISSNDATAPDUINORDER:
		long_return = Bool_to_TruthValue(session->data_pdu_in_order);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

    	case ISCSISSNERRORRECOVERYLEVEL:
		long_return = session->error_recovery_level;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	}

return_label:
	if (session_list != (iscsimgt_mapi_session_t *)0) {
		iscsimgt_mapi_helper_session_list_free(session_list);
		session_list = (iscsimgt_mapi_session_t *)0;
		session = (iscsimgt_mapi_session_t *)0;
	}
	*var_len = rval_len;
	return(rval);


return_null_label:
	if (session_list != (iscsimgt_mapi_session_t *)0) {
		iscsimgt_mapi_helper_session_list_free(session_list);
		session_list = (iscsimgt_mapi_session_t *)0;
		session = (iscsimgt_mapi_session_t *)0;
	}
	return(NULL);
}

/*
 * var_iscsiSessionStatsEntry()
 *
 * Invoked by the snmp agent to get the OID values of
 * MIB objects in the Session Stats Table.
 */
uint8_t *
var_iscsiSessionStatsEntry(SNMP_FM_ARGS)
{

	oid                       newname[SNMP_OID_MAX];
	int		          inst_index = 1;
	int		          node_index = 1;
	uint8_t *                  rval;
	int			  result;
	int                       rval_len;
	iscsimgt_mapi_session_t * session_list = NULL;
	iscsimgt_mapi_session_t * session = NULL;
	int			  sessctr,sessionno;
	int			 done = 0;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	if (compare(vp->name, vp->namelen, name, vp->namelen) != 0) {
		HEADER_SIMPLE_TABLE(10000);
	}

	bcopy((char *) vp->name, (char *) newname, (int) vp->namelen *
					sizeof(oid));

	newname[11] = inst_index;
	newname[12] = node_index;


	if (*length >= 14) {
		sessionno = name[13];
	} else  {
		newname[13] = sessionno = 1;
	}

	session_list = iscsimgt_mapi_session_get_list((zapi_error_info_t *)0);

	for (sessctr = 1; ((!done) && (sessctr <= sessionno)); sessctr++) {

		session = iscsi_snmp_get_nth_session(session_list, sessctr);

		if (!session) {
			goto return_null_label;
		}

		if (sessctr < sessionno) {
			continue;
		}

		newname[13] = (oid) sessctr;

		result = compare(name, *length, newname, 14);
		done = ((exact && result == 0) || (!exact && result < 0));

		if ((!done) && (!exact)) {
			sessionno++;
			name[13] = 0;
		}
	}

	if (!done || (session == NULL)) {
		goto return_null_label;
	}

	sessctr--;

	*length = 14;
	bcopy((char *) newname, (char *) name, (int) *length * sizeof(oid));
	*write_method = 0;


	switch(vp->magic) {
	case ISCSISSNCMDPDUS:
		long_return = session->stats.sc_cmd_pdus_xfrd;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSISSNRSPPDUS:
		long_return = session->stats.sc_rsp_pdus_xfrd;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSISSNLCTXDATAOCTETS:
		long_return = LOW32(session->stats.sc_data_out_blocks);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSISSNLCRXDATAOCTETS:
		long_return = LOW32(session->stats.sc_data_in_blocks);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	}

return_label:
	if (session_list != (iscsimgt_mapi_session_t *)0) {
		iscsimgt_mapi_helper_session_list_free(session_list);
		session_list = (iscsimgt_mapi_session_t *)0;
		session = (iscsimgt_mapi_session_t *)0;
	}
	*var_len = rval_len;
	return(rval);

return_null_label:
	if (session_list != (iscsimgt_mapi_session_t *)0) {
		iscsimgt_mapi_helper_session_list_free(session_list);
		session_list = (iscsimgt_mapi_session_t *)0;
		session = (iscsimgt_mapi_session_t *)0;
	}
	return(NULL);
}

/*
 * var_iscsiSessionCxnErrorStatsEntry()
 *
 * Invoked by the snmp agent to get the OID values
 * of MIB objects in the Session Connection Error Stats Table.
 */
uint8_t *
var_iscsiSessionCxnErrorStatsEntry(SNMP_FM_ARGS)
{
	oid                       newname[SNMP_OID_MAX];
	int                       result;
	int		          inst_index = 1;
	int		          node_index = 1;
	uint8_t *                  rval;
	int                       rval_len;
	iscsimgt_mapi_session_t * session_list = NULL;
	iscsimgt_mapi_session_t * session = NULL;
	int			  sessctr,sessionno;
	int			  done = 0;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	if (compare(vp->name, vp->namelen, name, vp->namelen) != 0) {
		HEADER_SIMPLE_TABLE(10000);
	}

	if (*length < 12) {
		goto return_null_label;
	}

	bcopy((char *) vp->name, (char *) newname, (int) vp->namelen *
					sizeof(oid));

	newname[11] = inst_index;
	newname[12] = node_index;


	if (*length >= 14) {
		sessionno = name[13];
	} else {
		newname[13] = sessionno = 1;
	}

	session_list = iscsimgt_mapi_session_get_list((zapi_error_info_t *)0);

	for ( sessctr = 1; ((!done) && (sessctr <= sessionno)) ; sessctr++ ) {

		session = iscsi_snmp_get_nth_session(session_list, sessctr);

		if (!session) {
			goto return_null_label;
		}

		if (sessctr < sessionno) {
			continue;
		}

		newname[13] = (oid) sessctr;

		result = compare(name, *length, newname, 14);
		done = ((exact && result == 0) || (!exact && result < 0));

		if ((!done) && (!exact)) {
			sessionno++;
			name[13] = 0;
		}
	}

	if (!done) {
		goto return_null_label;
	}

	if (!session) {
		goto return_null_label;
	}

	*length = 14;
	bcopy((char *) newname, (char *) name, (int) *length * sizeof(oid));
	*write_method = 0;

	switch(vp->magic) {

	case ISCSISSNDIGESTERRORS:
		long_return = LOW32(session->stats.sc_hdr_digest_errors +
		                    session->stats.sc_data_digest_errors);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSISSNCXNTIMEOUTERRORS:
		long_return = LOW32(session->stats.sc_conn_timeout_errors);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	}	

return_label:
	if (session_list != (iscsimgt_mapi_session_t *)0) {
		iscsimgt_mapi_helper_session_list_free(session_list);
		session_list = (iscsimgt_mapi_session_t *)0;
		session = (iscsimgt_mapi_session_t *)0;
	}
	*var_len = rval_len;
	return(rval);

return_null_label:
	if (session_list != (iscsimgt_mapi_session_t *)0) {
		iscsimgt_mapi_helper_session_list_free(session_list);
		session_list = (iscsimgt_mapi_session_t *)0;
		session = (iscsimgt_mapi_session_t *)0;
	}
	return(NULL);
}

/*
 * var_iscsiConnectionAttributesEntry()
 *
 * Invoked by the snmp agent to get the OID values
 * of MIB objects in the Connection Attributes Table.
 */
uint8_t *
var_iscsiConnectionAttributesEntry(SNMP_FM_ARGS)
{
	oid                            newname[SNMP_OID_MAX];
	int		               inst_index = 1;
	int		               node_index = 1;
	uint8_t *                       rval;
	int                            rval_len;
	int		               result;
	iscsimgt_mapi_session_t *      session_list = NULL;
	iscsimgt_mapi_session_t *      session = NULL;
	iscsimgt_mapi_session_conn_t * conn = NULL;
	struct sockaddr_storage *      sa;
	int				sessctr,sessionno;
	int 				index=0;
	int				done=0;
	int				count;

	return_buf[0] = 0;

	if (!is_licensed_and_enabled(LICENSE_ISCSI)) {
		return NULL;
	}

	if (compare(vp->name, vp->namelen, name, vp->namelen) != 0) {
		HEADER_SIMPLE_TABLE(10000);
	}

	if (*length < 12) {
		goto return_null_label;
	}

	bcopy((char *) vp->name, (char *) newname, (int) vp->namelen *
					sizeof(oid));
	/*
	 * We support only one instance
	 * For !exact we need to return ONLY if we are the NEXT oid i.e.
	 * if our OID(newname) > name
	 */
	newname[11] = inst_index;
	newname[12] = node_index;

	if (*length >= 14) {
		sessionno = name[13];
	} else {
		newname[13] = sessionno = 1;
	}
	
	session_list = iscsimgt_mapi_session_get_list((zapi_error_info_t *)0);

	for ( sessctr = 1; ((!done) && (sessctr <= sessionno)) ; sessctr++ ) {

		session = iscsi_snmp_get_nth_session(session_list, sessctr);

		if (!session) {
			goto return_null_label;
		}

		if (sessctr < sessionno) {
			continue;
		}

		newname[13] = (oid) sessctr;

		count = iscsi_snmp_get_session_conn_count(session);

		for (index = 0; !done && index < count; ) {
			index++;
			newname[14] = (oid) index;
			result = compare(name, *length, newname, 15);
			done = ((exact && result == 0) || 
				(!exact && result < 0));
		}

		if ((!done) && (!exact)) {
			sessionno++;
			name[14] = 0;
		}
	}

	if (!done || (session == NULL)) {
		goto return_null_label;
	}

	if (index < 1) {
		goto return_null_label;
	}

	sessctr--; 

	conn = iscsi_snmp_get_nth_session_conn(session, index);

	if (conn == NULL) {
		goto return_null_label;
	}

	*length = 15;
	bcopy((char *) newname, (char *) name, (int) *length * sizeof(oid));
	*write_method = 0;

	switch(vp->magic) {
	case ISCSICXNINDEX:
		long_return = index;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNCID:
		long_return = conn->cid;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNSTATE:
		long_return =
		    iscsimgt_mapi_helper_util_conn_snmp_state_get(
			    session->ism_id, conn->conn_state);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNPROTOCOL:
		long_return = ISCSI_SNMP_TRANSPORT_PROTOCOL_TCP;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNLOCALADDRTYPE:
		sa = &conn->local_sockaddr;
		long_return = (SA_FAM(sa) == AF_INET6) ?
			      ISCSI_SNMP_INET_ADDRESS_IPv6 :
			      ISCSI_SNMP_INET_ADDRESS_IPv4;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNLOCALADDR:
		sa = &conn->local_sockaddr;
		iscsimgt_util_inet_satop(SA(sa),
					&return_buf[0],
					MAX_IPADDRSTR_LEN);
		if (!return_buf[0]) {
			sprintf((char*)return_buf, "none");
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSICXNLOCALPORT:
		long_return = SA_PORT(&conn->local_sockaddr);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNREMOTEADDRTYPE:
		sa = &conn->remote_sockaddr;
		long_return = (SA_FAM(sa) == AF_INET6) ?
			      ISCSI_SNMP_INET_ADDRESS_IPv6 :
			      ISCSI_SNMP_INET_ADDRESS_IPv4;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNREMOTEADDR:
		sa = &conn->remote_sockaddr;
		iscsimgt_util_inet_satop(SA(sa),
					&return_buf[0],
					MAX_IPADDRSTR_LEN);
		if (!return_buf[0]) {
			sprintf((char*)return_buf, "none");
		}
		rval_len = strlen((const char *)return_buf);
		rval = (uint8_t *)return_buf;
		goto return_label;

	case ISCSICXNREMOTEPORT:
		long_return = SA_PORT(&conn->remote_sockaddr);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNMAXRECVDATASEGLENGTH:
		long_return = ISCSIMGT_MAX_RECV_DSEG_LENGTH_DEFAULT;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNMAXXMITDATASEGLENGTH:
		long_return = conn->initiator_mrdsl;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNHEADERINTEGRITY:
		long_return = ((conn->hdr_digest_enabled) ?
		               ISCSI_SNMP_DIGEST_VALUE_CRC32C :
		               ISCSI_SNMP_DIGEST_VALUE_NO_DIGEST);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNDATAINTEGRITY:
		long_return = ((conn->data_digest_enabled) ?
		               ISCSI_SNMP_DIGEST_VALUE_CRC32C :
		               ISCSI_SNMP_DIGEST_VALUE_NO_DIGEST);
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	case ISCSICXNRECVMARKER:
	case ISCSICXNSENDMARKER:
		long_return = ISCSI_SNMP_MARKER_SUPPORT;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	default:
		long_return = 0;
		rval_len = sizeof(long_return);
		rval = (uint8_t *)&long_return;
		goto return_label;

	}

return_label:
	if (session_list != (iscsimgt_mapi_session_t *)0) {
		iscsimgt_mapi_helper_session_list_free(session_list);
		session_list = (iscsimgt_mapi_session_t *)0;
		session = (iscsimgt_mapi_session_t *)0;
	}
	*var_len = rval_len;
	return(rval);

return_null_label:
	if (session_list != (iscsimgt_mapi_session_t *)0) {
		iscsimgt_mapi_helper_session_list_free(session_list);
		session_list = (iscsimgt_mapi_session_t *)0;
		session = (iscsimgt_mapi_session_t *)0;
	}
	return(NULL);
}

/*
 * Internal functions
 */
/*
 * iscsi_snmp_get_portal_count()
 *
 * Return the total number of portals available for the current
 * vfiler/host.
 */
static uint32_t
iscsi_snmp_get_portal_count(void)
{
	iscsimgt_mapi_portal_t * portal_list;
	iscsimgt_mapi_portal_t * portal;
	int count;

	portal_list = iscsimgt_mapi_portal_get_list((zapi_error_info_t *)0);

	count = 0;
	for (portal = portal_list;
	     portal != (iscsimgt_mapi_portal_t *)0;
	     portal = portal->next) {
		count++;
	}

	if (portal_list != (iscsimgt_mapi_portal_t *)0) {
		iscsimgt_mapi_helper_portal_list_free(portal_list);
		portal_list = (iscsimgt_mapi_portal_t *)0;
	}

	return(count);
}


/*
 * iscsi_snmp_get_nth_portal()
 *
 * Return the Nth portal (1-based) from the specified linked list
 * of portals.  Or return NULL if the portal_num is out of range.
 */
static iscsimgt_mapi_portal_t *
iscsi_snmp_get_nth_portal(
	iscsimgt_mapi_portal_t *portal_list,
	int portal_num)
{
	iscsimgt_mapi_portal_t *portal;
	int i;

	if (portal_num <= 0) {
		return((iscsimgt_mapi_portal_t *)0);
	}

	portal = portal_list;
	for (i = 1; i < portal_num; i++) {
		if (portal == (iscsimgt_mapi_portal_t *)0) {
			break;
		}
		portal = portal->next;
	}

	return(portal);
}


/*
 * iscsi_snmp_get_session_count()
 *
 * Return the total number of sessions active on the current
 * vfiler/host.
 */
static uint32_t
iscsi_snmp_get_session_count(void)
{
	iscsimgt_mapi_session_t * session_list;
	iscsimgt_mapi_session_t * session;
	int count;

	session_list = iscsimgt_mapi_session_get_list((zapi_error_info_t *)0);

	count = 0;
	for (session = session_list;
	     session != (iscsimgt_mapi_session_t *)0;
	     session = session->next) {
		count++;
	}

	if (session_list != (iscsimgt_mapi_session_t *)0) {
		iscsimgt_mapi_helper_session_list_free(session_list);
		session_list = (iscsimgt_mapi_session_t *)0;
	}

	return(count);
}

/*
 * iscsi_snmp_get_nth_session()
 *
 * Return the Nth session (1-based) from the specified linked list
 * of sessions.  Or return NULL if the session_num is out of range.
 */
static iscsimgt_mapi_session_t *
iscsi_snmp_get_nth_session(
	iscsimgt_mapi_session_t *session_list,
	int session_num)
{
	iscsimgt_mapi_session_t *session;
	int i;

	if (session_num <= 0) {
		return((iscsimgt_mapi_session_t *)0);
	}

	session = session_list;
	for (i = 1; i < session_num; i++) {
		if (session == (iscsimgt_mapi_session_t *)0) {
			break;
		}
		session = session->next;
	}

	return(session);
}


/*
 * iscsi_snmp_get_session_count()
 *
 * Return the total number of connections active in the specified
 * session.
 */
static uint32_t
iscsi_snmp_get_session_conn_count(
	iscsimgt_mapi_session_t *session)
{
	uint32_t count;
	iscsimgt_mapi_session_conn_t *conn;

	count = 0;
	for (conn = session->conn_list;
	     conn != (iscsimgt_mapi_session_conn_t *)0;
	     conn = conn->next) {
		count++;
	}

	return(count);
}


/*
 * iscsi_snmp_get_nth_session_conn()
 *
 * Return the Nth connection (1-based) from the specified connection.
 * Or return NULL if the conn_num is out of range.
 */
static iscsimgt_mapi_session_conn_t *
iscsi_snmp_get_nth_session_conn(
	iscsimgt_mapi_session_t *session,
	int conn_num)
{
	iscsimgt_mapi_session_conn_t *conn;
	int i;

	if (conn_num <= 0) {
		return((iscsimgt_mapi_session_conn_t *)0);
	}

	conn = session->conn_list;
	for (i = 1; i < conn_num; i++) {
		if (conn == (iscsimgt_mapi_session_conn_t *)0) {
			break;
		}
		conn = conn->next;
	}
	return (conn);
}


static snmpv2_TruthValue_t
Bool_to_TruthValue(bool param)
{
        return param ? SnmpTrue : SnmpFalse;
}
