/*
 *
 * Copyright (c) 1992-2012 Network Appliance, Inc.
 * All rights reserved.
 */

#ifndef _NETAPP_VARS_H_
#define _NETAPP_VARS_H_

#include <nfsd/nfs.h>
struct raid_disk_iterator;
#include <sk/sk_stats.h>
#include "snmpv3/net-snmp/library/asn1.h"
#include "snmpv3/net-snmp/library/snmp_impl.h"
#include "snmpv3/net-snmp/agent/snmp_vars.h"
#include "snmpv3/net-snmp/library/snmp.h"
#include "snmpv3/agent/mibgroup/mibparser_defines.h"
#include "snmpv3/agent/mibgroup/mibparser_structs.h"

#define MIB_DIR		0x01
#define MIB_VAR 	0x02
#define MIB_END         0x04
#define MIB_INDEX       0x08

#define MIB_RWSHIFT     4
#define MIB_RO          (RONLY << MIB_RWSHIFT)
#define MIB_RW          (RWRITE<< MIB_RWSHIFT)
#define MIB_NO          0
#define MIBTYPE_MASK    0x15
#define MIBACL_MASK     (0x07 << MIB_RWSHIFT)
	
/*
 * These are unit magic numbers for each variable.
 */

#define  NFSvxTOTALCALLS      1
#define  NFSvxSTARTVAL        NFSvxTOTALCALLS
#define  NFSvxTOTALPERCENT    2
#define  NFSvxCACHEDCALLS     3
#define  NFSvxCACHEDPERCENT   4
#define  NFSRPC         1
#define  NFSNFS         2
#define  NFSV2              3
#define         PRODTYPE_EISA   1
#define         PRODTYPE_PCI    2
#define         PRODTYPE_SIM    1000
#define  NFSV2STARTOID           V2CNULLS
#define  NFSV2ENDOID            V2CSTATFSS
#define  NFSV3STARTOID           V3CNULLS
#define  NFSV3ENDOID            V3CCOMMITS
#define  NFSREPLYCACHE      5
#define  TNFSRWSTATS  	    6		
#define  TV2READSTATS	    1
#define  TV2WRITESTATS      2
#define  TV3READSTATS       3
#define  TV3WRITESTATS      4
#define         QUOTASTATEOFF           1
#define         QUOTASTATEON            2
#define         QUOTASTATEINIT          3
#define         QUOTASTATECORRUPT       4
#define             QRTYPEUSER          1
#define             QRTYPEGROUP         2
#define             QRTYPETREE          3
#define             QRTYPEUNK           4
#define             QRVTYPEUSER          1
#define             QRVTYPEGROUP         2
#define             QRVTYPETREE          3
#define             QRVTYPEUSERDEFAULT   4
#define             QRVTYPEGROUPDEFAULT  5
#define             QRVTYPEUNK           6
#define             QRVTYPETREEDEFAULT   7
#define             QRVIDVALID         1
#define             QRVSIDVALID        2
#define  PCLNFSV2START          PCLNFSV2NULLS
#define  PCLNFSV2PCSTART        PCLNFSV2NULLPERCENT
#define  PCLNFSV3END            PCLNFSV3COMMITS
#define  PCLNFSV3PCEND          PCLNFSV3COMMITPERCENT
#define  PCLNFSV2RDSZEND	PCLNFSV2READ128KCALLS
#define  PCLNFSV2WRSZEND	PCLNFSV2WRITE128KCALLS
#define  PCLNFSV3RDSZEND	PCLNFSV3READ128KCALLS
#define  PCLNFSV3WRSZEND	PCLNFSV3WRITE128KCALLS
#define  NFSV2              3
#define  NFSV3              4
#define RAIDSTACTIVE       1
#define RAIDSTRECON        2
#define RAIDSTPARRECON     3
#define RAIDSTPARVERIF     4
#define RAIDSTSCRUB        5
#define RAIDSTFAILED       6
#define RAIDSTADDSPARE     7
#define RAIDSTSPARE        8
#define RAIDSTPREFAILED    9
#define RAIDSTOFFLINE      10 

#define SPARESTATSPARE     1
#define SPARESTATADDSPARE  2
#define ENCLINITIALIZING   1
#define ENCLTRANSITIONING  2
#define ENCLACTIVE		   3
#define ENCLINACTIVE	   4
#define ENCLRECONFIGURING  5
#define ENCLNONEXISTENT    6



/*
 * These must match the values in the MIB
 */
#define MISCGLOBALSTATUS_OTHER 1
#define MISCGLOBALSTATUS_UNKNOWN 2
#define MISCGLOBALSTATUS_OK 3
#define MISCGLOBALSTATUS_NONCRITICAL 4
#define MISCGLOBALSTATUS_CRITICAL 5
#define MISCGLOBALSTATUS_NONRECOVERABLE 6


/*
 * The following macro switches to the appropriate domain to call the
 * functions declared above (and switches back on return). 
 */
#define DOMAIN_SWITCH_NEST(domain, func, args)				\
        {                                                               \
	        sk_domain_t old_domain = sk_domain_switch_begin(domain);	\
	        uint8_t *value = func args;				\
	        SK_DOMAIN_REFERENCE(func, SK_REASON_MPSAFE);	       	\
	        sk_domain_switch_end(old_domain);			\
	        return value;                                           \
	}
	

extern int compare(oid *name1, int len1, oid *name2, int len2);
extern void snmp_fillname(oid *name, size_t *namelen, oid fill, int len);
extern void registerCifsDelay(uint32_t secs);
extern int32_t Interface_Scan_Get_Count(int af);
void *ifnet_inaddr(struct ifnet *ifp, int host, int af);

/*
 *  These values must match those in the MIB!
 */
typedef enum {
	VolumeOk = 1,
	VolumeNearlyFull = 2,
	VolumeFull = 3
} snmp_volume_status_t;


/*
 * Fix to BURT 255752:
 * Implement a new registry variable to control log messages issued when
 * volumes become nearly full or full by crossing fullness threshold
 * values. By enabling the new option, we will report each volume once in
 * its own message for each transition from one fullness state to another.
 */
#define MONITOR_EACH_VOLUME_CHANGE	"options.thresholds.fsReportEachFullVolumeChange"
/*
 * Fix to BURT 326786:
 * Implement Nearly full and Full Volume threshold monitoring for aggregates.
 */ 
#define MONITOR_AGGREGATES_FULL		"options.thresholds.fsReportFullAggregates"
/* For callouts to monitor.c */
typedef struct fv_volumes fv_volumes;

extern void monitor_add_to_fv_volumes(fv_volumes *my_full_volumes,
				      char *path,
				      bool is_aggregate,
				      int percent_full_blocks,
				      int percent_full_inodes,
				      snmp_volume_status_t onestatus);

extern size_t monitor_enumerate_full_volumes(fv_volumes *fvi,
					     char *message,
					     size_t size_msg);

extern int snmp_calculate_global_health(int, 
	char *, int *, int *, char *, int,
	fv_volumes *full_volumes_info);

extern snmp_volume_status_t snmp_check_for_full_volumes(
	snmp_volume_status_t level, int host, 
	char **reportp, int *sizep, 
	int *blocksp, int *inodesp, int *numFullp, 
	int *numNearlyFullp,
	fv_volumes *full_volumes_info);

/*
 *  We leave room for finer granularity later.
 */
typedef enum {
	CpuOk = 0,
	CpuBusy
} snmp_cpu_status_t;

extern snmp_cpu_status_t snmp_check_cpu_status(sk_system_stats_t *, char **, int *);

extern uint8_t  *var_nfsvx( oid vpn, int index );
extern uint8_t  *var_tnfsvx( oid vpn, int index );
extern uint64_t nfs_pclops( pclnt_nfs_Stats *pcl_sp );

extern FindVarMethod	var_unused;
extern FindVarMethod	var_nfs;
extern FindVarMethod	var_tnfs;

/* for nc compatability */
void snmp_get_sys_object_id(oid ** oidp, int * lenp);

#define MATCH_FAILED        (-1)
#define HEADER_GENERIC 	\
if (header_generic(vp, name, length, exact, var_len, write_method) == \
    MATCH_FAILED) {                                                   \
		return NULL;                                          \
	}                                                             

#define HEADER_SIMPLE_TABLE(max) 	\
if (header_simple_table(vp, name, length, exact, var_len, write_method, max) \
    == MATCH_FAILED) {                                                       \
	     return NULL;                                                    \
	}                                                                    
#define NAME_INDEX (name[*length - 1])

#endif /* _NETAPP_VARS_H_ */





