/*
 *
 * Copyright (c) 1992-2010 NetApp, Inc.
 * All rights reserved.
 */
/*
 * snmpd.c
 */
/** @defgroup agent The Net-SNMP agent
 * The snmp agent responds to SNMP queries from management stations
 */
/* Portions of this file are subject to the following copyrights.  See
 * the Net-SNMP's COPYING file for more details and other copyrights
 * that may apply:
 */
/*
 * Copyright 1988, 1989 by Carnegie Mellon University
 * 
 * All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * *****************************************************************
 */
/*
 * Copyright  2003 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms specified in the COPYING file
 * distributed with the Net-SNMP package.
 */
#include <net-snmp/net-snmp-config.h>
#include <sk/sk_param.h>


#include <sk/sk.h>
#include <sk/sk_signal.h>

#if HAVE_IO_H
#include <io.h>
#endif
#include <sys/errno.h>
#if HAVE_STRING_H
#include <legacy/string.h>
#else
#include <strings.h>
#endif
#if HAVE_STDLIB_H
#include <legacy/stdlib.h>
#endif
#if HAVE_UNISTD_H
#endif
#include <sys/types.h>
#if HAVE_NETINET_IN_H
#include <net7g/netinet/sk_in.h>
#endif
#if HAVE_ARPA_INET_H
#include <net7g/arpa/sk_inet.h>
#endif
#if TIME_WITH_SYS_TIME
# ifdef WIN32
#  include <sys/timeb.h>
# else
#  include <sk/time.h>
# endif
# include <legacy/time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sk/time.h>
# else
#  include <legacy/time.h>
# endif
#endif
#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#if HAVE_SYS_SOCKET_H
#include <net7g/sys/sk_socket.h>
#elif HAVE_WINSOCK_H
#include <winsock.h>
#endif
#if HAVE_NET_IF_H
#include <net7g/net/sk_if.h>
#endif
#if HAVE_INET_MIB2_H
#include <inet/mib2.h>
#endif
#if HAVE_SYS_IOCTL_H
#include <net7g/sys/sk_sockio.h>
#endif
#if HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#if HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#if HAVE_PROCESS_H              /* Win32-getpid */
#include <process.h>
#endif
#if HAVE_LIMITS_H
#include <sys/limits.h>
#endif
#if HAVE_PWD_H
#include <legacy/pwd.h>
#endif
#if HAVE_GRP_H
#include <legacy/grp.h>
#endif

#ifndef PATH_MAX
# ifdef _POSIX_PATH_MAX
#  define PATH_MAX _POSIX_PATH_MAX
# else
#  define PATH_MAX 255
# endif
#endif

#if HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include <net-snmp/library/keytools.h> /* USM_LENGTH_P_MIN */

#include "m2m.h"
#include <net-snmp/agent/mib_module_config.h>

#include "snmpd.h"
#include <net-snmp/agent/mib_modules.h>

#include "mibgroup/util_funcs.h"

#include <net-snmp/agent/agent_trap.h>
#include <net-snmp/library/snmpv3.h>

#include <net-snmp/agent/table.h>
#include <net-snmp/agent/table_iterator.h>
#include "agent/mibgroup/mib_module_includes.h"
#include <lib/na_auth.h>
#include <wafl/proto_ext.h>
#include <net-snmp/library/snmpusm.h>
#include <net-snmp/library/snmp.h>

#include <sk/sk_common.h>
#include <legacy/time.h>
#include <legacy/sys/syslog.h>
#include <machine/stdarg.h>
#include <sk/sk_debug.h>
#include <sk/sk_domain.h>
#include <sk/ha_host.h>
#include <qos/qos_classify.h>
#include <mgr/boot.h>
#include <registry/reg.h>
#include <ontap/ems/snmp_agent_ems.h>
#include <ha/fmrsrctbl.h>

extern void     netsnmp_init_helpers(void);

/*
 * Globals.
 */

#define TIMETICK         500000L

int             reconfig = 0;
int             Facility = LOG_DAEMON;

const char     *app_name = "snmpd";

extern int      netsnmp_running;
extern char   **argvrestartp;
extern char    *argvrestart;
extern char    *argvrestartname;
static int	snmp_on = 0;

#define SNMPD_STKSZ	128*1024
#define SIG_SELECT_TIMEOUT	0x80000000	/* XXX from uipc_socket.c */
#define SNMP_INFINITY           0x7FFFFFFF      /* "Infinite" timeout */
#define SNMP_INTERRUPT          RSH_INTERRUPT
sk_Pid	snmpd_pid;

/*
 * The value of snmp_enabled defaults to FALSE, since SNMP isn't initially
 * running, but it will (by default) be enabled after /etc/rc is read.
 * I.e. the registry value options.service.snmp.enable controls whether SNMP
 * will be enabled, and the default value of that is "on".
 */
int             snmp_enabled[ONTAP_NHOSTS] = { false, false };

/*
 * The value of the access specifications.
 */
char snmp_access_spec[ONTAP_NHOSTS] [PAC_MAX_ACCESS_SPEC_LEN] = {"legacy", "legacy"};

static int	snmp_processing[ONTAP_NHOSTS] = { false, false };

int  snmp_init_flag[ONTAP_NHOSTS] = {0, 0};       /* referenced in snmp_agent.c */

extern int      netsnmp_running;

int             snmp_read_packet(int);
static void     SnmpTrapNodeDown(void);
static int      receive(void);
int             snmpd_main(int, char **);

extern void vacm_parse_simple(const char *token, char *confline); 

void
snmpd_prerc(int host)
{
	int i;

#define SNMP_SERVICE_KEY "options.service.snmp.enable"

	/*
	 * In a freshly installed system/simulator, the registry key
	 * that controls the SNMP service isn't present yet - normally
	 * it comes from /etc/registry.default, but that might not
	 * be present if someone hasn't done a "complete" install
	 * (especially common with simulators), so here we make
	 * sure it's present if it's not already.
	 */
	if ( ! registry_key_exists(SNMP_SERVICE_KEY) ) {
		/*
		 * Default value of the registry value is on.
		 * This may appear to be inconsistent with the
		 * initial value of snmp_enabled[host] below,
		 * but the registry value is used later on
		 * in registry_startstopservices() to actually
		 * turn on the service.
		 */
		registry_set_and_commit(rootCred,SNMP_SERVICE_KEY,"on",NULL);
		
		// system_contact value is now kept in registry
		bzero(system_name[host], sizeof(system_name[host]));
		bzero(system_location[host], sizeof(system_location[host]));
		for (i = 0; i < NUM_TRAPHOSTS; i++)
	            bzero(&snmptraphosts[host][i], sizeof(struct snmptraphost));

	}
}

void
snmp_enable_change(void)
{
	/*
	 * For SNMPv3, we never use this function.  This VERIFY
	 * is so that we discover when someone tries to use it.
	 */
	VERIFY(0);
	return;
}

/*
 * Called to put SNMPV3 in a particular mode - enabled or disabled.
 */
void
snmp_enable_change_to(int newvalue)
{
	int host = sk_gethost();

	if (snmp_enabled[host] == false && newvalue == true) {
		snmp_enable(host);
	} else if (snmp_enabled[host] == true && newvalue == false) {
		snmp_disable(host);
	}
}

void
snmp_give_useradmin_rouser()
{
	naa_user_t * user_headP;
	naa_ret_values_t ret;
	naa_user_t * userP;
	char outbuf[128];
	char *password;

	ret = naa_user_list(NULL, &user_headP);
	if (ret != NAA_SUCCESS) {
		DEBUGMSGTL(("snmp_useradmin", naa_ret_values_string(ret)));
		return;
	}
	for (userP = user_headP; userP != NULL; userP = userP->next) {
		/*
		 * If username contains no spaces,
		 * add the user to the SNMP database.
		 * Otherwise, warn that this username
		 * is not appropriate for SNMP.
		 */
		if (strstr(userP->name, " ") == NULL) {
			vacm_parse_simple("rouser", userP->name);
		} else {
			EMS_snmp_agent_msg_access_denied(userP->name,
			    naa_ret_values_string(
				    NAA_USERNAME_DISALLOWED_CHAR));
		}
	}
	
	naa_user_list_free(user_headP);

	password = wafl_read_passwd();
	if ( password && (strlen(password) >= USM_LENGTH_P_MIN) ) {

		sprintf(outbuf, "root MD5 %s DES", password);
		usm_parse_create_usmUser("createUser", outbuf);
		vacm_parse_simple("rouser", "root");
	} else {
		EMS_snmp_agent_msg_access_denied(NAA_USER_ROOT,
			naa_ret_values_string(NAA_PASSWORD_TOO_SHORT));
	}
}

void
snmpv3_user_init(void) {
	static int snmpv3_user_init_done = 0;

	if (snmpv3_user_init_done) {
		return;
	}

	/*
	 * Loading of user information from SNMPpersist.conf file depends on
	 * the initialization of this routine. If not already initialized ,
	 * initialize it here.
	 */ 
	init_snmp("snmpd");
	snmp_give_useradmin_rouser();
	snmp_store("snmpd");
	snmpv3_user_init_done = 1;
}

void 
shutdown_snmpd(void) {
	netsnmp_running = 0;
	sk_send_signals(snmpd_pid, SNMP_INTERRUPT);
}

static void
SnmpTrapNodeDown(void)
{
    send_easy_trap(SNMP_TRAP_ENTERPRISESPECIFIC, 2);
    /*
     * XXX  2 - Node Down #define it as NODE_DOWN_TRAP 
     */
}

SK_DOMAIN_EXPORT(is_snmp_enabled,SK_REASON_SWITCH_DOMAIN);
int 
is_snmp_enabled(void)
{
	return snmp_enabled[sk_gethost()];
}

/*
 * Enable SNMP on the specified host.
 */
int
snmp_enable(int host)
{
	int old_host = sk_sethost(host);
	static int firsttime = true;

	if ( snmp_enabled[host]) {
		sk_sethost(old_host);
		return true;         /* no change*/
	}
//ntap	reconfig = B_TRUE;

	if ( host == SK_HOST_PARTNER ) {
		init_partner_agent();
	} else if ( host == SK_HOST_LOCAL ) {
		if (init_master_agent() != 0) {
			sk_sethost(old_host);
			return false;
		}
	}
	snmp_enabled[host] = true;

	/*
	 * The very first time the local host turns on SNMP,
	 * we sent a coldstart trap.
	 */
	if (firsttime && snmp_enabled[SK_HOST_LOCAL]) {
		/*
		 * Send coldstart trap if possible.  
		 */
		send_easy_trap(0, 0);
		firsttime = false;
	}

	/* Wake up the SNMP daemon to inform it that the world changed.*/
	sk_send_signals(snmpd_pid, SIG_SELECT_TIMEOUT);

	sk_sethost(old_host);
	return true;
}

/*
 * Disable SNMP on the specified host.
 */
int
snmp_disable(int host)
{
	int old_host = sk_sethost(host);

	if ( !snmp_enabled[host]) {
		sk_sethost(old_host);
		return true;         /* no change*/
	}
	snmp_enabled[host] = false;
  
	netsnmp_deregister_agent_nsap(SnmpAgentHandle[host]);
	SnmpAgentHandle[host] = 0;

	/*
	 * IPv6 agent handler is stored at offset 'host+2', where
	 * 'host' is SK_HOST_LOCAL or SK_HOST_PARTNER.
	 */
	if (ipv6_enabled) {
		netsnmp_deregister_agent_nsap(SnmpAgentHandle[host+2]);
		SnmpAgentHandle[host+2] = 0;
	}

	sk_send_signals(snmpd_pid, SIG_SELECT_TIMEOUT);
	sk_sethost(old_host);

	return true;
}

/*******************************************************************-o-******
 * receive
 *
 * Parameters:
 *      
 * Returns:
 *	0	On success.
 *	-1	System error.
 *
 * Infinite while-loop which monitors incoming messges for the agent.
 * Invoke the established message handlers for incoming messages on a per
 * port basis.  Handle timeouts.
 */
static int
receive(void)
{
	int numfds;
	struct socket  *readfds[NUM_SOCKETS];
	struct sec_micro  timeout, *tvp = &timeout;
	struct sec_micro  sched, *svp = &sched, now, *nvp = &now;
	int             count, block;
	int sockmask = 0;
	int32_t sigmask;
	int old_host;
	int found;
	
	/*
	 * Set the 'sched'uled timeout to the current time + one TIMETICK.
	 */
	gettod(nvp);
	svp->tv_usec = nvp->tv_usec + TIMETICK;
	svp->tv_sec = nvp->tv_sec;
	
	while (svp->tv_usec >= ONE_SEC) {
		svp->tv_usec -= ONE_SEC;
		svp->tv_sec++;
	}
	
	/*
	 * ignore early sighup during startup
	 */
	reconfig = 0;
	
	/*
	 * Loop-forever: execute message handlers for sockets with data,
	 * reset the 'sched'uler.
	 */
	while (1 /*no process can exit*/) {
			
		if (reconfig) {
			reconfig = 0;
			snmp_log(LOG_INFO, "SNMP restarted");
			update_config();
			send_easy_trap(SNMP_TRAP_ENTERPRISESPECIFIC, 3);
		}
		
		tvp = &timeout;
		tvp->tv_sec = 0;
		tvp->tv_usec = TIMETICK;
		
		numfds = 0;
		block = 0;
		count = snmp_select_info(&numfds, readfds, tvp, &block);
		if (block == 1) {
			tvp = NULL;         /* block without timeout */
		}
		numfds = count;
		
		DEBUGMSGTL(("snmpd/select", 
			    "select( numfds=%d, ..., tvp=%p)\n",
			    numfds, tvp));
		/*
		 * Wait for an SNMP request on a socket.
		 */
		sockmask = 0;
		sigmask = so_select_wait(readfds, NUM_SOCKETS, &sockmask, 
				 tvp == NULL ? SNMP_INFINITY : tvp->tv_usec);
		
		DEBUGMSGTL(("snmpd/select", 
			    "returned, sigmask = %x, sockmask = %x\n", 
			    sigmask, sockmask));
		
		if (sigmask == -1 || sigmask & SIG_SELECT_TIMEOUT) {
			DEBUGMSG(("snmpd/select", 
				  "timed out. Continuing...\n", 
				  sigmask, sockmask));
			continue;
		} else if (sigmask & SNMP_INTERRUPT) {
			DEBUGMSG(("snmpd/select", 
				  "interrupted. Continuing or exiting...\n", 
				  sigmask, sockmask));
			continue;
		}

		/*
		 * Only read from those file descriptors
		 * for which data is available (as determined by
		 * the sockmask value that so_select_wait returned).
		 */
		found = 0;
		/*
		 * Check if read is waiting on Primary node first.
		 */
		if ( (sockmask & (1 << SNMP_PRI_IPV6_SOCK))
			|| (sockmask & (1 << SNMP_PRI_SOCK)) ) {

			old_host = sk_sethost(SK_HOST_LOCAL);

			/*
			 * If an ipv6 socket read waiting, then clear out the
			 * select sockmask bit.
			 */
			if (sockmask & (1 << SNMP_PRI_IPV6_SOCK)) {

				snmp_processing[SK_HOST_LOCAL] = true;
				snmp_read(readfds[SNMP_PRI_IPV6_SOCK]);
				snmp_processing[SK_HOST_LOCAL] = false;

				sockmask &= ~(1 << SNMP_PRI_IPV6_SOCK);

			}
			/*
			 * If primary ipv4 socket read waiting, then clear out the
			 * select sockmask bit.
			 */

			if (sockmask & (1 << SNMP_PRI_SOCK)) {

				snmp_processing[SK_HOST_LOCAL] = true;
				snmp_read(readfds[SNMP_PRI_SOCK]);
				snmp_processing[SK_HOST_LOCAL] = false;

				sockmask &= ~(1 << SNMP_PRI_SOCK);
			}
			found = 1;
			sk_sethost(old_host);
		}
		/*
		 * Check if read is waiting on Partner node.
		 */
		if ( (sockmask & (1 << SNMP_SEC_IPV6_SOCK))
			|| (sockmask & (1 << SNMP_SEC_SOCK)) ) {

			old_host = sk_sethost(SK_HOST_PARTNER);

			/*
			 * If an ipv6 socket read waiting, then clear out the
			 * select sockmask bit.
			 */
			if (sockmask & (1 << SNMP_SEC_IPV6_SOCK)) {

				snmp_processing[SK_HOST_PARTNER] = true;
				snmp_read(readfds[SNMP_SEC_IPV6_SOCK]);
				snmp_processing[SK_HOST_PARTNER] = false;

				sockmask &= ~(1 << SNMP_SEC_IPV6_SOCK);
			}
			/*
			 * If primary ipv4 socket read waiting, then clear out the
			 * select sockmask bit.
			 */

			if (sockmask & (1 << SNMP_SEC_SOCK)) {

				snmp_processing[SK_HOST_PARTNER] = true;
				snmp_read(readfds[SNMP_SEC_SOCK]);
				snmp_processing[SK_HOST_PARTNER ] = false;

				sockmask &= ~(1 << SNMP_SEC_SOCK);
			}
			found = 1;
			sk_sethost(old_host);
		}

		if (found == 0 || sockmask != 0) {
			snmp_log(LOG_ERR, "SNMP sockmask=%d not handled?",
				sockmask);
			continue;
		}
		
		/*
		 * If the time 'now' is greater than the 'sched'uled time, then:
		 *
		 *    Check alarm and event timers.
		 *    Reset the 'sched'uled time to current time + one TIMETICK.
		 *    Age the cache network addresses (from whom messges have
		 *        been received).
		 */
		gettod(nvp);
		
		if (nvp->tv_sec > svp->tv_sec || (nvp->tv_sec == svp->tv_sec &&
						  nvp->tv_usec > svp->tv_usec)) {
			svp->tv_usec = nvp->tv_usec + TIMETICK;
			svp->tv_sec = nvp->tv_sec;
			
			while (svp->tv_usec >= ONE_SEC) {
				svp->tv_usec -= ONE_SEC;
				svp->tv_sec++;
			}
		}
		netsnmp_check_outstanding_agent_requests();
	}                           /* endwhile */
	
	return 0;

}                               /* end receive() */

static void
snmpd(long dummy1, long dummy2)
{
	qos_thread_classify(QOS_CLASS_SNMPD);

	/*
	 * wait here for the system to come up
	 */
	wait_for_network();

	/*
	 * Initialize SNMP. This routine will initialize
	 * agent, and MIB modules also.
	 */
	init_snmp("snmpd");

	netsnmp_addrcache_initialise();
	
	/*
	 * Forever monitor the dest_port for incoming PDUs.  
	 */
	DEBUGMSGTL(("snmpd/main", "We're up.  Starting to process data.\n"));
	
	if (!netsnmp_ds_get_boolean(NETSNMP_DS_APPLICATION_ID, 
				    NETSNMP_DS_AGENT_QUIT_IMMEDIATELY))
		receive();
	DEBUGMSGTL(("snmpd/main", "sending shutdown trap\n"));
	SnmpTrapNodeDown();
	DEBUGMSGTL(("snmpd/main", "Bye...\n"));
	snmp_shutdown("snmpd");
}

void
snmpd_init(void *dummy __unused)
{
	if (snmp_on) {
		return;
	}
	snmp_on = 1;
	snmpd_pid = SK_CREATE(snmpd, SNMPD_STKSZ);
}

struct sec_micro  starttime;


/**
 * Initialize the agent.  Calls into init_agent_read_config to set tha app's
 * configuration file in the appropriate default storage space,
 *  NETSNMP_DS_LIB_APPTYPE.  Need to call init_agent before calling init_snmp.
 *
 * @param app the configuration file to be read in, gets stored in default
 *        storage
 *
 * @return Returns non-zero on failure and zero on success.
 *
 * @see init_snmp
 */
int
init_agent(const char *app)
{
    static  int init_agent_done = 0;

    int             r = 0;

    if (init_agent_done) {
	    return 0;
    }
    /*
     * get current time (ie, the time the agent started) 
     */
    gettod(&starttime);
    starttime.tv_sec--;
    starttime.tv_usec += 1000000L;

    /*
     * we handle alarm signals ourselves in the select loop 
     */
    netsnmp_ds_set_boolean(NETSNMP_DS_LIBRARY_ID, 
			   NETSNMP_DS_LIB_ALARM_DONT_USE_SIG, 1);

    setup_tree();

    init_agent_read_config(app);

#ifdef TESTING
    auto_nlist_print_tree(-2, 0);
#endif

    netsnmp_init_helpers();
    init_traps();
    netsnmp_container_init_list();

    /*
     * Register configuration tokens from transport modules.  
     */
#ifdef SNMP_TRANSPORT_UDP_DOMAIN
    netsnmp_udp_agent_config_tokens_register();
#endif
#ifdef SNMP_TRANSPORT_UDPIPV6_DOMAIN
    netsnmp_udp6_agent_config_tokens_register();
#endif
#ifdef SNMP_TRANSPORT_UNIX_DOMAIN
    netsnmp_unix_agent_config_tokens_register();
#endif

#ifdef NETSNMP_EMBEDDED_PERL
    init_perl();
#endif
    init_agent_done = 1;

    return r;
}                               /* end init_agent() */

static fm_module_result_t 
snmp_startSecondary(void)
{
	/*
	 * Reinitialize the "global" information for the partner.
	 * NOTE: this must be done *before* /etc/rc (or the registry)
	 * are done, otherwise it wipes out any changes they might make!
	 */
	snmpd_prerc(SK_HOST_PARTNER);

	/*
	 * SNMP will be enabled by registry_postrc_startSecondary,
	 * when registry-enabled services are started.
	 */

	return FM_MODULE_OK;
}

static fm_module_result_t 
snmp_stopSecondary(void)
{
	snmp_disable(SK_HOST_PARTNER);

	snmp_init_flag[SK_HOST_PARTNER] = 0;
	/*
	 * engineID for partner host should get generated next time.
	 * so reset this global variable.
	 */
	snmpv3_reset_engineID(SK_HOST_PARTNER);

	/*
	 * Clear out partners user list created in snmp database.
	 */
	clear_user_list(SK_HOST_PARTNER);

	return FM_MODULE_OK;
}
FSM_ENTRY(
        FSM_snmp,

                "snmp",
                snmp_startSecondary,
                NULL,
                0,      /* veto */
                NULL,   /* No stats to dump */
                0,      /* No flags */
                0       /* No index needed */
        );

FSM_ENTRY(
        FSM_snmp_giveback,

                "snmp_giveback",
                NULL,
                snmp_stopSecondary,
                0,      /* veto */
                NULL,   /* No stats to dump */
                0,      /* No flags */
                0       /* No index needed */
        );

