/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */
/*
 * Definitions for SNMP (RFC 1067) agent variable finder.
 *
 */

#ifndef _SNMP_VARS_H_
#define _SNMP_VARS_H_

#include "net-snmp/library/data_list.h"
#include "net-snmp/library/snmp_api.h"
#include "net-snmp/agent/snmp_agent.h"

#ifdef __cplusplus
extern          "C" {
#endif

/* Portions of this file are subject to the following copyright(s).  See
 * the Net-SNMP's COPYING file for more details and other copyrights
 * that may apply:
 */
/***********************************************************
	Copyright 1988, 1989 by Carnegie Mellon University
	Copyright 1989	TGV, Incorporated

		      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU and TGV not be used
in advertising or publicity pertaining to distribution of the software
without specific, written prior permission.

CMU AND TGV DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL CMU OR TGV BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
******************************************************************/
/*
 * Portions of this file are copyrighted by:
 * Copyright  2003 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms specified in the COPYING file
 * distributed with the Net-SNMP package.
 */

#define HIGH32(x)	((uint32_t)((x)>>32))
#define LOW32(x)	((uint32_t)(x))

    struct variable;

    /*
     * Function pointer called by the master agent for writes. 
     */
    typedef int     (WriteMethod) (int action,
                                   uint8_t * var_val,
                                   uint8_t var_val_type,
                                   size_t var_val_len,
                                   uint8_t * statP,
                                   oid * name, size_t length);

    /*
     * Function pointer called by the master agent for mib information retrieval 
     */
    typedef uint8_t *(FindVarMethod) (struct variable * vp,
                                     oid * name,
                                     size_t * length,
                                     int exact,
                                     size_t * var_len,
                                     WriteMethod ** write_method);

	/*
	 * Dclare FindVarMethod function arguments.  All snmp 
	 * message processing functions take the same arguments:
	 *
	struct variable *vp,	   IN     - pointer to variable entry	
	                                    that points here 
	oid            *name,	   IN/OUT - input name requested, 
				            output name found 
	size_t         *length,	   IN/OUT - length of input and output oids
	int             exact,	   IN     - TRUE if exact match was requested.
	size_t         *var_len,   OUT    - length of variable or 0 if function
				            returned.
	WriteMethod **write_method OUT    - function pointer to write method
	                                    (Always NULL in OnTAP because no 
					    SET support is included).
	 */

#define SNMP_FM_ARGS               \
	struct variable *vp,	   \
	oid            *name,	   \
	size_t         *length,	   \
	int             exact,	   \
	size_t         *var_len,   \
	WriteMethod **write_method 

#define SNMP_FM_ARGS_FORWARD  vp, name, length, exact, var_len, write_method

    /*
     * Function pointer called by the master agent for setting up subagent requests 
     */
    typedef int     (AddVarMethod) (netsnmp_agent_session *asp,
                                    netsnmp_variable_list * vbp);

    struct nlist;

    extern int32_t     long_return;
    extern uint32_t    ipaddr_return;
    extern uint8_t     return_buf[];
    extern uint32_t    unsigned_return;
    extern uint64_t     big_return;

    extern oid                 nullOid[];
    extern int                 nullOidLen;
	extern struct counter64    counter64_return;

#define INST	0xFFFFFFFF      /* used to fill out the instance field of the variables table */
	
    struct variable {
        uint8_t          magic;  /* passed to function as a hint */
        char            type;   /* type of variable */
        /*
         * See important comment in snmp_vars.c relating to acl 
         */
        uint16_t         acl;    /* access control list for variable */
        FindVarMethod  *findVar;        /* function that finds variable */
        uint8_t          namelen;        /* length of above */
        oid             name[MAX_OID_LEN];      /* object identifier of variable */
    };

    int             init_agent(const char *);
    void            shutdown_agent(void);

    int             should_init(const char *module_name);
    void            add_to_init_list(char *module_list);

#ifdef USING_AGENTX_SUBAGENT_MODULE
    void            netsnmp_enable_subagent(void);
#endif

extern FindVarMethod	var_unused;

#ifndef _AGENT_REGISTRY_H
#include <net-snmp/agent/agent_handler.h>
#include <net-snmp/agent/var_struct.h>
#include <net-snmp/agent/agent_registry.h>
#endif

    /*
     * fail overloads non-negative integer value. it must be -1 ! 
     */
#define MATCH_FAILED	(-1)
#define MATCH_SUCCEEDED	0

#ifdef __cplusplus
}
#endif
#endif                          /* _SNMP_VARS_H_ */
