/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */
/*
 * Definitions for the variables as defined in the MIB
 *
 */
/***********************************************************
	Copyright 1988, 1989 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef _snmp_mib_h_
#define _snmp_mib_h_

    int             read_objid(const char *, oid *, size_t *);

struct netsnmp_variable_list;

struct mib_system {
	char            sysDescr[32];	/* textual description */
	uint8_t          sysObjectID[16];	/* OBJECT IDENTIFIER of
						 * system */
	uint8_t          ObjIDLen;	/* length of sysObjectID */
	uint32_t          sysUpTime;	/* Uptime in 100/s of a second */
};

struct mib_interface {
	int32_t            ifNumber;	/* number of interfaces */
};

struct mib_ifEntry {
	int32_t            ifIndex;/* index of this interface	 */
	char            ifDescr[32];	/* english description of interface	 */
	int32_t            ifType;	/* network type of device	 */
	int32_t            ifMtu;	/* size of largest packet in bytes	 */
	uint32_t          ifSpeed;/* bandwidth in bits/sec	 */
	uint8_t          ifPhysAddress[11];	/* interface's address */
	uint8_t          PhysAddrLen;	/* length of physAddr */
	int32_t            ifAdminStatus;	/* desired state of interface */
	int32_t            ifOperStatus;	/* current operational status */
	uint32_t          ifLastChange;	/* value of sysUpTime when current
					 * state entered */
	uint32_t          ifInOctets;	/* number of octets received on
					 * interface */
	uint32_t          ifInUcastPkts;	/* number of unicast packets
					 * delivered */
	uint32_t          ifInNUcastPkts;	/* number of broadcasts or multicasts */
	uint32_t          ifInDiscards;	/* number of packets discarded with
					 * no error */
	uint32_t          ifInErrors;	/* number of packets containing
					 * errors */
	uint32_t          ifInUnknownProtos;	/* number of packets with
						 * unknown protocol */
	uint32_t          ifOutOctets;	/* number of octets transmitted */
	uint32_t          ifOutUcastPkts;	/* number of unicast packets sent */
	uint32_t          ifOutNUcastPkts;	/* number of broadcast or
						 * multicast pkts */
	uint32_t          ifOutDiscards;	/* number of packets discarded with
					 * no error */
	uint32_t          ifOutErrors;	/* number of pkts discarded with an
					 * error */
	uint32_t          ifOutQLen;	/* number of packets in output queue */
};

struct mib_atEntry {
	int32_t            atIfIndex;	/* interface on which this entry maps */
	uint8_t          atPhysAddress[11];	/* physical address of
						 * destination */
	uint8_t          PhysAddressLen;	/* length of atPhysAddress */
	uint32_t          atNetAddress;	/* IP address of physical address */
};

struct mib_ip {
	int32_t            ipForwarding;	/* 1 if gateway, 2 if host */
	int32_t            ipDefaultTTL;	/* default TTL for pkts originating
					 * here */
	uint32_t          ipInReceives;	/* no. of IP packets received from
					 * interfaces */
	uint32_t          ipInHdrErrors;	/* number of pkts discarded due to
					 * header errors */
	uint32_t          ipInAddrErrors;	/* no. of pkts discarded due to bad
					 * address */
	uint32_t          ipForwDatagrams;	/* number pf pkts forwarded
						 * through this entity */
	uint32_t          ipInUnknownProtos;	/* no. of local-addressed
						 * pkts w/unknown proto */
	uint32_t          ipInDiscards;	/* number of error-free packets
					 * discarded */
	uint32_t          ipInDelivers;	/* number of datagrams delivered to
					 * upper level */
	uint32_t          ipOutRequests;	/* number of IP datagrams originating
					 * locally */
	uint32_t          ipOutDiscards;	/* number of error-free output IP
					 * pkts discarded */
	uint32_t          ipOutNoRoutes;	/* number of IP pkts discarded due to
					 * no route */
	int32_t            ipReasmTimeout;	/* seconds fragment is held awaiting
					 * reassembly */
	uint32_t          ipReasmReqds;	/* no. of fragments needing
					 * reassembly (here) */
	uint32_t          ipReasmOKs;	/* number of fragments reassembled */
	uint32_t          ipReasmFails;	/* number of failures in IP
					 * reassembly */
	uint32_t          ipFragOKs;	/* number of datagrams fragmented
					 * here */
	uint32_t          ipFragFails;	/* no. pkts unable to be fragmented
					 * here */
	uint32_t          ipFragCreates;	/* number of IP fragments created
					 * here */
};

struct mib_ipAddrEntry {
	uint32_t          ipAdEntAddr;	/* IP address of this entry */
	int32_t            ipAdEntIfIndex;	/* IF for this entry */
	uint32_t          ipAdEntNetMask;	/* subnet mask of this entry */
	int32_t            ipAdEntBcastAddr;	/* read the MIB for this one */
};

struct mib_ipRouteEntry {
	uint32_t          ipRouteDest;	/* destination IP addr for this route */
	int32_t            ipRouteIfIndex;	/* index of local IF for this route */
	int32_t            ipRouteMetric1;	/* Primary routing metric */
	int32_t            ipRouteMetric2;	/* Alternate routing metric */
	int32_t            ipRouteMetric3;	/* Alternate routing metric */
	int32_t            ipRouteMetric4;	/* Alternate routing metric */
	uint32_t          ipRouteNextHop;	/* IP addr of next hop */
	int32_t            ipRouteType;	/* Type of this route */
	int32_t            ipRouteProto;	/* How this route was learned */
	int32_t            ipRouteAge;	/* No. of seconds since updating this
					 * route */
};

struct mib_icmp {
	uint32_t          icmpInMsgs;	/* Total of ICMP msgs received */
	uint32_t          icmpInErrors;	/* Total of ICMP msgs received with
					 * errors */
	uint32_t          icmpInDestUnreachs;
	uint32_t          icmpInTimeExcds;
	uint32_t          icmpInParmProbs;
	uint32_t          icmpInSrcQuenchs;
	uint32_t          icmpInRedirects;
	uint32_t          icmpInEchos;
	uint32_t          icmpInEchoReps;
	uint32_t          icmpInTimestamps;
	uint32_t          icmpInTimestampReps;
	uint32_t          icmpInAddrMasks;
	uint32_t          icmpInAddrMaskReps;
	uint32_t          icmpOutMsgs;
	uint32_t          icmpOutErrors;
	uint32_t          icmpOutDestUnreachs;
	uint32_t          icmpOutTimeExcds;
	uint32_t          icmpOutParmProbs;
	uint32_t          icmpOutSrcQuenchs;
	uint32_t          icmpOutRedirects;
	uint32_t          icmpOutEchos;
	uint32_t          icmpOutEchoReps;
	uint32_t          icmpOutTimestamps;
	uint32_t          icmpOutTimestampReps;
	uint32_t          icmpOutAddrMasks;
	uint32_t          icmpOutAddrMaskReps;
};

struct mib_tcp {
	int32_t            tcpRtoAlgorithm;	/* retransmission timeout
						 * algorithm */
	int32_t            tcpRtoMin;	/* minimum retransmission timeout
					 * (mS) */
	int32_t            tcpRtoMax;	/* maximum retransmission timeout
					 * (mS) */
	int32_t            tcpMaxConn;	/* maximum tcp connections possible */
	uint32_t          tcpActiveOpens;	/* number of SYN-SENT -> CLOSED
					 * transitions */
	uint32_t          tcpPassiveOpens;	/* number of SYN-RCVD ->
						 * LISTEN transitions */
	uint32_t          tcpAttemptFails;	/* (SYN-SENT,SYN-RCVD)->CLOSED
						 *  or SYN-RCVD->LISTEN */
	uint32_t          tcpEstabResets;	/* (ESTABLISHED,CLOSE-WAIT) -> CLOSED */
	uint32_t          tcpCurrEstab;	/* number in ESTABLISHED or
					 * CLOSE-WAIT state */
	uint32_t          tcpInSegs;	/* number of segments received */
	uint32_t          tcpOutSegs;	/* number of segments sent */
	uint32_t          tcpRetransSegs;	/* number of retransmitted segments */
};

struct mib_tcpConnEntry {
	int32_t            tcpConnState;	/* State of this connection */
	uint32_t          tcpConnLocalAddress;	/* local IP address for this
						 * connection */
	int32_t            tcpConnLocalPort;	/* local port for this
						 * connection */
	uint32_t          tcpConnRemAddress;	/* remote IP address for this
						 * connection */
	int32_t            tcpConnRemPort;	/* remote port for this connection */
};

struct mib_udp {
	uint32_t          udpInDatagrams;	/* No. of UDP datagrams delivered to
					 * users */
	uint32_t          udpNoPorts;	/* No. of UDP datagrams to port with
					 * no listener */
	uint32_t          udpInErrors;	/* No. of UDP datagrams unable to be
					 * delivered */
	uint32_t          udpOutDatagrams;	/* No. of UDP datagrams sent
						 * from this entity */
};

struct mib_egp {
	uint32_t          egpInMsgs;	/* No. of EGP msgs received without
					 * error */
	uint32_t          egpInErrors;	/* No. of EGP msgs received with
					 * error */
	uint32_t          egpOutMsgs;	/* No. of EGP msgs sent */
	uint32_t          egpOutErrors;	/* No. of (outgoing) EGP msgs dropped
					 * due to error */
};

struct mib_egpNeighEntry {
	int32_t            egpNeighState;	/* local EGP state with this entry's
					 * neighbor */
	uint32_t          egpNeighAddr;	/* IP address of this entry's
					 * neighbor */
};

/* root of the network appliance enterprise */
#define NETAPP_ENTERPRISE 1, 3, 6, 1, 4, 1, 789
#define NETAPP_ENTERPRISE_LEN  7

#define MIB_IFTYPE_OTHER		    1
#define MIB_IFTYPE_REGULAR1822		    2
#define MIB_IFTYPE_HDH1822		    3
#define MIB_IFTYPE_DDNX25		    4
#define MIB_IFTYPE_RFC877X25		    5
#define MIB_IFTYPE_ETHERNETCSMACD	    6
#define MIB_IFTYPE_ISO88023CSMACD	    7
#define MIB_IFTYPE_ISO88024TOKENBUS	    8
#define MIB_IFTYPE_ISO88025TOKENRING	    9
#define MIB_IFTYPE_ISO88026MAN		    10
#define MIB_IFTYPE_STARLAN		    11
#define MIB_IFTYPE_PROTEON10MBIT	    12
#define MIB_IFTYPE_PROTEON80MBIT	    13
#define MIB_IFTYPE_HYPERCHANNEL		    14
#define MIB_IFTYPE_FDDI			    15
#define MIB_IFTYPE_LAPB			    16
#define MIB_IFTYPE_SDLC			    17
#define MIB_IFTYPE_T1CARRIER		    18
#define MIB_IFTYPE_CEPT			    19
#define MIB_IFTYPE_BASICISDN		    20
#define MIB_IFTYPE_PRIMARYISDN		    21
#define MIB_IFTYPE_PROPPOINTTOPOINTSERIAL   22

#define MIB_IFSTATUS_UP		1
#define MIB_IFSTATUS_DOWN	2
#define MIB_IFSTATUS_TESTING	3

#define MIB_FORWARD_GATEWAY	1
#define MIB_FORWARD_HOST	2

#define MIB_IPROUTETYPE_OTHER	1
#define MIB_IPROUTETYPE_INVALID	2
#define MIB_IPROUTETYPE_DIRECT	3
#define MIB_IPROUTETYPE_REMOTE	4

#define MIB_IPROUTEPROTO_OTHER	    1
#define MIB_IPROUTEPROTO_LOCAL	    2
#define MIB_IPROUTEPROTO_NETMGMT    3
#define MIB_IPROUTEPROTO_ICMP	    4
#define MIB_IPROUTEPROTO_EGP	    5
#define MIB_IPROUTEPROTO_GGP	    6
#define MIB_IPROUTEPROTO_HELLO	    7
#define MIB_IPROUTEPROTO_RIP	    8
#define MIB_IPROUTEPROTO_ISIS	    9
#define MIB_IPROUTEPROTO_ESIS	    10
#define MIB_IPROUTEPROTO_CISCOIGRP  11
#define MIB_IPROUTEPROTO_BBNSPFIGP  12
#define MIB_IPROUTEPROTO_OIGP	    13

#define MIB_TCPRTOALG_OTHER	1
#define MIB_TCPRTOALG_CONSTANT	2
#define MIB_TCPRTOALG_RSRE	3
#define MIB_TCPRTOALG_VANJ	4

#define MIB_TCPCONNSTATE_CLOSED		1
#define MIB_TCPCONNSTATE_LISTEN		2
#define MIB_TCPCONNSTATE_SYNSENT	3
#define MIB_TCPCONNSTATE_SYNRECEIVED	4
#define MIB_TCPCONNSTATE_ESTABLISHED	5
#define MIB_TCPCONNSTATE_FINWAIT1	6
#define MIB_TCPCONNSTATE_FINWAIT2	7
#define MIB_TCPCONNSTATE_CLOSEWAIT	8
#define MIB_TCPCONNSTATE_LASTACK	9
#define MIB_TCPCONNSTATE_CLOSING	10
#define MIB_TCPCONNSTATE_TIMEWAIT	11

#define MIB_EGPNEIGHSTATE_IDLE		1
#define MIB_EGPNEIGHSTATE_AQUISITION	2
#define MIB_EGPNEIGHSTATE_DOWN		3
#define MIB_EGPNEIGHSTATE_UP		4
#define MIB_EGPNEIGHSTATE_CEASE		5

#ifdef NET_SNMP_CONFIG_H
    struct variable_list;
    struct enum_list;

    void            print_mib(FILE * fp);
    void            print_ascii_dump(FILE *);
    int             read_objid(const char *, oid *, size_t *);
    void            register_mib_handlers(void);
    void            netsnmp_set_mib_directory(const char *dir);
    char            *netsnmp_get_mib_directory(void);
    void            netsnmp_fixup_mib_directory(void);
    void            init_mib(void);
    void            shutdown_mib(void);
    void            print_description(oid *, size_t, int);
    void            fprint_description(FILE *, oid *, size_t, int);
    int             snprint_description(char *, size_t, oid *, size_t, int);
    int             sprint_realloc_description(uint8_t **, size_t *, size_t *,
                                               int, oid *, size_t, int);
    int             get_module_node(const char *, const char *, oid *,
                                    size_t *);
    int             get_wild_node(const char *, oid *, size_t *);
    int             get_node(const char *, oid *, size_t *);
extern    oid            *snmp_parse_oid(const char *, oid *, size_t *);
    struct tree    *get_symbol(oid *, size_t, struct tree *, char *);
    struct tree    *get_tree(const oid *, size_t, struct tree *);
    struct tree    *get_tree_head(void);
    void            set_function(struct tree *);

extern    int             parse_one_oid_index(oid ** oidStart, size_t * oidLen,
					      netsnmp_variable_list * data,
					      int complete);
    int             parse_oid_indexes(oid * oidIndex, size_t oidLen,
                                      netsnmp_variable_list * data);
    int             build_oid_noalloc(oid * in, size_t in_len,
                                      size_t * out_len, oid * prefix,
                                      size_t prefix_len,
                                      netsnmp_variable_list * indexes);
    int             build_oid(oid ** out, size_t * out_len, oid * prefix,
                              size_t prefix_len,
                              netsnmp_variable_list * indexes);
    int             build_oid_segment(netsnmp_variable_list * var);


    void            print_variable(const oid * objid, size_t objidlen,
                                   const netsnmp_variable_list * variable);

    void            fprint_variable(FILE * fp,
                                    const oid * objid, size_t objidlen,
                                    const netsnmp_variable_list * variable);

    int             snprint_variable(char *buf, size_t buf_len,
                                     const oid * objid, size_t objidlen,
                                     const netsnmp_variable_list * variable);

    int             sprint_realloc_variable(uint8_t ** buf,
                                            size_t * buf_len,
                                            size_t * out_len,
                                            int allow_realloc,
                                            const oid * objid,
                                            size_t objidlen,
                                            const netsnmp_variable_list *
                                            variable);

#ifndef DISABLE_MIB_LOADING
    struct tree    *
#else
    void
#endif
                    netsnmp_sprint_realloc_objid_tree(uint8_t ** buf,
                                                      size_t * buf_len,
                                                      size_t * out_len,
                                                      int allow_realloc,
                                                      int *buf_overflow,
                                                      const oid * objid,
                                                      size_t objidlen);

    void
                    netsnmp_sprint_realloc_objid(uint8_t ** buf,
                                                 size_t * buf_len,
                                                 size_t * out_len,
                                                 int allow_realloc,
                                                 int *buf_overflow,
                                                 const oid * objid,
                                                 size_t objidlen);

    void            print_value(const oid * objid, size_t objidlen,
                                const netsnmp_variable_list * variable);

    void            fprint_value(FILE * fp,
                                 const oid * objid, size_t objidlen,
                                 const netsnmp_variable_list * variable);

    int             snprint_value(char *buf, size_t buf_len,
                                  const oid * objid, size_t objidlen,
                                  const netsnmp_variable_list * variable);

    int             sprint_realloc_value(uint8_t ** buf, size_t * buf_len,
                                         size_t * out_len,
                                         int allow_realloc,
                                         const oid * objid,
                                         size_t objidlen,
                                         const netsnmp_variable_list * variable);


    void            print_objid(const oid * objid, size_t objidlen);

    void            fprint_objid(FILE * fp,
                                 const oid * objid, size_t objidlen);

    int             snprint_objid(char *buf, size_t buf_len,
                                  const oid * objid, size_t objidlen);

    int             sprint_realloc_objid(uint8_t ** buf, size_t * buf_len,
                                         size_t * out_len,
                                         int allow_realloc,
                                         const oid * objid,
                                         size_t objidlen);


    int             sprint_realloc_by_type(uint8_t ** buf, size_t * buf_len,
                                           size_t * out_len,
                                           int allow_realloc,
                                           const netsnmp_variable_list * var,
                                           const struct enum_list *enums,
                                           const char *hint,
                                           const char *units);

    int             sprint_realloc_hexstring(uint8_t ** buf,
                                             size_t * buf_len,
                                             size_t * out_len,
                                             int allow_realloc,
                                             const uint8_t *, size_t);

    int             sprint_realloc_asciistring(uint8_t ** buf,
                                               size_t * buf_len,
                                               size_t * out_len,
                                               int allow_realloc,
                                               const uint8_t * cp,
                                               size_t len);

    int             sprint_realloc_octet_string(uint8_t ** buf,
                                                size_t * buf_len,
                                                size_t * out_len,
                                                int allow_realloc,
                                                const netsnmp_variable_list *,
                                                const struct enum_list *,
                                                const char *,
                                                const char *);

    int             sprint_realloc_opaque(uint8_t ** buf, size_t * buf_len,
                                          size_t * out_len,
                                          int allow_realloc,
                                          const netsnmp_variable_list *,
                                          const struct enum_list *, const char *,
                                          const char *);

    int             sprint_realloc_object_identifier(uint8_t ** buf,
                                                     size_t * buf_len,
                                                     size_t * out_len,
                                                     int allow_realloc,
                                                     const netsnmp_variable_list
                                                     *, const struct enum_list *,
                                                     const char *,
                                                     const char *);

    int             sprint_realloc_timeticks(uint8_t ** buf,
                                             size_t * buf_len,
                                             size_t * out_len,
                                             int allow_realloc,
                                             const netsnmp_variable_list *,
                                             const struct enum_list *,
                                             const char *, const char *);

    int             sprint_realloc_hinted_integer(uint8_t ** buf,
                                                  size_t * buf_len,
                                                  size_t * out_len,
                                                  int allow_realloc, long,
                                                  const char, const char *,
                                                  const char *);

    int             sprint_realloc_integer(uint8_t ** buf, size_t * buf_len,
                                           size_t * out_len,
                                           int allow_realloc,
                                           const netsnmp_variable_list *,
                                           const struct enum_list *,
                                           const char *, const char *);

    int             sprint_realloc_uinteger(uint8_t ** buf,
                                            size_t * buf_len,
                                            size_t * out_len,
                                            int allow_realloc,
                                            const netsnmp_variable_list *,
                                            const struct enum_list *,
                                            const char *, const char *);

    int             sprint_realloc_gauge(uint8_t ** buf, size_t * buf_len,
                                         size_t * out_len,
                                         int allow_realloc,
                                         const netsnmp_variable_list *,
                                         const struct enum_list *, const char *,
                                         const char *);

    int             sprint_realloc_counter(uint8_t ** buf, size_t * buf_len,
                                           size_t * out_len,
                                           int allow_realloc,
                                           const netsnmp_variable_list *,
                                           const struct enum_list *,
                                           const char *, const char *);

    int             sprint_realloc_networkaddress(uint8_t ** buf,
                                                  size_t * buf_len,
                                                  size_t * out_len,
                                                  int allow_realloc,
                                                  const netsnmp_variable_list *,
                                                  const struct enum_list *,
                                                  const char *,
                                                  const char *);

    int             sprint_realloc_ipaddress(uint8_t ** buf,
                                             size_t * buf_len,
                                             size_t * out_len,
                                             int allow_realloc,
                                             const netsnmp_variable_list *,
                                             const struct enum_list *,
                                             const char *, const char *);

    int             sprint_realloc_null(uint8_t ** buf, size_t * buf_len,
                                        size_t * out_len,
                                        int allow_realloc,
                                        const netsnmp_variable_list *,
                                        const struct enum_list *, const char *,
                                        const char *);

    int             sprint_realloc_bitstring(uint8_t ** buf,
                                             size_t * buf_len,
                                             size_t * out_len,
                                             int allow_realloc,
                                             const netsnmp_variable_list *,
                                             const struct enum_list *,
                                             const char *, const char *);

    int             sprint_realloc_nsapaddress(uint8_t ** buf,
                                               size_t * buf_len,
                                               size_t * out_len,
                                               int allow_realloc,
                                               const netsnmp_variable_list *,
                                               const struct enum_list *,
                                               const char *, const char *);

    int             sprint_realloc_counter64(uint8_t ** buf,
                                             size_t * buf_len,
                                             size_t * out_len,
                                             int allow_realloc,
                                             const netsnmp_variable_list *,
                                             const struct enum_list *,
                                             const char *, const char *);

    int             sprint_realloc_badtype(uint8_t ** buf, size_t * buf_len,
                                           size_t * out_len,
                                           int allow_realloc,
                                           const netsnmp_variable_list *,
                                           const struct enum_list *,
                                           const char *, const char *);

#ifdef OPAQUE_SPECIAL_TYPES
    int             sprint_realloc_float(uint8_t ** buf, size_t * buf_len,
                                         size_t * out_len,
                                         int allow_realloc,
                                         const netsnmp_variable_list *,
                                         const struct enum_list *, const char *,
                                         const char *);

    int             sprint_realloc_double(uint8_t ** buf, size_t * buf_len,
                                          size_t * out_len,
                                          int allow_realloc,
                                          const netsnmp_variable_list *,
                                          const struct enum_list *, const char *,
                                          const char *);
#endif


    int             snprint_by_type(char *buf, size_t buf_len,
                                    netsnmp_variable_list * var,
                                    const struct enum_list *enums,
                                    const char *hint, const char *units);

    int             snprint_hexstring(char *buf, size_t buf_len,
                                      const uint8_t *, size_t);

    int             snprint_asciistring(char *buf, size_t buf_len,
                                        const uint8_t * cp, size_t len);

    int             snprint_octet_string(char *buf, size_t buf_len,
                                         const netsnmp_variable_list *,
                                         const struct enum_list *, const char *,
                                         const char *);

    int             snprint_opaque(char *buf, size_t buf_len,
                                   const netsnmp_variable_list *,
                                   const struct enum_list *, const char *,
                                   const char *);

    int             snprint_object_identifier(char *buf, size_t buf_len,
                                              const netsnmp_variable_list *,
                                              const struct enum_list *,
                                              const char *, const char *);

    int             snprint_timeticks(char *buf, size_t buf_len,
                                      const netsnmp_variable_list *,
                                      const struct enum_list *, const char *,
                                      const char *);

    int             snprint_hinted_integer(char *buf, size_t buf_len,
                                           long, const char *,
                                           const char *);

    int             snprint_integer(char *buf, size_t buf_len,
                                    const netsnmp_variable_list *,
                                    const struct enum_list *, const char *,
                                    const char *);

    int             snprint_uinteger(char *buf, size_t buf_len,
                                     const netsnmp_variable_list *,
                                     const struct enum_list *, const char *,
                                     const char *);

    int             snprint_gauge(char *buf, size_t buf_len,
                                  const netsnmp_variable_list *,
                                  const struct enum_list *, const char *,
                                  const char *);

    int             snprint_counter(char *buf, size_t buf_len,
                                    const netsnmp_variable_list *,
                                    const struct enum_list *, const char *,
                                    const char *);

    int             snprint_networkaddress(char *buf, size_t buf_len,
                                           const netsnmp_variable_list *,
                                           const struct enum_list *,
                                           const char *, const char *);

    int             snprint_ipaddress(char *buf, size_t buf_len,
                                      const netsnmp_variable_list *,
                                      const struct enum_list *, const char *,
                                      const char *);

    int             snprint_null(char *buf, size_t buf_len,
                                 const netsnmp_variable_list *,
                                 const struct enum_list *, const char *,
                                 const char *);

    int             snprint_bitstring(char *buf, size_t buf_len,
                                      const netsnmp_variable_list *,
                                      const struct enum_list *, const char *,
                                      const char *);

    int             snprint_nsapaddress(char *buf, size_t buf_len,
                                        const netsnmp_variable_list *,
                                        const struct enum_list *, const char *,
                                        const char *);

    int             snprint_counter64(char *buf, size_t buf_len,
                                      const netsnmp_variable_list *,
                                      const struct enum_list *, const char *,
                                      const char *);

    int             snprint_badtype(char *buf, size_t buf_len,
                                    const netsnmp_variable_list *,
                                    const struct enum_list *, const char *,
                                    const char *);

#ifdef OPAQUE_SPECIAL_TYPES
    int             snprint_float(char *buf, size_t buf_len,
                                  const netsnmp_variable_list *,
                                  const struct enum_list *, const char *,
                                  const char *);

    int             snprint_double(char *buf, size_t buf_len,
                                   const netsnmp_variable_list *,
                                   const struct enum_list *, const char *,
                                   const char *);
#endif

    void            print_oid_report(FILE *);
    void            print_oid_report_enable_labeledoid(void);
    void            print_oid_report_enable_oid(void);
    void            print_oid_report_enable_suffix(void);
    void            print_oid_report_enable_symbolic(void);

    const char     *parse_octet_hint(const char *hint, const char *value,
	                             unsigned char **new_val, int *new_val_len);

    void            clear_tree_flags(register struct tree *tp);

    char           *snmp_out_toggle_options(char *);
    void            snmp_out_toggle_options_usage(const char *, FILE *);
    char           *snmp_in_toggle_options(char *);
    char           *snmp_in_options(char *, int, char * const *);
    void            snmp_in_toggle_options_usage(const char *, FILE *);
    uint8_t          mib_to_asn_type(int mib_type);


    int             netsnmp_str2oid(const char *S, oid * O, int L);


#define NETSNMP_STRING_OUTPUT_GUESS  1
#define NETSNMP_STRING_OUTPUT_ASCII  2
#define NETSNMP_STRING_OUTPUT_HEX    3

#define NETSNMP_OID_OUTPUT_SUFFIX  1
#define NETSNMP_OID_OUTPUT_MODULE  2
#define NETSNMP_OID_OUTPUT_FULL    3
#define NETSNMP_OID_OUTPUT_NUMERIC 4
#define NETSNMP_OID_OUTPUT_UCD     5
#define NETSNMP_OID_OUTPUT_NONE    6

#endif /* NET_SNMP_CONFIG_H */
#endif /* _snmp_mib_h_ */







