/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */
/*
 * scapi.h
 */

#ifndef _SCAPI_H
#define _SCAPI_H

#ifdef __cplusplus
extern          "C" {
#endif

    /*
     * Authentication/privacy transform bitlengths.
     */
#define SNMP_TRANS_AUTHLEN_HMACMD5	128
#define SNMP_TRANS_AUTHLEN_HMACSHA1	160

#define SNMP_TRANS_AUTHLEN_HMAC96	96

#define SNMP_TRANS_PRIVLEN_1DES		64
#define SNMP_TRANS_PRIVLEN_1DES_IV	64

#define SNMP_TRANS_PRIVLEN_AES128	128
#define SNMP_TRANS_PRIVLEN_AES128_IV	128
#define SNMP_TRANS_AES_AES128_PADSIZE   128

#define SNMP_TRANS_PRIVLEN_AES192	192
#define SNMP_TRANS_PRIVLEN_AES192_IV	192
#define SNMP_TRANS_AES_AES192_PADSIZE   128

#define SNMP_TRANS_PRIVLEN_AES256	256
#define SNMP_TRANS_PRIVLEN_AES256_IV	256
#define SNMP_TRANS_AES_AES256_PADSIZE   128

    /*
     * Prototypes.
     */
    int             sc_get_properlength(const oid * hashtype,
                                        uint32_t hashtype_len);

    int             sc_init(void);
    int             sc_shutdown(int majorID, int minorID, void *serverarg,
                                void *clientarg);

    int             sc_random(uint8_t * buf, size_t * buflen);

    int             sc_generate_keyed_hash(const oid * authtype,
                                           size_t authtypelen,
                                           uint8_t * key, uint32_t keylen,
                                           uint8_t * message, uint32_t msglen,
                                           uint8_t * MAC, size_t * maclen);

    int             sc_check_keyed_hash(const oid * authtype,
                                        size_t authtypelen, uint8_t * key,
                                        uint32_t keylen, uint8_t * message,
                                        uint32_t msglen, uint8_t * MAC,
                                        uint32_t maclen);

    int             sc_encrypt(const oid * privtype, size_t privtypelen,
                               uint8_t * key, uint32_t keylen,
                               uint8_t * iv, uint32_t ivlen,
                               uint8_t * plaintext, uint32_t ptlen,
                               uint8_t * ciphertext, size_t * ctlen);

    int             sc_decrypt(const oid * privtype, size_t privtypelen,
                               uint8_t * key, uint32_t keylen,
                               uint8_t * iv, uint32_t ivlen,
                               uint8_t * ciphertext, uint32_t ctlen,
                               uint8_t * plaintext, size_t * ptlen);

    int             sc_hash(const oid * hashtype, size_t hashtypelen,
                            uint8_t * buf, size_t buf_len,
                            uint8_t * MAC, size_t * MAC_len);

    int             sc_get_transform_type(oid * hashtype,
                                          uint32_t hashtype_len,
                                          int (**hash_fn) (const int mode,
                                                           void **context,
                                                           const uint8_t *
                                                           data,
                                                           const int
                                                           data_len,
                                                           uint8_t **
                                                           digest,
                                                           size_t *
                                                           digest_len));


    /*
     * All functions devolve to the following block if we can't do cryptography
     */
#define	_SCAPI_NOT_CONFIGURED					\
{								\
        snmp_log(LOG_ERR, "Encryption support not enabled.\n"); \
        DEBUGMSGTL(("scapi", "SCAPI not configured"));		\
	return SNMPERR_SC_NOT_CONFIGURED;			\
}

    /*
     * define a transform type if we're using the internal md5 support 
     */
#ifdef USE_INTERNAL_MD5
#define INTERNAL_MD5 1
#endif

#ifdef __cplusplus
}
#endif
#endif                          /* _SCAPI_H */
