/*
 *
 * Copyright (c) 1992-2010 NetApp, Inc.
 * All rights reserved.
 */

#ifndef _SNMP_IMPL_H_
#define _SNMP_IMPL_H_

#ifdef __cplusplus
extern          "C" {
#endif
    /*
     * * file: snmp_impl.h
     */

    /*
     * Definitions for SNMP implementation.
     *
     *
     */
/***********************************************************
	Copyright 1988, 1989 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <sk/sk_debug.h>
#include <sk/ha_host.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/netinet/sk_in.h>
#include <net7g/arpa/sk_inet.h>
#include <snmp/trapdefs.h>

/*
 * Error codes:
 */
/*
 * These must not clash with SNMP error codes (all positive).
 */
#define PARSE_ERROR	-1
#define BUILD_ERROR	-2

#define SID_MAX_LEN	64
#define SNMP_OID_MAX    128	/* number of subid's in a objid */

#define COMMUNITY_MAX_LEN	256

    /*
     * Space for character representation of an object identifier 
     */
#define SPRINT_MAX_LEN		2560

#include <sys/_null.h>

#ifndef true
#define true	1
#endif
#ifndef false
#define false	0
#endif

#define READ	    1
#define WRITE	    0

#define SNMP_IMPL_RESERVE1    0
#define SNMP_IMPL_RESERVE2    1
#define SNMP_IMPL_ACTION	    2
#define SNMP_IMPL_COMMIT      3
#define SNMP_IMPL_FREE        4
#define SNMP_IMPL_UNDO        5
#define SNMP_IMPL_FINISHED_SUCCESS        9
#define SNMP_IMPL_FINISHED_FAILURE	10

#define RONLY	0x1
#define RWRITE	0x3
#define NOACCESS 0x0

    /*
     * defined types (from the SMI, RFC 1157) 
     */
#define ASN_IPADDRESS   (ASN_APPLICATION | 0)
#define ASN_COUNTER	(ASN_APPLICATION | 1)
#define ASN_GAUGE	(ASN_APPLICATION | 2)
#define ASN_UNSIGNED    (ASN_APPLICATION | 2)   /* RFC 1902 - same as GAUGE */
#define ASN_TIMETICKS   (ASN_APPLICATION | 3)
#define ASN_OPAQUE	(ASN_APPLICATION | 4)   /* changed so no conflict with other includes */

    /*
     * defined types (from the SMI, RFC 1442) 
     */
#define ASN_NSAP	(ASN_APPLICATION | 5)   /* historic - don't use */
#define ASN_COUNTER64   (ASN_APPLICATION | 6)
#define ASN_UINTEGER    (ASN_APPLICATION | 7)   /* historic - don't use */

#define INTEGER		         ASN_INTEGER
#define STRING		         ASN_OCTET_STR
#define OCTET		         ASN_OCTET_STR
#define OBJID		         ASN_OBJECT_ID
#define NULLOBJ		         ASN_NULL
#define DISPLAYSTRING	         STRING 
#define PHYSADDRESS	         STRING 
#define NETWORKADDRESS	         IPADDRESS 
#define TRUTHVALUE               INTEGER
#define VARIABLEPOINTER          ASN_OBJECT_ID
#define ROWPOINTER               ASN_OBJECT_ID
#define IPV6ADDRESS              STRING
#define IPV6ADDRESSPREFIX        STRING
#define IPV6ADDRESSIFIDENTIFIER  STRING
#define IPV6IFINDEXNUM           INTEGER
#define IPV6IFINDEXORZERO        INTEGER 
#define COUNTER64                ASN_COUNTER64
#define DATEANDTIME              ASN_OCTET_STR
#define COUNTERBASEDGAUGE64      ASN_COUNTER64
#define TIMETICKS64		 ASN_COUNTER64

/*
 * iSCSI specific 
 */
#define INETADDRESS     	STRING
#define SNMPADMINSTRING 	STRING
#define ISCSINAME		STRING
#define AUTONOMOUSTYPE  	STRING
#define ROWSTATUS       	INTEGER
#define INETADDRESSTYPE 	INTEGER
#define ISCSITRANSPORTPROTOCOLS INTEGER
#define ISCSIDIGESTMETHOD       INTEGER
#define UNSIGNED32  		GAUGE
#define TIMESTAMP		(ASN_APPLICATION | 1)



#define MIB_DIR		0x01
#define MIB_VAR 	0x02
#define MIB_END         0x04
#define MIB_INDEX       0x08

#define MIB_RWSHIFT     4
#define MIB_RO          (RONLY << MIB_RWSHIFT)
#define MIB_RW          (RWRITE<< MIB_RWSHIFT)
#define MIB_NO          0
#define MIBTYPE_MASK    0x15
#define MIBACL_MASK     (0x07 << MIB_RWSHIFT)
	
/* defined types (from the SMI, RFC 1065) */
#define IPADDRESS   (ASN_APPLICATION | 0)
#define COUNTER	    (ASN_APPLICATION | 1)
#define GAUGE	    (ASN_APPLICATION | 2)
#define TIMETICKS   (ASN_APPLICATION | 3)
#define OPAQUE	    (ASN_APPLICATION | 4)

/* defined types from SNMPv2, RFC 1155 */
#define COUNTER32   (ASN_APPLICATION | 1)
#define GAUGE32     (ASN_APPLICATION | 2)
#define INTEGER32    ASN_INTEGER

#ifdef DEBUG
#define ERROR(string) D(snmp, 100, ("snmp: %s\n", string));
#else
#define ERROR(string)
#endif
#define ERROR_MSG(string) syslog(LOG_WARNING, string)

struct dispTab {
	uint8_t	mibType;
	uint32_t	lookupVal;
        uint32_t   type;
};

/* from snmp.c */
extern uint8_t   sid[];		/* size SID_MAX_LEN */

extern uint8_t         *snmp_parse_var_op(uint8_t *, oid *, size_t *, uint8_t *,
                                      size_t *, uint8_t **, uint8_t **, size_t *);
extern uint8_t         *snmp_build_var_op(uint8_t *, oid *, size_t *, uint8_t,
                                      size_t, uint8_t *, size_t *);

extern void shift_array(uint8_t *begin, int length, int shift_amount);

extern int snmp_build_trap(uint8_t *out_data, int *length, void *sysOid,
    int sysOidLen, uint32_t myAddr, int trapType, int specificType, uint32_t time,
    oid *varName, int varNameLen, uint8_t varType, int varLen, uint8_t *varVal);

extern uint8_t *snmp_auth_parse(uint8_t *data, size_t *length, uint8_t *sid, 
			       size_t *slen, int32_t *version);
extern uint8_t *snmp_auth_build(uint8_t *data, size_t *length, uint8_t *sid,
    int *slen, int32_t *version, int messagelen);

extern int snmp_agent_parse(uint8_t *data, int length, uint8_t *out_data,
    int *out_length, uint32_t sourceip);
extern void snmp_trap(int trapType, int specificType);

extern char* sprintoid(oid *name, int len);
extern char* sprintoidto(oid *name, int len, char *buff, int bufflen);
extern void printoid(oid *name, int len);
extern void snmp_dump_bytes(int level, uint8_t * data, int length);
extern boolean_t snmp_registry_key_to_oid(const char *key, oid *oidbuf,
	int *oidbuflen);

/*
 * Get this system's sysObjectID.  Set the *lenp to the length
 * in bytes of the OID.
 */
extern void snmp_get_sys_object_id(oid ** oidp, int * lenp);

#define NUM_COMMUNITIES	8
#define NUM_TRAPHOSTS	8
#define COMMUNITYSTRLEN	15
#define DEFAULT_NAMELEN	63
#define SYSTEM_NAMELEN	255

/*
 * snmptraphost struct contains an Alias, Common name, and an ip address.
 * When the a lookup is required, the traphost ip address == INADDR_ANY.
 * Use the alias to pick up the common name and ip address from dns.
 */
struct snmptraphost {
	char    aName[DEFAULT_NAMELEN + 1];
	char    cName[DEFAULT_NAMELEN + 1];
	struct in_addr addr;
	struct in6_addr ipv6_addr;
	bool bResolved;
};

struct	community {
	char name[COMMUNITYSTRLEN + 1];
	uint32_t	acl;
};
	/* system_descr is same for both local and partner */
extern char		system_descr[SYSTEM_NAMELEN + 1];
extern char		system_contact[ONTAP_NHOSTS][SYSTEM_NAMELEN + 1];
extern char		system_name[ONTAP_NHOSTS][SYSTEM_NAMELEN + 1];
extern char		system_location[ONTAP_NHOSTS][SYSTEM_NAMELEN + 1];
extern struct snmptraphost snmptraphosts[ONTAP_NHOSTS][NUM_TRAPHOSTS];
extern struct community	communities[ONTAP_NHOSTS][NUM_COMMUNITIES];
extern struct community	communitiesProto[NUM_COMMUNITIES];

/*
 * XXX - have to declare "struct variable" before declaring "find_method_t";
 * however, a full declaration of a "struct variable" includes, as a
 * member, a "find_method_t *".  This works around that annoying bit of
 * recursion.
 */
struct variable;

    /*
     * For calling secauth_build, FIRST_PASS is an indication that a new nonce
     * and lastTimeStamp should be recorded.  LAST_PASS is an indication that
     * the packet should be checksummed and encrypted if applicable, in
     * preparation for transmission.
     * 0 means do neither, FIRST_PASS | LAST_PASS means do both.
     * For secauth_parse, FIRST_PASS means decrypt the packet, otherwise leave it
     * alone.  LAST_PASS is ignored.
     */
#define FIRST_PASS	1
#define	LAST_PASS	2
    uint8_t         *snmp_comstr_parse(uint8_t *, size_t *, uint8_t *,
                                      size_t *, int32_t *);
    uint8_t         *snmp_comstr_build(uint8_t *, size_t *, uint8_t *,
                                      size_t *, int32_t *, size_t);

    int             has_access(uint8_t, int, int, int);

#ifdef __cplusplus
}
#endif
#endif /* _SNMP_IMPL_H_ */

