/*
 *
 * Copyright (c) 1992-2010 NetApp, Inc.
 * All rights reserved.
 */

/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <cmds/cmds.h>
#include <sk/ha_host.h>

#include <cmds/cmd_entry.h>

#include <sk/sk_param.h>
#include <sk/sk_errno.h>
#include <sys/param.h>
#include <mgr/boot.h>
#include <net7g/sys/sk_mbuf.h>
#include <net7g/netinet/sk_in.h>
#include <net7g/sk_netdb.h>
#include <net7g/arpa/sk_inet.h>
#include <net7g/arpa/sk_nameser.h>
#include <net-snmp/library/asn1.h>
#include <net-snmp/library/snmp_impl.h>
#include <net-snmp/library/snmp.h>
#include <net-snmp/library/callback.h>
#include <net-snmp/agent/agent_callbacks.h>
#include <net-snmp/library/snmp_api.h>
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include <net-snmp/agent/agent_registry.h>
#include <net-snmp/library/snmp_api.h>
#include <legacy/ctype.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/net/sk_if.h>     /* ipv6_enabled */
#include <driver/lcd/lcd.h>
#include <registry/reg.h>
#include <registry/reg_impl.h>
#include <registry/reg_private.h>
#include <rsh/rsh.h>
#undef SNMP_FREE
//SNMP_FREE is redefined in trap.h
#include <snmp/trap.h>
#include <fio/fio_stream.h>
#include <sk/sk_common.h>
#include <cmds/options.h>
#include <adminapi/zapi.h>
#include <ontap/zapi/zapierrno.h>
#include <ontap/ems/snmpd_ems.h>
#include <lib/access_check.h>
#include <agent/snmpd.h>

#include <driver/rlm/sp_config.h>

static int      snmp_set_community(char * action, char * acl, 
				   char * community, na_elem_t *xo);
static int      snmp_set_traphost(char *name, char *value, na_elem_t *xo);
static int      snmp_do_init(char * value, na_elem_t *xo);
static void     snmp_print_options(na_elem_t *xo);
static int      snmp_print_communities(na_elem_t *xo);
static void     snmp_print_location(na_elem_t *xo);
static void     snmp_print_contact(na_elem_t *xo);
static void     snmp_print_init(na_elem_t *xo);
static void     snmp_print_enableauthentraps(na_elem_t * xo);
static int      snmp_print_traphosts(na_elem_t *xo);
static int	snmp_get_enableauthentraps_value(void);

STATIC const char * options_snmp_registry_walkinit(regImpl* impl, regCred a,
		const char* prefix, regWalkType w);
STATIC boolean_t options_snmp_registry_get(regImpl* impl, regCred a,
		const char* key, char* valbuff, int leng);
STATIC void snmp_registry_get_communities(char *buff,int host);

extern void vacm_parse_simple(const char *token, char *confline);
extern void netsnmp_udp_com2SecList_free(void);

extern int             snmp_enableauthentraps[ONTAP_NHOSTS];
struct snmptraphost    snmptraphosts[ONTAP_NHOSTS][NUM_TRAPHOSTS];

SK_DOMAIN_IMPORT(ipv6_enabled, SK_REASON_READONLY);
/*
 * Maximum value lengths, for registry things.
 * The +4 on COMMUNITIES is for the commas and the ":rw" suffixes.
 * The +1 on TRAPHOSTS is for the commas.
 */
#define COMMUNITIES_MAX_VALSIZE (NUM_COMMUNITIES*(COMMUNITYSTRLEN+4))
#define TRAPHOSTS_MAX_VALSIZE (NUM_TRAPHOSTS*(ONTAP_MAXHOSTNAMELEN+1))
#define SNMP_MAX_VALSIZE (MAX(COMMUNITIES_MAX_VALSIZE,TRAPHOSTS_MAX_VALSIZE))
#define SNMP_TRAPHOST_LOOKUP_TIMEOUT	5


#define SNMP_CONTACT_REGKEY   "options.snmp.contact"
#define SNMP_LOCATION_REGKEY  "options.snmp.location"
#define SNMP_AUTHTRAP_REGKEY  "options.snmp.authtrap"
#define SNMP_INIT_REGKEY      "options.snmp.init"
#define SNMP_TRAPHOST_REGKEY  "options.snmp.traphosts"
#define SNMP_COMMUNITY_REGKEY "options.snmp.communities"
#define SYS_STRING_LEN       (SYSTEM_NAMELEN + 1)

			
static void
snmp_usage(void)
{
	cmd_printf("usage: snmp options\n");
	cmd_printf(
	"Options are:\n"
	"	           (no option => display current settings)\n"
	"	authtrap   [ 0|1 ]       (0 => disabled, 1 => enabled)\n"
	"	community  [ add ro <community> ]\n"
	"	community  [ delete { all | ro <community> } ]\n"
	"	contact    [ <contact> ]\n"
	"	init       [ 1 ]         (0 => uninitialized, 1 => initialized)\n"
	"	location   [ <location> ]\n"
	"	traphost   [ add|delete <hostname|ipaddress> ]\n"
	"	traps	   [ enable|disable|reset|delete ] [ <trapname> ] \n"
	"	traps	   <trapname>.<parm> <value>\n"
	"	traps	   walk <prefix>\n"
	"	traps	   load <filename>\n");
	if (cmd_my_priv_includes(SH_PRIV_DIAG)) {
		cmd_printf("	list	   counters\n");
	}
	cmd_printf("\n");
	cmd_printf("Use single quotes (i.e. 'abc def') to enclose strings with spaces or tabs\n");
}

static void
na_failed_or_print(na_elem_t *xo, const char *format, ...) 
{
	va_list         ap;

	va_start(ap, format);
	if (xo == NULL) {
		cmd_errvprintf(format, ap);
	} else {
		na_results_set_failed_v(xo, EINVAL, format, ap);
	}
        va_end(ap);
}

static void snmp_print_contact(na_elem_t * xo) 
{
	char * dstname = system_contact[sk_gethost()];
	if (xo != NULL) {
		na_child_add_string(xo, "contact", dstname);
	} else {
		cmd_printf("\t%s\n", dstname);
	}
}

static void snmp_set_contact(char * name)
{
	int namelen;
	namelen = strlen(name);
	if (namelen < 0 || namelen > SYS_STRING_LEN ) {
		cmd_printf("snmp: \"%s\" too long.  Must be < %d characters\n",
		      name, SYS_STRING_LEN);
		return;
	}
	if (!registry_set_and_commit(rootCred, SNMP_CONTACT_REGKEY, 
				     name, NULL)) {
		cmd_errprintf("snmp: unable to modify the contact name.\n");
		return;
	}
}

static void snmp_print_location(na_elem_t * xo) 
{
	char * dstname = system_location[sk_gethost()];
	if (xo != NULL) {
		na_child_add_string(xo, "location", dstname);
	} else {
		cmd_printf("\t%s\n", dstname);
	}
}

static void snmp_set_location(char * name)
{
	int namelen;
	namelen = strlen(name);
	if (namelen < 0 || namelen > SYS_STRING_LEN ) {
		cmd_printf("snmp: \"%s\" too long.  Must be < %d characters\n",
		      name, SYS_STRING_LEN);
		return;
	}
	if (!registry_set_and_commit(rootCred, SNMP_LOCATION_REGKEY, 
				     name, NULL)) {
		cmd_errprintf("snmp: unable to modify the location name.\n");
		return;
	}
}

/*
 * Internally SNMP stores enabled as 1 and disabled as 2.
 */
static int snmp_get_enableauthentraps_value() {
	int host = sk_gethost();

	if (snmp_enableauthentraps[host] == SNMP_AUTHENTICATED_TRAPS_DISABLED) {
		return 0;
	} else {
		return 1;
	}
}

static void snmp_print_enableauthentraps(na_elem_t * xo) 
{
	if (xo != NULL) {
		return; //not supported for apis
	} else {
		cmd_printf("\t%d\n", snmp_get_enableauthentraps_value());
	}
}

static void snmp_set_enableauthentraps(char * name)
{
	uint32_t readval;

	readval = (uint32_t) strtoul(name, (char **) NULL, 10);
	if (readval != 1 && readval != 0) {
		cmd_errprintf("snmp: illegal value passed.  Legal values are "
			      "0 or 1.\n");
		return;
	} 
	if (!registry_set_and_commit(rootCred, SNMP_AUTHTRAP_REGKEY, 
				     name, NULL)) {
		cmd_errprintf("snmp: unable to modify the authtrap value.\n");
		return;
	}
}

static int 
snmp_print_communities(na_elem_t *xo)
{
	na_elem_t *coms = NULL;
	na_elem_t *com_info;
	int i;
	int host = sk_gethost();
	
	if (xo != NULL) {
		coms = na_elem_new("communities");
	}
	for (i = 0; i < NUM_COMMUNITIES; i++) {
		if (communities[host][i].name[0] != '\0') {
			if (xo == NULL) {
				cmd_printf("\t%s %s\n",
					   communities[host][i].acl == RONLY ?
					   "ro" : "rw", communities[host][i].name);
			} else {
				com_info = na_elem_new("community-info");
				na_child_add_string(com_info, "community",
						    communities[host][i].name); 
				na_child_add_string(com_info,
						    "access-control", 
						    communities[host][i].acl == RONLY ?
						    "ro" : "rw");
				na_child_add(coms, com_info);
			}
		}
	}
	if (xo != NULL)
		na_child_add(xo, coms);

	return 0;
}

static int
snmp_print_traphosts(na_elem_t *xo)
{
	na_elem_t *thosts = NULL;
	na_elem_t *thost_info;
	int i;
	int		hostid = sk_gethost();
	
	struct in6_addr in6_addr_any = (struct in6_addr) IN6ADDR_ANY_INIT;
	char addr[INET6_ADDRSTRLEN];	
	if (xo != NULL) 
		thosts = na_elem_new("traphosts");
	
	for (i = 0; i < NUM_TRAPHOSTS; i++) {
		if (snmptraphosts[hostid][i].addr.s_addr == INADDR_ANY && 
			(bcmp(&snmptraphosts[hostid][i].ipv6_addr, 
				&in6_addr_any, 
				sizeof(struct in6_addr)) == 0)){
			continue;
		}	
		if (xo == NULL) {
			if (snmptraphosts[hostid][i].addr.s_addr != 
							INADDR_ANY ) {
				cmd_printf("\t%s (%s) <%s>\n",
				   snmptraphosts[hostid][i].aName,
				   snmptraphosts[hostid][i].cName,  
				   inet_ntoa(snmptraphosts[hostid][i].addr));
			}
			else {
				cmd_printf("\t%s (%s) <%s>\n",
				   snmptraphosts[hostid][i].aName,
				   snmptraphosts[hostid][i].cName, 
				   inet_ntop(AF_INET6, 
					&snmptraphosts[hostid][i].ipv6_addr, 
					addr, sizeof(addr))); 
			}
		} else {
			thost_info = na_elem_new("traphost-info");
			na_child_add_string(thost_info, "host-name",
					    snmptraphosts[hostid][i].aName);
			if (snmptraphosts[hostid][i].addr.s_addr != 
							INADDR_ANY ) {
				na_child_add_string(thost_info,
				   "ip-address", 
				   inet_ntoa(snmptraphosts[hostid][i].addr));
			}
			else {
				na_child_add_string(thost_info,
				  "ip-address", 
				   inet_ntop(AF_INET6, 
					&snmptraphosts[hostid][i].ipv6_addr, 
					addr, sizeof(addr))); 
			}
			na_child_add(thosts, thost_info);
		}
	}
	
	if (xo != NULL)
		na_child_add(xo, thosts);
	
	return 0;
}

static void snmp_print_init(na_elem_t * xo) 
{
	if (xo == NULL) {
		cmd_printf("\t%d\n", snmp_init_flag[sk_gethost()]);
	} else {
		na_child_add_string(xo, "is-trap-enabled", 
				    snmp_init_flag[sk_gethost()] ? "true" : "false");
	}
}

/*
 * Warn if a changing snmp value conflicts with a line already in /etc/rc.
 */
static void
warn_snmp_vs_rc(char *name)
{
	char line[MAX_LINELENGTH_ETCRC];
	FIO_FILE* f;
	char* lp;
	char** av;
	char* rshargs[RSH_MAXARGS];
	enum rsh_tokens t;

	/*
	 * Open /etc/rc and look for an snmp line for this name.
	 */
	f = fio_fopen("/etc/rc","r");
	if ( f == NULL )
		return;
	while ( fio_fgets(line,sizeof(line),f) != NULL ) {
		lp = line;
		av = rshargs;
		t = rsh_tokenize_line(&lp,(char**)&rshargs);
		if ( t != RSH_LASTTOK && t != RSH_EOF && t != RSH_EOL )
			continue;
		if ( *av == NULL || ! streq(*av,"snmp") )
			continue;
		av++;
		if ( *av == NULL || ! streq(*av,name) )
			continue;
		av++;
		if ( *av == NULL )
			continue;
	}
	fio_fclose(f);
}


/*
 * Set appropriate variables for snmp.
 */
SK_DOMAIN_EXPORT(cmd_snmp,SK_REASON_SWITCH_DOMAIN);
CSMP_STATIC cmd_result_t cmd_snmp(char **);
CSMP_STATIC cmd_result_t
cmd_snmp(char **av)
{
	cmd_result_t cmdResult;
	sk_domain_t calling_domain = sk_domain_switch_begin(sk_domain_kahuna);
	cmdResult = cmd_snmp_impl(av, NULL);
	sk_domain_switch_end(calling_domain);
	return cmdResult;
}
FILER_RSH_ONCMD(cmd_snmp, snmp, CLASSIC, "set/get snmp variables", NULL);

cmd_result_t
cmd_snmp_impl(char **av, na_elem_t *xo)
{
	cmd_result_t cmdResult = cmdDefaultReturn;
	char * option = NULL;

	++av;
	if (*av == NULL) {
		snmp_print_options(xo);
		cmdResult = cmdOkReturn;
		goto done;	
	}
	if (**av == '?' || **av == '-') {
		snmp_usage();
		cmdResult = cmdOkReturn;
		goto done;	
	}

	/* 
	 * The traps command is a subcommand of snmp.
	 * Check to see if that's what this is and
	 * jump to the cmd_traps() routine (in trap_cmds.c)
	 * if it is.
	 */
	if (strcmp(*av, SNMP_TRAPS) == 0) {
		cmdResult = cmd_traps(av, xo);
		goto done;	
	}

	option = *av;
	av++;

	if (strcmp(option, "contact") == 0) {
		if (*av == NULL) {
			snmp_print_contact(xo);
			return cmdOkReturn;
		}
		snmp_set_contact(*av);
	} else if (strcmp(option, "location") == 0) {
		if (*av == NULL) {
			snmp_print_location(xo);
			return cmdOkReturn;
		}
		snmp_set_location(*av);
	} else if (strcmp(option, "authtrap") == 0) {
		if (*av == NULL) {
			snmp_print_enableauthentraps(xo);
			return cmdOkReturn;
		}
		snmp_set_enableauthentraps(*av);
	} else if (strcmp(option, "init") == 0) {
		if (*av == NULL) {
			snmp_print_init(xo);
			return cmdOkReturn;
		}
		snmp_do_init(*av, xo);
	} else if (strcmp(option, "traphost") == 0) {
		if (*av == NULL) {
			snmp_print_traphosts(xo);
			return cmdOkReturn;
		}
		snmp_set_traphost(av[0], av[1], xo);
	} else if (strcmp(option, "community") == 0) {
		if (*av == NULL) {
			snmp_print_communities(xo);
			return cmdOkReturn;
		}
		if (av[0] == NULL) {
			cmd_printf("snmp: No action specified. Use add or delete\n");
			return cmdBadArgReturn;
		}
		if (av[1]  == NULL) {
			cmd_printf("snmp: No permission argument given.  Use ro.\n");
			return cmdBadArgReturn;
		}
		snmp_set_community(av[0], av[1], av[2], xo);
	} else if ((strcmp(option, "list") == 0) &&
		   *av != NULL &&
		   (strcmp(*av, "counters") == 0) &&
		   (cmd_my_priv_includes(SH_PRIV_DIAG))) {
		print_reginfo_counters();
		return cmdOkReturn;
	} else {
		cmd_printf("snmp: Did not recognize option \"%s\".\n", option);
		snmp_usage();
		cmdResult = cmdBadArgReturn;
		goto done;	
	}

	/*
	 * If we're not reading /etc/rc or the registry, we want
	 * to produce a warning if there is a line in /etc/rc that
	 * conflicts with the value that we're now setting.
	 *
	 */
	if ( !run_rc_InProgress[sk_gethost()] && !VV_registryLoading && !xo ) {
		warn_snmp_vs_rc(*av);
	}
	
done:
	return cmdResult; 
}

/*
 * initialize snmp it by calling "snmp_init()", or set it as
 * uninitialized, depending on whether the argument is "1" or "0".
 *
 * NOTE: "initialized" is different from "enabled".  The SNMP daemon could
 * be "enabled" but not "initialized"; that just means it doesn't do
 * traps.
 */
static int
snmp_do_init(char * value, na_elem_t *xo)
{
	int	host = sk_gethost();
	int     i;

	if (value[0] == '1' || strcmp(value,"on") == 0) {
		if (!registry_set_and_commit(rootCred, SNMP_INIT_REGKEY, 
					     "1", NULL)) {
			na_failed_or_print(xo, 
					   "Unable to enable snmp traps.\n");
			return errno;
		}

		/*
		 * Need to reset & resolve the traphosts
		 */
		for (i = 0; i < NUM_TRAPHOSTS; i ++) {
			snmptraphosts[host][i].addr.s_addr = INADDR_ANY;
			snmptraphosts[host][i].bResolved = false;
			snmptraphosts[host][i].ipv6_addr = 
				(struct in6_addr) IN6ADDR_ANY_INIT;
		}
		snmp_resolve_traphosts();

		/*
		 * We only send traps out if SNMP is currently enabled.
		 * When snmp_init_flag is set, and SNMP is subsequently
		 * enabled, the traps will be sent out at that time.
		 * So, the scenario at boot time is usually as follows:
		 * snmp_init_flag gets set (either by restoring the
		 * registry value "options.snmp.init" or by executing
		 * "snmp init 1" from /etc/rc).  Then, in registry_postrc_nonjava(),
		 * the options.service.snmp.enable value (if set to on) causes
		 * SNMP to be enabled, and the fact that snmp_init_flag is
		 * set will cause traps to be sent at that time.
		 */
		if (snmp_enabled[host])
			snmp_init_sendtraps();
		/*
		 * Setting init to on needs to load the trap
		 * queue if it has not already been loaded.
		 * ASSUME: the maintrapqueue variable is never
		 * NULL once the trap queue is loaded.
		 * NOTE: the reason not to just load it is that
		 * some traps may be in various states of backoff;
		 * the user should use "snmp traps reset" to force
		 * a load of the trap database.
		 */
		if (maintrapqueue == NULL && boot_is_done != 0) {
			snmp_load_queue(traproot); 
		}
	} else if (value[0] == '0'|| strcmp(value,"off") == 0) {
		if (!registry_set_and_commit(rootCred, SNMP_INIT_REGKEY, 
					     "0", NULL)) {
			na_failed_or_print(xo, 
					   "Unable to disable snmp traps.\n");
			return errno;
		}
	} else {
		cmd_errprintf("snmp: illegal value passed to snmp init\n");
		return EINVAL;
	}
	return 0;
}

#define SNMP_PRINT_VARIABLE(xo, name) { if (xo == NULL) {	\
						cmd_printf("%s:\n", name);	\
					}	\
				}				
/*
 * print value of all current snmp options
 */
static void
snmp_print_options(na_elem_t *xo)
{

	SNMP_PRINT_VARIABLE(xo, "contact");
	snmp_print_contact(xo);

	SNMP_PRINT_VARIABLE(xo, "location");
	snmp_print_location(xo);

	SNMP_PRINT_VARIABLE(xo, "authtrap");
	snmp_print_enableauthentraps(xo);

	SNMP_PRINT_VARIABLE(xo, "init");
	snmp_print_init(xo);
	
	SNMP_PRINT_VARIABLE(xo, "traphosts");
	snmp_print_traphosts(xo);
	
	SNMP_PRINT_VARIABLE(xo, "community");
	snmp_print_communities(xo);
}

/*
 * create/delete snmp community
 */
static int
snmp_set_community(char * action, char * acl, char * community, na_elem_t *xo)
{
	int             i, add = -1;
	int		host = sk_gethost();
	int		delete_all = 0;
	char		val[SNMP_MAX_VALSIZE];
	int 		commulen;
	bool 	before;
	int 		result = 0;
	char		*ptrcom = NULL;

	ASSERT(action != NULL && acl != NULL);

	if (strcmp("add", action) == 0) {
		add = 1;
	} else if (strcmp("delete", action) == 0) {
		add = 0;
	} else {
		na_failed_or_print(xo, "Action \"%s\" isn't valid. Use add or "
				   "delete\n", action);
		return EINVAL;
	}

	if (add == 0 && strcmp("all", acl) == 0) {
		delete_all = 1;
	} else if (strcmp("ro", acl) != 0) {
		na_failed_or_print(xo, "Permissions \"%s\" "
				   "aren't valid. Only \"ro\" supported.\n", acl);
		return EINVAL;
	}

	if (community == NULL && !delete_all) {
		na_failed_or_print(xo, "No community name specified.\n");
		return EINVAL;
	}

	if (delete_all) {
		if (!registry_set_and_commit(rootCred, 
					     SNMP_COMMUNITY_REGKEY, 
					     "", NULL)) {
			na_failed_or_print(xo, 
					   "Unable to delete all the "
					   "communities.\n");
			return errno;
		}
		goto end;
	}
	/*
	 * "delete all" is handled above, so handle deleting a single community now.
	 */
	if (!add) { 
		for (i = 0; i < NUM_COMMUNITIES; i++) {
			if (strcmp(community, communities[host][i].name) == 0){
				break;
			}
		}

		if (i == NUM_COMMUNITIES) {
			na_failed_or_print(xo, "Did not find community \"%s\" "
					   "to delete\n", community);
			return EINVAL;
		} 
		communities[host][i].name[0] = '\0';
	}  else {  /* Process 'add' case here. */

		/*
		 * So, first check for duplicates.
		 */
		for (i = 0; i < NUM_COMMUNITIES; i++) {
			if (communities[host][i].name[0] == '\0') {
				continue;
			}
			if (strcmp(community, communities[host][i].name) == 0) {
				na_failed_or_print (xo, 
						    "\"%s\" already exists in "
						    "list of communities\n", 
						    community);
				return EINVAL;
			}
		}
		for (i = 0; i < NUM_COMMUNITIES; i++) {
			if (communities[host][i].name[0] == '\0')
				break;
		}
		if (i == NUM_COMMUNITIES) {
			na_failed_or_print (xo, "No more space for new "
					    "communities.\nCan support a "
					    "maximum of %d communities\n", 
					    NUM_COMMUNITIES);
			
			return EINVAL;
		} 

		commulen = strlen(community);
		if (commulen < 0 || commulen > COMMUNITYSTRLEN ) {
			na_failed_or_print (xo, "\"%s\" too long. Must"
					    " be < %d characters\n",
					    community, COMMUNITYSTRLEN + 1);
			return EINVAL;
		}

		/*BURT 747734*/
		for (ptrcom = community; *ptrcom ; ptrcom++) {
			if ((*ptrcom == ' ') || (*ptrcom == '\t')) {
				na_failed_or_print (xo, "Error: invalid character \"%c\" in community"
						" \"%s\". The community must not contain spaces or Tab"
						" characters.\n", *ptrcom, community);
				return EINVAL;
			}
		}

		strlcpy(communities[host][i].name, community, COMMUNITYSTRLEN + 1);
		communities[host][i].acl = RONLY;
	}
	/*
	 * In the above code, we couldn't use the snmp registry interface
	 * to set the values because registry interface doesn't support
	 * adding or deleting a single entry. So, we've modified the global
	 * variables directly. Now, we need to get the current value and
	 * set it using registry interface. Otherwise, changes wouldn't
	 * get into registry.local file.
	 */
	before = registry_get(rootCred, SNMP_COMMUNITY_REGKEY, val, sizeof(val));
	if (before == false) {
		na_failed_or_print(xo, "Unable to get the value for %s "
				   "from the filer registry.\n", 
				   SNMP_COMMUNITY_REGKEY);
		return EINVAL;
	}
	if (!registry_set_and_commit(rootCred, SNMP_COMMUNITY_REGKEY, 
				     val, NULL)) {
		/*
		 * Restore back the old values by modifying the global
		 * variables directly.
		 */
		if (!add) {
			commulen = strlen(community);
			strlcpy(communities[host][i].name, community,
			    commulen + 1);
			na_failed_or_print(xo, "Unable to delete "
					   "the %s community.\n", 
					   community);
		} else {
			communities[host][i].name[0] = '\0';
			na_failed_or_print(xo, "Unable to add %s "
					   "as a community.\n", 
					   community);
		}
		result = errno;
	}
end:
	return result;
}

extern void snmpd_free_trapsinks(void);

static int
snmp_add_traphost(char * host, na_elem_t * xo)
{
	int		i;
	const char      *host_cName = NULL;
	char            cName[MAXDNAME + 1];
	uint32_t	ipaddr = INADDR_ANY;
	struct hostent 	*hp;
	struct hostent 	hbuf;
	char 		buf[MAX_LINELENGTH_HOSTS+1];
	int 		err;
	int		hostid = sk_gethost();
	char		val[SNMP_MAX_VALSIZE];
	bool	 	before;
	int             empty = NUM_TRAPHOSTS;
	struct addrinfo hints;
	struct addrinfo *res = NULL;
	int		status = 0;

	struct in6_addr ip6addr = (struct in6_addr) IN6ADDR_ANY_INIT;
	bool isipv6 = false;
	bool isipv4 = false;

	if (host == NULL) {
		na_failed_or_print(xo, "Specify a host name or ipaddress\n");
		return EINVAL;
	}

	host_cName = host;

	/* 
	 * check if the host address is ipv4 or ipv6 address
	 */
	if (inet_pton(AF_INET6, host, &ip6addr)) {	
		isipv6 = true;
		/*
                 * host was valid ipv6 representation - see if we can dig up
                 * name.  If not, leave name = ipaddr
                 */
		if (!ipv6_enabled) {
			na_failed_or_print(xo, "IPv6 is not enabled." 
			" SNMP traphost cannot be added.\n");
			return EINVAL;
		}
                if ((hp = gethostbyaddr_r((char *) &ip6addr, 16, AF_INET6, &hbuf,
                                        buf, sizeof(buf), &err)) != NULL) {
                        host_cName = hp->h_name;
                }

	}
	else if (inet_pton(AF_INET, host, &ipaddr)) {
		isipv4 = true;
		/*
                 * host was valid ipv4 representation - see if we can dig up
                 * name.  If not, leave name = ipaddr
                 */
                if ((hp = gethostbyaddr_r((char *) &ipaddr, 4, AF_INET, &hbuf,
                                        buf, sizeof(buf), &err)) != NULL) {
                        host_cName = hp->h_name;
                }

	}	
	if (ipaddr == 0 && !isipv6) {
		bzero(&hints, sizeof(hints));
		hints.ai_family = AF_INET6;
		hints.ai_flags |= AI_CANONNAME;
		SK_DOMAIN_IMPORT(getaddrinfo, SK_REASON_SWITCH_DOMAIN);
		status = getaddrinfo(host, NULL, &hints, &res);
		if (status) {
			if (res) {
				freeaddrinfo(res);
			}
			bzero(&hints, sizeof(hints));
			hints.ai_family = AF_INET;
			hints.ai_flags |= AI_CANONNAME;
			res = 0;
			status = 0;
			status = getaddrinfo(host, NULL, &hints, &res);
			if (status) {
				na_failed_or_print(xo,
						"Unknown host \"%s\"\n", host);
				if (res) {
					freeaddrinfo(res);
				}
				return EINVAL;
			}
		}
		if (res->ai_family == AF_INET6) {
			if (!ipv6_enabled) {
				na_failed_or_print(xo, "IPv6 is not enabled."
				" SNMP traphost cannot be added.\n");
				freeaddrinfo(res);
				return EINVAL;
			}
			bcopy(&((struct sockaddr_in6 *)res->ai_addr)->sin6_addr, &ip6addr, sizeof(struct in6_addr));
		} else if (res->ai_family == AF_INET) {
			bcopy(&((struct sockaddr_in *)res->ai_addr)->sin_addr, &ipaddr, sizeof(struct in_addr));
		} else {
			na_failed_or_print(xo,
					"Unknown host \"%s\"\n", host);
			freeaddrinfo(res);
			return EINVAL;
		}
		if (res->ai_canonname) {
			strlcpy(cName, res->ai_canonname, sizeof(cName));
			host_cName = cName;
		}
		freeaddrinfo(res);
	}

	/* first check for duplicates */
	for (i = 0; i < NUM_TRAPHOSTS; i++) {
		if (snmptraphosts[hostid][i].aName[0] == '\0') {
			empty = empty < i ? empty : i;
			continue;
		}
		if (strcmp(host, snmptraphosts[hostid][i].aName) == 0 ||
		    strcmp(host_cName, snmptraphosts[hostid][i].cName) == 0 ||
		    (isipv4 && snmptraphosts[hostid][i].addr.s_addr == ipaddr) || 
		    (isipv6 && (bcmp(&snmptraphosts[hostid][i].ipv6_addr,
		    	&ip6addr, 
			sizeof(struct in6_addr)) == 0))) {
			na_failed_or_print(xo, 
					   "\"%s\" already exists in "
					   "list of trap hosts.\n",
					   host);
			return EINVAL;
		}
	}
	
	if (empty == NUM_TRAPHOSTS) {
		na_failed_or_print(xo, 
				   "No more space for new trap hosts. "
				   "Can support a maximum of %d "
				   "traphosts.\n", NUM_TRAPHOSTS);
		return EINVAL;
	} 
	
	if (strlen(host) > DEFAULT_NAMELEN) {
		na_failed_or_print(xo, "\"%s\" too long. "
				   "Must be < %d characters.\n",
				   host, DEFAULT_NAMELEN);
		return EINVAL;		
	} else if (strlen(host_cName) > DEFAULT_NAMELEN) {
		na_failed_or_print(xo, "\"%s\" too long. "
				   "Must be < %d characters.\n",
				   host_cName, DEFAULT_NAMELEN);
		return EINVAL;		
	}
	
	(void) strlcpy(snmptraphosts[hostid][empty].aName, host, 
				DEFAULT_NAMELEN); 
	(void) strlcpy(snmptraphosts[hostid][empty].cName, host_cName, 
				DEFAULT_NAMELEN); 
	
	snmptraphosts[hostid][empty].addr.s_addr = ipaddr;
	snmptraphosts[hostid][empty].bResolved = false;

	bcopy(&ip6addr, &snmptraphosts[hostid][empty].ipv6_addr, 
			sizeof(struct in6_addr));

	/*
	 * In the above code, we couldn't use the snmp registry interface
	 * to set the values because registry interface doesn't support
	 * adding or deleting a single entry. So, we've modified the global
	 * variables directly. Now, we need to get the current value and
	 * set it using registry interface. Otherwise, changes wouldn't
	 * get into registry.local file.
	 */
	before = registry_get(rootCred, SNMP_TRAPHOST_REGKEY, val, 
			      sizeof(val));
	if (before == false) {
		/*
		 * Restore back the old values by modifying the global
		 * variables directly.
		 */
		snmptraphosts[hostid][empty].addr.s_addr = 
			INADDR_ANY;
		snmptraphosts[hostid][empty].ipv6_addr = (struct in6_addr)
			IN6ADDR_ANY_INIT; 
		snmptraphosts[hostid][empty].aName[0] = '\0';
		snmptraphosts[hostid][empty].cName[0] = '\0';
		snmptraphosts[hostid][empty].bResolved = false;
		na_failed_or_print(xo, 
				   "Unable to get value for %s from "
				   "filer registry.\n",
				   SNMP_TRAPHOST_REGKEY);
		return EINVAL;
	} 
	if (!registry_set_and_commit(rootCred, SNMP_TRAPHOST_REGKEY, 
				     val, NULL)) {
		/*
		 * Restore back the old values by modifying the global
		 * variables directly.
		 */
		snmptraphosts[hostid][empty].addr.s_addr = 
			INADDR_ANY;
		snmptraphosts[hostid][empty].ipv6_addr = (struct in6_addr)
			IN6ADDR_ANY_INIT; 
		snmptraphosts[hostid][empty].aName[0] = '\0';
		snmptraphosts[hostid][empty].cName[0] = '\0';
		snmptraphosts[hostid][empty].bResolved = false;
		na_failed_or_print(xo, "Unable to add %s as a "
				   "trap host.\n", host);
		return errno;
	}
	return 0;
}

/* Flag to indicate that the unregister handler to be invoked */
static bool traphost_del_flag = false;

static int
snmp_delete_traphost(char * host, na_elem_t * xo)
{
	int		i;
	const char      *host_cName = NULL;
	char            cName[MAXDNAME + 1];
	uint32_t	ipaddr = INADDR_ANY;
	struct hostent 	*hp;
	struct hostent 	hbuf;
	char 		buf[MAX_LINELENGTH_HOSTS+1];
	int 		err;
	int		hostid = sk_gethost();
	char		val[SNMP_MAX_VALSIZE];
	bool		before;
	struct snmptraphost trap_entry_save;

	struct in6_addr ip6addr = (struct in6_addr) IN6ADDR_ANY_INIT;
	bool isipv6 = false;
	struct addrinfo hints;
	struct addrinfo *res = NULL;
	int		status = 0;

	if (host == NULL) {
		na_failed_or_print(xo, "Specify a host name or ipaddress\n");
		return EINVAL;
	}

	/*
	 * If the aName matches or cName matches, there's no need to do
	 * an IP lookup.
	 */
	for (i = 0; i < NUM_TRAPHOSTS; i++) {
		if (strcmp(host, snmptraphosts[hostid][i].aName) == 0 ||
		    strcmp(host, snmptraphosts[hostid][i].cName) == 0)
			break;
	}

	if (i != NUM_TRAPHOSTS) {
		goto foundTrap;
	}

	/*
	 * check if the host address is ipv4 or ipv6 address
	 */
	if (inet_pton(AF_INET6, host, &ip6addr)) {
		isipv6 = true;
		/*
		 * host was valid ipv6 representation - see if we can dig up
		 * name.  If not, leave name = ipaddr
		 */
		if ((hp = gethostbyaddr_r((char *) &ip6addr, 16, AF_INET6, &hbuf,
				buf, sizeof(buf), &err)) != NULL) {
			host_cName = hp->h_name;
		}
	}
	else if (inet_pton(AF_INET, host, &ipaddr)) {
		/*
		 * host was valid ipv4 representation - see if we can dig up
		 * name.  If not, leave name = ipaddr
		 */
		if ((hp = gethostbyaddr_r((char *) &ipaddr, 4, AF_INET, &hbuf,
				buf, sizeof(buf), &err)) != NULL) {
			host_cName = hp->h_name;
		}
	}

	if (ipaddr == 0 && !isipv6) {
		/* host was not a valid ip representation. check for chars */
		bzero(&hints, sizeof(hints));
		hints.ai_family = AF_INET6;
		hints.ai_flags |= AI_CANONNAME;
		SK_DOMAIN_IMPORT(getaddrinfo, SK_REASON_SWITCH_DOMAIN);
		status = getaddrinfo(host, NULL, &hints, &res);
		if (status) {
			if (res) {
				freeaddrinfo(res);
			}
			bzero(&hints, sizeof(hints));
			hints.ai_family = AF_INET;
			hints.ai_flags |= AI_CANONNAME;
			res = 0;
			status = 0;
			status = getaddrinfo(host, NULL, &hints, &res);
			if (status) {
				na_failed_or_print(xo, "Unknown host \"%s\"\n", 
								host);
				if (res) {
					freeaddrinfo(res);
				}
				return EINVAL;
			}
		}
		/* 
		 * if DNS resolves hostname to ipv6 address or if we ask DNS 
		 * to resolve hostname into ipv6 address.
		 */
		if (res->ai_family == AF_INET6) {
			bcopy(&((struct sockaddr_in6 *)res->ai_addr)->sin6_addr, &ip6addr, sizeof(struct in6_addr));
		} else if (res->ai_family == AF_INET) {
			bcopy(&((struct sockaddr_in *)res->ai_addr)->sin_addr, &ipaddr, sizeof(struct in_addr));
		} else {
			na_failed_or_print(xo, "Unknown host \"%s\"\n",
							host);
			freeaddrinfo(res);
			return EINVAL;
		}
		if (res->ai_canonname) {
			strlcpy(cName, res->ai_canonname, sizeof(cName));
			host_cName = cName;
		}
		freeaddrinfo(res);
	}

	for (i = 0; i < NUM_TRAPHOSTS; i++) {
		if (snmptraphosts[hostid][i].addr.s_addr == ipaddr && 
			bcmp(&snmptraphosts[hostid][i].ipv6_addr,
				&ip6addr, sizeof(struct in6_addr)) == 0) 
			break;
	}
	if (i == NUM_TRAPHOSTS) {
		na_failed_or_print(xo, 
				   "Did not find the host name %s for "
				   "deletion.\n", host);
		return EINVAL;
	}

  foundTrap:
	/*
	 * save the entry in case attempts to update the
	 * registry fail and there is a need to put the
	 * entry back the way it started.
	 */
	bcopy(&snmptraphosts[hostid][i], &trap_entry_save, 
	      sizeof(struct snmptraphost));
	snmptraphosts[hostid][i].addr.s_addr = INADDR_ANY;
	snmptraphosts[hostid][i].ipv6_addr = (struct in6_addr) IN6ADDR_ANY_INIT;
	snmptraphosts[hostid][i].aName[0] = '\0'; 
	snmptraphosts[hostid][i].cName[0] = '\0';
	snmptraphosts[hostid][i].bResolved = false;


	/*
	 * In the above code, we couldn't use the snmp registry interface
	 * to set the values because registry interface doesn't support
	 * adding or deleting a single entry. So, we've modified the global
	 * variables directly. Now, we need to get the current value and
	 * set it using registry interface. Otherwise, changes wouldn't
	 * get into registry.local file.
	 */
	before = registry_get(rootCred, SNMP_TRAPHOST_REGKEY, val, sizeof(val));
	if (before == false) {
		/*
		 * Restore back the old values by modifying the global
		 * variables directly.
		 */
		bcopy(&trap_entry_save, &snmptraphosts[hostid][i],
		      sizeof(struct snmptraphost));
		na_failed_or_print(xo, 
				   "Unable to get value for %s from "
				   "filer registry.\n",
				   SNMP_TRAPHOST_REGKEY);
		return EINVAL;
	} 
	/*
         * Set the flag to indicate that the unregister handler
         * to be invoked by the callback in snmp_resolve_traphost()
         */
        traphost_del_flag = true;
 
	if (!registry_set_and_commit(rootCred, SNMP_TRAPHOST_REGKEY, val, 
				     NULL)) {
		/*
		 * Restore back the old values by modifying the global
		 * variables directly.
		 */
		bcopy(&trap_entry_save, &snmptraphosts[hostid][i],
		      sizeof(struct snmptraphost));
		na_failed_or_print(xo, 
				   "Unable to delete %s "
				   "from the trap hosts list.\n", 
				   host);
		traphost_del_flag = false;
		return errno;
	}

	return 0;
}

/*
 * add/delete snmp trap hosts
 */
static int
snmp_set_traphost(char * action, char *host, na_elem_t *xo)
{
	if (action == NULL) {
		action = "";
	}
	if (strcmp("add", action) == 0) {
		return snmp_add_traphost(host, xo);
	} else if (strcmp("delete", action) == 0) {
		return snmp_delete_traphost(host, xo);
	} else {
		na_failed_or_print(xo, "Did not recognize \"%s\".\n	"
				  "Use either add or delete:\n", action);
		return EINVAL;
	}
	
	return 0;
}

/*
 * The code below provides a registry implementation for the options.snmp.*
 * namespace, providing read/write access to snmp options through the registry.
 */
SK_DOMAIN_EXPORT(snmp_registry_get_communities, SK_REASON_MPSAFE);
STATIC void
snmp_registry_get_communities(char * buff,int host)
{
	char* sep = "";
	int i;
	char* suffix;

	buff[0] = '\0';
	for ( i=0; i<NUM_COMMUNITIES; i++ ) {
		const char* nm = communities[host][i].name;
		if ( nm == NULL || *nm == '\0' )
			continue;
		/*
		 * RWRITE communities aren't really supported, but when they
		 * are, they are distinguished by an ":rw" suffix on the
		 * community name.
		 */
		switch (communities[host][i].acl) {
		case RONLY:
			suffix = "";
			break;
		case RWRITE:
			suffix = ":rw";
			break;
		default:
			suffix = "??";
			break;
		}
		sprintf(strchr(buff,'\0'),"%s%s%s",sep,nm,suffix);
		sep = ",";
	}
}

static int
registry_snmp_set_string(char *name, char *dstname, int dstlen)
{
	int namelen = strlen(name);
	if (namelen < 0 || namelen > dstlen - 1)
		return EINVAL;
	bcopy(name, dstname, namelen);
	dstname[namelen] = 0;
	return 0;
}

#define	REGISTRY_SNMP_SET_STRING(name, dst)	\
	registry_snmp_set_string(name, dst, sizeof dst)

/*
 * Set the snmp.communities registry value, where the value is a
 * comma-separated list of community names.
 */
static int
snmp_registry_set_communities(const char *value, int host)
{
	char valbuff[SNMP_MAX_VALSIZE];
	char* seps = ",\t\n\r";
	char* p;
	char* lasts;
	int i;

	snprintf(valbuff,sizeof(valbuff),"%s",value);
	for ( i=0,p=strtok_r(valbuff,seps,&lasts); p!=NULL; i++,p=strtok_r(NULL,seps,&lasts) ) {
		/*
		 * To indicate a "read-write" community, you append a ":rw"
		 * suffix to the community name.  Communities without a suffix
		 * are "read-only", which is really all we currently
		 * support anyway.
		 */
		char* suff = strchr(p,':');
		if ( suff != NULL )
			*suff++ = '\0';
		snprintf(communities[host][i].name,
			sizeof(communities[host][i].name),"%s",p);
		if ( suff != NULL && strcmp(suff,"rw")==0 ) {
			communities[host][i].acl = RWRITE;
		}else{
			communities[host][i].acl = RONLY;
		}
	}
	/*
	 * clear out the rest
	 */
	for ( ; i<NUM_COMMUNITIES; i++ ) {
		communities[host][i].name[0] = '\0';
	}
	/* 
	 * This is perfect time to resolve communities as we just loaded registry
	 * values in to memory. 
	 */
	netsnmp_udp_com2SecList_free();
	if (ipv6_enabled) {
		netsnmp_udp6_com2Sec6List_free();
	}
	snmp_resolve_communities();
	return 0;
}

extern void snmpd_parse_config_trapsess(const char *word, char *cptr);

int
snmp_resolve_traphosts(void)
{
	int h = sk_gethost();
	int i;
	struct hostent *hp = NULL;
	struct hostent hbuf;
	char buf[MAX_LINELENGTH_HOSTS+1];
	int err; 
        uint32_t ipaddr = INADDR_ANY;
	bool trapsess_update_needed = false ;

	struct in6_addr in6_addr_any = (struct in6_addr) IN6ADDR_ANY_INIT;
	struct in6_addr ip6addr = (struct in6_addr) IN6ADDR_ANY_INIT;
	bool isipv6 = false;
	struct addrinfo hints;
	struct addrinfo *res;
	int status;
	int check = 0;
        char *host = NULL;

	for ( i=0; i < NUM_TRAPHOSTS; i++) {
		host = snmptraphosts[h][i].aName;
		if ( *host == '\0' ) {
			continue;
		}
		isipv6 = false;
		ip6addr = (struct in6_addr) IN6ADDR_ANY_INIT;
		ipaddr = INADDR_ANY;

		if (snmptraphosts[h][i].addr.s_addr != INADDR_ANY || 
			bcmp(&snmptraphosts[h][i].ipv6_addr, 
				&in6_addr_any, 
				sizeof(struct in6_addr)) != 0) {
			continue;
		}
		strcpy(host, snmptraphosts[h][i].cName);
		
		/*
         	 * check if the host address is ipv4 or ipv6 address
         	 */
	        if (inet_pton(AF_INET, host, &ipaddr)) {
			hp = gethostbyaddr_r_timeout((char *) &ipaddr, 4, 
				AF_INET, &hbuf, buf, sizeof(buf), &err, 
				SNMP_TRAPHOST_LOOKUP_TIMEOUT);
        	}
        	else if (inet_pton(AF_INET6, host, &ip6addr)) {
                	isipv6 = true;
                	hp = gethostbyaddr_r_timeout((char *) &ip6addr, 16, 
				AF_INET6, &hbuf, buf, sizeof(buf), &err, 
				SNMP_TRAPHOST_LOOKUP_TIMEOUT);  
                }

	        if (ipaddr == 0 && !isipv6) {
			/* 
			 * host was not a valid ipv4 or ipv6 representation. 
			 * check for chars. Check for IPv4 before IPv6.
			 */
			bzero(&hints, sizeof(hints));
			hints.ai_family = AF_INET;
			hints.ai_flags |= AI_CANONNAME;
			SK_DOMAIN_IMPORT(getaddrinfo_timeout1, SK_REASON_SWITCH_DOMAIN);
			res = 0;
			status = 0;
			status = getaddrinfo_timeout1(host, NULL, &hints, &res, SNMP_TRAPHOST_LOOKUP_TIMEOUT);
			if (status) {
				if (res) {
					freeaddrinfo(res);
				}
				bzero(&hints, sizeof(hints));
				hints.ai_family = AF_INET6;
				hints.ai_flags |= AI_CANONNAME;
				res = 0;
				status = 0;
				status = getaddrinfo_timeout1(host, NULL, &hints, &res, SNMP_TRAPHOST_LOOKUP_TIMEOUT);
				if (status) {
					EMS_snmp_traphost_resolve_failed(host);
					if (res) {
						freeaddrinfo(res);
					}
					continue;
				}
			}
			if (res->ai_family == AF_INET6) {
				bcopy(&((struct sockaddr_in6 *)res->ai_addr)->sin6_addr, &ip6addr, sizeof(struct in6_addr));
			} else if (res->ai_family == AF_INET) {
				bcopy(&((struct sockaddr_in *)res->ai_addr)->sin_addr, &ipaddr, sizeof(struct in_addr));
			} else {
				EMS_snmp_traphost_resolve_failed(host);
				freeaddrinfo(res);
				continue;
			}
			if (res->ai_canonname) {
				strlcpy(snmptraphosts[h][i].cName, res->ai_canonname, sizeof(snmptraphosts[h][i].cName));
			}
			freeaddrinfo(res);
		} else {
			 if (hp != NULL) {
				 strcpy(snmptraphosts[h][i].cName, hp->h_name);
			 }
		}

		snmptraphosts[h][i].addr.s_addr = ipaddr;
		bcopy(&ip6addr, &snmptraphosts[h][i].ipv6_addr, 
			sizeof(struct in6_addr));
		snmptraphosts[h][i].bResolved = false;
		trapsess_update_needed = true;
	}

	/*
	 * Restart all of the snmp trap sessions.
	 */
	if (trapsess_update_needed || traphost_del_flag) {
		char trap_args[256];
		traphost_del_flag = false;
		check =	snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
				SNMPD_CALLBACK_UNREGISTER_NOTIFICATIONS, NULL);

		if (check == SNMPERR_GENERR ) {
			return SNMPERR_GENERR;
		} 
		for (i = 0; i < NUM_TRAPHOSTS; i ++) {
			if (((bcmp(&snmptraphosts[h][i].ipv6_addr, 
				&in6_addr_any, 
				sizeof(struct in6_addr)) == 0) && 
			    snmptraphosts[h][i].addr.s_addr == INADDR_ANY)) {
				continue;
			}
			if (bcmp(&snmptraphosts[h][i].ipv6_addr, 
					&in6_addr_any, 
					sizeof(struct in6_addr)) == 0) {
				snprintf(trap_args, sizeof(trap_args),
                                 "-c %s -v 1 %s:162",
                                 communities[h][0].name[0] == '\0' ?
                                      "public" : communities[h][0].name,
                                 snmptraphosts[h][i].cName);
			}
			else {
				snprintf(trap_args, sizeof(trap_args), 
				 "-c %s -v 1 udpipv6:%s:162",
				 communities[h][0].name[0] == '\0' ?
				      "public" : communities[h][0].name, 
				 snmptraphosts[h][i].cName);
			}
			snmpd_parse_config_trapsess("trapsess", trap_args);
			snmptraphosts[h][i].bResolved = true;
		}
	}
	
	{
		if(VFILER_IS_DEFAULT(sk_get_vfiler())) {
			sp_config_snmp_update();
		}
	}
	return 0;
}

void
snmp_resolve_communities(void)
{
	int i;
	int host = sk_gethost();

	/*
	 * For each valid community string, resolve security, group, access,
	 * and view. If ipv6 enabled, then let udp6 resolve GAV.
	 */
	for (i = 0 ; i<NUM_COMMUNITIES; i++ ) {
		if (communities[host][i].name[0] != '\0') {
			vacm_parse_simple("rocommunity", communities[host][i].name);
			if (ipv6_enabled) {
				vacm_parse_simple("rocommunity6", communities[host][i].name);
			}
		}
	}
}

void
snmp_autosupport(const char *output_file)
{
    int i = 0;
    int curhost = sk_gethost();
    char buf[2048] = { 0 };
    FIO_FILE * ofp = fio_fopen((char *)output_file, "w");

    if (ofp) {

        if (system_contact[curhost][0] != '\0') {
            snprintf(buf, 
                     sizeof buf, 
                     "SNMP Contact : %s\n", 
                     system_contact[curhost]);
            (void) fio_fputs(buf, ofp);
        }

        if (system_location[curhost][0] != '\0') {
            snprintf(buf, 
                     sizeof buf, 
                     "SNMP Location : %s\n", 
                     system_location[curhost]);
            (void) fio_fputs(buf, ofp);
        }

        snprintf(buf, 
                 sizeof buf, 
                 "SNMP Authtrap : %d\n",
                  snmp_enableauthentraps[curhost] == SNMP_AUTHENTICATED_TRAPS_DISABLED ? 0 : 1);
        (void) fio_fputs(buf, ofp);

        for (i = 0; i < NUM_TRAPHOSTS; i++) {

            if (snmptraphosts[curhost][i].addr.s_addr == INADDR_ANY) {
                continue;
            }

            snprintf(buf, 
                     sizeof buf, 
                     "SNMP Traphost %d : %s\n", 
                     i+1, 
                     snmptraphosts[curhost][i].aName);
            (void) fio_fputs(buf, ofp);
        }

        snprintf(buf, 
                 sizeof buf, 
                 "SNMP init = %d\n", 
                 snmp_init_flag[curhost]);

        (void) fio_fputs(buf, ofp);

        snprintf(buf, 
                 sizeof buf, 
                 "SNMP Counters\n");

        (void) fio_fputs(buf, ofp);

        read_reginfo_counters(buf, sizeof buf);

        (void) fio_fputs(buf, ofp);

        fio_fclose(ofp);
    }
}

/*
 * Set the options.snmp.traphosts registry value, where the value is a
 * comma-separated list of host names.  If the registry is loading,
 * we're being called before /etc/rc and hence the network
 * interfaces aren't up and hence we can't resolve (via DNS or NIS, anyway)
 * hostnames.  So, in that case we merely set the hostnames and
 * delay the resolution until the boot is finished (see snmp_trap_main()).
 */
static int
snmp_registry_set_traphosts(const char *value, int h)
{
	struct snmptraphost    copytraphosts[NUM_TRAPHOSTS];
	char valbuff[SNMP_MAX_VALSIZE];
	char* seps = ", \t\n\r";
	char* lasts;
	char* host;
	int i, j;
	int check = 0;
	snprintf(valbuff,sizeof(valbuff),"%s",value);

	for ( i=0,host=strtok_r(valbuff,seps,&lasts);
		host!=NULL && i<NUM_TRAPHOSTS;
		i++,host=strtok_r(NULL,seps,&lasts) ) {

		strcpy(copytraphosts[i].aName, host);
		strcpy(copytraphosts[i].cName, host);
		copytraphosts[i].addr.s_addr = INADDR_ANY;
		copytraphosts[i].ipv6_addr = (struct in6_addr)
						IN6ADDR_ANY_INIT;
		copytraphosts[i].bResolved = false;
	}
	/*
	 * Clear out the rest
	 */
	for (; i < NUM_TRAPHOSTS; i++) {
		copytraphosts[i].aName[0] = '\0';
		copytraphosts[i].cName[0] = '\0';
		copytraphosts[i].addr.s_addr = INADDR_ANY;
		copytraphosts[i].ipv6_addr = (struct in6_addr)
						IN6ADDR_ANY_INIT;
		copytraphosts[i].bResolved = false;
	}

	/*
	 * compare copy and original. Keep the hosts with same name.
	 */
	for(i=0; i<NUM_TRAPHOSTS; i++)
	{
		if (copytraphosts[i].aName[0] == '\0')
		{
		   break;
		}
		for (j=0; j<NUM_TRAPHOSTS; j++)
		{
		   if (snmptraphosts[h][j].aName[0] == '\0')
		   {
		      continue;
		   }
		   if (strcmp(copytraphosts[i].aName, snmptraphosts[h][j].aName) == 0
		      || strcmp(copytraphosts[i].cName, snmptraphosts[h][j].cName) == 0)
		   {
		      if (snmptraphosts[h][j].bResolved )
		      {
		         memcpy(&copytraphosts[i], &snmptraphosts[h][j],
		            sizeof(struct snmptraphost));
		      }
		      break;
		   }
		}
	}
	/*
	 * Copy now contains the newly updated host list. Move it to original list.
	 */
	memcpy(&snmptraphosts[h][0],&copytraphosts[0], sizeof(snmptraphosts[h]));

	/*
	 * resolve traphosts only if the registry value is modified directly.
	 */
	if ( (! VV_registryLoading) &&
	     (! VV_registryPeriodicUpdating[sk_gethost()]) ) {
		check = snmp_resolve_traphosts();
		if( check == SNMPERR_GENERR) {
			return -1;
		}
	}
	return 0;
}

/*
 * options_snmp_registry_get
 */
SK_DOMAIN_EXPORT(options_snmp_registry_get, SK_REASON_MPSAFE);
STATIC boolean_t
options_snmp_registry_get(regImpl* impl, regCred a, const char* key, char* valbuff,
	int leng)
{
	const char* result = NULL;
	int host = sk_gethost();
	char buff[SNMP_MAX_VALSIZE];

	D(registry, 20, ("options_snmp_registry_get: key=%s\n",key) );

	if ( strcmp("access",key) == 0 ) {
		result = &(snmp_access_spec[host][0]);
	} else if ( strcmp("authtrap",key) == 0 ) {
		snprintf(buff,sizeof(buff),"%d",
			snmp_get_enableauthentraps_value());
		result = buff;
	} else if ( strcmp("communities",key) == 0 ) {
		snmp_registry_get_communities(buff,host);
		result = buff;
	} else if ( strcmp("contact",key) == 0 ) {
		result = system_contact[host];
	} else if ( strcmp("enable",key) == 0 ) {
		result = snmp_enabled[host] ? "on" : "off";
	} else if ( strcmp("init",key) == 0 ) {
		snprintf(buff,sizeof(buff),"%d",snmp_init_flag[host]);
		result = buff;
	} else if ( strcmp("location",key) == 0 ) {
		result = system_location[host];
	} else if ( strcmp("traphosts",key) == 0 ) {
		char *sep = "";
		int len = 0;
		int size = sizeof(buff);
		int i;

		buff[0] = '\0';
		for ( i = 0; i < NUM_TRAPHOSTS; i++ ) {
			if ( snmptraphosts[host][i].aName[0] != '\0' ) {
				len = strlen(buff);
				snprintf(buff+len,size-len,"%s%s",
					 sep,snmptraphosts[host][i].aName);
				sep = ",";
			}
		}
		result = buff;
	} else {
		errno = ENOENT;
		return false;
	}
	return registry_returnbuff(result,valbuff,leng);
}
/*
 * options_snmp_registry_set - set an snmp option.
//ntap This doesn't make sense. Sometimes the options are set here, and
//sometimes they are set in options.c (Look for getval_snmp_enable().)
 */
static boolean_t
options_snmp_registry_set(regImpl* impl, regCred a, const char* key, const char* value)
{
	int host = sk_gethost();
	int r = 0;
	char* v = (char*)value;

	D(registry, 20, ("options_snmp_registry_set: key=%s value=%s\n",key,value) );

	/*
	 * options_snmp_registry_set can be called before snmpd thread can
	 * init mainly during boot process. Multiple re-initialization of 
	 * init_snmp() is avoided using a static "done_init" variable. 
	 * So initialize it here safely.
	 */
	init_snmp("snmpd");


	if ( !VFILER_IS_DEFAULT(sk_get_vfiler()) ) {
		return false;
	}

	if ( strcmp("access",key) == 0 ) {
		char prev_val[MAX_OPT_VALUE_LEN + 1];
		AccessRet ret;
		char err_buf[128];
		char* new_val = &(snmp_access_spec[host] [0]);

		strcpy(prev_val, &(snmp_access_spec[host][0]));

		SK_DOMAIN_REFERENCE(access_check_syntax, SK_REASON_SWITCH_DOMAIN);
		ret = access_check_syntax(v, err_buf, sizeof(err_buf));
		if ( ret == AR_SYNTAX_OK ) {
			SK_DOMAIN_REFERENCE(pac_set_access_specs,
				SK_REASON_SWITCH_DOMAIN);
			pac_set_access_specs(new_val, v);
		} else {
			errno = EINVAL;
			return false;
		}

		/* 
		 * Apply the HA constraints to the old+new
		 * setting combination.
		 */
		check_set_in_HA("snmp.access", prev_val, new_val,
				SAME_REQD); 

	} else if ( strcmp("authtrap",key) == 0 ) {
		uint32_t t = (uint32_t) strtoul(v, (char **) NULL, 10);
		/* SNMP internally stores enabled as 1 and disabled as 2 */
		if (t == 0) {
			snmp_enableauthentraps[host] = SNMP_AUTHENTICATED_TRAPS_DISABLED;
		} else {
			snmp_enableauthentraps[host] = SNMP_AUTHENTICATED_TRAPS_ENABLED;
		}
	} else if ( strcmp("communities",key) == 0 ) {
		r = snmp_registry_set_communities(v,host);
	} else if ( strcmp("contact",key) == 0 ) {
		r = REGISTRY_SNMP_SET_STRING(v, system_contact[host]);
	} else if ( strcmp("enable",key) == 0 ) {
		int value = check_boolean_value(v, "snmp.enable");
		const char* prev_val = (snmp_enabled[host] ? "on" : "off");

		if ( value == -1 ) {
			errno = EINVAL;
			return false;
		}

		if (value == 1) {
			snmp_enable(host);
		} else {
			snmp_disable(host);
		}
		
		/*
		 * Tell the registry to make it persistent.
		 */
		registry_service_persistent("snmp",snmp_enabled[host]);

		/* 
		 * Apply the HA constraints to the old+new
		 * setting combination.
		 */
		const char* new_setting = (snmp_enabled[host] ? "on" : "off");
		
		check_set_in_HA("snmp.enable", prev_val, new_setting,
				SAME_REQD); 
	} else if ( strcmp("init",key) == 0 ) {
		snmp_init_flag[host] = (int) strtol(v, (char **) NULL, 10);
	} else if ( strcmp("location",key) == 0 ) {
		r = REGISTRY_SNMP_SET_STRING(v, system_location[host]);
	} else if ( strcmp("traphosts",key) == 0 ) {
		r = snmp_registry_set_traphosts(v,host);
	} else {
		errno = ENOENT;
		return false;
	}
	if ( r != 0 ) {
		errno = EINVAL;
		return false;
	}
	return true;
}

/*
 * options_snmp_registry_delete - delete an snmp option
 */
static boolean_t
options_snmp_registry_delete(regImpl* impl, regCred a, const char* key)
{
	VERIFY(key != NULL);

	D(registry, 20, ("options_snmp_registry_delete: key=%s\n",key) );

	if ( (strcmp("access",key) == 0) ||
		(strcmp("authtrap",key) == 0) ||
		(strcmp("communities",key) == 0) ||
		(strcmp("contact",key) == 0) ||
		(strcmp("enable",key) == 0) ||
		(strcmp("init",key) == 0) ||
		(strcmp("location",key) == 0) ||
		(strcmp("traphosts",key) == 0) ) {
		/*
		 * XXX - Key should be set to it's default value.
		 */
		return true;
	} else {
		errno = ENOENT;
		return false;
	}
}

/*
 * options_snmp_registry_walkinit - Start walking the options values from the
 * first that matches a specified prefix.  Return a cookie that can be given
 * back to snmp_registry_walknext() to actually do the walk.
 */
SK_DOMAIN_EXPORT(options_snmp_registry_walkinit, SK_REASON_MPSAFE);
STATIC  const char *
options_snmp_registry_walkinit(regImpl* impl, regCred a, const char* prefix, regWalkType w)
{
	regCookie* c;

	VERIFY(prefix != NULL);

	D(registry, 20, ("options_snmp_registry_walkinit: prefix=%s\n",prefix) );

	c = registry_listwalker_newcookie(impl,prefix,w);
	if ( c == NULL ) {
		return NULL;
	}

	/*
	 * Here we enumerate all the keys that we're going to walk,
	 * and add them to the list maintained by the registry_listwalker_*
	 * functions.  These keys are then returned one at a time by
	 * the walknext function.
	 *
	 * "snmp.access" is a valid option in Redhook and later releases only.
	 */
	registry_listwalker_additem(c,"access");
	registry_listwalker_additem(c,"authtrap");
	registry_listwalker_additem(c,"communities");
	registry_listwalker_additem(c,"contact");
	registry_listwalker_additem(c,"init");	/* actually controls traps */
	registry_listwalker_additem(c,"location");
	registry_listwalker_additem(c,"traphosts");
	return c->cookie;
}

regImpl optionsSnmpRegImpl = {
	WALK_ALL | WALK_DEFAULTPERSISTENCE,
	0,
	0,			/* load */
	0,
	options_snmp_registry_get,
	options_snmp_registry_set,
	options_snmp_registry_delete,
	options_snmp_registry_walkinit,	/* walkinit */
	default_registry_walknext_using_listwalker,	/* walknext */
	default_registry_walkend_using_listwalker,	/* walkend */
	0,			/* forward */
	0,			/* rawinfo */
	REGIMPL_DEFLT_INIT	/* other default methods and things */
};
//
// The remainder of this file implements the methods required for the options
// implementation.
// It also defiend the options themselves
//

static char *
getval_snmp_access(void *value_arg, char *return_str, int leng)
{
        char *specp = &(snmp_access_spec[sk_gethost()] [0]);
        return (specp);
}

static void
setval_snmp_access(void *value_arg, char *entry,
                  uint32_t mmin, uint64_t mmax, char *name)
{
        AccessRet ret;
        char err_buf[128];
        /* int host = sk_gethost(); */
        char *specp = &(snmp_access_spec[sk_gethost()] [0]);

        /*
         * check syntax and set if valid
         */
        SK_DOMAIN_REFERENCE(access_check_syntax,SK_REASON_SWITCH_DOMAIN);
        ret = access_check_syntax(entry, err_buf, sizeof(err_buf));
        if ( ret == AR_SYNTAX_OK ) {
                SK_DOMAIN_REFERENCE(pac_set_access_specs,
                                    SK_REASON_SWITCH_DOMAIN);
                pac_set_access_specs(specp, entry);
        }
        else {
                cmd_errprintf("%s\n", err_buf );
                D( pac, 5,
                   ("pac: setval_snmp_access() syntax error: %d:%s.\n",
                   ret, err_buf) );
        }

}


        OPTION_DEFINITION_VIS( "snmp.access", getval_snmp_access, setval_snmp_access,
           NULL, NULL, DIFF_OK, OPT_FILER_WIDE, VISIBLE(ALL), 0, 0, PERSIST);


static char *
getval_snmp_enable(void *value_arg, char *return_str, int leng)
{
        return (snmp_enabled[sk_gethost()] ? "on" : "off");
}

static void
setval_snmp_enable(void *value_arg, char *entry,
                                        uint32_t mmin, uint64_t mmax, char *name)
{
        int value = check_boolean_value(entry,name);
        int host = sk_gethost();

        if (value == -1) {
                return;
	}

        if ((value == 1 && snmp_enabled[host]) ||
                (value == 0 && !snmp_enabled[host])) {
                return;         /* no change*/
	}

        snmp_enable_change_to(value);

        /*
         * Tell the registry to make it persistent.
         */
        registry_service_persistent("snmp",snmp_enabled[host]);
        /*
         * We assume the options command is going to call registry_commit()
         */

        /*
         * Keep any remote management controllers in sync, but only
         * if this was actually changed (and not just being initially
         * loaded).
         */
        if (!VV_registryLoading) {
                sp_config_snmp_update();
        }

}

        OPTION_DEFINITION_VIS( "snmp.enable", getval_snmp_enable, setval_snmp_enable,
           NULL, NULL, DIFF_OK, OPT_FILER_WIDE, VISIBLE(ALL), 0, 0, NOPERSIST);
