/*
 *
 * Copyright (c) 1992-2010 NetApp, Inc.
 * All rights reserved.
 */
/* Portions of this file are subject to the following copyright(s).  See
 * the Net-SNMP's COPYING file for more details and other copyrights
 * that may apply:
 */
/*
 * Portions of this file are copyrighted by:
 * Copyright Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms specified in the COPYING file
 * distributed with the Net-SNMP package.
 */

#include <net-snmp/net-snmp-config.h>
#include <sk/ha_host.h>


#include <sk/sk_param.h>
#include <sk/sk_errno.h>

#include <sys/types.h>
#include <legacy/ctype.h>
#include <sys/errno.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/sys/sk_socketvar.h>
#include <net7g/sys/sk_socketfunc.h>
#include<net7g/net/sk_if.h>
#include <net7g/netinet/sk_in_var.h>
#include <net7g/netinet/sk_in.h>
#include <net7g/netinet/sk_ip6.h>
#include <net7g/netinet6/sk_in6_var.h>
#include <net7g/netinet/sk_in_systm.h>
#include <net7g/netinet/sk_ip.h>
#include <lib/access_check.h>
#include <net7g/sys/sk_mbuf.h>
#include <cmds/ifconfig.h> /* for getifnamewithaddr */
#include <ontap/ems/snmpd_ems.h> /* for EMS messages */
#include <sk/sk_mutex.h>

#ifdef HAVE_STRING_H
#include <legacy/string.h>
#else
#include <strings.h>
#endif
#ifdef HAVE_STDLIB_H
#include <legacy/stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <net7g/sys/sk_socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <net7g/netinet/sk_in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <net7g/arpa/sk_inet.h>
#endif
#ifdef HAVE_NETDB_H
#include <net7g/sk_netdb.h>
#endif

#ifdef HAVE_WINSOCK_H
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

#ifdef HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

#include <net-snmp/types.h>
#include <net-snmp/output_api.h>
#include <net-snmp/config_api.h>

#include <net-snmp/library/snmp_transport.h>
#include <net-snmp/library/snmpUDPDomain.h>

#ifndef INADDR_NONE
#define INADDR_NONE	-1
#endif

static netsnmp_tdomain udpDomain;

/*
 * Return a string representing the address in data, or else the "far end"
 * address if data is NULL.  
 */

static char *
netsnmp_udp_fmtaddr(netsnmp_transport *t, void *data, int len)
{
    struct sockaddr_in *to = NULL;

    if (data != NULL && len == sizeof(struct sockaddr_in)) {
        to = (struct sockaddr_in *) data;
    } else if (t != NULL && t->data != NULL) {
        to = (struct sockaddr_in *) t->data;
    }
    if (to == NULL) {
        return strdup("UDP: unknown");
    } else {
	char tmp[64];
	char addrbuf[INET_ADDRSTRLEN];

	/*
         * Here we just print the IP address of the peer for compatibility
         * purposes.  It would be nice if we could include the port number and
         * some indication of the domain (c.f. AAL5PVC).  
         */

	sprintf(tmp, "%s", inet_ntop(AF_INET, (void *) &(to->sin_addr),
				addrbuf, INET_ADDRSTRLEN));
        return strdup(tmp);
    }
}

/*
 * You can write something into opaque that will subsequently get passed back 
 * to your send function if you like.  For instance, you might want to
 * remember where a PDU came from, so that you can send a reply there...  
 */
SK_DOMAIN_IMPORT(getifnamewithaddr, SK_REASON_SWITCH_DOMAIN);
static int
netsnmp_udp_recv(netsnmp_transport *t, void *buf, int size,
		 void **opaque, int *olength)
{
	int             rc = -1;
	socklen_t       fromlen = sizeof(struct sockaddr);
	struct sockaddr *from;
	struct msghdr msg;
	struct cmsghdr *cp;
	struct in_addr *destaddr;
	char ifname[IFNAMSIZ];
	struct sockaddr_in destsockaddr_in;
	char control[sizeof(struct cmsghdr) + sizeof(struct in_addr)];
	sk_domain_t calling_domain;
	PacRet p_ret;

	calling_domain = sk_domain_switch_begin(sk_domain_network);

	if (t != NULL && t->sock >= 0) {
		from = (struct sockaddr *) malloc(sizeof(struct sockaddr_in));
 		if (from == NULL) {
			goto done;
		} else {
			memset(from, 0, fromlen);
		}
		msg.msg_control = control;
		msg.msg_controllen = sizeof(control);
		rc = recvmsgfrom(t->sock, buf, size, 0, from, &fromlen, &msg); 
		if (rc < 0) {
			free(from);
			goto done;
	 	}
		if (msg.msg_controllen != sizeof(control)) {
			free(from);
			rc = -1;
			goto done;
		}
		cp = (struct cmsghdr *)msg.msg_control;
		VERIFY(cp->cmsg_level == IPPROTO_IP);
		VERIFY(cp->cmsg_len == sizeof (struct cmsghdr) +
					 sizeof (struct in_addr));
		destaddr = (struct in_addr *)(cp + 1);
		
		bzero(&destsockaddr_in, sizeof(destsockaddr_in));
		destsockaddr_in.sin_len = sizeof(destsockaddr_in);
		destsockaddr_in.sin_family = AF_INET;
		ifname[0] = 0;
		destsockaddr_in.sin_addr = *destaddr;
		SK_DOMAIN_REFERENCE(getifnamewithaddr,SK_REASON_SWITCH_DOMAIN);
		getifnamewithaddr((struct sockaddr *)&destsockaddr_in,ifname);

		/*
		 * For the filer we use protocol access control to allow
		 * or deny snmp access.
	 	 * Had to use pac_net() here instead of pac_socket(), because
		 * snmp is a udp connection and getpeername() did not work
		 * inside pac_socket().  Changed toaddr to orig_toaddr
		 * because if the packet is a broadcast packet, the
		 * toaddr is NULLed out.  See burt 49408.  PR_LEGACY means
		 * to use old behavior which is to allow everything.
		 */
		struct sockaddr_in *peersock = (struct sockaddr_in *)from;

		SK_DOMAIN_REFERENCE(pac_net,SK_REASON_SWITCH_DOMAIN);
		p_ret = pac_net("snmp",
				snmp_access_spec[sk_gethost()],
				PAC_CAN_BLOCK,
				ifname,
				(struct sockaddr *)peersock,
				(struct sockaddr *)(&destsockaddr_in));
		if (p_ret != PR_ACCESS_ALLOWED && p_ret != PR_LEGACY) {
			/* connection refused and already logged */
			free(from);
			rc = -1;
			goto done;
		}

		sk_domain_switch_end(calling_domain);
		if (rc >= 0) {
			char *str = netsnmp_udp_fmtaddr(NULL, from, fromlen);
			DEBUGMSGTL(("netsnmp_udp",
				"recvfrom fd %x got %d bytes (from %s)\n",
				t->sock, rc, str));
			free(str);
		} else {
			DEBUGMSGTL(("netsnmp_udp", "recvfrom fd %x err %d (\"%s\")\n",
			t->sock, errno, strerror(errno)));
		}
		*opaque = (void *)from;
		*olength = sizeof(struct sockaddr_in);
		return rc;
	}
  done:
	*opaque = NULL;
	*olength = 0;
	sk_domain_switch_end(calling_domain);
	return rc;
}

static int
netsnmp_udp_send(netsnmp_transport *t, void *buf, int size,
		 void **opaque, int *olength)
{
    int rc = -1;
    struct sockaddr *to = NULL;

    if (opaque != NULL && *opaque != NULL &&
        *olength == sizeof(struct sockaddr_in)) {
        to = (struct sockaddr *) (*opaque);
    } else if (t != NULL && t->data != NULL &&
               t->data_length == sizeof(struct sockaddr_in)) {
        to = (struct sockaddr *) (t->data);
    }

    if (to != NULL && t != NULL && t->sock >= 0) {
        char *str = netsnmp_udp_fmtaddr(NULL, (void *) to,
					sizeof(struct sockaddr_in));
        DEBUGMSGTL(("netsnmp_udp", "send %d bytes from %p to %s on fd %x\n",
                    size, buf, str, t->sock));
        free(str);
	while (rc < 0) {
	    rc = sendto(t->sock, buf, size, 0, to, sizeof(struct sockaddr));
	    if (rc < 0 && errno != EINTR) {
		break;
	    }
	}
    }
    return rc;
}



static int
netsnmp_udp_close(netsnmp_transport *t)
{
    int rc = -1;
    if (t->sock >= 0) {
#ifndef HAVE_CLOSESOCKET
        rc = soclose(t->sock);
#else
        rc = closesocket(t->sock);
#endif
        t->sock = NULL;
    }
    return rc;
}



/*
 * Open a UDP-based transport for SNMP.  Local is TRUE if addr is the local
 * address to bind to (i.e. this is a server-type session); otherwise addr is 
 * the remote address to send things to.  
 */

netsnmp_transport *
netsnmp_udp_transport(struct sockaddr_in *addr, int local)
{
    netsnmp_transport *t = NULL;
    int             rc = 0, udpbuf = (1 << 17);
    char           *str = NULL;
    char           *client_socket = NULL;
    int on = 1;
    inp_filter_func_t filter = &snmp_input_filter;

    if (addr == NULL || addr->sin_family != AF_INET) {
        return NULL;
    }

    t = (netsnmp_transport *) malloc(sizeof(netsnmp_transport));
    if (t == NULL) {
        return NULL;
    }

    str = netsnmp_udp_fmtaddr(NULL, (void *)addr, 
				 sizeof(struct sockaddr_in));
    DEBUGMSGTL(("netsnmp_udp", "open %s %s\n", local ? "local" : "remote",
                str));
    free(str);

    memset(t, 0, sizeof(netsnmp_transport));

    t->domain = netsnmpUDPDomain;
    t->domain_length = netsnmpUDPDomain_len;

    t->sock = socket(PF_INET, SOCK_DGRAM, 0, "snmp");
    if (t->sock == 0) {
        netsnmp_transport_free(t);
        return NULL;
    }
#ifdef  SO_BSDCOMPAT
    /*
     * Patch for Linux.  Without this, UDP packets that fail get an ICMP
     * response.  Linux turns the failed ICMP response into an error message
     * and return value, unlike all other OS's.  
     */
    if (0 == netsnmp_os_prematch("Linux","2.4"))
    {
        int             one = 1;
        setsockopt(t->sock, SOL_SOCKET, SO_BSDCOMPAT, (void *) &one,
                   sizeof(one));
    }
#endif                          /*SO_BSDCOMPAT */
    /*
     * SO_REUSEADDR will allow multiple apps to open the same port at
     * the same time. Only the last one to open the socket will get
     * data. Obviously, for an agent, this is a bad thing. There should
     * only be one listener.
     */
#ifdef ALLOW_PORT_HIJACKING
#ifdef  SO_REUSEADDR
    /*
     * Allow the same port to be specified multiple times without failing.
     *    (useful for a listener)
     */
    {
        int             one = 1;
        setsockopt(t->sock, SOL_SOCKET, SO_REUSEADDR, (void *) &one,
                   sizeof(one));
    }
#endif                          /*SO_REUSEADDR */
#endif
    /*
     * Try to set the send and receive buffers to a reasonably large value, so
     * that we can send and receive big PDUs (defaults to 8192 bytes (!) on
     * Solaris, for instance).  Don't worry too much about errors -- just
     * plough on regardless.  
     */

#ifdef  SO_SNDBUF
    if (setsockopt
        (t->sock, SOL_SOCKET, SO_SNDBUF, (void *) &udpbuf,
         sizeof(int)) != 0) {
        DEBUGMSGTL(("netsnmp_udp", "couldn't set SO_SNDBUF to %d bytes: %s\n",
                    udpbuf, strerror(errno)));
    }
#endif                          /*SO_SNDBUF */

#ifdef  SO_RCVBUF
    if (setsockopt
        (t->sock, SOL_SOCKET, SO_RCVBUF, (void *) &udpbuf,
         sizeof(int)) != 0) {
        DEBUGMSGTL(("netsnmp_udp", "couldn't set SO_RCVBUF to %d bytes: %s\n",
                    udpbuf, strerror(errno)));
    }
#endif                          /*SO_RCVBUF */

    if (local) {
        /*
         * This session is inteneded as a server, so we must bind on to the
         * given IP address, which may include an interface address, or could
         * be INADDR_ANY, but certainly includes a port number.
         */

//ntap Taking the hint from old snmpd.c	    
/*
	 * Set the "receive destination address" option, so that we can
	 * determine the address of incoming packets.  We will always
	 * try to set the outgoing source address to match the incoming
	 * destination address.  This allows the SNMP Java applets
	 * to work properly.
	 */
	    if (setsockopt(t->sock, IPPROTO_IP, IP_RECVDSTADDR, &on,
			   sizeof(on))) {
	 	EMS_snmp_daemon_socket_receive_option_failed(strerror(errno));
		netsnmp_udp_close(t);
		netsnmp_transport_free(t);
		return NULL;
	}
        /*
         * Install an IP_FILTER on the socket.  The socket listens
         * for connections at all times.
         */
        if (setsockopt(t->sock, IPPROTO_IP, IP_FILTER, &filter,
            sizeof(filter))) {
		EMS_snmp_daemon_socket_ipfilter_option_failed(strerror(errno));
		netsnmp_udp_close(t);
		netsnmp_transport_free(t);
                return NULL;
        }

        t->local = malloc(6);
        if (t->local == NULL) {
            netsnmp_udp_close(t);
            netsnmp_transport_free(t);
            return NULL;
        }
        memcpy(t->local, (uint8_t *) & (addr->sin_addr.s_addr), 4);
        t->local[4] = (htons(addr->sin_port) & 0xff00) >> 8;
        t->local[5] = (htons(addr->sin_port) & 0x00ff) >> 0;
        t->local_length = 6;

        rc = bind(t->sock, (struct sockaddr *) addr,
                  sizeof(struct sockaddr));
        if (rc != 0) {
	    EMS_snmp_daemon_socket_bind_failed(ntohs(addr->sin_port), 
							strerror(errno));
            netsnmp_udp_close(t);
            netsnmp_transport_free(t);
            return NULL;
        }
        t->data = NULL;
        t->data_length = 0;
    } else {
        /*
         * This is a client session.  If we've been given a
         * client address to send from, then bind to that.
         * Otherwise the send will use "something sensible".
         */
        client_socket = netsnmp_ds_get_string(NETSNMP_DS_LIBRARY_ID,
                                              NETSNMP_DS_LIB_CLIENT_ADDR);
        if (client_socket) {
            struct sockaddr_in client_addr;
            netsnmp_sockaddr_in( &client_addr, client_socket, 0);
            client_addr.sin_port = 0;
            rc = bind(t->sock, (struct sockaddr *)&client_addr,
                  sizeof(struct sockaddr));
	    if (rc != 0) {
		EMS_snmp_daemon_socket_bind_failed(ntohs(client_addr.sin_port),
						strerror(errno));
	    }
        }
        /*
         * Save the (remote) address in the
         * transport-specific data pointer for later use by netsnmp_udp_send.
         */

        t->data = malloc(sizeof(struct sockaddr_in));
        t->remote = malloc(6);
        if (t->data == NULL || t->remote == NULL) {
            netsnmp_transport_free(t);
            return NULL;
        }
        memcpy(t->remote, (uint8_t *) & (addr->sin_addr.s_addr), 4);
        t->remote[4] = (htons(addr->sin_port) & 0xff00) >> 8;
        t->remote[5] = (htons(addr->sin_port) & 0x00ff) >> 0;
        t->remote_length = 6;
        memcpy(t->data, addr, sizeof(struct sockaddr_in));
        t->data_length = sizeof(struct sockaddr_in);
    }

    /*
     * 16-bit length field, 8 byte UDP header, 20 byte IPv4 header  
     */

    t->msgMaxSize = 0xffff - 8 - 20;
    t->f_recv     = netsnmp_udp_recv;
    t->f_send     = netsnmp_udp_send;
    t->f_close    = netsnmp_udp_close;
    t->f_accept   = NULL;
    t->f_fmtaddr  = netsnmp_udp_fmtaddr;

    return t;
}



int
netsnmp_sockaddr_in(struct sockaddr_in *addr,
                    const char *inpeername, int remote_port)
{
    char           *cp = NULL, *peername = NULL;

    if (addr == NULL) {
        return 0;
    }
    memset(addr, 0, sizeof(struct sockaddr_in));

    DEBUGMSGTL(("netsnmp_sockaddr_in", "addr %p, peername \"%s\"\n",
                addr, inpeername ? inpeername : "[NIL]"));

    addr->sin_addr.s_addr = htonl(INADDR_ANY);
    addr->sin_family = AF_INET;
    if (remote_port > 0) {
        addr->sin_port = htons((uint16_t)remote_port);
    } else if (netsnmp_ds_get_int(NETSNMP_DS_LIBRARY_ID, 
				  NETSNMP_DS_LIB_DEFAULT_PORT) > 0) {
        addr->sin_port = htons((uint16_t)netsnmp_ds_get_int(NETSNMP_DS_LIBRARY_ID, 
						 NETSNMP_DS_LIB_DEFAULT_PORT));
    } else {
        addr->sin_port = htons(SNMP_PORT);
    }

    if (inpeername != NULL) {
        /*
         * Duplicate the peername because we might want to mank around with
         * it.  
         */

        peername = strdup(inpeername);
        if (peername == NULL) {
            return 0;
        }

        /*
         * Try and extract an appended port number.  
         */
        cp = strchr(peername, ':');
        if (cp != NULL) {
            *cp = '\0';
            cp++;
            if (atoi(cp) != 0) {
                DEBUGMSGTL(("netsnmp_sockaddr_in",
                            "port number suffix :%d\n", atoi(cp)));
                addr->sin_port = htons((uint16_t)atoi(cp));
            }
        }

        for (cp = peername; *cp && isdigit((int) *cp); cp++);
        if (!*cp && atoi(peername) != 0) {
            /*
             * Okay, it looks like just a port number.  
             */
            DEBUGMSGTL(("netsnmp_sockaddr_in", "totally numeric: %d\n",
                        atoi(peername)));
            addr->sin_port = htons((uint16_t)atoi(peername));
        } else if (inet_addr(peername) != INADDR_NONE) {
            /*
             * It looks like an IP address.  
             */
            DEBUGMSGTL(("netsnmp_sockaddr_in", "IP address\n"));
            addr->sin_addr.s_addr = inet_addr(peername);
        } else {
            /*
             * Well, it must be a hostname then.  
             */
#ifdef  HAVE_GETHOSTBYNAME
	    char buf[MAX_LINELENGTH_HOSTS+1];
	    struct hostent hbuf;
	    int err = 0;
	    struct hostent *hp = NULL;
	    hbuf.h_addrtype = AF_UNSPEC;

	    bzero(buf, MAX_LINELENGTH_HOSTS+1);
	    hp = gethostbyname_r(peername,&hbuf, buf, MAX_LINELENGTH_HOSTS, &err);
            if ((hp == NULL) || (hp->h_addr == NULL)) {
                DEBUGMSGTL(("netsnmp_sockaddr_in",
                            "hostname (couldn't resolve)\n"));
                free(peername);
                return 0;
            } else {
                if (hp->h_addrtype != AF_INET) {
                    DEBUGMSGTL(("netsnmp_sockaddr_in",
                                "hostname (not AF_INET!)\n"));
                    free(peername);
                    return 0;
                } else {
                    DEBUGMSGTL(("netsnmp_sockaddr_in",
                                "hostname (resolved okay)\n"));
                    memcpy(&(addr->sin_addr), hp->h_addr, hp->h_length);
                }
            }
#else                           /*HAVE_GETHOSTBYNAME */
            DEBUGMSGTL(("netsnmp_sockaddr_in",
                        "hostname (no gethostbyname)\n"));
            free(peername);
            return 0;
#endif                          /*HAVE_GETHOSTBYNAME */
        }
    } else {
        DEBUGMSGTL(("netsnmp_sockaddr_in", "NULL peername"));
        return 0;
    }
    DEBUGMSGTL(("netsnmp_sockaddr_in", "return { AF_INET, %s:%hu }\n",
                inet_ntoa(addr->sin_addr), ntohs(addr->sin_port)));
    free(peername);
    return 1;
}


/*
 * The following functions provide the "com2sec" configuration token
 * functionality for compatibility.  
 */

#define EXAMPLE_NETWORK		"NETWORK"
#define EXAMPLE_COMMUNITY	"COMMUNITY"

typedef struct _com2SecEntry {
    char            community[VACMSTRINGLEN];
    unsigned long   network;
    unsigned long   mask;
    char            secName[VACMSTRINGLEN];
    struct _com2SecEntry *next;
} com2SecEntry;

com2SecEntry   *com2SecList[ONTAP_NHOSTS] = {NULL, NULL};
sk_mutex_t     com2sec_mutex[ONTAP_NHOSTS];

void
netsnmp_udp_parse_security(const char *token, char *param)
{
    char           secName[VACMSTRINGLEN];
    char           source[VACMSTRINGLEN];
    char           community[VACMSTRINGLEN];
    char           *cp = NULL;
    const char     *strmask = NULL;
    com2SecEntry   *e = NULL;
    in_addr_t   network = 0, mask = 0;
    int32_t     host = sk_gethost();
    bool        is_locked = false;

    /*
     * Get security, source address/netmask and community strings.  
     */

    cp = copy_nword(param, secName, sizeof(secName));
    if (secName[0] == '\0') {
        config_perror("missing NAME parameter");
        return;
    } else if (strlen(secName) > (VACMSTRINGLEN - 1)) {
        config_perror("security name too long");
        return;
    }
    cp = copy_nword(cp, source, sizeof(source));
    if (source[0] == '\0') {
        config_perror("missing SOURCE parameter");
        return;
    } else if (strncmp(source, EXAMPLE_NETWORK, strlen(EXAMPLE_NETWORK)) ==
               0) {
        config_perror("example config NETWORK not properly configured");
        return;
    }
    cp = copy_nword(cp, community, sizeof(community));
    if (community[0] == '\0') {
        config_perror("missing COMMUNITY parameter\n");
        return;
    } else
        if (strncmp
            (community, EXAMPLE_COMMUNITY, strlen(EXAMPLE_COMMUNITY))
            == 0) {
        config_perror("example config COMMUNITY not properly configured");
        return;
    } else if (strlen(community) > (VACMSTRINGLEN - 1)) {
        config_perror("community name too long");
        return;
    }

    /*
     * Process the source address/netmask string.  
     */

    cp = strchr(source, '/');
    if (cp != NULL) {
        /*
         * Mask given.  
         */
        *cp = '\0';
        strmask = cp + 1;
    }

    /*
     * Deal with the network part first.  
     */

    if ((strcmp(source, "default") == 0)
        || (strcmp(source, "0.0.0.0") == 0)) {
        network = 0;
        strmask = "0.0.0.0";
    } else {
        /*
         * Try interpreting as a dotted int64_t.  
         */
        network = inet_addr(source);

        if (network == (in_addr_t) -1) {
            /*
             * Nope, wasn't a dotted int64_t.  Must be a hostname.  
             */
#ifdef  HAVE_GETHOSTBYNAME
	    char buf[MAX_LINELENGTH_HOSTS+1];
	    struct hostent hbuf;
	    int err = 0;
	    struct hostent *hp = NULL;
	    hbuf.h_addrtype = AF_UNSPEC;

	    bzero(buf, MAX_LINELENGTH_HOSTS+1);
	    hp = gethostbyname_r(source,&hbuf, buf, MAX_LINELENGTH_HOSTS, &err);
            if ((hp == NULL) || (hp->h_addr == NULL)) {
                config_perror("bad source address");
                return;
            } else {
                if (hp->h_addrtype != AF_INET) {
                    config_perror("no IP address for source hostname");
                    return;
                }
                network = *((in_addr_t *) hp->h_addr);
            }
#else                           /*HAVE_GETHOSTBYNAME */
            /*
             * Oh dear.  
             */
            config_perror("cannot resolve source hostname");
            return;
#endif                          /*HAVE_GETHOSTBYNAME */
        }
    }

    /*
     * Now work out the mask.  
     */

    if (strmask == NULL || *strmask == '\0') {
        /*
         * No mask was given.  Use 255.255.255.255.  
         */
        //Burt 545293: Lint warning 570
        mask = 0xffffffffU;
    } else {
        if (strchr(strmask, '.')) {
            /*
             * Try to interpret mask as a dotted int64_t.  
             */
            mask = inet_addr(strmask);
            if (mask == (in_addr_t) -1 &&
                strncmp(strmask, "255.255.255.255", 15) != 0) {
                config_perror("bad mask");
                return;
            }
        } else {
            /*
             * Try to interpret mask as a "number of 1 bits".  
             */
            int             maskLen = atoi(strmask), maskBit = 0x80000000L;
            if (maskLen <= 0 || maskLen > 32) {
                config_perror("bad mask length");
                return;
            }
            while (maskLen--) {
                mask |= maskBit;
                maskBit >>= 1;
            }
            mask = htonl(mask);
        }
    }

    /*
     * Check that the network and mask are consistent.  
     */

    if (network & ~mask) {
        config_perror("source/mask mismatch");
        return;
    }

    e = (com2SecEntry *) malloc(sizeof(com2SecEntry));
    if (e == NULL) {
        config_perror("memory error");
        return;
    }

    /*
     * Everything is okay.  Copy the parameters to the structure allocated
     * above and add it to END of the list.  
     */

    DEBUGMSGTL(("netsnmp_udp_parse_security",
                "<\"%s\", 0x%08x/0x%08x> => \"%s\"\n", community, network,
                mask, secName));

    strcpy(e->secName, secName);
    strcpy(e->community, community);
    e->network = network;
    e->mask = mask;
    e->next = NULL;
    
    /*
     * Since vacm_parse_simple() is called recursively, make
     * sure not to acquire the lock when holding it.
     */
    if (!sk_is_mutex_mine(&com2sec_mutex[host])) {
        sk_mutex_lock(&com2sec_mutex[host]);
        is_locked = true;
    }
    e->next = com2SecList[host];
    com2SecList[host] = e;
    
    if (is_locked) {
        sk_mutex_unlock(&com2sec_mutex[host]);
    }
}


void
netsnmp_udp_com2SecList_free(void)
{
    int32_t     host = sk_gethost();
    bool        is_locked = false;
    com2SecEntry   *e = NULL;

    if (!sk_is_mutex_mine(&com2sec_mutex[host])) {
        sk_mutex_lock(&com2sec_mutex[host]);
        is_locked = true;
    }
    e = com2SecList[host];
    com2SecList[host] = NULL;
    if (is_locked) {
        sk_mutex_unlock(&com2sec_mutex[host]);
    }

    /* Clean up associated vac resources. 
     * Find and delete group (v1 & v2c), access, and view.
     */
    while (e != NULL) {
        com2SecEntry   *tmp = e;

	vacm_free_GAV(SNMP_SEC_MODEL_SNMPv1, e->secName);
	vacm_free_GAV(SNMP_SEC_MODEL_SNMPv2c, e->secName);

        e = e->next;
        free(tmp);
    }
}


void
netsnmp_udp_agent_config_tokens_register(void)
{
    register_app_config_handler("com2sec", netsnmp_udp_parse_security,
                                netsnmp_udp_com2SecList_free,
                                "name IPv4-network-address[/netmask] community");
}



/*
 * Return 0 if there are no com2sec entries, or return 1 if there ARE com2sec 
 * entries.  On return, if a com2sec entry matched the passed parameters,
 * then *secName points at the appropriate security name, or is NULL if the
 * parameters did not match any com2sec entry.  
 */

int
netsnmp_udp_getSecName(void *opaque, int olength,
                       const char *community,
                       size_t community_len, char **secName)
{
    com2SecEntry   *c;
    struct sockaddr_in *from = (struct sockaddr_in *) opaque;
    char           *ztcommunity = NULL;
    int32_t     host = sk_gethost();
    bool        is_locked = false;


    if (secName != NULL) {
        *secName = NULL;  /* Haven't found anything yet */
    }

    if (community == NULL || community_len == 0) {
        DEBUGMSGTL(("netsnmp_udp_getSecName", "Zero-length community\n"));
	return 0;
    }

    /*
     * Special case if there are NO entries (as opposed to no MATCHING
     * entries).  
     */
    if (!sk_is_mutex_mine(&com2sec_mutex[host])) {
        sk_mutex_lock(&com2sec_mutex[host]);
        is_locked = true;
    }
    if (com2SecList[host] == NULL) {
        DEBUGMSGTL(("netsnmp_udp_getSecName", "no com2sec entries\n"));
        if (is_locked) {
            sk_mutex_unlock(&com2sec_mutex[host]);
        }
        return 0;
    }

    /*
     * If there is no IPv4 source address, then there can be no valid security
     * name.  
     */

    if (opaque == NULL || olength != sizeof(struct sockaddr_in) ||
        from->sin_family != AF_INET) {
        DEBUGMSGTL(("netsnmp_udp_getSecName",
		    "no IPv4 source address in PDU?\n"));
        if (is_locked) {
            sk_mutex_unlock(&com2sec_mutex[host]);
        }
        return 1;
    }

    DEBUGIF("netsnmp_udp_getSecName") {
	ztcommunity = (char *)malloc(community_len + 1);
	if (ztcommunity != NULL) {
	    memcpy(ztcommunity, community, community_len);
	    ztcommunity[community_len] = '\0';
	}

	DEBUGMSGTL(("netsnmp_udp_getSecName", "resolve <\"%s\", 0x%08x>\n",
		    ztcommunity ? ztcommunity : "<malloc error>",
		    from->sin_addr.s_addr));
    }

    for (c = com2SecList[host]; c != NULL; c = c->next) {
        DEBUGMSGTL(("netsnmp_udp_getSecName","compare <\"%s\", 0x%08x/0x%08x>",
		    c->community, c->network, c->mask));
        if ((community_len == strlen(c->community)) &&
	    (memcmp(community, c->community, community_len) == 0) &&
            ((from->sin_addr.s_addr & c->mask) == c->network)) {
            DEBUGMSG(("netsnmp_udp_getSecName", "... SUCCESS\n"));
            if (secName != NULL) {
                *secName = c->secName;
            }
            break;
        }
        DEBUGMSG(("netsnmp_udp_getSecName", "... nope\n"));
    }
    if (is_locked) {
        sk_mutex_unlock(&com2sec_mutex[host]);
    }
    if (ztcommunity != NULL) {
        free(ztcommunity);
    }
    return 1;
}



static netsnmp_transport *
netsnmp_udp_create_tstring(const char *str, int local)
{
    struct sockaddr_in addr;

    if (netsnmp_sockaddr_in(&addr, str, 0)) {
        return netsnmp_udp_transport(&addr, local);
    } else {
        return NULL;
    }
}


static netsnmp_transport *
netsnmp_udp_create_ostring(const uint8_t * o, size_t o_len, int local)
{
    struct sockaddr_in addr;

    if (o_len == 6) {
        unsigned short porttmp = (o[4] << 8) + o[5];
        addr.sin_family = AF_INET;
        memcpy((uint8_t *) & (addr.sin_addr.s_addr), o, 4);
        addr.sin_port = htons(porttmp);
        return netsnmp_udp_transport(&addr, local);
    }
    return NULL;
}


void
netsnmp_udp_ctor(void)
{
    udpDomain.name = netsnmpUDPDomain;
    udpDomain.name_length = netsnmpUDPDomain_len;
    udpDomain.prefix = calloc(2, sizeof(char *));
    udpDomain.prefix[0] = "udp";

    udpDomain.f_create_from_tstring = netsnmp_udp_create_tstring;
    udpDomain.f_create_from_ostring = netsnmp_udp_create_ostring;

    netsnmp_tdomain_register(&udpDomain);
}

SK_DOMAIN_EXPORT(snmp_input_filter, SK_DOMAIN_NETWORK);	/* network calls */
int
snmp_input_filter(struct inpcb *inp, struct in_ifaddr *ia, void *ifpv,
	struct mbuf *m)
{
	struct ip* ipp = NULL;
	struct ip6_hdr* ip6p = NULL;
	struct ifnet *ifp = (struct ifnet*)ifpv;

	struct sockaddr_storage src;
	struct sockaddr_storage dst;
	PacRet p_ret;
	bool   isipv6;

	isipv6 = (mtod(m, struct ip *)->ip_v == 6);

	if (isipv6) {
		ip6p = mtod(m, struct ip6_hdr *);
		D(pac, 10, ("snmp: if=%s  dest IP=%x:%x:%x:%x  src IP=%x:%x:%x:%x\n",
			ifp->if_fullname, ip6p->ip6_dst.s6_addr32[0],
		 	ip6p->ip6_dst.s6_addr32[1], ip6p->ip6_dst.s6_addr32[2],
			ip6p->ip6_dst.s6_addr32[3], ip6p->ip6_src.s6_addr32[0],
			ip6p->ip6_src.s6_addr32[1], ip6p->ip6_src.s6_addr32[2],
			ip6p->ip6_src.s6_addr32[3]));
	} else {
		ipp = mtod(m, struct ip *);
		D(pac, 10, ("snmp: if=%s  dest IP=%x  src IP=%x\n",
			ifp->if_fullname, ipp->ip_dst.s_addr,
			ipp->ip_src.s_addr) );
	}

	/*
	 * if a loopback packet, then access allowed
	 */
	if (ifp->if_flags & IFF_LOOPBACK)
		return true;
	
	/*
	 * if snmp not enabled then access is denied
	 */
	if (!snmp_enabled[sk_gethost()])
		return false;

	if (isipv6) {
		SAS_SetIPv6Addr(&src, &(ip6p->ip6_src));
		SAS_SetIPv6Addr(&dst, &(ip6p->ip6_dst));
	} else {
		SAS_SetIPv4Addr(&src, ipp->ip_src.s_addr);
		SAS_SetIPv4Addr(&dst, ipp->ip_dst.s_addr);
	}
	/*
	 * snmp is enabled, so check the specifications against the
	 * credentials. 
	 */
	SK_DOMAIN_REFERENCE(pac_net,SK_REASON_SWITCH_DOMAIN);
	p_ret = pac_net( "snmp",
			 snmp_access_spec[sk_gethost()],
			 PAC_CANNOT_BLOCK,
			 ifp->if_fullname,
			(struct sockaddr *)&src,
			(struct sockaddr *)&dst);

	if ( p_ret == PR_ACCESS_DENIED ) {
		return false;
	}

	/* return TRUE for access allowed and would block */
	return true;
}
